@@ -15,7 +15,6 @@
 from __future__ import annotations
 
 from collections.abc import Callable, Sequence
-import functools
 from functools import partial
 import logging
 from typing import Any
@@ -27,6 +26,7 @@ from jax._src import ad_util
 from jax._src import api
 from jax._src import config
 from jax._src import core
+from jax._src import deprecations
 from jax._src import dtypes
 from jax._src import linear_util as lu
 from jax._src import effects
@@ -43,7 +43,10 @@ from jax._src.lib.mlir.dialects import hlo
 from jax._src.state import discharge
 from jax._src.state.types import AbstractRef
 from jax._src.traceback_util import api_boundary
-from jax._src.tree_util import PyTreeDef, tree_flatten, tree_unflatten, tree_structure
+from jax._src.tree_util import (
+    PyTreeDef, tree_flatten, tree_unflatten, tree_structure, broadcast_prefix,
+    tree_map)
+from jax._src.typing import DeprecatedArg
 from jax._src.util import (unzip2, wraps, split_list, partition_list, safe_map,
                            safe_zip, merge_lists, weakref_lru_cache)
 
@@ -58,30 +61,48 @@ logger = logging.getLogger(__name__)
 ### Policies
 
 def everything_saveable(*_, **__) -> bool:
-  # This is the effective policy without any use of jax.remat.
+  """The default strategy, as if ``jax.checkpoint`` were not being used at all.
+
+  This is the effective policy without any use of jax.remat."""
   return True
 
 def nothing_saveable(*_, **__) -> bool:
-  # This is the effective policy when using jax.remat without explicit policy.
+  """Rematerialize everything, as if a custom policy were not being used at all.
+
+  This is the effective policy when using jax.remat without explicit policy."""
   return False
 
 def dots_saveable(prim, *_, **__) -> bool:
   # Matrix multiplies are expensive, so let's save them (and nothing else).
-  return prim in {lax_internal.dot_general_p,
-                  lax_convolution.conv_general_dilated_p}
+  # We check for scaled matmul by name because we want to avoid importing
+  # cudnn here.
+  return (prim in {lax_internal.dot_general_p,
+                  lax_convolution.conv_general_dilated_p} or
+          prim.name == "scaled_matmul_wrapper")
 checkpoint_dots = dots_saveable
 
-def dots_with_no_batch_dims_saveable(prim, *_, **params) -> bool:
-  # This is a useful heuristic for transformers.
+def dots_with_no_batch_dims_saveable(prim, *args, **params) -> bool:
+  """This is a useful heuristic for transformers."""
   if prim is lax_internal.dot_general_p:
     (_, _), (lhs_b, rhs_b) = params['dimension_numbers']
     if not lhs_b and not rhs_b:
       return True
+
+  # We check for scaled matmul by name because we want to avoid importing
+  # cudnn here.
+  if prim.name == "scaled_matmul_wrapper":
+    lhs = args[0]
+    # Only save the dot if its batch dim is of size 1.
+    return lhs.shape[0] == 1
+
   return False
 
 def offload_dot_with_no_batch_dims(offload_src, offload_dst):
+  """Same as ``dots_with_no_batch_dims_saveable``, but offload to CPU memory
+  instead of recomputing.
+
+  This is a useful heuristic for transformers."""
   def policy(prim, *_, **params):
-    # This is a useful heuristic for transformers.
     if prim is lax_internal.dot_general_p:
       (_, _), (lhs_b, rhs_b) = params['dimension_numbers']
       if not lhs_b and not rhs_b:
@@ -102,7 +123,8 @@ def save_anything_except_these_names(*names_not_to_save):
   return policy
 
 def save_any_names_but_these(*names_not_to_save):
-  """Save only named values, excluding the names given."""
+  """Save only named values, i.e. any outputs of `checkpoint_name`, excluding
+  the names given."""
   names_not_to_save = frozenset(names_not_to_save)
   def policy(prim, *_, **params):
     if prim is name_p:
@@ -122,6 +144,8 @@ def save_only_these_names(*names_which_can_be_saved):
 def save_and_offload_only_these_names(
     *, names_which_can_be_saved, names_which_can_be_offloaded,
     offload_src, offload_dst):
+  """Same as ``save_only_these_names``, but offload to CPU memory instead of
+  recomputing."""
   names_which_can_be_saved = set(names_which_can_be_saved)
   names_which_can_be_offloaded = set(names_which_can_be_offloaded)
   intersection = names_which_can_be_saved.intersection(names_which_can_be_offloaded)
@@ -142,7 +166,9 @@ def save_and_offload_only_these_names(
 
 
 def save_from_both_policies(policy_1, policy_2):
+  """Logical OR of the given policies.
 
+  A residual is saveable iff it is saveable according to either policy."""
   def policy(prim, *args, **params):
     out1 = policy_1(prim, *args, **params)
     out2 = policy_2(prim, *args, **params)
@@ -174,11 +200,11 @@ checkpoint_policies = types.SimpleNamespace(
 
 ### Main API
 
-@api_boundary
+@partial(api_boundary, repro_api_name="jax.checkpoint")
 def checkpoint(fun: Callable, *, prevent_cse: bool = True,
                policy: Callable[..., bool] | None = None,
                static_argnums: int | tuple[int, ...] = (),
-               ) -> Callable:
+               concrete: bool | DeprecatedArg = DeprecatedArg()) -> Callable:
   """Make ``fun`` recompute internal linearization points when differentiated.
 
   The :func:`jax.checkpoint` decorator, aliased to :func:`jax.remat`, provides a
@@ -232,6 +258,8 @@ def checkpoint(fun: Callable, *, prevent_cse: bool = True,
       returns a boolean indicating whether the corresponding output value(s) can
       be saved as residuals (or instead must be recomputed in the (co)tangent
       computation if needed).
+    concrete: Optional boolean; deprecated. Will raise a DeprecationWarning if
+      used, and passing True will result in a NotImplementedError.
 
   Returns:
     A function (callable) with the same input/output behavior as ``fun`` but
@@ -319,8 +347,23 @@ def checkpoint(fun: Callable, *, prevent_cse: bool = True,
   ``jax.ensure_compile_time_eval``), it may be easier to compute some values
   outside the :func:`jax.checkpoint`-decorated function and then close over them.
   """
+  if not isinstance(concrete, DeprecatedArg):
+    concrete_msg = (
+        "The `concrete` option to `jax.checkpoint` has been deprecated."
+        " In its place please use `static_argnums`; for details refer to"
+        " https://docs.jax.dev/en/latest/jep/11830-new-remat-checkpoint.html."
+    )
+    deprecations.warn("jax-checkpoint-concrete", concrete_msg, stacklevel=2)
+    if concrete:
+      raise NotImplementedError(concrete_msg)
+
   if isinstance(static_argnums, int):
     static_argnums = static_argnums,
+  if isinstance(prevent_cse, list):
+    prevent_cse = tuple(prevent_cse)
+  if not isinstance(prevent_cse, (tuple, bool)):
+    raise TypeError("prevent_cse must be a bool or tuple of bools, got "
+                    f"{type(prevent_cse)=}")
 
   @wraps(fun)
   @api_boundary
@@ -332,13 +375,25 @@ def checkpoint(fun: Callable, *, prevent_cse: bool = True,
     args_flat, in_tree = tree_flatten((args, kwargs))
     in_avals = [core.shaped_abstractify(x) for x in args_flat]
     jaxpr, consts, out_tree = _trace_to_jaxpr(fun_, in_tree, tuple(in_avals), debug)
+    if isinstance(prevent_cse, tuple):
+      cse_args = (tuple(args), kwargs) if kwargs else tuple(args)
+      cse = (False,) * len(consts) + tuple(broadcast_prefix(prevent_cse, cse_args))
+    else:
+      cse = prevent_cse
     out_flat = remat_p.bind(
-        *consts, *args_flat, jaxpr=jaxpr, prevent_cse=prevent_cse,
-        differentiated=False, policy=policy)
+        *consts, *args_flat, jaxpr=jaxpr, prevent_cse=cse, differentiated=False,
+        policy=policy)
     return tree_unflatten(out_tree, out_flat)
   return fun_remat
 
-remat = checkpoint  # alias
+
+def remat(fun: Callable, *, prevent_cse: bool = True,
+          policy: Callable[..., bool] | None = None,
+          static_argnums: int | tuple[int, ...] = (),
+          concrete: bool | DeprecatedArg = DeprecatedArg()) -> Callable:
+  """Alias of :func:`jax.checkpoint`."""
+  return checkpoint(fun, prevent_cse=prevent_cse, policy=policy,
+                    static_argnums=static_argnums, concrete=concrete)
 
 # This function is similar to api_util.argnums_partial, except the error
 # messages are specific to jax.remat (and thus more actionable), the
@@ -457,10 +512,11 @@ def saved_residuals(f: Callable,
   jaxpr = jaxpr_.jaxpr
   out_shape = out_shape_[1]
   num_res = tree_structure(out_shape).num_leaves
-  jaxpr = jaxpr.replace(outvars=jaxpr.outvars[len(jaxpr.outvars) - num_res:])
-  out_tree = lambda: tree_structure(out_shape)
+  jaxpr = jaxpr.replace(
+      outvars=jaxpr.outvars[len(jaxpr.outvars) - num_res:],
+      debug_info=debug_info._replace(result_paths=None))
   assert len(jaxpr.invars) == len(in_leaves)
-  return _saved_residuals(jaxpr, debug_info.arg_names)
+  return _saved_residuals(jaxpr, debug_info.arg_names or ("unknown",) * len(jaxpr.invars))
 
 def _saved_residuals(jaxpr: core.Jaxpr,
                      arg_names: Sequence[str]) -> list[tuple[core.AbstractValue, str]]:
@@ -513,6 +569,12 @@ def print_saved_residuals(f, *args, **kwargs):
 remat_p = core.Primitive('remat2')
 remat_p.multiple_results = True
 
+def _remat_bind(*args, jaxpr, prevent_cse, differentiated, policy):
+  assert isinstance(prevent_cse, bool) or len(prevent_cse) == len(args)
+  return core.Primitive.bind(remat_p, *args, jaxpr=jaxpr, prevent_cse=prevent_cse,
+                             differentiated=differentiated, policy=policy)
+remat_p.bind = _remat_bind  # type: ignore
+
 @remat_p.def_impl
 def remat_impl(*args, jaxpr, prevent_cse, differentiated, policy):
   del prevent_cse, differentiated, policy  # Unused.
@@ -521,7 +583,7 @@ def remat_impl(*args, jaxpr, prevent_cse, differentiated, policy):
 @remat_p.def_effectful_abstract_eval
 def remat_abstract_eval(*args, jaxpr, prevent_cse, differentiated, policy):
   del args, prevent_cse, differentiated, policy  # Unused.
-  return [v.aval for v in jaxpr.outvars], jaxpr.effects
+  return [v.aval for v in jaxpr.outvars], core.eqn_effects(jaxpr)
 
 def remat_jvp(primals, tangents, jaxpr, prevent_cse, differentiated, policy):
   assert not jaxpr.constvars
@@ -529,6 +591,8 @@ def remat_jvp(primals, tangents, jaxpr, prevent_cse, differentiated, policy):
   jaxpr_jvp_, out_nz = ad.jvp_jaxpr(pe.close_jaxpr(jaxpr), in_nonzeros, False)
   nonzero_tangents = [t for t in tangents if type(t) is not ad_util.Zero]
   jaxpr_jvp = pe.convert_constvars_jaxpr(jaxpr_jvp_.jaxpr)
+  if isinstance(prevent_cse, tuple):
+    prevent_cse += (True,) * len(nonzero_tangents)
   outs = remat_p.bind(
       *jaxpr_jvp_.consts, *primals, *nonzero_tangents, jaxpr=jaxpr_jvp,
       prevent_cse=prevent_cse, differentiated=differentiated, policy=policy)
@@ -540,7 +604,7 @@ def remat_jvp(primals, tangents, jaxpr, prevent_cse, differentiated, policy):
 ad.primitive_jvps[remat_p] = remat_jvp
 
 def remat_partial_eval(trace: pe.JaxprTrace, *tracers: core.Tracer,
-                       jaxpr: core.Jaxpr, **params):
+                       jaxpr: core.Jaxpr, prevent_cse, **params):
   assert not jaxpr.constvars
   disallowed_effects = effects.remat_allowed_effects.filter_not_in(jaxpr.effects)
   if disallowed_effects:
@@ -568,7 +632,7 @@ def remat_partial_eval(trace: pe.JaxprTrace, *tracers: core.Tracer,
   # on producers of any residuals. See https://github.com/jax-ml/jax/pull/22244.
   jaxpr_known_ = _insert_reduce_precision(jaxpr_known, num_res)
 
-  # compute known outputs and residuals (hoisted out of remat primitive)
+  # Compute known outputs and residuals (hoisted out of remat primitive)
   _, in_consts_ = unzip2(t.pval for t in tracers if t.pval.is_known())
   _, in_consts = partition_list(in_used_known, in_consts_)
   out_consts = core.eval_jaxpr(jaxpr_known_, (), *in_consts)
@@ -580,9 +644,13 @@ def remat_partial_eval(trace: pe.JaxprTrace, *tracers: core.Tracer,
   in_jaxpr_tracers = res_tracers + map(trace.instantiate_const, tracers_staged)  # type: ignore
   out_jaxpr_tracers = [pe.JaxprTracer(trace, pe.PartialVal.unknown(x.aval), None)
                        for x in jaxpr_unknown.outvars]
-  new_params = dict(params, jaxpr=jaxpr_unknown, differentiated=True)
+  if isinstance(prevent_cse, tuple):
+    _, prevent_cse_ = partition_list(in_used_staged, prevent_cse)
+    prevent_cse = (True,) * len(res_tracers) + tuple(prevent_cse_)
+  new_params = dict(params, jaxpr=jaxpr_unknown, differentiated=True,
+                    prevent_cse=prevent_cse)
   recipe = pe.new_eqn_recipe(trace, in_jaxpr_tracers, out_jaxpr_tracers, remat_p,
-                             new_params, jaxpr_unknown.effects,
+                             new_params, core.eqn_effects(jaxpr_unknown),
                              source_info_util.current())
 
   # log info about saved residuals
@@ -618,7 +686,7 @@ def _insert_reduce_precision(jaxpr: core.Jaxpr, num_res: int) -> core.Jaxpr:
   used_vars = {x for e in jaxpr.eqns for x in e.invars if isinstance(x, core.Var)}
   invars, constvars, eqns = jaxpr.invars[:], jaxpr.constvars[:], jaxpr.eqns[:]
   for v in res_vars:
-    if (not isinstance(v.aval, core.UnshapedArray) or
+    if (not isinstance(v.aval, core.ShapedArray) or
         not dtypes.issubdtype(v.aval.dtype, np.inexact)):
       continue
     if v not in used_vars:
@@ -650,29 +718,42 @@ def _insert_reduce_precision(jaxpr: core.Jaxpr, num_res: int) -> core.Jaxpr:
   return new_jaxpr
 
 def remat_partial_eval_custom_params_updater(*args):
-  *_, params_known, params_staged = args
+  unks_in, inst_in, *_, params_known, params_staged = args
+  prevent_cse = params_known['prevent_cse']
+  assert prevent_cse == params_staged['prevent_cse']
+  if isinstance(prevent_cse, tuple):
+    prevent_cse_known, _ = partition_list(unks_in, prevent_cse)
+    _, prevent_cse_staged = partition_list(inst_in, prevent_cse)
+    params_known = dict(params_known, prevent_cse=tuple(prevent_cse_known))
+    params_staged = dict(params_staged, prevent_cse=tuple(prevent_cse_staged))
   return params_known, dict(params_staged, differentiated=True)
 pe.partial_eval_jaxpr_custom_rules[remat_p] = \
     partial(pe.call_partial_eval_custom_rule, 'jaxpr',
             remat_partial_eval_custom_params_updater)
 
-def remat_transpose(out_cts, *in_primals, jaxpr, **params):
+def remat_transpose(out_cts, *args, jaxpr, prevent_cse, **params):
+  # TODO(mattjj): avoid round-tripping into UndefinedPrimals
+  args_ = [ad.UndefinedPrimal(x.aval) if isinstance(x, ad.ValAccum) else x
+           for x in args]
+
   assert not jaxpr.constvars
-  in_linear = [ad.is_undefined_primal(x) for x in in_primals]
+  in_linear = [ad.is_undefined_primal(x) for x in args_]
   out_zeros = [type(ct) is ad_util.Zero for ct in out_cts]
   transposed_jaxpr_, in_zeros = transpose_jaxpr(
       pe.close_jaxpr(jaxpr), in_linear, out_zeros)
   transposed_jaxpr, consts = transposed_jaxpr_.jaxpr, transposed_jaxpr_.consts
   transposed_jaxpr = pe.convert_constvars_jaxpr(transposed_jaxpr)
-  args, _ = tree_flatten((in_primals, out_cts))
-  in_cts_nz = remat_p.bind(*consts, *args, jaxpr=transposed_jaxpr, **params)
+  flat_args, _ = tree_flatten((args_, out_cts))
+  if isinstance(prevent_cse, tuple):
+    prevent_cse_, _ = partition_list(in_linear, prevent_cse)
+    prevent_cse = tuple(prevent_cse_) + (True,) * (len(out_zeros) - sum(out_zeros))
+  in_cts_nz = remat_p.bind(*consts, *flat_args, jaxpr=transposed_jaxpr,
+                           prevent_cse=prevent_cse, **params)
   in_cts_nz_, in_zeros_ = iter(in_cts_nz), iter(in_zeros)
-  in_cts = [None if not ad.is_undefined_primal(x) else
-            ad_util.Zero(x.aval) if next(in_zeros_) else next(in_cts_nz_)
-            for x in in_primals]
-  assert next(in_cts_nz_, None) is next(in_zeros_, None) is None
-  return in_cts
-ad.primitive_transposes[remat_p] = remat_transpose
+  for x in args:
+    if isinstance(x, ad.ValAccum) and not next(in_zeros_):
+      x.accum(next(in_cts_nz_))
+ad.fancy_transposes[remat_p] = remat_transpose
 
 # TODO(mattjj): move this to ad.py
 def transpose_jaxpr(jaxpr: core.ClosedJaxpr, in_linear: bool | Sequence[bool],
@@ -724,7 +805,8 @@ def _transpose_jaxpr(jaxpr: core.ClosedJaxpr,
     in_cts_nz, _ = partition_list(in_zeros, in_cts)
     return in_cts_nz
 
-  transposed_wrapped = lu.wrap_init(transposed, debug_info=jaxpr.jaxpr.debug_info)
+  dbg = jaxpr.jaxpr.debug_info.with_unknown_names()
+  transposed_wrapped = lu.wrap_init(transposed, debug_info=dbg)
   transposed_jaxpr_, _, consts = pe.trace_to_jaxpr_dynamic(
       transposed_wrapped, in_avals)
   transposed_jaxpr = core.ClosedJaxpr(transposed_jaxpr_, consts)
@@ -748,7 +830,10 @@ def remat_dce(used_outputs: list[bool], eqn: core.JaxprEqn
   if not any(used_outputs) and not pe.has_effects(eqn):
     return [False] * len(eqn.invars), None
   new_jaxpr, used_inputs = pe.dce_jaxpr(eqn.params['jaxpr'], used_outputs)
-  new_params = dict(eqn.params, jaxpr=new_jaxpr)
+  prevent_cse = eqn.params['prevent_cse']
+  if isinstance(prevent_cse, tuple):
+    prevent_cse = tuple(p for p, u in zip(prevent_cse, used_inputs) if u)
+  new_params = dict(eqn.params, jaxpr=new_jaxpr, prevent_cse=prevent_cse)
   if (not any(used_inputs) and not any(used_outputs) and
       _has_effects(new_jaxpr.effects)):
     return used_inputs, None
@@ -756,12 +841,14 @@ def remat_dce(used_outputs: list[bool], eqn: core.JaxprEqn
     new_eqn = pe.new_jaxpr_eqn(
         [v for v, used in zip(eqn.invars, used_inputs) if used],
         [v for v, used in zip(eqn.outvars, used_outputs) if used],
-        eqn.primitive, new_params, new_jaxpr.effects, eqn.source_info, eqn.ctx)
+        eqn.primitive, new_params, core.eqn_effects(new_jaxpr),
+        eqn.source_info, eqn.ctx)
     return used_inputs, new_eqn
 pe.dce_rules[remat_p] = remat_dce
 
 def _has_effects(effects) -> bool:
-  return bool({e for e in effects if not isinstance(e, core.NamedAxisEffect)})
+  not_really_effects = (core.NamedAxisEffect, core.InternalMutableArrayEffect)
+  return any(not isinstance(e, not_really_effects) for e in effects)
 
 
 def remat_expansion(
@@ -783,34 +870,76 @@ def _remat_translation_using_opt_barrier(*args, jaxpr: core.Jaxpr):
 
 
 def _remat_lowering(
-    ctx,
+    ctx: mlir.LoweringRuleContext,
     *args,
     jaxpr: core.Jaxpr,
     prevent_cse: bool,
     differentiated: bool,
     policy,
 ):
-  jaxpr_args: Sequence[mlir.IrValues]
-  if differentiated and prevent_cse:
-    arg_types = map(mlir.aval_to_ir_type, ctx.avals_in)
-    flat_args = mlir.flatten_ir_values(args)
-    barrier_op = hlo.OptimizationBarrierOp(flat_args)
-    jaxpr_args = mlir.unflatten_ir_values_like_types(
-      barrier_op.results, arg_types)
-  else:
-    jaxpr_args = args
+  if isinstance(prevent_cse, bool):
+    prevent_cse = (prevent_cse,) * len(ctx.avals_in)  # type: ignore
+  assert isinstance(prevent_cse, tuple)
+  if differentiated and any(prevent_cse):
+    _, barrier_avals = partition_list(prevent_cse, ctx.avals_in)
+    other_args, barrier_args = partition_list(prevent_cse, args)
+    barrier_op = hlo.OptimizationBarrierOp(
+        mlir.flatten_ir_values(barrier_args))
+    barrier_results = mlir.unflatten_ir_values_like_types(
+        barrier_op.results, map(mlir.aval_to_ir_type, barrier_avals))
+    args = merge_lists(prevent_cse, other_args, barrier_results)  # type: ignore
   outs, tokens_out = mlir.jaxpr_subcomp(
       ctx.module_context, jaxpr, ctx.name_stack.extend('checkpoint'),
-      ctx.tokens_in, (), *jaxpr_args, dim_var_values=ctx.dim_var_values)
+      ctx.tokens_in, (), *args, dim_var_values=ctx.dim_var_values,
+      const_lowering=ctx.const_lowering)
   ctx.set_tokens_out(tokens_out)
   return outs
 
-
 mlir.register_lowering(remat_p, _remat_lowering)
 
 
 def checkpoint_name(x, name):
-  return name_p.bind(x, name=name)
+  """Identifies a value with a name within :func:`jax.checkpoint`.
+
+  This function acts as an identity function at runtime (returning ``x``
+  unchanged) but attaches a string name to the value in the JAX trace.
+  These names can be targeted by specific checkpointing policies (see
+  :ref:`checkpoint-policies`) to control which intermediate values
+  are saved during the forward pass and which are recomputed during the
+  backward pass.
+
+  Args:
+    x: array or PyTree of arrays to be named.
+    name: A string name to associate with the value ``x``.
+
+  Returns:
+    The input ``x``, unchanged.
+
+  See Also:
+    - :func:`jax.checkpoint` (alias: :func:`jax.remat`): decorator to
+      enable checkpointing.
+    - :mod:`jax.checkpoint_policies`: a namespace containing policies
+      that use names marked via ``checkpoint_name`` to determine behavior.
+
+  Example:
+    >>> import jax
+    >>> import jax.numpy as jnp
+    >>> from jax.ad_checkpoint import checkpoint_name
+
+    >>> # Define a function where we explicitly name an intermediate value
+    >>> def f(x):
+    ...   y = jnp.sin(x)
+    ...   z = checkpoint_name(y, "my_intermediate")
+    ...   return jnp.cos(z)
+
+    >>> # Use a policy that saves only the named value
+    >>> policy = jax.checkpoint_policies.save_only_these_names("my_intermediate")
+    >>> f_checkpointed = jax.checkpoint(f, policy=policy)
+
+    For further examples, see the `remat example notebook
+    <https://docs.jax.dev/en/latest/notebooks/autodiff_remat.html>`_.
+  """
+  return tree_map(partial(name_p.bind, name=name), x)
 
 name_p.def_impl(lambda x, *, name: x)
 name_p.def_abstract_eval(lambda x, *, name: x)
@@ -828,66 +957,6 @@ def name_batcher(args, dims, *, name):
 batching.primitive_batchers[name_p] = name_batcher
 
 
-@functools.wraps(checkpoint)
-def checkpoint_wrapper(
-    fun: Callable,
-    *,
-    concrete: bool = False,
-    prevent_cse: bool = True,
-    static_argnums: int | tuple[int, ...] = (),
-    policy: Callable[..., bool] | None = None,
-) -> Callable:
-  if concrete:
-    msg = ("The 'concrete' option to jax.checkpoint / jax.remat is deprecated; "
-           "in its place, you can use its `static_argnums` option, and if "
-           "necessary the `jax.ensure_compile_time_eval()` context manager.\n"
-           "\n"
-           "For example, if using `concrete=True` for an `is_training` flag:\n"
-           "\n"
-           "  from functools import partial\n"
-           "\n"
-           "  @partial(jax.checkpoint, concrete=True)\n"
-           "  def foo(x, is_training):\n"
-           "    if is_training:\n"
-           "      return f(x)\n"
-           "    else:\n"
-           "      return g(x)\n"
-           "\n"
-           "replace it with a use of `static_argnums`:\n"
-           "\n"
-           "  @partial(jax.checkpoint, static_argnums=(1,))\n"
-           "  def foo(x, is_training):\n"
-           "    ...\n"
-           "\n"
-           "If jax.numpy operations need to be performed on static arguments, "
-           "we can use the `jax.ensure_compile_time_eval()` context manager. "
-           "For example, we can replace this use of `concrete=True`\n:"
-           "\n"
-           "  @partial(jax.checkpoint, concrete=True)\n"
-           "  def foo(x, y):\n"
-           "    if y > 0:\n"
-           "      return f(x)\n"
-           "    else:\n"
-           "      return g(x)\n"
-           "\n"
-           "with this combination of `static_argnums` and "
-           "`jax.ensure_compile_time_eval()`:\n"
-           "\n"
-           "  @partial(jax.checkpoint, static_argnums=(1,))\n"
-           "  def foo(x, y):\n"
-           "    with jax.ensure_compile_time_eval():\n"
-           "      y_pos = y > 0\n"
-           "    if y_pos:\n"
-           "      return f(x)\n"
-           "    else:\n"
-           "      return g(x)\n"
-           "\n"
-           "See https://docs.jax.dev/en/latest/jep/11830-new-remat-checkpoint.html\n")
-    raise NotImplementedError(msg)
-  return checkpoint(fun, prevent_cse=prevent_cse, policy=policy,
-                    static_argnums=static_argnums)
-
-
 @discharge.register_discharge_rule(remat_p)
 def _remat_state_discharge_rule(
     in_avals, out_avals, *args, jaxpr, **params):
