@@ -31,6 +31,10 @@ T = TypeVar('T')
 map = safe_map
 
 def add_jaxvals(x: ArrayLike, y: ArrayLike) -> Array:
+  ty = core.typeof(x)
+  if hasattr(ty, 'vspace_add'):  # TODO(mattjj,dougalm): revise away hasattr
+    return ty.vspace_add(x, y)
+  x, y = core.standard_insert_pvary(x, y)
   return add_jaxvals_p.bind(x, y)
 
 add_jaxvals_p = Primitive('add_any')
@@ -43,10 +47,12 @@ raw_jaxval_adders = {}  # type: ignore
 
 @add_jaxvals_p.def_abstract_eval
 def add_abstract(x, y):
-  assert core.typematch(x, y)
+  assert core.typematch(x, y), (x, y)
   return x
 
 def zeros_like_aval(aval: core.AbstractValue) -> Array:
+  if hasattr(aval, 'vspace_zero'):  # TODO(mattjj,dougalm): revise away hasattr
+    return aval.vspace_zero()
   return aval_zeros_likers[type(aval)](aval)
 aval_zeros_likers: dict[type, Callable[[Any], Array]] = {}
 
@@ -67,7 +73,10 @@ class Zero:
     return f'Zero({self.aval})'
   @staticmethod
   def from_primal_value(val: Any) -> Zero:
+    # TODO(mattjj,yashkatariya): sometimes we want to_cotangent_aval...
     return Zero(get_aval(val).to_tangent_aval())
+  def instantiate(self):
+    return zeros_like_aval(self.aval)
 
 register_pytree_node(Zero, lambda z: ((), z.aval), lambda aval, _: Zero(aval))
 
