@@ -18,7 +18,7 @@ from collections.abc import Callable, Sequence
 import dataclasses
 import functools
 import logging
-from typing import Any
+from typing import Any, cast
 
 from jax._src import api
 from jax._src import config
@@ -35,7 +35,6 @@ from jax._src import xla_bridge as xb
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
-from jax._src.interpreters import xla
 from jax._src.lib import xla_client as xc
 from jax._src.lib.mlir import ir
 from jax._src.lib.mlir.dialects import hlo
@@ -135,6 +134,17 @@ batching.primitive_batchers[pure_callback_p] = functools.partial(
     ffi.ffi_batching_rule, pure_callback_p
 )
 
+def _get_sdy_array_list_for_callbacks(avals: Sequence[core.ShapedArray]) -> SdyArrayList:
+  """Returns an SdyArrayList with `max(1, len(avals))` replicated shardings."""
+  ndims = [0]
+  if avals:
+    ndims = [x.ndim for x in avals if isinstance(x, core.ShapedArray)]
+  return SdyArrayList([
+      SdyArray(
+          mesh_shape=(),
+          dim_shardings=[SdyDim(axes=[], is_open=False)] * ndim,
+          logical_device_ids=()) for ndim in ndims])
+
 
 def _callback_op_sharding(
     axis_context, sharding: SingleDeviceSharding | None, avals_out
@@ -153,16 +163,7 @@ def _callback_op_sharding(
           " computations"
       )
     if config.use_shardy_partitioner.value:
-      ndim = 0
-      if avals_out and isinstance(avals_out[0], core.ShapedArray):
-        ndim = avals_out[0].ndim
-      op_sharding = SdyArrayList([
-          SdyArray(
-              mesh_shape=(),
-              dim_shardings=[
-                  SdyDim(axes=[], is_open=False)
-              ] * ndim,
-              logical_device_ids=())])
+      op_sharding = _get_sdy_array_list_for_callbacks(avals_out)
     else:
       op_sharding = xc.OpSharding()  # type: ignore[assignment]
       op_sharding.type = xc.OpSharding.Type.MANUAL
@@ -237,6 +238,7 @@ def pure_callback_lowering(
       ctx.avals_in,
       ctx.avals_out,
       has_side_effect=False,
+      returns_token=False,
       sharding=op_sharding,
   )
   return result
@@ -513,6 +515,7 @@ def io_callback_lowering(ctx, *args, callback, sharding, ordered, **params):
         ctx.avals_in,
         ctx.avals_out,
         has_side_effect=True,
+        returns_token=True,
         sharding=op_sharding,
     )
     ctx.set_tokens_out(mlir.TokenSet({_OrderedIOEffect: token}))
@@ -525,6 +528,7 @@ def io_callback_lowering(ctx, *args, callback, sharding, ordered, **params):
         ctx.avals_in,
         ctx.avals_out,
         has_side_effect=True,
+        returns_token=False,
         sharding=op_sharding,
     )
   return result
@@ -640,7 +644,7 @@ def receive_from_host(
     if config.use_shardy_partitioner.value:
       assert isinstance(sharding, SdyArrayList)
       assert len(sharding.shardings) >= 1
-       # `RecvOp`'s last argument is a `TokenType`. Since Shardy requires the
+      # `RecvOp`'s last argument is a `TokenType`. Since Shardy requires the
       # number of shardings to match the number of results, but JAX only sees
       # the array result, we need to add an equivalent sharding for the token.
       # Note that even if a function returns N results, we will end up with N
@@ -657,7 +661,6 @@ def receive_from_host(
   return token, result
 
 
-
 def _aval_to_xla_shape(aval: core.AbstractValue) -> xc.Shape:
   try:
     return _xla_shape_handlers[type(aval)](aval)
@@ -677,7 +680,7 @@ _xla_shape_handlers[core.AbstractToken] = lambda _: xc.Shape.token_shape()
 
 
 def _emit_tpu_python_callback(
-    backend: xb.XlaBackend,
+    backend: xc.Client,
     ctx: mlir.LoweringRuleContext,
     callback,
     token: Any | None,
@@ -687,6 +690,7 @@ def _emit_tpu_python_callback(
     result_avals: Sequence[core.ShapedArray],
     result_shapes: Sequence[xc.Shape],
     *,
+    returns_token: bool,
     sharding: SdyArrayList | xc.OpSharding | None = None,
 ) -> tuple[Sequence[ir.Value], Any]:
   token = token or hlo.create_token()
@@ -718,13 +722,28 @@ def _emit_tpu_python_callback(
 
   recv_channels = []
   outputs = []
-  for result_aval in result_avals:
+  if returns_token and not result_avals:
+    # If the caller expects a token, we need at least one result so that the
+    # token from the recv is used as an indication that the callback is
+    # complete. Without this, we would only wait for the send to finish.
+    callback_without_results = _wrapped_callback
+    def _wrapped_callback(*args):  # pylint: disable=function-redefined
+      callback_without_results(*args)
+      return 0.0,
+    dummy_recv_aval = core.ShapedArray((), np.float32)
+    result_shapes = [_aval_to_xla_shape(dummy_recv_aval)]
     channel = ctx.module_context.new_channel()
-    assert isinstance(result_aval, core.ShapedArray)
-    token, out = receive_from_host(channel, token, result_aval,
-                                   callback.__name__, sharding=sharding)
-    outputs.append(out)
+    token, _ = receive_from_host(
+        channel, token, dummy_recv_aval, callback.__name__, sharding=sharding)
     recv_channels.append(channel)
+  else:
+    for result_aval in result_avals:
+      channel = ctx.module_context.new_channel()
+      assert isinstance(result_aval, core.ShapedArray)
+      token, out = receive_from_host(channel, token, result_aval,
+                                     callback.__name__, sharding=sharding)
+      outputs.append(out)
+      recv_channels.append(channel)
   ifrt_callback = backend.make_python_callback_from_host_send_and_recv(
       _wrapped_callback, operand_shapes, result_shapes, send_channels,
       recv_channels, pickle_util.dumps)
@@ -741,6 +760,7 @@ def emit_python_callback(
     result_avals: Sequence[core.ShapedArray],
     *,
     has_side_effect: bool,
+    returns_token: bool = True,
     partitioned: bool = False,
     sharding: SdyArrayList | xc.OpSharding | None = None,
 ) -> tuple[Sequence[mlir.IrValues], Any, Any]:
@@ -754,6 +774,7 @@ def emit_python_callback(
     operand_avals: The abstract values of the operands.
     result_avals: The abstract values of the results.
     has_side_effect: Whether the callback has side effects.
+    returns_token: Whether the callback should return a token.
     partitioned: If True, then `callback` is called on local shards only. If
       False, then `callback` is called on all shards.
     sharding: The sharding of the callback.
@@ -774,7 +795,7 @@ def emit_python_callback(
           f"Partitioned callback not implemented on {platform} backend.")
     if result_avals:
       raise ValueError("Partitioned callback not supported with return values.")
-  backend = ctx.module_context.get_backend()
+  backend: xc.Client = cast(xc.Client, ctx.module_context.get_backend())
   result_shapes = [_aval_to_xla_shape(aval) for aval in result_avals]
   operand_shapes = [_aval_to_xla_shape(aval) for aval in operand_avals]
 
@@ -787,7 +808,7 @@ def emit_python_callback(
           "Mismatched number of outputs from callback. "
           "Expected: {}, Actual: {}".format(len(result_avals), len(out_vals)))
     # Handle Python literals, and custom arrays, e.g., tf.Tensor.
-    out_vals = tuple(xla.canonicalize_dtype(np.asarray(a)) for a in out_vals)
+    out_vals = tuple(dtypes.canonicalize_value(np.asarray(a)) for a in out_vals)
     for i, (out_val, out_aval) in enumerate(zip(out_vals, result_avals)):
       if out_val.shape != out_aval.shape:
         raise RuntimeError(
@@ -820,7 +841,7 @@ def emit_python_callback(
         backend, ctx, _wrapped_callback,  token,
         operands, operand_avals, operand_shapes,
         non_empty_result_avals, non_empty_result_shapes,
-        sharding=sharding)
+        returns_token=returns_token, sharding=sharding)
     non_empty_outputs_iter = iter(non_empty_outputs)
     outputs = [
         mlir.ir_constant(np.zeros(result_aval.shape, dtype=result_aval.dtype))
