@@ -34,6 +34,7 @@ from jax._src import monitoring
 from jax._src import path as pathlib
 from jax._src import profiler
 from jax._src import traceback_util
+from jax._src import util
 from jax._src.interpreters import mlir
 from jax._src.lib import xla_client as xc
 from jax._src.lib import _jax
@@ -290,16 +291,6 @@ def backend_compile(
 ) -> xc.Executable:
   sym_name = module.operation.attributes['sym_name']
   module_name = ir.StringAttr(sym_name).value
-  # Convert ir.Module to a string representation, unless the backend
-  # explicitly flags the ability to handle a module directly (avoiding the
-  # overhead of back and forth conversions).
-  # TODO(slebedev): Change the backend.compile() to accept ir.Module.
-  built_c: Any
-  if getattr(backend, "needs_str_ir", True):
-    built_c = mlir.module_to_bytecode(module)
-  else:
-    built_c = module
-
   if (options.executable_build_options.fdo_profile is not None
       and len(options.executable_build_options.fdo_profile)):
     logger.debug(
@@ -309,8 +300,8 @@ def backend_compile(
     )
 
   try:
-    return backend.compile(built_c, executable_devices, options)
-  except xc.XlaRuntimeError as e:
+    return backend.compile(module, executable_devices, options)
+  except _jax.JaxRuntimeError as e:
     for error_handler in _XLA_RUNTIME_ERROR_HANDLERS:
       handler_result = error_handler(e)
       if handler_result is not None:
@@ -328,15 +319,6 @@ def backend_compile_and_load(
 ) -> xc.LoadedExecutable:
   sym_name = module.operation.attributes['sym_name']
   module_name = ir.StringAttr(sym_name).value
-  # Convert ir.Module to a string representation, unless the backend
-  # explicitly flags the ability to handle a module directly (avoiding the
-  # overhead of back and forth conversions).
-  # TODO(slebedev): Change the backend.compile() to accept ir.Module.
-  built_c: Any
-  if getattr(backend, "needs_str_ir", True):
-    built_c = mlir.module_to_bytecode(module)
-  else:
-    built_c = module
 
   if (options.executable_build_options.fdo_profile is not None
       and len(options.executable_build_options.fdo_profile)):
@@ -353,7 +335,7 @@ def backend_compile_and_load(
     if isinstance(backend, _jax.CompileOnlyPyClient):
       if host_callbacks:
         return backend.compile(
-            built_c,
+            module,
             executable_devices=executable_devices,  # type: ignore
             compile_options=options,
             host_callbacks=host_callbacks,  # type: ignore
@@ -361,12 +343,15 @@ def backend_compile_and_load(
       # Some backends don't have `host_callbacks` option yet
       # TODO(sharadmv): remove this fallback when all backends allow `compile`
       # to take in `host_callbacks`
-      return backend.compile(
-          built_c, executable_devices=executable_devices, compile_options=options)  # type: ignore
+      return backend.compile(  # type: ignore
+          module,
+          executable_devices=executable_devices,
+          compile_options=options,
+      )
     else:
       if host_callbacks:
         return backend.compile_and_load(
-            built_c,
+            module,
             executable_devices=executable_devices,
             compile_options=options,
             host_callbacks=host_callbacks,
@@ -375,11 +360,11 @@ def backend_compile_and_load(
       # TODO(sharadmv): remove this fallback when all backends allow `compile`
       # to take in `host_callbacks`
       return backend.compile_and_load(
-          built_c,
+          module,
           executable_devices=executable_devices,
           compile_options=options,
       )
-  except xc.XlaRuntimeError as e:
+  except _jax.JaxRuntimeError as e:
     for error_handler in _XLA_RUNTIME_ERROR_HANDLERS:
       handler_result = error_handler(e)
       if handler_result is not None:
@@ -391,7 +376,7 @@ _XLA_RUNTIME_ERROR_HANDLERS = []
 
 
 def register_xla_runtime_error_handler(
-    handler_fn: Callable[[xc.XlaRuntimeError], Exception | None],
+    handler_fn: Callable[[_jax.JaxRuntimeError], Exception | None],
 ):
   """Registers a custom exception handler for XLA runtime errors.
 
@@ -467,7 +452,8 @@ def compile_or_get_cached(
         "/jax/compilation_cache/cache_retrieval_time_sec", cache_retrieval_time)
 
     return retrieved_executable
-  elif (
+  util.test_event("compile_after_persistent_compilation_miss")
+  if (
       config.share_binary_between_hosts.value
       and is_multi_process
       and distributed.global_state.client is not None
@@ -624,7 +610,7 @@ def _get_cache_key(
         backend,
         ignore_callbacks,
     )
-  except xc._xla.XlaRuntimeError as ex:
+  except _jax.JaxRuntimeError as ex:
     logger.error("compile_or_get_cached: unable to generate cache key, "
                   "skipping the cache: %s", ex)
   return None
@@ -638,11 +624,11 @@ def _share_fdo_profiles(
     backend: xc.Client,
     global_client: lib._jax.DistributedRuntimeClient,
     min_process_id
-) -> bytes | None:
+) -> bytes:
   sym_name = computation.operation.attributes['sym_name']
   module_name = ir.StringAttr(sym_name).value
   fdo_profile = compile_options.executable_build_options.fdo_profile
-  if fdo_profile is None or len(fdo_profile) == 0:
+  if len(fdo_profile) == 0:
     return fdo_profile
 
   compile_options.executable_build_options.fdo_profile = b""
@@ -657,7 +643,7 @@ def _share_fdo_profiles(
         )
         + "_fdo_sync"
     )
-  except xc._xla.XlaRuntimeError as ex:
+  except _jax.JaxRuntimeError as ex:
     logger.error(
         "compile_or_get_cached: unable to generate cache key, "
         "skipping the fdo profile sharing: %s",
