@@ -66,7 +66,7 @@ def _scaled_matmul_impl(a, b, a_scale, b_scale, preferred_element_type):
   )
 
 
-def _scaled_matmul_cuda_lowering(
+def _scaled_matmul_gpu_lowering(
     ctx, a, b, a_scales, b_scales, preferred_element_type
   ):
   lhs_type = ir.RankedTensorType(a.type)
@@ -105,7 +105,6 @@ def _scaled_matmul_cuda_lowering(
 
 
 def _scaled_matmul_abstract(a, b, a_scale, b_scale, *, preferred_element_type):
-  a_dtype = dtypes.canonicalize_dtype(a.dtype)
   batch, non_contracting_lhs, contracting_lhs = a.shape
   _, non_contracting_rhs, _ = b.shape
   output_shape = (batch, non_contracting_lhs, non_contracting_rhs)
@@ -120,9 +119,14 @@ _scaled_matmul_p.def_abstract_eval(_scaled_matmul_abstract)
 
 mlir.register_lowering(
     _scaled_matmul_p,
-    _scaled_matmul_cuda_lowering,
+    _scaled_matmul_gpu_lowering,
     platform="cuda",
 )
+mlir.register_lowering(
+    _scaled_matmul_p,
+    _scaled_matmul_gpu_lowering,
+    platform="rocm",
+)
 
 _scaled_matmul_p_wrapper = core.Primitive("scaled_matmul_wrapper")
 _scaled_matmul_p_wrapper.multiple_results = True
@@ -176,6 +180,12 @@ def _enable_all_reduce(lhs, rhs):
   _, n_spec, rhs_k_spec = rhs.spec
   return lhs_k_spec != None and lhs_k_spec == rhs_k_spec and n_spec == None
 
+def _are_specs_overlapping(lhs, rhs):
+  if lhs is None or rhs is None:
+    return False
+  lhs = (lhs,) if isinstance(lhs, str) else lhs
+  rhs = (rhs,) if isinstance(rhs, str) else rhs
+  return not set(lhs).isdisjoint(rhs)
 
 def _get_output_sharding(shardings):
   lhs, rhs = shardings[0], shardings[1]
@@ -186,7 +196,9 @@ def _get_output_sharding(shardings):
     return [NamedSharding(lhs.mesh, P(*lhs.spec))]
 
   output_specs = (batch_spec, m_spec)
-  output_specs += (n_spec,) if m_spec != n_spec else (None,)
+  # If the m and n specs are overlapping, we cannot keep both -
+  # we (arbitrarily) pick m and replicate for n.
+  output_specs += (None,) if _are_specs_overlapping(m_spec, n_spec) else (n_spec,)
   return [NamedSharding(lhs.mesh, P(*output_specs))]
 
 
@@ -217,53 +229,50 @@ def _get_reduce_scatter_dim(lhs, rhs, output):
   return None
 
 
-def supported_in_sharding(shardings, reduce_scatter_dim):
-  lhs_sharding, rhs_sharding = shardings[0], shardings[1]
+def _supported_in_out_sharding(lhs_sharding, rhs_sharding, out_sharding, reduce_scatter_dim):
   use_all_reduce = _enable_all_reduce(lhs_sharding, rhs_sharding)
 
-  lhs_specs, rhs_specs = list(lhs_sharding.spec), list(rhs_sharding.spec)
+  batch_spec, m_spec, k_spec = lhs_sharding.spec
+  batch_spec_rhs, n_spec, _ = rhs_sharding.spec
 
-  def named_sharding(lhs, rhs, lhs_specs, rhs_specs):
-    lhs_sharding = NamedSharding(lhs.mesh, P(*lhs_specs))
-    rhs_sharding = NamedSharding(rhs.mesh, P(*rhs_specs))
-    return (lhs_sharding, rhs_sharding, lhs_sharding, rhs_sharding)
+  # This is checked by the caller, assert here for documentation.
+  assert batch_spec == batch_spec_rhs
+
+  def named_sharding(lhs_specs, rhs_specs, out_specs):
+    lhs = NamedSharding(lhs_sharding.mesh, P(*lhs_specs))
+    rhs = NamedSharding(rhs_sharding.mesh, P(*rhs_specs))
+    out = NamedSharding(lhs_sharding.mesh, P(*out_specs))
+    return ((lhs, rhs, lhs, rhs), [out])
 
   if reduce_scatter_dim == 1:
-    lhs_specs[1] = None
-    return named_sharding(lhs_sharding, rhs_sharding, lhs_specs, rhs_specs)
+    lhs_specs = (batch_spec, None, k_spec)
+    rhs_specs = (batch_spec, n_spec, k_spec)
+    out_specs = (batch_spec, k_spec, n_spec)
+    return named_sharding(lhs_specs, rhs_specs, out_specs)
 
   if reduce_scatter_dim == 2:
-    rhs_specs[1] = None
-    return named_sharding(lhs_sharding, rhs_sharding, lhs_specs, rhs_specs)
+    lhs_specs = (batch_spec, m_spec, k_spec)
+    rhs_specs = (batch_spec, None, k_spec)
+    out_specs = (batch_spec, m_spec, k_spec)
+    return named_sharding(lhs_specs, rhs_specs, out_specs)
 
-  if use_all_reduce:
-    return named_sharding(lhs_sharding, rhs_sharding, lhs_specs, rhs_specs)
+  if not use_all_reduce:
+    k_spec = None
 
-  lhs_specs[2] = None
-  rhs_specs[2] = None
-  m_spec, n_spec = lhs_specs[1], rhs_specs[1]
-  if m_spec == n_spec:
-    rhs_specs[1] = None
-
-  return named_sharding(lhs_sharding, rhs_sharding, lhs_specs, rhs_specs)
-
-
-def _supported_out_sharding(lhs, rhs, reduce_scatter_dim):
-  batch_spec, m_spec, k_spec = lhs.spec
-  _, n_spec, _ = rhs.spec
-
-  if reduce_scatter_dim == 1:
-    out_m_spec = k_spec
-    out_n_spec = n_spec
-  elif reduce_scatter_dim == 2:
-    out_m_spec = m_spec
-    out_n_spec = k_spec
-  else:
-    out_m_spec = m_spec
-    out_n_spec = n_spec if m_spec != n_spec else None
-
-  return [NamedSharding(lhs.mesh, P(batch_spec, out_m_spec, out_n_spec))]
+  if _are_specs_overlapping(m_spec, n_spec):
+    # We have m and n specs that share an axis, so we can't keep both.
+    # Let us keep the one that was inferred in the output.
+    if n_spec == out_sharding.spec[2]:
+      # Output has n spec, so we get rid of m.
+      m_spec = None
+    else:
+      # Otherwise, we get rid of n.
+      n_spec = None
 
+  lhs_specs = (batch_spec, m_spec, k_spec)
+  rhs_specs = (batch_spec, n_spec, k_spec)
+  out_specs = (batch_spec, m_spec, n_spec)
+  return named_sharding(lhs_specs, rhs_specs, out_specs)
 
 def _scaled_matmul_partition(
     preferred_element_type, mesh, shapes, output_shape
@@ -287,8 +296,7 @@ def _scaled_matmul_partition(
       z = lax_parallel.psum(z, lhs_k_spec)
     return z
 
-  out_shardings = _supported_out_sharding(lhs, rhs, reduce_scatter_dim)
-  arg_shardings = supported_in_sharding(shardings, reduce_scatter_dim)
+  arg_shardings, out_shardings = _supported_in_out_sharding(lhs, rhs, out, reduce_scatter_dim)
   return mesh, _scaled_matmul_impl_partition, out_shardings, arg_shardings
 
 
@@ -299,7 +307,7 @@ _scaled_matmul_lower = custom_partitioning(
 _scaled_matmul_lower.def_partition(
     infer_sharding_from_operands=_scaled_matmul_infer_sharding_from_operands,
     partition=_scaled_matmul_partition,
-    sharding_rule='b m k, b n k, b m x, b n y -> b m n',
+    sharding_rule='b m k, b n k, b m k, b n k -> b m n',
 )
 
 
@@ -352,7 +360,7 @@ batching.primitive_batchers[_scaled_matmul_p_wrapper] = _scaled_matmul_batcher
 batching.primitive_batchers[_scaled_matmul_p] = _scaled_matmul_batcher
 
 
-@partial(api.jit, static_argnames=("preferred_element_type",))
+@api.jit(static_argnames=("preferred_element_type",))
 def _scaled_matmul(
     lhs: Array,
     rhs: Array,
@@ -405,9 +413,8 @@ def scaled_matmul_wrapper(
     assert lhs_K == rhs_K
     _, _, K_block = lhs_scales.shape
 
-    preferred_element_type = dtypes.canonicalize_dtype(
-        np.dtype(preferred_element_type)
-    )
+    preferred_element_type = dtypes.check_and_canonicalize_user_dtype(
+        preferred_element_type, "scaled_matmul_wrapper")
 
     out = _scaled_matmul(
         lhs,
@@ -521,21 +528,26 @@ def quantize(x, config):
   assert contract_dim >= block_size and contract_dim % block_size == 0
   x_new_shape = x_shape[:-1] + (x_shape[-1] // block_size, block_size)
   x = x.reshape(x_new_shape)  # shape = (B, M, K / block_size, block_size)
-
-  amax = jnp.max(jnp.abs(x), axis=-1, keepdims=True)
   MAX = dtypes.finfo(config.data_type).max.astype(x.dtype)
-  scales = amax / MAX  # shape = (B, M, K / block_size, 1)
+
+  def get_scales_per_block(values):
+    # shape = (B, M, K / block_size, 1)
+    return jnp.max(jnp.abs(values), axis=-1, keepdims=True) / MAX
 
   if config.mode == "mxfp8":
+    assert config.global_scale is None
     assert config.scale_type == dtypes.float8_e8m0fnu
-    scales_q = cast_to_e8m0_with_rounding_up(scales)
-    scaled_x = x / e8m0_to_dtype(scales_q, scales.dtype)
+
+    scales_q = cast_to_e8m0_with_rounding_up(get_scales_per_block(x))
+    scaled_x = x / e8m0_to_dtype(scales_q, x.dtype)
   elif config.mode == "nvfp4":
     assert config.scale_type == dtypes.float8_e4m3fn
     assert config.global_scale.dtype == np.float32
+
     SCALE_MAX = dtypes.finfo(config.scale_type).max.astype(x.dtype)
 
-    scales_q = jnp.clip(scales / config.global_scale, 0, SCALE_MAX)
+    x /= config.global_scale
+    scales_q = jnp.clip(get_scales_per_block(x), 0, SCALE_MAX)
     scales_q = lax.optimization_barrier(scales_q.astype(config.scale_type))
     scaled_x = x / scales_q.astype(np.float32)
   else:
@@ -557,9 +569,8 @@ def scaled_dot_impl(lhs, rhs, dimension_numbers, preferred_element_type,
         lhs, rhs, return_weak_type_flag=False
     )
   else:
-    preferred_element_type = dtypes.canonicalize_dtype(
-        np.dtype(preferred_element_type)
-    )
+    preferred_element_type = dtypes.check_and_canonicalize_user_dtype(
+        preferred_element_type, "scaled_dot_impl")
 
   (lhs_contract, rhs_contract), (lhs_batch, rhs_batch) = dimension_numbers
   lhs_dn = (lhs_contract, lhs_batch)
