@@ -18,9 +18,7 @@ from __future__ import annotations
 import atexit
 from collections.abc import Sequence
 import dataclasses
-import enum
 from functools import partial
-import itertools
 import logging
 import threading
 import time
@@ -32,7 +30,8 @@ from jax._src import basearray
 from jax._src import config
 from jax._src import core
 from jax._src import dtypes
-from jax._src import lib
+
+from jax._src import literals
 from jax._src import pjit
 from jax._src import traceback_util
 from jax._src import util
@@ -42,8 +41,8 @@ from jax._src.abstract_arrays import array_types
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
+from jax._src.interpreters import partial_eval
 from jax._src.interpreters import pxla
-from jax._src.interpreters import xla
 from jax._src.api_util import InternalFloatingPointError
 from jax._src.layout import Layout, Format
 from jax._src.lib import xla_client as xc
@@ -52,7 +51,7 @@ from jax._src.monitoring import record_scalar, record_event_duration_secs, recor
 from jax._src.partition_spec import PartitionSpec
 from jax._src.sharding import Sharding
 from jax._src.sharding_impls import (
-    NamedSharding, SingleDeviceSharding, TransferToMemoryKind, GSPMDSharding,
+    NamedSharding, SingleDeviceSharding, GSPMDSharding,
     is_single_device_sharding)
 from jax._src.stages import SourceInfo
 import numpy as np
@@ -68,6 +67,7 @@ xe = xc._xla
 
 Backend = xe.Client
 Device = xc.Device
+ArrayCopySemantics = xc.ArrayCopySemantics
 
 CompileOptions = xc.CompileOptions
 
@@ -86,13 +86,18 @@ def apply_primitive(prim, *args, **params):
   fun = xla_primitive_callable(prim, **params)
   # TODO(yashkatariya): Investigate adding is_primitive to jit and never
   # triggering the disable jit path instead of messing around with it here.
-  prev = lib.jax_jit.swap_thread_local_state_disable_jit(False)
+  prev = config.disable_jit.swap_local(False)
   try:
     outs = fun(*args)
   finally:
-    lib.jax_jit.swap_thread_local_state_disable_jit(prev)
+    config.disable_jit.set_local(prev)
   return outs
 
+# TODO(necula): this cache will contain strong references to all
+# Jaxprs in `params` (for higher-order primitives).
+# This is not immediately fixable by using
+# util.multi_weakref_lru_cache, because the `params` (including the Jaxpr)
+# are closed over in the `prim_fun` lambda. Leaving this fix for a later PR.
 @util.cache()
 def xla_primitive_callable(prim: core.Primitive, **params):
   util.test_event("xla_primitive_callable_cache_miss")
@@ -141,7 +146,11 @@ class RuntimeTokenSet(threading.local):
     # We only use replicated sharding for the first time when the token for the
     # order effect hasn't been created.
     s = GSPMDSharding.get_replicated(devices)
-    sharded_tok = core.Token(pxla.shard_args([s], [None], [None], [tok])[0])
+    sharded_tok = core.Token(
+        pxla.shard_args(
+            [s], [None], [xc.ArrayCopySemantics.REUSE_INPUT], [tok]
+        )[0]
+    )
     self.current_tokens[eff] = sharded_tok
     return sharded_tok
 
@@ -276,53 +285,12 @@ def get_intermediate_shardings(
   return out
 
 
-def jaxpr_has_bints(jaxpr: core.Jaxpr) -> bool:
-  return (any(type(v.aval.dtype) is core.bint for v in jaxpr.invars
-              if isinstance(v.aval, core.UnshapedArray)) or
-          any(_is_bint_axis_size(d)
-              for j in itertools.chain([jaxpr], core.subjaxprs(jaxpr))
-              for e in j.eqns for v in e.outvars
-              if isinstance(v.aval, core.DShapedArray) for d in v.aval.shape))
-
-def _is_bint_axis_size(d: core.AxisSize) -> bool:
-  if isinstance(d, core.DArray):
-    assert not d.shape
-    return type(d.dtype) is core.bint
-  elif isinstance(d, core.Var):
-    return (isinstance(d.aval, core.DShapedArray) and
-            type(d.aval.dtype) is core.bint)
-  return False
-
-
 def check_arg(arg: Any):
   if not (isinstance(arg, core.Tracer) or core.valid_jaxtype(arg)):
     raise TypeError(f"Argument '{arg}' of type {type(arg)} is not a valid "
                     "JAX type.")
 
 
-def jaxpr_replicas(jaxpr: core.Jaxpr) -> int:
-  """The number of replicas needed for a jaxpr.
-
-  For a eqn, multiply the `axis_size` with the `jaxpr_replicas` of the
-  subjaxprs. For a list of eqns, take the maximum number of replicas.
-  """
-  return max(unsafe_map(_eqn_replicas, jaxpr.eqns), default=1)
-
-# TODO(mattjj): this function assumes that only pmap has a parameter named
-# axis_size, and that it corresponds to cross-replica mapping
-def _eqn_replicas(eqn: core.JaxprEqn) -> int:
-  call_jaxpr = eqn.params.get("call_jaxpr")
-  if call_jaxpr:
-    return eqn.params.get('axis_size', 1) * jaxpr_replicas(call_jaxpr)
-  elif eqn.primitive in xla.initial_style_primitives:
-    return _initial_style_primitive_replicas(eqn.params)
-  else:
-    return 1
-
-def _initial_style_primitive_replicas(params: dict[str, Any]) -> int:
-  return max(core.traverse_jaxpr_params(jaxpr_replicas, params).values(),
-             default=1)
-
 def needs_check_special() -> bool:
   return config.debug_infs.value or config.debug_nans.value
 
@@ -331,6 +299,15 @@ def check_special(name: str, bufs: Sequence[basearray.Array]) -> None:
     for buf in bufs:
       _check_special(name, buf.dtype, buf)
 
+
+def check_special_array(name: str, arr: array.ArrayImpl) -> array.ArrayImpl:
+  if needs_check_special():
+    if dtypes.issubdtype(arr.dtype, np.inexact):
+      for buf in arr._arrays:
+        _check_special(name, buf.dtype, buf)
+  return arr
+
+
 def _check_special(name: str, dtype: np.dtype, buf: basearray.Array) -> None:
   if dtypes.issubdtype(dtype, np.inexact):
     if config.debug_nans.value and np.any(np.isnan(np.asarray(buf))):
@@ -338,20 +315,18 @@ def _check_special(name: str, dtype: np.dtype, buf: basearray.Array) -> None:
     if config.debug_infs.value and np.any(np.isinf(np.asarray(buf))):
       raise InternalFloatingPointError(name, "inf")
 
-class CopySemantics(enum.Enum):
-  ALIAS = enum.auto()
-  COPY = enum.auto()
-  DONATE = enum.auto()
-
 def _identity_fn(x):
   return x
 
-def _different_device_order_reshard(x, target_sharding, copy: CopySemantics):
+
+def _different_device_order_reshard(
+    x: array.ArrayImpl, target_sharding: NamedSharding, copy: ArrayCopySemantics
+) -> array.ArrayImpl:
   x._check_if_deleted()
   inp_sharding = x.sharding
   assert isinstance(inp_sharding, NamedSharding)
 
-  donate_argnums = 0 if copy == CopySemantics.DONATE else None
+  donate_argnums = 0 if copy == ArrayCopySemantics.DONATE_INPUT else None
   if inp_sharding._device_assignment == target_sharding._device_assignment:
     return api.jit(_identity_fn, out_shardings=target_sharding,
                    donate_argnums=donate_argnums)(x)
@@ -368,22 +343,14 @@ def _different_device_order_reshard(x, target_sharding, copy: CopySemantics):
       new_mesh, inp_sharding.spec, memory_kind=target_sharding.memory_kind,
       _logical_device_ids=(None if permute_order is None else
                             tuple(permute_order.tolist())))
-  new_x = _reorder_shards(x, new_s, CopySemantics.ALIAS)
+  new_x = xc.reorder_shards(x, new_s, ArrayCopySemantics.REUSE_INPUT)  # type: ignore
   return api.jit(_identity_fn, out_shardings=target_sharding,
                 donate_argnums=donate_argnums)(new_x)
 
 
-def _reorder_shards(x, new_s, copy_semantics: CopySemantics):
-  """Reorders array shards to match the order indicated by the new sharding."""
-  xc_copy_semantics = pxla.to_xc_copy_semantics([copy_semantics])[0]
-  return xc.reorder_shards(x, new_s, xc_copy_semantics)  # type: ignore
-
-
-@util.cache()
+@util.cache(max_size=2048, trace_context_in_key=False)
 def _is_supported_cross_host_transfer(ndim, src_sharding, dst_sharding):
   """Returns True if src->dst is a supported cross-host transfer."""
-  if xla_bridge.process_count() == 1:
-    return False
   if (src_sharding._internal_device_list.device_kind !=
       dst_sharding._internal_device_list.device_kind):
     return False
@@ -423,49 +390,62 @@ class _DeferredShardArg:
   s: Sharding
   aval: core.AbstractValue
   committed: bool
-  copy_semantics: CopySemantics
+  copy_semantics: ArrayCopySemantics
 
   def result_handler(self, shard_arg_result):
     return pxla.global_aval_to_result_handler(
         self.aval, self.s, self.committed)(shard_arg_result)
 
 
-def _device_put_sharding_impl(x, aval, device, copy):
+def _device_put_sharding_impl(
+    x: Any,
+    aval: core.ShapedArray,
+    device: Device | Sharding | None,
+    copy: ArrayCopySemantics,
+):
   from jax.experimental import multihost_utils  # pytype: disable=import-error
 
+  if isinstance(x, array.ArrayImpl):
+    x_is_jax_array = True
+    x_is_fully_addressable, x_sharding = x.is_fully_addressable, x.sharding
+  else:
+    x_is_jax_array = False
+    x_is_fully_addressable, x_sharding = None, None
+
   if isinstance(device, Sharding):
     s = device
+    s_is_fully_addressable = s.is_fully_addressable
     if (getattr(x, 'sharding', None) == s and getattr(x, '_committed', False)
-        and copy == CopySemantics.ALIAS):
+        and copy == ArrayCopySemantics.REUSE_INPUT):
       return x
 
-    if (not s.is_fully_addressable and
-        isinstance(x, array.ArrayImpl) and not x.is_fully_addressable and
-        s.device_set == x.sharding.device_set):
-      assert isinstance(s, Sharding)
+    if (not s_is_fully_addressable and
+        x_is_jax_array and not x_is_fully_addressable and
+        s.device_set == x_sharding.device_set):
+      assert isinstance(s, NamedSharding), s
       return _different_device_order_reshard(x, s, copy)
 
-    if (s.is_fully_addressable and isinstance(x, array.ArrayImpl) and
-        x.is_fully_addressable and s.num_devices > 1 and
-        s._internal_device_list != x.sharding._internal_device_list and  # pytype: disable=attribute-error
-        s.device_set == x.sharding.device_set):
-      assert isinstance(s, Sharding)
+    if (s_is_fully_addressable and x_is_jax_array and
+        x_is_fully_addressable and s.num_devices > 1 and
+        s._internal_device_list != x_sharding._internal_device_list and  # pytype: disable=attribute-error
+        s.device_set == x_sharding.device_set):
+      assert isinstance(s, NamedSharding), s
       return _different_device_order_reshard(x, s, copy)
 
-    if (isinstance(x, array.ArrayImpl) and x._committed
-        and _is_supported_cross_host_transfer(x.ndim, x.sharding, s)):
+    if (x_is_jax_array and x._committed and xla_bridge.process_count() > 1
+        and _is_supported_cross_host_transfer(x.ndim, x_sharding, s)):
       return xc.batched_copy_array_to_devices_with_sharding(
           [x], [s._internal_device_list], [s],  # pytype: disable=attribute-error
-          pxla.to_xc_copy_semantics([copy]))[0]
+          [copy])[0]
 
-    if not s.is_fully_addressable:
+    if not s_is_fully_addressable:
       # If both the source and target shardings are not fully addressable and
       # one of the above conditions has not been met, then assume that the user
       # is attempting a different device order reshard.
-      if (isinstance(x, array.ArrayImpl) and not x.is_fully_addressable
-          and s.device_set != x.sharding.device_set):
-        inp_ids = [d.id for d in x.sharding._device_assignment]
-        inp_plat = x.sharding._device_assignment[0].platform.upper()
+      if (x_is_jax_array and not x_is_fully_addressable
+          and s.device_set != x_sharding.device_set):
+        inp_ids = [d.id for d in x_sharding._device_assignment]
+        inp_plat = x_sharding._device_assignment[0].platform.upper()
         target_ids = [d.id for d in s._device_assignment]
         target_plat = s._device_assignment[0].platform.upper()
         raise ValueError(
@@ -477,8 +457,8 @@ def _device_put_sharding_impl(x, aval, device, copy):
             "different device sets, when input/output shardings have the same "
             "indices and layouts, in the TFRT TPU runtime only.")
 
-      if ((isinstance(x, array.ArrayImpl) and not x._committed) or
-          type(x) in array_types or type(x) in dtypes.python_scalar_dtypes):
+      if ((x_is_jax_array and not x._committed) or
+          type(x) in array_types or type(x) in dtypes.python_scalar_types):
         # If all hosts participate in the sharding, assert that the input is the
         # same on all hosts. If some hosts have no addressable devices in the
         # sharding, bypass the check, since we can't easily distinguish between
@@ -502,19 +482,19 @@ def _device_put_sharding_impl(x, aval, device, copy):
     return _DeferredShardArg(x, s, aval, True, copy)
 
   # Only `Device` exists below. `Sharding` instance is handled above.
-  if isinstance(x, array.ArrayImpl):
-    if not x.is_fully_addressable:
+  if x_is_jax_array:
+    if not x_is_fully_addressable:
       raise ValueError(
           "device_put's first argument must be a fully addressable array, but "
           f"got value with devices {x.devices()}")
     if device is None:
-      if copy == CopySemantics.ALIAS:
+      if copy == ArrayCopySemantics.REUSE_INPUT:
         return x
       else:
-        return _DeferredShardArg(x, x.sharding, aval, x.committed, copy)
-    elif is_single_device_sharding(x.sharding):
-      device = x.sharding._device_assignment[0] if device is None else device
-      if copy == CopySemantics.COPY:
+        return _DeferredShardArg(x, x_sharding, aval, x.committed, copy)
+    elif is_single_device_sharding(x_sharding):
+      device = x_sharding._device_assignment[0] if device is None else device
+      if copy == ArrayCopySemantics.ALWAYS_COPY:
         return xc.batched_device_put(aval, SingleDeviceSharding(device), [x],
                                      [device], True, True)
       return pxla.batched_device_put(aval, SingleDeviceSharding(device), [x],
@@ -527,19 +507,18 @@ def _device_put_sharding_impl(x, aval, device, copy):
 
 def _device_put_impl(
     x, *, device: Device | Sharding | Format | None,
-    src: Device | Sharding | Format | None, copy: CopySemantics):
-  if (isinstance(device, TransferToMemoryKind) or
-      isinstance(src, TransferToMemoryKind)):
-    raise ValueError(
-        "TransferToMemoryKind argument to jax.device_put can only be used"
-        " inside jax.jit. If you are using device_put outside jax.jit, then"
-        " please provide a concrete Sharding with memory_kind.")
+    src: Device | Sharding | Format | None, copy: ArrayCopySemantics, aval):
+  if aval is None:
+    try:
+      aval = core.abstractify(x)
+      aval = update_dp_aval(aval, device)
+    except TypeError as err:
+      raise TypeError(
+          f"Argument '{x}' of type {type(x)} is not a valid JAX type") from err
 
-  try:
-    aval = core.abstractify(x)
-  except TypeError as err:
-    raise TypeError(
-        f"Argument '{x}' of type {type(x)} is not a valid JAX type") from err
+  if isinstance(device, core.MemorySpace):
+    return apply_primitive(device_put_p, x, devices=(device,), srcs=(src,),
+                           copy_semantics=(copy,))[0]
 
   if isinstance(device, Format):
     l = device
@@ -553,13 +532,15 @@ def _device_put_impl(
           "sharding and layout in `Layout` instance should be"
           f" concrete. Got layout: {l} for input {aval.str_short()}")
     if (getattr(x, 'format', None) == l and getattr(x, '_committed', False) and
-        copy == CopySemantics.ALIAS):
+        copy == ArrayCopySemantics.REUSE_INPUT):
       return x
     if x_dll is None and dll is None:
       return _device_put_sharding_impl(x, aval, l.sharding, copy)
     return api.jit(
-        _identity_fn, out_shardings=l,
-        donate_argnums=(0 if copy == CopySemantics.DONATE else None))(x)
+        _identity_fn,
+        out_shardings=l,
+        donate_argnums=(0 if copy == ArrayCopySemantics.DONATE_INPUT else None),
+    )(x)
 
   return _device_put_sharding_impl(x, aval, device, copy)
 
@@ -568,11 +549,13 @@ def _batched_device_put_impl(
     *xs,
     devices: Sequence[Device | Sharding | Format | None],
     srcs: Sequence[Device | Sharding | Format | None],
-    copy_semantics: Sequence[CopySemantics]):
+    copy_semantics: Sequence[ArrayCopySemantics],
+    dst_avals: Sequence[core.ShapedArray | None]):
   ys = []
   dsa_indices, dsa_xs, dsa_shardings, dsa_copy_semantics = [], [], [], []
-  for i, (x, device, src, cp) in enumerate(zip(xs, devices, srcs, copy_semantics)):
-    y = _device_put_impl(x, device=device, src=src, copy=cp)
+  for i, (x, device, src, cp, aval) in enumerate(
+      zip(xs, devices, srcs, copy_semantics, dst_avals)):
+    y = _device_put_impl(x, device=device, src=src, copy=cp, aval=aval)
     if isinstance(y, _DeferredShardArg):
       dsa_indices.append(i)
       dsa_xs.append(y.x)
@@ -590,16 +573,51 @@ def _batched_device_put_impl(
     for i, shard_arg_result in zip(dsa_indices, shard_arg_results):
       assert isinstance(ys[i], _DeferredShardArg)
       ys[i] = ys[i].result_handler(shard_arg_result)
-
   return ys
 
+def batched_device_put_impl(
+    *xs,
+    devices: Sequence[Device | Sharding | Format | None],
+    srcs: Sequence[Device | Sharding | Format | None],
+    copy_semantics: Sequence[ArrayCopySemantics]):
+  return _batched_device_put_impl(
+      *xs, devices=devices, srcs=srcs, copy_semantics=copy_semantics,
+      dst_avals=[None] * len(devices))
+
 
 device_put_p = core.Primitive('device_put')
 device_put_p.multiple_results = True
-device_put_p.def_impl(_batched_device_put_impl)
+device_put_p.def_impl(batched_device_put_impl)
+
+
+def _device_put_folding_rule(consts, params, out_avals):
+  # We elide device_puts that do nothing; these can be generated by jnp.array,
+  # for example.
+  if (all(x is None for x in params["devices"])
+      and all(isinstance(x, literals.TypedNdArray) for x in consts)
+      and all(x == ArrayCopySemantics.REUSE_INPUT for x in params["copy_semantics"])):
+    return consts
+  return None
+
+partial_eval.const_fold_rules[device_put_p] = _device_put_folding_rule
+
+
+def update_dp_aval(aval, d):
+  if not isinstance(aval, core.ShapedArray):
+    return aval
+  if isinstance(d, Sharding):
+    aval = (aval.update(sharding=aval.sharding.update(mesh=d.mesh.abstract_mesh,
+                                                      spec=d.spec))
+            if isinstance(d, NamedSharding) else aval.update(sharding=None))
+    if d.memory_kind is not None:
+      aval = aval.update(memory_space=core.mem_kind_to_space(d.memory_kind))
+    return aval
+  elif isinstance(d, core.MemorySpace):
+    return aval.update(memory_space=d)
+  return aval
 
 def _device_put_abstract_eval(*xs, devices, srcs, copy_semantics):
-  return xs
+  return [update_dp_aval(x, d) for x, d in zip(xs, devices)]
 device_put_p.def_abstract_eval(_device_put_abstract_eval)
 
 def _device_put_transpose(cts, *_, devices, srcs, copy_semantics):
@@ -612,15 +630,15 @@ def _device_put_transpose(cts, *_, devices, srcs, copy_semantics):
     indices, args, devices, srcs, copy_semantics = list(zip(*dp_args))
     new_copy_semantics = []
     for cp in copy_semantics:
-      if cp == CopySemantics.DONATE:
+      if cp == ArrayCopySemantics.DONATE_INPUT:
         raise ValueError(
             "donate=True is not allowed during tranposition of device_put."
             " Please file an issue if you want this to be supported.")
-      elif cp == CopySemantics.ALIAS:
-        new_copy_semantics.append(CopySemantics.COPY)
+      elif cp == ArrayCopySemantics.REUSE_INPUT:
+        new_copy_semantics.append(ArrayCopySemantics.ALWAYS_COPY)
       else:
-        assert cp == CopySemantics.COPY
-        new_copy_semantics.append(CopySemantics.COPY)
+        assert cp == ArrayCopySemantics.ALWAYS_COPY
+        new_copy_semantics.append(ArrayCopySemantics.ALWAYS_COPY)
     ys = device_put_p.bind(*args, devices=srcs, srcs=devices,
                            copy_semantics=tuple(new_copy_semantics))
     for i, y in zip(indices, ys):
@@ -643,8 +661,8 @@ def _tpu_gpu_device_put_lowering(ctx, *xs, devices, srcs, copy_semantics):
   if ctx.module_context.all_default_mem_kind:
     return xs
   def lower(x, device, aval, out_aval):
-    if (isinstance(device, (Sharding, TransferToMemoryKind)) and
-        device.memory_kind is not None):
+    if ((isinstance(device, Sharding) and device.memory_kind is not None) or
+        isinstance(device, core.MemorySpace)):
       if isinstance(device, Sharding):
         if config.use_shardy_partitioner.value:
           x = mlir.wrap_with_sharding_op(
@@ -654,7 +672,9 @@ def _tpu_gpu_device_put_lowering(ctx, *xs, devices, srcs, copy_semantics):
           x = mlir.wrap_with_sharding_op(
               ctx, x, out_aval,
               device._to_xla_hlo_sharding(aval.ndim).to_proto())
-      x = mlir.wrap_with_memory_kind(x, device.memory_kind, out_aval)
+      mem_kind = (core.mem_space_to_kind(device)
+                  if isinstance(device, core.MemorySpace) else device.memory_kind)
+      x = mlir.wrap_with_memory_kind(x, mem_kind, out_aval)
       return x
     return x
   return list(map(lower, xs, devices, ctx.avals_in, ctx.avals_out))
@@ -668,13 +688,3 @@ mlir.register_lowering(
 def _common_device_put_lowering(ctx, *xs, devices, srcs, copy_semantics):
   return xs
 mlir.register_lowering(device_put_p, _common_device_put_lowering)
-
-def _propagate_mem_kind_dp(*xm, devices, srcs, copy_semantics):
-  memory_kinds = []
-  for device in devices:
-    if isinstance(device, (Sharding, TransferToMemoryKind)):
-      memory_kinds.append(device.memory_kind)
-    else:
-      memory_kinds.append(None)
-  return memory_kinds
-pxla.memory_kind_propagate_rule[device_put_p] = _propagate_mem_kind_dp
