@@ -17,10 +17,11 @@ from __future__ import annotations
 from typing import Any
 
 from jax._src import array
-from jax._src import deprecations
+from jax._src import dtypes
 from jax._src import xla_bridge
 from jax._src.api import device_put
 from jax._src.lax.lax import _array_copy
+from jax._src.lib import _jax
 from jax._src.lib import xla_client
 from jax._src.numpy import lax_numpy as jnp
 from jax._src.numpy import scalar_types as jnp_types
@@ -59,8 +60,8 @@ def is_supported_dtype(dtype: DTypeLike) -> bool:
 
 
 def _to_dlpack(x: Array, stream: int | Any | None,
-               src_device: xla_client.Device | None = None,
-               device: xla_client.Device | None = None,
+               src_device: _jax.Device | None = None,
+               device: _jax.Device | None = None,
                copy: bool | None = None):
 
   if src_device is None:
@@ -76,7 +77,7 @@ def _to_dlpack(x: Array, stream: int | Any | None,
       arr = device_put(x, device)
   else:
     arr = _array_copy(x) if copy else x
-  return xla_client._xla.buffer_to_dlpack_managed_tensor(
+  return _jax.buffer_to_dlpack_managed_tensor(
     arr.addressable_data(0), stream=stream
   )
 
@@ -89,7 +90,7 @@ _DL_DEVICE_TO_PLATFORM = {
 
 
 def to_dlpack(x: Array, stream: int | Any | None = None,
-              src_device: xla_client.Device | None = None,
+              src_device: _jax.Device | None = None,
               dl_device: tuple[DLDeviceType, int] | None = None,
               max_version: tuple[int, int] | None = None,
               copy : bool | None = None):
@@ -170,7 +171,7 @@ def to_dlpack(x: Array, stream: int | Any | None = None,
       f"version ({max_version}) was requested."
     )
 
-def _place_array(_arr, device, dlpack_device, copy):
+def _check_device(device, dlpack_device, copy):
   if device and dlpack_device != device:
     if copy is not None and not copy:
       raise ValueError(
@@ -178,75 +179,23 @@ def _place_array(_arr, device, dlpack_device, copy):
         f"is {repr(dlpack_device)}, however copy=False. Set copy=True or "
         "copy=None to perform the requested operation."
       )
-    else:
-      return device_put(_arr, device)
+
+def _place_array(_arr, device, dlpack_device, copy):
+  if device and dlpack_device != device:
+    return device_put(_arr, device)
   if copy:
     return jnp.array(_arr, copy=True)
   return _arr
 
-def _legacy_from_dlpack(dlpack, device: xla_client.Device | None = None,
-                        copy: bool | None = None):
-  preferred_platform = getattr(device, "platform", None)
-  if device and preferred_platform == "gpu":
-    preferred_platform = "cuda" if "cuda" in device.client.platform_version else "rocm"
-
-  cpu_backend = xla_bridge.get_backend("cpu")
-  gpu_backend = None
-
-  if preferred_platform in {"cuda", "rocm"}:
-    try:
-      gpu_backend = xla_bridge.get_backend(preferred_platform)
-    except RuntimeError:
-      raise TypeError(
-        f"A {str.upper(preferred_platform)} device was specified, however no "
-        f"{str.upper(preferred_platform)} backend was found."
-      )
-
-  if preferred_platform is None:
-    try:
-      gpu_backend = xla_bridge.get_backend("cuda")
-    except RuntimeError:
-      pass
-    # Try ROCm if CUDA backend not found
-    if gpu_backend is None:
-      try:
-        gpu_backend = xla_bridge.get_backend("rocm")
-      except RuntimeError:
-        pass
-
-  _arr = jnp.asarray(xla_client._xla.dlpack_managed_tensor_to_buffer(
-      dlpack, cpu_backend, gpu_backend))
-  dlpack_device, = _arr.devices()
-  return _place_array(_arr, device, dlpack_device, copy)
-
-def _from_dlpack(external_array, device: xla_client.Device | None = None,
-                 copy: bool | None = None):
-  dl_device_type, device_id = external_array.__dlpack_device__()
-  try:
-    dl_device_platform = _DL_DEVICE_TO_PLATFORM[dl_device_type]
-  except KeyError:
-    raise TypeError(
-        "Array passed to from_dlpack is on unsupported device type "
-        f"(DLDeviceType: {dl_device_type}, array: {external_array}"
-    ) from None
-
-  backend = xla_bridge.get_backend(dl_device_platform)
-  dlpack_device = backend.device_from_local_hardware_id(device_id)
-  try:
-    stream = dlpack_device.get_stream_for_external_ready_events()
-  except xla_client.XlaRuntimeError as err:
-    if "UNIMPLEMENTED" in str(err):
-      stream = None
-    else:
-      raise
-  dlpack = external_array.__dlpack__(stream=stream)
-
-  _arr = jnp.asarray(xla_client._xla.dlpack_managed_tensor_to_buffer(
-      dlpack, dlpack_device, stream))
-  return _place_array(_arr, device, dlpack_device, copy)
+def _is_tensorflow_tensor(external_array):
+  t = type(external_array)
+  return (
+      t.__qualname__ == "EagerTensor"
+      and t.__module__.endswith("tensorflow.python.framework.ops")
+  )
 
 def from_dlpack(external_array,
-                device: xla_client.Device | Sharding | None = None,
+                device: _jax.Device | Sharding | None = None,
                 copy: bool | None = None):
   """Returns a :class:`~jax.Array` representation of a DLPack tensor.
 
@@ -286,18 +235,54 @@ def from_dlpack(external_array,
         f"a Sharding with {len(device_set)} devices was provided."
       )
     device, = device_set
-  if hasattr(external_array, "__dlpack__"):
-    return _from_dlpack(external_array, device, copy)
+  if not hasattr(external_array, "__dlpack__") or not hasattr(external_array, "__dlpack_device__"):
+    raise TypeError(
+        "The array passed to from_dlpack must have __dlpack__ and __dlpack_device__ methods."
+    )
 
-  # Deprecated legacy path.
-  # TODO(slebedev): Remove on or after December 3rd 2023.
-  deprecations.warn(
-      "jax-dlpack-import-legacy",
-      (
-          "Calling from_dlpack with a DLPack tensor is deprecated. The argument"
-          " to from_dlpack should be an array from another framework that"
-          " implements the __dlpack__ protocol."
-      ),
-      stacklevel=2,
-  )
-  return _legacy_from_dlpack(external_array, device, copy)
+  dl_device_type, device_id = external_array.__dlpack_device__()
+  try:
+    dl_device_platform = _DL_DEVICE_TO_PLATFORM[dl_device_type]
+  except KeyError:
+    raise TypeError(
+        "Array passed to from_dlpack is on unsupported device type "
+        f"(DLDeviceType: {dl_device_type}, array: {external_array}"
+    ) from None
+
+  backend = xla_bridge.get_backend(dl_device_platform)
+  dlpack_device = backend.device_from_local_hardware_id(device_id)
+  _check_device(device, dlpack_device, copy)
+  if _is_tensorflow_tensor(external_array):
+    # TensorFlow does not support stream=.
+    stream = None
+  else:
+    try:
+      stream = dlpack_device.get_stream_for_external_ready_events()
+    except _jax.JaxRuntimeError as err:
+      if "UNIMPLEMENTED" in str(err):
+        stream = None
+      else:
+        raise
+  dlpack = external_array.__dlpack__(stream=stream)
+
+  try:
+    arr = _jax.dlpack_managed_tensor_to_buffer(
+      dlpack, dlpack_device, stream, copy)
+  except xla_client.XlaRuntimeError as e:
+    se = str(e)
+    if "is not aligned to" in se:
+      i = se.index("is not aligned to")
+      raise ValueError(
+        "Specified input which requires a copy since the source data "
+        f"buffer {se[i:]} However copy=False. Set copy=True or "
+        "copy=None to perform the requested operation."
+      )
+    else:
+      raise
+  # TODO(phawkins): when we are ready to support x64 arrays in
+  # non-x64 mode, change the semantics to not canonicalize here.
+  arr = jnp.asarray(arr, dtype=dtypes.canonicalize_dtype(arr.dtype))
+  if copy:
+    # copy was already handled by dlpack_managed_tensor_to_buffer.
+    copy = None
+  return _place_array(arr, device, dlpack_device, copy)
