@@ -16,9 +16,10 @@
 
 from __future__ import annotations
 
-from typing import Callable, TypeVar
-from collections.abc import Sequence
+import types
+from collections.abc import Callable, Sequence
 from functools import partial
+from typing import Any, TypeVar
 
 try:
   import flatbuffers
@@ -45,10 +46,17 @@ SerT = TypeVar("SerT")
 # even if the change is backwards compatible.
 # Version 1, Nov 2023, first version.
 # Version 2, Dec 16th, 2023, adds the f0 dtype.
-_SERIALIZATION_VERSION = 2
+# Version 3, October 16th, 2024, adds serialization for namedtuple and custom types
+#   This version is backwards compatible with Version 2.
+# Version 4, April 7th, 2025, adds serialization for PRNGs key types.
+#   This version is backwards compatible with Version 2 and 3.
+# Version 5, November 23rd, 2025, adds serialization for aval memory_space,
+#   upgrade num_devices to a 32 bit value.
+#   This version is backwards compatible with Version 2 to 4.
+_SERIALIZATION_VERSION = 5
 
 def serialize(exp: _export.Exported, vjp_order: int = 0) -> bytearray:
-  """Serialize an Exported.
+  """Serializes an Exported.
 
   Args:
     exp: the Exported to serialize.
@@ -64,7 +72,7 @@ def serialize(exp: _export.Exported, vjp_order: int = 0) -> bytearray:
 
 
 def deserialize(ser: bytearray) -> _export.Exported:
-  """Deserialize an Exported."""
+  """Deserializes an Exported."""
   exp = ser_flatbuf.Exported.GetRootAsExported(ser)
   return _deserialize_exported(exp)
 
@@ -93,8 +101,8 @@ def _serialize_exported(
   disabled_safety_checks = _serialize_array(
       builder, _serialize_disabled_safety_check, exp.disabled_safety_checks
   )
-  lowering_platforms = _serialize_array(
-      builder, lambda b, p: b.CreateString(p), exp.lowering_platforms
+  platforms = _serialize_array(
+      builder, lambda b, p: b.CreateString(p), exp.platforms
   )
   mlir_module_serialized = builder.CreateByteVector(exp.mlir_module_serialized)
   module_kept_var_idx = builder.CreateNumpyVector(
@@ -112,26 +120,32 @@ def _serialize_exported(
     vjp = _serialize_exported(builder, exp.vjp(), vjp_order - 1)
 
   ser_flatbuf.ExportedStart(builder)
-  ser_flatbuf.ExportedAddSerializationVersion(builder, _SERIALIZATION_VERSION)
+  # TODO(necula): we cannot really store the actual serialization_version
+  # in the flatbuffer because prior to 11/25/2025 deserializers checked
+  # if the version is 2 or 3. I have now removed that check, but for the
+  # sake of old deserializers we can only store version 3. Starting
+  # on January 2026 we can store the actual version.
+  ser_flatbuf.ExportedAddSerializationVersion(builder, 3)
   ser_flatbuf.ExportedAddFunctionName(builder, fun_name)
   ser_flatbuf.ExportedAddInTree(builder, in_tree)
   ser_flatbuf.ExportedAddInAvals(builder, in_avals)
   ser_flatbuf.ExportedAddOutTree(builder, out_tree)
   ser_flatbuf.ExportedAddOutAvals(builder, out_avals)
   ser_flatbuf.ExportedAddNrDevices(builder, exp.nr_devices)
+  ser_flatbuf.ExportedAddNrDevicesShort(builder, exp.nr_devices)  # For forward compatibility, can remove after January 2026
   ser_flatbuf.ExportedAddInShardings(builder, in_shardings)
   ser_flatbuf.ExportedAddOutShardings(builder, out_shardings)
-  ser_flatbuf.ExportedAddLoweringPlatforms(builder, lowering_platforms)
+  ser_flatbuf.ExportedAddPlatforms(builder, platforms)
   ser_flatbuf.ExportedAddOrderedEffects(builder, ordered_effects)
   ser_flatbuf.ExportedAddUnorderedEffects(builder, unordered_effects)
   ser_flatbuf.ExportedAddDisabledChecks(builder, disabled_safety_checks)
   ser_flatbuf.ExportedAddMlirModuleSerialized(builder, mlir_module_serialized)
-  ser_flatbuf.ExportedAddMlirModuleSerializationVersion(
-      builder, exp.mlir_module_serialization_version
+  ser_flatbuf.ExportedAddCallingConventionVersion(
+      builder, exp.calling_convention_version
   )
   ser_flatbuf.ExportedAddModuleKeptVarIdx(builder, module_kept_var_idx)
-  ser_flatbuf.ExportedAddUsesShapePolymorphism(
-      builder, exp.uses_shape_polymorphism
+  ser_flatbuf.ExportedAddUsesGlobalConstants(
+      builder, exp.uses_global_constants
   )
   if vjp is not None:
     ser_flatbuf.ExportedAddVjp(builder, vjp)
@@ -152,13 +166,9 @@ def _serialize_array(
 
 def _deserialize_exported(exp: ser_flatbuf.Exported) -> _export.Exported:
   serialization_version = exp.SerializationVersion()
-  if serialization_version != _SERIALIZATION_VERSION:
-    raise NotImplementedError(
-        f"deserialize unsupported version {serialization_version}"
-    )
 
   fun_name = exp.FunctionName().decode("utf-8")
-  _, in_tree = tree_util.tree_flatten(
+  in_tree = tree_util.tree_structure(
       _deserialize_pytreedef_to_pytree(exp.InTree())
   )
   scope = shape_poly.SymbolicScope(())  # TODO: serialize the constraints
@@ -166,22 +176,25 @@ def _deserialize_exported(exp: ser_flatbuf.Exported) -> _export.Exported:
   in_avals = _deserialize_tuple(
       exp.InAvalsLength, exp.InAvals, deser_aval
   )
-  _, out_tree = tree_util.tree_flatten(
+  out_tree = tree_util.tree_structure(
       _deserialize_pytreedef_to_pytree(exp.OutTree())
   )
   out_avals = _deserialize_tuple(
       exp.OutAvalsLength, exp.OutAvals, deser_aval
   )
-  nr_devices = exp.NrDevices()
+  # TODO(necula): remove the fallback to NrDevicesShort and mark
+  # the field "deprecated" once we abandon the old
+  # serialization format (6 months after 11/24/2025).
+  nr_devices = exp.NrDevices() or exp.NrDevicesShort()
   in_shardings = _deserialize_tuple(
       exp.InShardingsLength, exp.InShardings, _deserialize_sharding
   )
   out_shardings = _deserialize_tuple(
       exp.OutShardingsLength, exp.OutShardings, _deserialize_sharding
   )
-  lowering_platforms = _deserialize_tuple(
-      exp.LoweringPlatformsLength,
-      exp.LoweringPlatforms,
+  platforms = _deserialize_tuple(
+      exp.PlatformsLength,
+      exp.Platforms,
       lambda v: v.decode("utf-8"),
   )
   ordered_effects = _deserialize_tuple(
@@ -197,9 +210,9 @@ def _deserialize_exported(exp: ser_flatbuf.Exported) -> _export.Exported:
   )
 
   mlir_module_serialized = exp.MlirModuleSerializedAsNumpy().tobytes()
-  mlir_module_serialization_version = exp.MlirModuleSerializationVersion()
+  calling_convention_version = exp.CallingConventionVersion()
   module_kept_var_idx = tuple(exp.ModuleKeptVarIdxAsNumpy().tolist())
-  uses_shape_polymorphism = exp.UsesShapePolymorphism()
+  uses_global_constants = exp.UsesGlobalConstants()
 
   _get_vjp = None
   if vjp := exp.Vjp():
@@ -214,14 +227,14 @@ def _deserialize_exported(exp: ser_flatbuf.Exported) -> _export.Exported:
       nr_devices=nr_devices,
       in_shardings_hlo=in_shardings,
       out_shardings_hlo=out_shardings,
-      lowering_platforms=lowering_platforms,
+      platforms=platforms,
       ordered_effects=ordered_effects,
       unordered_effects=unordered_effects,
       disabled_safety_checks=disabled_safety_checks,
       mlir_module_serialized=mlir_module_serialized,
-      mlir_module_serialization_version=mlir_module_serialization_version,
+      calling_convention_version=calling_convention_version,
       module_kept_var_idx=module_kept_var_idx,
-      uses_shape_polymorphism=uses_shape_polymorphism,
+      uses_global_constants=uses_global_constants,
       _get_vjp=_get_vjp,
   )
 
@@ -246,23 +259,46 @@ def _serialize_pytreedef(
     children_vector_offset = _serialize_array(
         builder, _serialize_pytreedef, children
     )
+  custom_name = None
+  custom_auxdata = None
+  node_type = node_data and node_data[0]
 
   if node_data is None:  # leaf
     kind = ser_flatbuf.PyTreeDefKind.leaf
-  elif node_data[0] is type(None):
+  elif node_type is types.NoneType:
     kind = ser_flatbuf.PyTreeDefKind.none
-  elif node_data[0] is tuple:
+  elif node_type is tuple:
     kind = ser_flatbuf.PyTreeDefKind.tuple
-  elif node_data[0] is list:
+  elif node_type is list:
     kind = ser_flatbuf.PyTreeDefKind.list
-  elif node_data[0] is dict:
+  elif node_type is dict:
     kind = ser_flatbuf.PyTreeDefKind.dict
     assert len(node_data[1]) == len(children)
+    def serialize_key(builder, k):
+      if not isinstance(k, str):
+        raise TypeError(
+            "Serialization is supported only for dictionaries with string keys."
+            f" Found key {k} of type {type(k)}.")
+      return builder.CreateString(k)
     children_names_vector_offset = _serialize_array(
-        builder, lambda b, s: b.CreateString(s), node_data[1]
+        builder, serialize_key, node_data[1]
     )
+  elif node_type in _export.serialization_registry:
+    kind = ser_flatbuf.PyTreeDefKind.custom
+    serialized_name, serialize_auxdata = _export.serialization_registry[node_type]
+    custom_name = builder.CreateString(serialized_name)
+    serialized_auxdata = serialize_auxdata(node_data[1])
+    if not isinstance(serialized_auxdata, (bytes, bytearray)):
+      raise ValueError(
+          "The custom serialization function for `node_type` must "
+          f"return a `bytes` object. It returned a {type(serialized_auxdata)}.")
+    custom_auxdata = builder.CreateByteVector(serialized_auxdata)
   else:
-    raise NotImplementedError(f"serializing PyTreeDef {node_data}")
+    raise ValueError(
+        "Cannot serialize PyTreeDef containing an "
+        f"unregistered type `{node_type}`. "
+        "Use `export.register_pytree_node_serialization` or "
+        "`export.register_namedtuple_serialization`.")
 
   ser_flatbuf.PyTreeDefStart(builder)
   ser_flatbuf.PyTreeDefAddKind(builder, kind)
@@ -270,6 +306,10 @@ def _serialize_pytreedef(
     ser_flatbuf.PyTreeDefAddChildren(builder, children_vector_offset)
   if children_names_vector_offset:
     ser_flatbuf.PyTreeDefAddChildrenNames(builder, children_names_vector_offset)
+  if custom_name is not None:
+    ser_flatbuf.PyTreeDefAddCustomName(builder, custom_name)
+  if custom_auxdata is not None:
+    ser_flatbuf.PyTreeDefAddCustomAuxdata(builder, custom_auxdata)
   return ser_flatbuf.PyTreeDefEnd(builder)
 
 
@@ -294,6 +334,17 @@ def _deserialize_pytreedef_to_pytree(p: ser_flatbuf.PyTreeDef):
     assert p.ChildrenNamesLength() == nr_children
     keys = [p.ChildrenNames(i).decode("utf-8") for i in range(nr_children)]
     return dict(zip(keys, children))
+  elif kind == ser_flatbuf.PyTreeDefKind.custom:
+    serialized_name = p.CustomName().decode("utf-8")
+    if serialized_name not in _export.deserialization_registry:
+      raise ValueError(
+          "Cannot deserialize a PyTreeDef containing an "
+          f"unregistered type `{serialized_name}`. "
+          "Use `export.register_pytree_node_serialization` or "
+          "`export.register_namedtuple_serialization`.")
+    nodetype, deserialize_auxdata, from_iter = _export.deserialization_registry[serialized_name]
+    auxdata = deserialize_auxdata(p.CustomAuxdataAsNumpy().tobytes())
+    return from_iter(auxdata, children)
   else:
     assert False, kind
 
@@ -322,37 +373,50 @@ _dtype_to_dtype_kind = {
     dtypes._float8_e4m3fnuz_dtype: ser_flatbuf.DType.f8_e4m3fnuz,
     dtypes._float8_e5m2_dtype: ser_flatbuf.DType.f8_e5m2,
     dtypes._float8_e5m2fnuz_dtype: ser_flatbuf.DType.f8_e5m2fnuz,
+    dtypes._float8_e3m4_dtype: ser_flatbuf.DType.f8_e3m4,
+    dtypes._float8_e4m3_dtype: ser_flatbuf.DType.f8_e4m3,
+    dtypes._float8_e8m0fnu_dtype: ser_flatbuf.DType.f8_e8m0fnu,
+    dtypes._float4_e2m1fn_dtype: ser_flatbuf.DType.f4_e2m1fn,
 }
 
-
 _dtype_kind_to_dtype = {
     kind: dtype for dtype, kind in _dtype_to_dtype_kind.items()
 }
 
 
-def _serialize_aval(
-    builder: flatbuffers.Builder, aval: core.AbstractValue
-) -> int:
-  aval_type = type(aval)
-  if aval_type is core.ShapedArray:
-    aval_kind = ser_flatbuf.AbstractValueKind.shapedArray
-    shape_offsets = [builder.CreateString(str(d)) for d in aval.shape]
-    ser_flatbuf.AbstractValueStartShapeVector(builder, len(aval.shape))
-    for d in reversed(shape_offsets):
-      builder.PrependUOffsetTRelative(d)
-    shape_vector_offset = builder.EndVector()
+def register_dtype_kind(dtype: Any, kind: int):
+  _dtype_to_dtype_kind[dtype] = kind
+  _dtype_kind_to_dtype[kind] = dtype
 
-    ser_flatbuf.AbstractValueStart(builder)
-    ser_flatbuf.AbstractValueAddKind(builder, aval_kind)
-    ser_flatbuf.AbstractValueAddShape(builder, shape_vector_offset)
-    ser_flatbuf.AbstractValueAddDtype(builder, _dtype_to_dtype_kind[aval.dtype])
-    return ser_flatbuf.AbstractValueEnd(builder)
-  else:
-    raise NotImplementedError(f"serializing AbstractValue: {aval}")
+
+_memory_space_to_enum = {
+    core.MemorySpace.Device: ser_flatbuf.MemorySpace.Device,
+    core.MemorySpace.Host: ser_flatbuf.MemorySpace.Host,
+    core.MemorySpace.Any: ser_flatbuf.MemorySpace.Any,
+}
+_memory_space_from_enum = {v: k for k, v in _memory_space_to_enum.items()}
+
+
+def _serialize_aval(
+    builder: flatbuffers.Builder, aval: core.ShapedArray
+) -> int:
+  aval_kind = ser_flatbuf.AbstractValueKind.shapedArray
+  shape_offsets = [builder.CreateString(str(d)) for d in aval.shape]
+  ser_flatbuf.AbstractValueStartShapeVector(builder, len(aval.shape))
+  for d in reversed(shape_offsets):
+    builder.PrependUOffsetTRelative(d)
+  shape_vector_offset = builder.EndVector()
+
+  ser_flatbuf.AbstractValueStart(builder)
+  ser_flatbuf.AbstractValueAddKind(builder, aval_kind)
+  ser_flatbuf.AbstractValueAddShape(builder, shape_vector_offset)
+  ser_flatbuf.AbstractValueAddDtype(builder, _dtype_to_dtype_kind[aval.dtype])
+  ser_flatbuf.AbstractValueAddMemorySpace(builder, _memory_space_to_enum[aval.memory_space])
+  return ser_flatbuf.AbstractValueEnd(builder)
 
 
 def _deserialize_aval(aval: ser_flatbuf.AbstractValue,
-                      scope) -> core.AbstractValue:
+                      scope) -> core.ShapedArray:
   aval_kind = aval.Kind()
   if aval_kind == ser_flatbuf.AbstractValueKind.shapedArray:
     dtype = _dtype_kind_to_dtype[aval.Dtype()]
@@ -362,7 +426,8 @@ def _deserialize_aval(aval: ser_flatbuf.AbstractValue,
         ),
         scope=scope
     )
-    return core.ShapedArray(shape, dtype)
+    mem_space = aval.MemorySpace() or ser_flatbuf.MemorySpace.Device
+    return core.ShapedArray(shape, dtype, memory_space=_memory_space_from_enum[mem_space])
   else:
     assert False, aval_kind
 
@@ -454,8 +519,6 @@ def _serialize_disabled_safety_check(
     custom_call_target = builder.CreateString(custom_call_target_str)
   elif check == _export.DisabledSafetyCheck.platform():
     kind = ser_flatbuf.DisabledSafetyCheckKind.platform
-  elif check == _export.DisabledSafetyCheck.shape_assertions():
-    kind = ser_flatbuf.DisabledSafetyCheckKind.shape_assertions
   else:
     raise NotImplementedError(f"serializing DisabledSafetyCheck: {check}")
 
@@ -479,5 +542,10 @@ def _deserialize_disabled_safety_check(
   if kind == ser_flatbuf.DisabledSafetyCheckKind.platform:
     return _export.DisabledSafetyCheck.platform()
   if kind == ser_flatbuf.DisabledSafetyCheckKind.shape_assertions:
-    return _export.DisabledSafetyCheck.shape_assertions()
+    # shape_assertions has been deprecated in June 2024 (turned into a no-op),
+    # and removed in November 2024. We deserialize it to a DisabledSafetyCheck
+    # that has no effect.
+    # TODO(necula): remove this after June 2025, when we should not have any
+    # more serialized artifacts with shape_assertions.
+    return _export.DisabledSafetyCheck.custom_call("no op")
   assert False, kind
