@@ -15,55 +15,70 @@
 from __future__ import annotations
 
 from collections.abc import Sequence
+import enum
 from functools import partial
 import math
 
 import numpy as np
 
-from jax import lax
-
 from jax._src import dispatch
+from jax._src import dtypes
 from jax._src.api import jit, linear_transpose, ShapeDtypeStruct
 from jax._src.core import Primitive, is_constant_shape
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
+from jax._src.lax import lax
 from jax._src.lib.mlir.dialects import hlo
-from jax._src.lib import xla_client
-from jax._src.numpy.util import promote_dtypes_complex, promote_dtypes_inexact
 
 __all__ = [
   "fft",
   "fft_p",
 ]
 
-def _str_to_fft_type(s: str) -> xla_client.FftType:
+class FftType(enum.IntEnum):
+  "Describes which FFT operation to perform."
+
+  FFT = 0
+  "Forward complex-to-complex FFT."
+
+  IFFT = 1
+  "Inverse complex-to-complex FFT."
+
+  RFFT = 2
+  "Forward real-to-complex FFT."
+
+  IRFFT = 3
+  "Inverse real-to-complex FFT."
+
+
+def _str_to_fft_type(s: str) -> FftType:
   if s in ("fft", "FFT"):
-    return xla_client.FftType.FFT
+    return FftType.FFT
   elif s in ("ifft", "IFFT"):
-    return xla_client.FftType.IFFT
+    return FftType.IFFT
   elif s in ("rfft", "RFFT"):
-    return xla_client.FftType.RFFT
+    return FftType.RFFT
   elif s in ("irfft", "IRFFT"):
-    return xla_client.FftType.IRFFT
+    return FftType.IRFFT
   else:
     raise ValueError(f"Unknown FFT type '{s}'")
 
-@partial(jit, static_argnums=(1, 2))
-def fft(x, fft_type: xla_client.FftType | str, fft_lengths: Sequence[int]):
+@jit(static_argnums=(1, 2))
+def fft(x, fft_type: FftType | str, fft_lengths: Sequence[int]):
   if isinstance(fft_type, str):
     typ = _str_to_fft_type(fft_type)
-  elif isinstance(fft_type, xla_client.FftType):
+  elif isinstance(fft_type, FftType):
     typ = fft_type
   else:
     raise TypeError(f"Unknown FFT type value '{fft_type}'")
 
-  if typ == xla_client.FftType.RFFT:
+  if typ == FftType.RFFT:
     if np.iscomplexobj(x):
       raise ValueError("only real valued inputs supported for rfft")
-    x, = promote_dtypes_inexact(x)
+    x = lax.convert_element_type(x, dtypes.to_inexact_dtype(dtypes.dtype(x)))
   else:
-    x, = promote_dtypes_complex(x)
+    x = lax.convert_element_type(x, dtypes.to_complex_dtype(dtypes.dtype(x)))
   if len(fft_lengths) == 0:
     # XLA FFT doesn't support 0-rank.
     return x
@@ -80,7 +95,7 @@ def fft_abstract_eval(x, fft_type, fft_lengths):
   if len(fft_lengths) > x.ndim:
     raise ValueError(f"FFT input shape {x.shape} must have at least as many "
                     f"input dimensions as fft_lengths {fft_lengths}.")
-  if fft_type == xla_client.FftType.RFFT:
+  if fft_type == FftType.RFFT:
     if x.dtype not in (np.float32, np.float64):
       raise ValueError(f"RFFT input must be float32 or float64, got {x.dtype}")
     if x.shape[-len(fft_lengths):] != fft_lengths:
@@ -89,7 +104,7 @@ def fft_abstract_eval(x, fft_type, fft_lengths):
     shape = (x.shape[:-len(fft_lengths)] + fft_lengths[:-1]
              + (fft_lengths[-1] // 2 + 1,))
     dtype = _complex_dtype(x.dtype)
-  elif fft_type == xla_client.FftType.IRFFT:
+  elif fft_type == FftType.IRFFT:
     if not np.issubdtype(x.dtype, np.complexfloating):
       raise ValueError("IRFFT input must be complex64 or complex128, got "
                        f"{x.dtype}")
@@ -108,7 +123,7 @@ def fft_abstract_eval(x, fft_type, fft_lengths):
                       f"be equal to fft_lengths {fft_lengths}")
     shape = x.shape
     dtype = x.dtype
-  return x.update(shape=shape, dtype=dtype)
+  return x.update(shape=shape, dtype=dtype, vma=x.vma)
 
 def _fft_lowering(ctx, x, *, fft_type, fft_lengths):
   if not is_constant_shape(fft_lengths):
@@ -121,11 +136,11 @@ def _fft_lowering(ctx, x, *, fft_type, fft_lengths):
 
 
 def _naive_rfft(x, fft_lengths):
-  y = fft(x, xla_client.FftType.FFT, fft_lengths)
+  y = fft(x, FftType.FFT, fft_lengths)
   n = fft_lengths[-1]
   return y[..., : n//2 + 1]
 
-@partial(jit, static_argnums=1)
+@jit(static_argnums=1)
 def _rfft_transpose(t, fft_lengths):
   # The transpose of RFFT can't be expressed only in terms of irfft. Instead of
   # manually building up larger twiddle matrices (which would increase the
@@ -144,7 +159,7 @@ def _irfft_transpose(t, fft_lengths):
   # factor and a mask. The mask scales the cotangent for the Hermitian
   # symmetric components of the RFFT by a factor of two, since these components
   # are de-duplicated in the RFFT.
-  x = fft(t, xla_client.FftType.RFFT, fft_lengths)
+  x = fft(t, FftType.RFFT, fft_lengths)
   n = x.shape[-1]
   is_odd = fft_lengths[-1] % 2
   full = partial(lax.full_like, t, dtype=x.dtype)
@@ -157,13 +172,13 @@ def _irfft_transpose(t, fft_lengths):
   out = scale * lax.expand_dims(mask, range(x.ndim - 1)) * x
   assert out.dtype == _complex_dtype(t.dtype), (out.dtype, t.dtype)
   # Use JAX's convention for complex gradients
-  # https://github.com/google/jax/issues/6223#issuecomment-807740707
+  # https://github.com/jax-ml/jax/issues/6223#issuecomment-807740707
   return lax.conj(out)
 
 def _fft_transpose_rule(t, operand, fft_type, fft_lengths):
-  if fft_type == xla_client.FftType.RFFT:
+  if fft_type == FftType.RFFT:
     result = _rfft_transpose(t, fft_lengths)
-  elif fft_type == xla_client.FftType.IRFFT:
+  elif fft_type == FftType.IRFFT:
     result = _irfft_transpose(t, fft_lengths)
   else:
     result = fft(t, fft_type, fft_lengths)
