@@ -189,8 +189,8 @@ def _bitcast_uint8_to_uint4(operand):
 
 def bitcast_convert_type(operand, dtype):
   operand = np.asarray(operand)
-  nbits_in = dtypes.bit_width(operand.dtype)
-  nbits_out = dtypes.bit_width(dtype)
+  nbits_in = dtypes.itemsize_bits(operand.dtype)
+  nbits_out = dtypes.itemsize_bits(dtype)
 
   if nbits_out > nbits_in:
     assert operand.shape[-1] == nbits_out // nbits_in
@@ -285,7 +285,8 @@ def ragged_dot(
 
   out = np.zeros((m, n), dtype=lhs.dtype)
   result_iota = np.expand_dims(np.arange(out.shape[0]), list(range(1, out.ndim)))
-  start = 0
+  result_iota = result_iota.astype(group_sizes.dtype)
+  start = np.asarray(0, dtype=group_sizes.dtype)
   for i, size in enumerate(group_sizes):
     out += np.where(
         np.logical_and(start <= result_iota, result_iota < (start + size)),
@@ -528,3 +529,14 @@ def _reducer_from_pyfunc(py_binop, init_val):
       result[out_idx] = py_binop(result[out_idx], operand[idx])
     return result
   return reducer
+
+def top_k(operand, k, axis=-1):
+  if axis < 0:
+    axis = operand.ndim + axis
+  assert 0 <= axis < operand.ndim
+  operand_flipped = np.flip(operand, axis)
+  indices_flipped = np.argsort(operand_flipped, axis=axis, kind="stable")
+  indices_all = (operand.shape[axis] - 1 - np.flip(indices_flipped, axis)).astype(np.int32)
+  indices = indices_all[(_slice(None),) * axis + (_slice(k),)]
+  values = np.take_along_axis(operand, indices, axis=axis)
+  return values, indices
