@@ -20,7 +20,6 @@ import dataclasses
 import functools
 from typing import Any, Union
 
-from jax._src import config
 from jax._src.util import use_cpp_class, cache, use_cpp_method
 from jax._src.lib import xla_client as xc
 from jax._src.lib.mlir.dialects import sdy
@@ -47,6 +46,11 @@ class AUTO:
     return SdyArray(mesh_shape=self.mesh.shape_tuple,
                     dim_shardings=dim_shardings)
 
+  @property
+  def _device_assignment(self):
+    return self.mesh._flat_devices_tuple
+
+
 class UnspecifiedValue:
   def __repr__(self):
     return "UnspecifiedValue"
@@ -198,10 +202,8 @@ class NamedSharding(JSharding.Sharding):
     if isinstance(self.mesh, mesh_lib.AbstractMesh):
       raise ValueError('is_fully_addressable is not implemented for '
                        '`jax.sharding.AbstractMesh`.')
-    if config.enable_empty_arrays.value:
-      # return False if addressable_device_list is empty.
-      return self._internal_device_list.is_fully_addressable  # type: ignore
-    return not self.mesh.is_multi_process
+    # return False if addressable_device_list is empty.
+    return self._internal_device_list.is_fully_addressable  # type: ignore
 
   @property
   def _is_concrete(self) -> bool:
@@ -229,6 +231,14 @@ class NamedSharding(JSharding.Sharding):
       num_partitions *= mesh_shape[name]
     return num_partitions == 1
 
+  @functools.cached_property
+  def replicated_axes(self) -> frozenset[MeshAxisName]:
+    flat_spec = frozenset(
+        s for s in flatten_spec(self.spec)
+        if s is not None and s is not PartitionSpec.UNCONSTRAINED)
+    return frozenset(self.mesh.axis_names) - (
+        flat_spec | self.spec.unreduced | self.spec.reduced)
+
   def with_memory_kind(self, kind: str) -> NamedSharding:
     return self.update(memory_kind=kind)
 
@@ -265,6 +275,16 @@ class NamedSharding(JSharding.Sharding):
 
 NamedSharding.__module__ = 'jax.sharding'
 
+def flatten_spec(spec):
+  out = []
+  for s in spec:
+    if isinstance(s, tuple):
+      out.extend(s)
+    else:
+      out.append(s)
+  return out
+
+
 def get_array_mapping(
     axis_resources: PartitionSpec | AUTO | UnspecifiedValue
 ) -> ArrayMappingOrAutoOrUnspecified:
@@ -364,11 +384,6 @@ def modify_sdy_sharding_wrt_axis_types(sdy_sharding: SdyArray, mesh):
 @cache(max_size=4096, trace_context_in_key=False)
 def named_sharding_to_xla_hlo_sharding(
     self, num_dimensions: int) -> xc.HloSharding:
-  if self.spec.unreduced or self.spec.reduced:
-    raise ValueError(
-        'unreduced/reduced only works with the shardy partitioner. Please use'
-        " `jax.config.update('jax_use_shardy_partitioner', True)` to switch"
-        ' shardy on.')
   mesh_shape = self.mesh.shape
   array_mapping = get_array_mapping(self.spec)
   mesh_axis_pos = {name: i for i, name in enumerate(self.mesh.axis_names)}
@@ -380,6 +395,12 @@ def named_sharding_to_xla_hlo_sharding(
     for manual_axis in manual_axes:
       special_axes[axis_names.index(manual_axis)] = xc.OpSharding.Type.MANUAL
 
+  unreduced_axes = self.spec.unreduced
+  if unreduced_axes:
+    axis_names = self.mesh.axis_names
+    for u in unreduced_axes:
+      special_axes[axis_names.index(u)] = xc.OpSharding.Type.UNREDUCED
+
   replicated_mesh_axes = []
   for i, (axis_name, axis_val) in enumerate(mesh_shape.items()):
     if axis_name not in array_mapping:  # type: ignore
@@ -518,12 +539,12 @@ def _check_mesh_resource_axis(mesh, pspec):
             f"Resource axis: {r} of {pspec} "
             f"is not found in mesh: {tuple(mesh.shape.keys())}.")
   check_pspec_mix_axis_type(mesh, pspec)
-  if (AxisType.Auto not in mesh._axis_types_dict and
+  if (AxisType.Auto not in mesh.axis_types and
       PartitionSpec.UNCONSTRAINED in pspec):
     raise ValueError(
         f'{pspec} cannot contain'
         ' `P.UNCONSTRAINED` when no mesh axis_types are `Auto`. Got mesh'
-        f' axis_types: {mesh._axis_types_dict}')
+        f' axis_types: {mesh.axis_types}')
 
 def _check_mesh_unreduced(mesh, pspec):
   for u in pspec.unreduced:
@@ -531,10 +552,10 @@ def _check_mesh_unreduced(mesh, pspec):
       raise ValueError(
           f'Unreduced axes {u} is not found in {mesh.axis_names=}. '
           f'Got {pspec=}')
-    if mesh._name_to_type[u] in (AxisType.Auto, AxisType.Manual):
+    if mesh._name_to_type[u] == AxisType.Auto:
       raise ValueError(
-          'Unreduced axes can only refer to mesh axes that is of type'
-          f' `Explicit`. Got unreduced axes: {pspec.unreduced} and'
+          'Unreduced axes can only refer to mesh axes that are of type'
+          f' `Explicit` or `Manual`. Got unreduced axes: {pspec.unreduced} and'
           f' mesh: {mesh}')
 
   for u in pspec.reduced:
@@ -542,8 +563,8 @@ def _check_mesh_unreduced(mesh, pspec):
       raise ValueError(
           f'Reduced axes {u} is not found in {mesh.axis_names=}. '
           f'Got {pspec=}')
-    if mesh._name_to_type[u] in (AxisType.Auto, AxisType.Manual):
+    if mesh._name_to_type[u] == AxisType.Auto:
       raise ValueError(
-          'Reduced axes can only refer to mesh axes that is of type'
-          f' `Explicit`. Got reduced axes: {pspec.reduced} and'
+          'Reduced axes can only refer to mesh axes that are of type'
+          f' `Explicit` or `Manual`. Got reduced axes: {pspec.reduced} and'
           f' mesh: {mesh}')
