@@ -24,7 +24,7 @@ from __future__ import annotations
 __all__ = ['register_jax_array_methods']
 
 import abc
-from functools import partial, wraps
+from functools import wraps
 import math
 from typing import Any
 from collections.abc import Callable, Sequence
@@ -34,6 +34,7 @@ import numpy as np
 from jax._src import api
 from jax._src import core
 from jax._src import dtypes
+from jax._src import literals
 from jax._src.api_util import _ensure_index_tuple
 from jax._src.array import ArrayImpl
 from jax._src.lax import lax
@@ -48,7 +49,6 @@ from jax._src.numpy import reductions
 from jax._src.numpy import ufuncs
 from jax._src.pjit import PartitionSpec
 from jax._src.sharding import Sharding
-from jax._src import mesh as mesh_lib
 from jax._src.sharding_impls import canonicalize_sharding, NamedSharding
 from jax._src.ops import scatter
 from jax._src.typing import Array, ArrayLike, DimSize, DTypeLike, Shape, StaticScalar
@@ -220,7 +220,7 @@ def _item(self: Array, *args: int) -> bool | int | float | complex:
 
 def _itemsize_property(self: Array) -> int:
   """Length of one array element in bytes."""
-  return dtypes.dtype(self, canonicalize=True).itemsize
+  return self.dtype.itemsize
 
 def _matrix_transpose_property(self: Array):
   """Compute the (batched) matrix transpose.
@@ -262,7 +262,7 @@ def _min(self: Array, axis: reductions.Axis = None, out: None = None,
 
 def _nbytes_property(self: Array) -> int:
   """Total bytes consumed by the elements of the array."""
-  return np.size(self) * dtypes.dtype(self, canonicalize=True).itemsize
+  return np.size(self) * self.dtype.itemsize
 
 def _nonzero(self: Array, *, fill_value: None | ArrayLike | tuple[ArrayLike, ...] = None,
              size: int | None = None) -> tuple[Array, ...]:
@@ -318,6 +318,10 @@ def _reshape(self: Array, *args: Any, order: str = "C", out_sharding=None
     return lax.reshape(self, newshape, None, out_sharding=out_sharding)
   elif order == "F":
     dims = list(range(self.ndim)[::-1])
+    out_sharding = canonicalize_sharding(out_sharding, "jnp.reshape")
+    out_sharding = (
+        None if out_sharding is None else out_sharding.update(
+            spec=out_sharding.spec.update(partitions=out_sharding.spec[::-1])))
     return lax.reshape(self, newshape[::-1], dims, out_sharding=out_sharding).T
   elif order == "A":
     raise NotImplementedError("np.reshape order=A is not implemented.")
@@ -506,22 +510,32 @@ def _view(self: Array, dtype: DTypeLike | None = None, type: None = None) -> Arr
     Array([ True, False,  True], dtype=bool)
 
   However, there are no guarantees about the results of any expression involving
-  a view such as this: `jnp.array([1, 2, 3], dtype=jnp.int8).view(jnp.bool_)`.
+  a view such as this: ``jnp.array([1, 2, 3], dtype=jnp.int8).view(jnp.bool_)``.
   In particular, the results may change between JAX releases and depending on
   the platform. To safely convert such an array to a boolean array, compare it
   with `0`::
 
     >>> jnp.array([1, 2, 0], dtype=jnp.int8) != 0
     Array([ True,  True, False], dtype=bool)
+
+  Args:
+    dtype: An optional output dtype. If not specified, the output dtype is the
+      same as the input dtype.
+    type: Not implemented; accepted for NumPy compatibility.
+  Returns:
+    The array, viewed as the new dtype. Unlike NumPy, the array may or may not
+    be a copy of the input array.
   """
   if type is not None:
     raise NotImplementedError("`type` argument of array.view() is not supported.")
 
-  dtypes.check_user_dtype_supported(dtype, "view")
-  dtype = dtypes.canonicalize_dtype(dtype)
+  if dtype is None:
+    return self
 
-  nbits_in = dtypes.bit_width(self.dtype)
-  nbits_out = dtypes.bit_width(dtype)
+  dtype = dtypes.check_and_canonicalize_user_dtype(dtype, "view")
+
+  nbits_in = dtypes.itemsize_bits(self.dtype)
+  nbits_out = dtypes.itemsize_bits(dtype)
 
   if self.ndim == 0:
     if nbits_in != nbits_out:
@@ -543,9 +557,10 @@ def _view(self: Array, dtype: DTypeLike | None = None, type: None = None) -> Arr
   if lax_numpy.issubdtype(self.dtype, np.complexfloating):
     new_shape = (*self.shape[:-1], self.shape[-1] * 2)
     new_dtype = lax_numpy.finfo(self.dtype).dtype
-    self = (array_creation.zeros(new_shape, new_dtype)
-             .at[..., 0::2].set(self.real)
-             .at[..., 1::2].set(self.imag))
+    new_sharding = core.typeof(self).sharding
+    self = (array_creation.zeros(new_shape, new_dtype, out_sharding=new_sharding)
+            .at[..., 0::2].set(self.real)
+            .at[..., 1::2].set(self.imag))
     return _view(self, dtype)
 
   if dtype == bool:
@@ -573,7 +588,15 @@ def _notimplemented_flat(self):
   raise NotImplementedError("JAX Arrays do not implement the arr.flat property: "
                             "consider arr.flatten() instead.")
 
-_accepted_binop_types = (int, float, complex, np.generic, np.ndarray, Array)
+_accepted_binop_types = (
+    int,
+    float,
+    complex,
+    np.generic,
+    np.ndarray,
+    Array,
+    literals.TypedNdArray,
+)
 _rejected_binop_types = (list, tuple, set, dict)
 
 def _defer_to_unrecognized_arg(opchar, binary_op, swap=False):
@@ -621,7 +644,7 @@ def __array_module__(self, types):
     return NotImplemented
 
 
-@partial(api.jit, static_argnums=(1,2,3))
+@api.jit(static_argnums=(1,2,3))
 def _multi_slice(self: Array,
                  start_indices: tuple[tuple[int, ...]],
                  limit_indices: tuple[tuple[int, ...]],
@@ -659,16 +682,6 @@ def _chunk_iter(x, size):
 def _getitem(self, item):
   return indexing.rewriting_take(self, item)
 
-def _val_sharding(val):
-  cur_mesh = mesh_lib.get_abstract_mesh()
-  val_s = core.typeof(val).sharding
-  if val_s.mesh._are_all_axes_auto_or_manual:
-    return None
-  if val_s.mesh.empty:
-    return (NamedSharding(cur_mesh, PartitionSpec())
-            if cur_mesh._are_all_axes_explicit else None)
-  return val_s
-
 
 # Syntactic sugar for scatter operations.
 class _IndexUpdateHelper:
@@ -807,7 +820,7 @@ class _IndexUpdateRef:
   def get(self, *, indices_are_sorted: bool = False, unique_indices: bool = False,
           mode: str | lax_slicing.GatherScatterMode | None = None,
           fill_value: ArrayLike | None = None,
-          out_sharding: Sharding | PartitionSpec | None = None,
+          out_sharding: NamedSharding | PartitionSpec | None = None,
           wrap_negative_indices: bool = True):
     """Equivalent to ``x[idx]``.
 
@@ -831,6 +844,7 @@ class _IndexUpdateRef:
   def set(self, values: ArrayLike, *, indices_are_sorted: bool = False,
           unique_indices: bool = False,
           mode: str | lax_slicing.GatherScatterMode | None = None,
+          out_sharding: NamedSharding | PartitionSpec | None = None,
           wrap_negative_indices: bool = True) -> None:
     """Pure equivalent of ``x[idx] = y``.
 
@@ -839,11 +853,14 @@ class _IndexUpdateRef:
 
     See :func:`jax.numpy.ndarray.at` for details.
     """
-    out_s = _val_sharding(self.array)
+    if out_sharding is not None:
+      assert isinstance(out_sharding, (NamedSharding, PartitionSpec))
+      out_sharding = canonicalize_sharding(out_sharding, '.set')
     return scatter._scatter_update(
         self.array, self.index, values, lax_slicing.scatter,
         indices_are_sorted=indices_are_sorted, unique_indices=unique_indices,
-        mode=mode, out_sharding=out_s, normalize_indices=wrap_negative_indices)
+        mode=mode, out_sharding=out_sharding,  # type: ignore
+        normalize_indices=wrap_negative_indices)
 
   def apply(self, func: Callable[[ArrayLike], Array], *,
             indices_are_sorted: bool = False, unique_indices: bool = False,
@@ -872,6 +889,7 @@ class _IndexUpdateRef:
   def add(self, values: ArrayLike, *,
           indices_are_sorted: bool = False, unique_indices: bool = False,
           mode: str | lax_slicing.GatherScatterMode | None = None,
+          out_sharding: NamedSharding | PartitionSpec | None = None,
           wrap_negative_indices: bool = True) -> Array:
     """Pure equivalent of ``x[idx] += y``.
 
@@ -880,11 +898,14 @@ class _IndexUpdateRef:
 
     See :func:`jax.numpy.ndarray.at` for details.
     """
-    out_s = _val_sharding(self.array)
+    if out_sharding is not None:
+      assert isinstance(out_sharding, (NamedSharding, PartitionSpec))
+      out_sharding = canonicalize_sharding(out_sharding, '.add')
     return scatter._scatter_update(
         self.array, self.index, values, lax_slicing.scatter_add,
         indices_are_sorted=indices_are_sorted, unique_indices=unique_indices,
-        mode=mode, normalize_indices=wrap_negative_indices, out_sharding=out_s)
+        mode=mode, out_sharding=out_sharding,  # type: ignore
+        normalize_indices=wrap_negative_indices)
 
   def subtract(self, values: ArrayLike, *,
                indices_are_sorted: bool = False, unique_indices: bool = False,
@@ -1191,7 +1212,6 @@ def _set_array_abstract_methods(basearray):
 def register_jax_array_methods():
   """Call this function once to register methods of JAX arrays"""
   _set_shaped_array_attributes(core.ShapedArray)
-  _set_shaped_array_attributes(core.DShapedArray)
 
   _set_array_base_attributes(ArrayImpl, exclude={'__getitem__'})
   _set_tracer_aval_forwarding(core.Tracer, exclude={*_impl_only_array_methods, "at"})
