@@ -14,7 +14,6 @@
 
 from __future__ import annotations
 
-from functools import partial
 import math
 import operator
 from typing import cast, NamedTuple
@@ -26,6 +25,7 @@ from jax._src import core
 from jax._src import dtypes
 from jax._src.lax import lax
 from jax._src.lax import slicing as lax_slicing
+from jax._src.lax import utils as lax_utils
 from jax._src.numpy.array_creation import empty, full, full_like, ones, zeros
 from jax._src.numpy.lax_numpy import (
     append, arange, concatenate, diff,
@@ -41,7 +41,7 @@ from jax._src.typing import Array, ArrayLike
 export = set_module('jax.numpy')
 
 
-@partial(api.jit, static_argnames=('assume_unique', 'invert', 'method'))
+@api.jit(static_argnames=('assume_unique', 'invert', 'method'))
 def _in1d(ar1: ArrayLike, ar2: ArrayLike, invert: bool,
           method='auto', assume_unique=False) -> Array:
   ar1, ar2 = ensure_arraylike("in1d", ar1, ar2)
@@ -156,6 +156,9 @@ def setdiff1d(ar1: ArrayLike, ar2: ArrayLike, assume_unique: bool = False,
     Array([1, 2, 0, 0], dtype=int32)
   """
   arr1, arr2 = ensure_arraylike("setdiff1d", ar1, ar2)
+  arr1 = arr1.ravel()
+  arr2 = arr2.ravel()
+
   if size is None:
     core.concrete_or_error(None, ar1, "The error arose in setdiff1d()")
   else:
@@ -250,7 +253,7 @@ def union1d(ar1: ArrayLike, ar2: ArrayLike,
   return cast(Array, out)
 
 
-@partial(api.jit, static_argnames=['assume_unique', 'size'])
+@api.jit(static_argnames=['assume_unique', 'size'])
 def _setxor1d_size(arr1: Array, arr2: Array, fill_value: ArrayLike | None, *,
                    assume_unique: bool, size: int, ) -> Array:
   # Ensured by caller
@@ -337,14 +340,15 @@ def setxor1d(ar1: ArrayLike, ar2: ArrayLike, assume_unique: bool = False, *,
   return aux[flag[1:] & flag[:-1]]
 
 
-@partial(api.jit, static_argnames=['return_indices'])
+@api.jit(static_argnames=['return_indices'])
 def _intersect1d_sorted_mask(arr1: Array, arr2: Array,
                              return_indices: bool) -> tuple[Array, Array, Array | None]:
   """JIT-compatible helper function for intersect1d"""
   assert arr1.ndim == arr2.ndim == 1
   arr = concatenate((arr1, arr2))
   if return_indices:
-    iota = lax.broadcasted_iota(np.int64, np.shape(arr), dimension=0)
+    idx_dtype = lax_utils.int_dtype_for_dim(arr.shape[0], signed=True)
+    iota = lax.broadcasted_iota(idx_dtype, np.shape(arr), dimension=0)
     aux, indices = lax.sort_key_val(arr, iota)
   else:
     aux = sort(arr)
@@ -353,7 +357,7 @@ def _intersect1d_sorted_mask(arr1: Array, arr2: Array,
   return aux, mask, indices
 
 
-@partial(api.jit, static_argnames=['fill_value', 'assume_unique', 'size', 'return_indices'])
+@api.jit(static_argnames=['fill_value', 'assume_unique', 'size', 'return_indices'])
 def _intersect1d_size(arr1: Array, arr2: Array, fill_value: ArrayLike | None, assume_unique: bool,
                       size: int, return_indices: bool) -> Array | tuple[Array, Array, Array]:
   """Jit-compatible helper function for intersect1d with size specified."""
@@ -570,7 +574,7 @@ UNIQUE_SIZE_HINT = (
   "To make jnp.unique() compatible with JIT and other transforms, you can specify "
   "a concrete value for the size argument, which will determine the output size.")
 
-@partial(api.jit, static_argnames=['axis', 'equal_nan'])
+@api.jit(static_argnames=['axis', 'equal_nan'])
 def _unique_sorted_mask(ar: Array, axis: int, equal_nan: bool) -> tuple[Array, Array, Array]:
   aux = moveaxis(ar, axis, 0)
   if np.issubdtype(aux.dtype, np.complexfloating):
@@ -635,7 +639,7 @@ def _unique(ar: Array, axis: int, return_index: bool = False, return_inverse: bo
   if return_inverse:
     if aux.size:
       imask = cumsum(mask) - 1
-      inv_idx = zeros(mask.shape, dtype=dtypes.canonicalize_dtype(dtypes.int_))
+      inv_idx = zeros(mask.shape, dtype=int)
       inv_idx = inv_idx.at[perm].set(imask)
     else:
       inv_idx = zeros(ar.shape[axis], dtype=int)
@@ -649,7 +653,7 @@ def _unique(ar: Array, axis: int, return_index: bool = False, return_inverse: bo
         idx = idx.at[1:].set(where(idx[1:], idx[1:], mask.size))
       ret += (diff(idx),)
     elif ar.shape[axis]:
-      ret += (full((1,), ar.shape[axis], dtype=dtypes.canonicalize_dtype(dtypes.int_)),)
+      ret += (full((1,), ar.shape[axis], dtype=int),)
     else:
       ret += (empty(0, dtype=int),)
   if return_true_size:
