@@ -29,10 +29,12 @@ from jax import lax
 from jax._src import ad_util
 from jax._src import core
 from jax._src import custom_derivatives
+from jax._src import hijax
 from jax._src import pjit
 from jax._src import prng
 from jax._src import state
 from jax._src import tree_util
+from jax._src import typing
 from jax._src import util
 from jax._src.interpreters import partial_eval as pe
 from jax._src.pallas import core as pallas_core
@@ -40,6 +42,7 @@ from jax._src.pallas import utils as pallas_utils
 from jax._src.pallas.fuser import fuser_utils
 from jax._src.state import indexing
 from jax._src.state import primitives as state_primitives
+from jax._src.traceback_util import api_boundary
 import jax.numpy as jnp
 import numpy as np
 
@@ -73,6 +76,7 @@ class PushRuleContext:
   avals_out: tuple[core.AbstractValue, ...]
 
 
+@functools.partial(api_boundary, repro_api_name="fuser.make_scalar_prefetch_handler")
 def make_scalar_prefetch_handler(*args):
   def scalar_prefetch_getter(*sp_inputs):
     result = sp_inputs
@@ -147,8 +151,8 @@ class KernelEvalContext:
   program_ids: tuple[int | jax.Array, ...] | None
   avals_in: tuple[core.AbstractValue, ...] | None
   avals_out: tuple[core.AbstractValue, ...] | None
-  in_block_specs: tuple[pallas_core.BlockSpec, ...] | None
-  out_block_specs: tuple[pallas_core.BlockSpec, ...] | None
+  in_block_specs: tuple[pallas_core.BlockSpec, ...]
+  out_block_specs: tuple[pallas_core.BlockSpec, ...]
   grid: tuple[int | jax.Array, ...] | None
   scalar_prefetch_handler: Any | None
   out_usages: tuple[set[Usage], ...] | None
@@ -234,6 +238,7 @@ def _unwrap_block_spec_scalar_prefetch(
   return out_block_spec
 
 
+@functools.partial(api_boundary, repro_api_name="fuser.pull_block_spec")
 def pull_block_spec(
     f: Callable,
     out_block_specs: pallas_core.BlockSpec | tuple[pallas_core.BlockSpec, ...],
@@ -291,6 +296,34 @@ def pull_block_spec(
   return wrapped
 
 
+def _block_dim_equal(
+    b1: int | pallas_core.BlockDim | None, b2: int | pallas_core.BlockDim | None
+) -> bool:
+  block_size1 = pallas_core.get_block_size(b1)
+  block_size2 = pallas_core.get_block_size(b2)
+  match (b1, b2):
+    case (None, _) | (_, None):
+      return b1 == b2
+    case (
+        (pallas_core.Blocked(), int())
+        | (int(), pallas_core.Blocked())
+        | (pallas_core.Blocked(), pallas_core.Blocked())
+        | (int(), int())
+    ):
+      return block_size1 == block_size2
+    case _:
+      return type(b1) == type(b2) and (block_size1 == block_size2)
+
+
+def _block_shapes_equal(
+    bs1: tuple[int | pallas_core.BlockDim | None] | None,
+    bs2: tuple[int | pallas_core.BlockDim | None] | None,
+) -> bool:
+  if bs1 is None or bs2 is None:
+    return bs1 == bs2
+  return all(_block_dim_equal(b1, b2) for b1, b2 in zip(bs1, bs2))
+
+
 def _pull_block_spec(
     jaxpr: core.Jaxpr,
     out_block_specs: tuple[pallas_core.BlockSpec, ...],
@@ -358,7 +391,7 @@ def _pull_block_spec(
           jaxpr.invars,
           needed_invars,
           jaxpr.eqns[: jaxpr.eqns.index(eqn)],
-          debug_info=jaxpr.debug_info,
+          debug_info=jaxpr.debug_info._replace(result_paths=None),
       )
       scalar_prefetch_jaxpr, used_consts, used_invars = pe.dce_jaxpr_consts(
           scalar_prefetch_jaxpr_no_dce,
@@ -368,6 +401,7 @@ def _pull_block_spec(
       scalar_prefetch_jaxpr = scalar_prefetch_jaxpr.replace(
           constvars=[],
           invars=jaxpr.constvars,
+          debug_info=scalar_prefetch_jaxpr.debug_info.with_unknown_names(),
       )
 
       def _scalar_prefetch_fn(jaxpr):
@@ -387,7 +421,8 @@ def _pull_block_spec(
       if (
           not isinstance(v, core.Literal)
           and v in env
-          and env[v].block_shape != in_block_spec.block_shape
+          and not _block_shapes_equal(env[v].block_shape,
+                                      in_block_spec.block_shape)
       ):
         in_block_spec = pallas_core.BlockSpec(_illegal, _illegal)  # pytype: disable=wrong-arg-types
       _write_block_spec(v, in_block_spec)
@@ -430,7 +465,6 @@ def make_kernel_function(
   def _remove_nones(
       shape: tuple[pallas_core.BlockDim | int | None, ...] | None,
   ) -> tuple[int, ...]:
-    assert shape is not None
     new_shape = tuple(_block_size(s) for s in shape)
     return tuple(s for s in new_shape if s is not None)
 
@@ -452,18 +486,8 @@ def make_kernel_function(
   unflat_in_block_arg_avals, unflat_in_block_kwarg_avals = (
       tree_util.tree_unflatten(in_tree, in_block_avals)
   )
-  unflat_arg_usages, unflat_kwarg_usages = tree_util.tree_unflatten(
-      in_tree, invar_usages
-  )
 
-  def sds_like(x):
-    if x is _no_aval:
-      return _no_aval
-    return jax.ShapeDtypeStruct(x.shape, x.dtype)
-
-  kernel_in_type = jax.tree.map(
-      sds_like, (unflat_in_block_arg_avals, unflat_in_block_kwarg_avals)
-  )
+  kernel_in_type = (unflat_in_block_arg_avals, unflat_in_block_kwarg_avals)
 
   def _read_block_spec(atom: core.Atom) -> pallas_core.BlockSpec | Any:
     if isinstance(atom, core.Literal):
@@ -471,39 +495,6 @@ def make_kernel_function(
     return bs_env.get(atom, pallas_core.no_block_spec)
 
   def kernel_fn(program_ids, scalar_prefetch, *args, **kwargs):
-    def _check_args(prefix, path, x, y, usage):
-      if usage == {Usage.SCALAR_PREFETCH}:
-        return
-      if y is _no_aval:
-        return
-      x_aval, y_aval = core.get_aval(x), core.get_aval(y)
-      if x_aval.shape != y_aval.shape:
-        raise ValueError(
-            f'Shapes do not match: actual={x_aval.shape} !='
-            f' expected={y_aval.shape}. Path:'
-            f' {prefix}{jax.tree_util.keystr(path)}. Expected type:'
-            f' {kernel_in_type}. Actual args: {(args, kwargs)}'
-        )
-      if x_aval.dtype != y_aval.dtype:
-        raise ValueError(
-            f'DTypes do not match: actual={x_aval.dtype} !='
-            f' expected={y_aval.dtype}. Path:'
-            f' {prefix}{jax.tree_util.keystr(path)}. Expected type:'
-            f' {kernel_in_type}. Actual args: {(args, kwargs)}'
-        )
-
-    jax.tree_util.tree_map_with_path(
-        functools.partial(_check_args, 'args'),
-        args,
-        kernel_in_type[0],
-        unflat_arg_usages,
-    )
-    jax.tree_util.tree_map_with_path(
-        functools.partial(_check_args, 'kwargs'),
-        kwargs,
-        kernel_in_type[1],
-        unflat_kwarg_usages,
-    )
     flat_args, in_tree_ = tree_util.tree_flatten((args, kwargs))
     if in_tree_ != tree_util.tree_structure(kernel_in_type):
       raise ValueError(f'Expected {kernel_in_type} PyTree, got {in_tree_}')
@@ -575,9 +566,12 @@ def make_kernel_function(
   return kernel_fn
 
 
+@functools.partial(api_boundary, repro_api_name="fuser.get_fusion_values")
 def get_fusion_values(
     fusion: Callable, *args, **kwargs
-) -> tuple[Callable, tuple[jax.Array, ...], tuple[jax.Array, ...]]:
+) -> tuple[
+    Callable, tuple[typing.SupportsShape, ...], tuple[typing.SupportsShape, ...]
+]:
   jaxpr, values, in_tree, out_tree = fuser_utils.make_jaxpr(
       fusion, *args, **kwargs
   )
@@ -724,7 +718,7 @@ def _eltwise_usage_rule(
   return [used_out]
 
 
-def _bcast_block_spec(
+def _pull_bcast_block_spec(
     block_spec: pallas_core.BlockSpec, i: int
 ) -> pallas_core.BlockSpec:
   def new_index_map(*args):
@@ -747,6 +741,21 @@ def _bcast_block_spec(
   return pallas_core.BlockSpec(new_block_shape, new_index_map)
 
 
+def _push_bcast_block_spec(
+    block_spec: pallas_core.BlockSpec,
+    i: int,
+    size: int,
+) -> pallas_core.BlockSpec:
+
+  bcast_dim_block_shape = size
+  if isinstance(block_spec.block_shape[i], pallas_core.Element):
+    bcast_dim_block_shape = pallas_core.Element(size)
+  new_block_shape = util.tuple_update(  # pytype: disable=wrong-arg-types
+      block_spec.block_shape, i, bcast_dim_block_shape
+  )
+  return pallas_core.BlockSpec(new_block_shape, block_spec.index_map)
+
+
 def _binop_usage_rule(prim, ctx, used_out: set[Usage]):
   del prim
   if used_out == {Usage.SCALAR_PREFETCH}:
@@ -788,9 +797,9 @@ def _binop_pull_rule(prim, ctx: PullRuleContext, block_spec):
       zip(left_aval.shape, right_aval.shape, strict=True)
   ):
     if l == 1 and r != 1:
-      l_block_spec = _bcast_block_spec(l_block_spec, i)
+      l_block_spec = _pull_bcast_block_spec(l_block_spec, i)
     if r == 1 and l != 1:
-      r_block_spec = _bcast_block_spec(r_block_spec, i)
+      r_block_spec = _pull_bcast_block_spec(r_block_spec, i)
 
   return [l_block_spec, r_block_spec]
 
@@ -958,7 +967,7 @@ def _slice_rule(
     strides: tuple[int, ...] | None,
 ):
   del ctx
-  if strides is not None:
+  if strides is not None and not all(stride == 1 for stride in strides):
     raise NotImplementedError('strides are not supported yet')
   slice_sizes = tuple(
       int(end - start) for start, end in zip(start_indices, limit_indices)
@@ -1083,7 +1092,10 @@ def _swap_eval_rule(ctx: KernelEvalContext, ref, val, *idx, tree):
   assert hasattr(ref_aval, 'shape')
   if len(indexers) > 1:
     raise NotImplementedError('swap not supported yet')
-  indexer_aval = indexers_avals[0]
+  if not indexers_avals:
+    indexer_aval = indexing.NDIndexer.make_trivial_indexer(ref_aval.shape)
+  else:
+    indexer_aval = indexers_avals[0]
   for idx_aval, size in zip(indexer_aval.indices, ref_aval.shape, strict=True):
     if not isinstance(idx_aval, indexing.Slice):
       raise NotImplementedError('swap not supported yet')
@@ -1123,11 +1135,15 @@ def _get_pull_rule(
   indexers_avals = tree_util.tree_unflatten(tree, ctx.avals_in[1:])
   if len(indexers_avals) > 1:
     raise NotImplementedError('get not supported yet')
-  indexer_aval = indexers_avals[0]
+  if not indexers_avals:
+    indexer_aval = indexing.NDIndexer.make_trivial_indexer(ref_aval.shape)
+  else:
+    indexer_aval = indexers_avals[0]
   block_shape_iter = iter(block_spec.block_shape)
   block_shape = []
   if not all(
-      isinstance(bd, (int, pallas_core.Blocked, pallas_core.Squeezed, None))
+      bd is None
+      or isinstance(bd, (int, pallas_core.Blocked, pallas_core.Squeezed))
       for bd in block_spec.block_shape
   ):
     raise NotImplementedError('get not supported yet')
@@ -1162,8 +1178,9 @@ def _get_pull_rule(
     assert next(idx_iter, None) is None
     return indices
 
-  block_spec = pallas_core.BlockSpec(block_shape, new_index_map)
-  return [block_spec] + [pallas_core.no_block_spec] * (len(ctx.avals_in) - 1)
+  new_block_spec = pallas_core.BlockSpec(block_shape, new_index_map)
+  return ([new_block_spec]
+          + [pallas_core.no_block_spec] * (len(ctx.avals_in) - 1))
 
 
 @register_eval_rule(state_primitives.get_p)
@@ -1175,8 +1192,12 @@ def _get_eval_rule(ctx: KernelEvalContext, ref, *idx, tree):
   assert hasattr(ref_aval, 'shape')
   if len(indexers) > 1:
     raise NotImplementedError('get not supported yet')
-  indexer = indexers[0]
-  indexer_aval = indexers_avals[0]
+  if not indexers:
+    indexer = indexing.NDIndexer.make_trivial_indexer(ref_aval.shape)
+    indexer_aval = indexer
+  else:
+    indexer = indexers[0]
+    indexer_aval = indexers_avals[0]
   block_indexer = []
 
   def _slice(i, b):
@@ -1534,10 +1555,16 @@ def _iota_eval_rule(
   block_spec = eval_ctx.out_block_specs[0]
   block_idx = eval_ctx.get_out_block_indices()[0]
   assert len(block_idx) == len(shape)
-  iota_shape = tuple(s for s in block_spec.block_shape if s is not None)
-  dim_ = dimension - sum(s is None for s in block_spec.block_shape[:dimension])
+  iota_shape = tuple(
+      _block_size(s) for s in block_spec.block_shape if s is not None
+  )
+  dim_ = dimension - sum(
+      _block_size(s) is None for s in block_spec.block_shape[:dimension]
+  )
   local_iota = jax.lax.broadcasted_iota(dtype, iota_shape, dim_)
-  return local_iota + block_idx[dimension] * block_spec.block_shape[dimension]
+  return local_iota + block_idx[dimension] * _block_size(
+      block_spec.block_shape[dimension]
+  )
 
 
 @register_pull_block_spec_rule(lax.iota_p)
@@ -1593,7 +1620,7 @@ def _reshape_pull_rule(
   aval_out = ctx.avals_out[0]
   assert isinstance(aval_out, core.ShapedArray)
 
-  block_shape = tuple(map(_block_size, block_spec.block_shape))
+  block_shape = block_spec.block_shape
   shape_in = aval_in.shape
   shape_out = aval_out.shape
   assert np.prod(shape_in) == np.prod(shape_out)
@@ -1620,24 +1647,34 @@ def _reshape_pull_rule(
     new_grids = []
 
     for d, bd, merged in zip(shape_out, block_shape, merged_dims):
+      bs = pallas_core.get_block_size(bd)
+
+      if len(merged) == 1:
+        new_grids.append((merged[0] // bs,))
+        new_block_shape.append(bd if bd is not None else 1)
+        continue
+
       if not isinstance(bd, (int, pallas_core.Blocked)):
         raise NotImplementedError('reshape merge must use `Blocked` block size')
 
-      num_blocks = pallas_utils.cdiv(d, bd)
+      num_blocks = pallas_utils.cdiv(d, bs)
       new_block_dims = []
       for md in reversed(merged):
-        if bd % md == 0:
+        if bs % md == 0:
           new_block_dims.append(md)
-          bd //= md
-        elif md % bd == 0:
-          new_block_dims.append(bd)
-          bd = 1
+          bs //= md
+        elif md % bs == 0:
+          new_block_dims.append(bs)
+          bs = 1
         else:
           raise NotImplementedError('unsupported reshape merge')
 
       new_block_dims.reverse()
       new_block_shape.extend(new_block_dims)
-      new_grid = [np.int32(md // bd) for md, bd in zip(merged, new_block_dims)]
+      new_grid = [
+          np.int32(md // pallas_core.get_block_size(bd))
+          for md, bd in zip(merged, new_block_dims)
+      ]
       new_grids.append(tuple(new_grid))
 
       if np.prod(new_grid) != num_blocks:
@@ -1679,6 +1716,7 @@ def _reshape_pull_rule(
           f' {block_shape}'
       )
     new_block_shape = block_shape[:-2] + (total_block_size,)
+
     def new_index_map(*args):  # pylint: disable=function-redefined
       *idx, second_to_last, last = block_spec.index_map(*args)
       # last should always be 0
@@ -1687,6 +1725,7 @@ def _reshape_pull_rule(
             'Must select entire block on last dimension for reshape'
         )
       return *idx, second_to_last
+
     return [pallas_core.BlockSpec(new_block_shape, new_index_map)]
 
   raise NotImplementedError(f'reshape not supported yet: {aval_in}, {aval_out}')
@@ -1707,6 +1746,63 @@ def _reshape_eval_rule(
   return x
 
 
+@register_pull_block_spec_rule(lax.reduce_sum_p)
+def _reduce_sum_pull_rule(
+    ctx: PullRuleContext,
+    block_spec: pallas_core.BlockSpec,
+    *,
+    axes: tuple[int, ...],
+    out_sharding,
+):
+  aval_in = ctx.avals_in[0]
+  assert isinstance(aval_in, core.ShapedArray)
+  new_block_shape = []
+  block_shape = iter(block_spec.block_shape)
+  for i, d in enumerate(aval_in.shape):
+    if i in axes:
+      new_block_shape.append(pallas_core.Blocked(d))
+    else:
+      new_block_shape.append(next(block_shape))
+  assert next(block_shape, None) is None
+
+  def new_index_map(*args):
+    idx = block_spec.index_map(*args)
+    new_idx = []
+    idx_iter = iter(idx)
+    for i in range(len(aval_in.shape)):
+      if i in axes:
+        new_idx.append(0)
+      else:
+        new_idx.append(next(idx_iter))
+    assert next(idx_iter, None) is None
+    return tuple(new_idx)
+
+  new_block_spec = block_spec.replace(
+      block_shape=tuple(new_block_shape), index_map=new_index_map
+  )
+  return [new_block_spec]
+
+
+@register_eval_rule(lax.reduce_sum_p)
+def _reduce_sum_eval_rule(
+    ctx: KernelEvalContext,
+    x,
+    *,
+    axes: tuple[int, ...],
+    out_sharding,
+):
+  aval_in = ctx.avals_in[0]
+  assert isinstance(aval_in, core.ShapedArray)
+  block_shape = tuple(ctx.in_block_specs[0].block_shape)
+  for i in axes:
+    if _block_size(block_shape[i]) != aval_in.shape[i]:
+      raise NotImplementedError(
+          f'reduce_sum on partial blocks not supported: {aval_in=},'
+          f' {block_shape=}'
+      )
+  return jax.lax.reduce_sum(x, axes=axes)
+
+
 # Higher order primitives
 
 
@@ -1902,6 +1998,20 @@ def _custom_vjp_call_pull_block_spec_rule(
   return in_block_specs
 
 
+@register_pull_block_spec_rule(hijax.call_hi_primitive_p)
+def _custom_call_hi_primitive_pull_block_spec_rule(
+    ctx: PullRuleContext, out_block_specs, *, prim
+):
+  return prim.pull_block_spec_rule(ctx, out_block_specs)
+
+@register_eval_rule(hijax.call_hi_primitive_p)
+def _custom_call_hi_primitive_eval_rule(
+    ctx: KernelEvalContext, *args, prim
+):
+  return jax.tree.leaves(prim.block_eval_rule(ctx, *args))
+
+
+@functools.partial(api_boundary, repro_api_name="fuser.push_block_spec")
 def push_block_spec(
     f: Callable,
     *in_spec_args,
@@ -2022,6 +2132,23 @@ def _binop_push_rule(
   left_aval, right_aval = ctx.avals_in
   assert isinstance(left_aval, core.ShapedArray)
   assert isinstance(right_aval, core.ShapedArray)
+  if not right_aval.shape:
+    return left_block_spec
+  if not left_aval.shape:
+    return right_block_spec
+  lhs_has_block_spec = left_block_spec is not pallas_core.no_block_spec
+  rhs_has_block_spec = right_block_spec is not pallas_core.no_block_spec
+  if not (lhs_has_block_spec ^ rhs_has_block_spec):
+    # We can only do a push if one of the block specs is unspecified
+    # or they are identical.
+    if left_block_spec is right_block_spec:
+      return left_block_spec
+    raise ValueError('Illegal binary push. One of the block specs must be no_block_spec.')
+  for l, r in zip(left_aval.shape, right_aval.shape, strict=True):
+    if l == 1 and r != 1 and lhs_has_block_spec:
+      raise ValueError('Cannot propagate block spec through LHS broadcast.')
+    if r == 1 and l != 1 and rhs_has_block_spec:
+      raise ValueError('Cannot propagate block spec through RHS broadcast.')
   if left_block_spec is pallas_core.no_block_spec:
     return right_block_spec
   if right_block_spec is pallas_core.no_block_spec:
@@ -2066,7 +2193,7 @@ def _transpose_push_rule(
 ) -> pallas_core.BlockSpec:
   del ctx
   block_shape = block_spec.block_shape
-  new_shape = [block_shape[i] for i in permutation]
+  new_shape = tuple(block_shape[i] for i in permutation)
   if set(permutation[-2:]) != {permutation[-1], permutation[-2]}:
     raise NotImplementedError(
         'Cannot permute last two dimensions with leading dimensions.'
@@ -2131,6 +2258,12 @@ def _custom_vjp_call_push_rule(
   del ctx, num_consts, fwd_jaxpr_thunk, bwd, out_trees, symbolic_zeros
   return _push_block_spec_jaxpr(call_jaxpr.jaxpr, *block_specs)
 
+@register_push_block_spec_rule(hijax.call_hi_primitive_p)
+def _custom_call_hi_primitive_push_block_spec_rule(
+    ctx: PullRuleContext, *block_specs, prim
+):
+  return prim.push_block_spec_rule(ctx, block_specs)
+
 
 @register_push_block_spec_rule(pjit.jit_p)
 def _pjit_push_rule(ctx, *block_specs, jaxpr: core.ClosedJaxpr, **_):
@@ -2155,9 +2288,11 @@ register_eltwise_rule(lax.sin_p)
 register_eltwise_rule(lax.cos_p)
 register_eltwise_rule(lax.sqrt_p)
 register_eltwise_rule(lax.rsqrt_p)
+register_eltwise_rule(lax.square_p)
 register_eltwise_rule(lax.log_p)
 register_eltwise_rule(lax.integer_pow_p)
 
+
 @register_push_block_spec_rule(lax.reshape_p)
 def _reshape_push_rule(
     ctx: PullRuleContext,
@@ -2199,3 +2334,127 @@ def _reshape_push_rule(
 
     return pallas_core.BlockSpec(new_block_shape, new_index_map)
   raise NotImplementedError(f'reshape not supported yet: {aval_in}, {aval_out}')
+
+
+@register_push_block_spec_rule(lax.reduce_sum_p)
+def _reduce_sum_push_rule(
+    ctx: PushRuleContext,
+    block_spec: pallas_core.BlockSpec,
+    *,
+    axes: tuple[int, ...],
+    out_sharding,
+):
+  del out_sharding
+  aval_in = ctx.avals_in[0]
+  assert isinstance(aval_in, core.ShapedArray)
+  if not all(
+      aval_in.shape[i] == pallas_core.get_block_size(block_spec.block_shape[i])
+      for i in axes
+  ):
+    raise NotImplementedError(
+        f'reduce_sum over partial blocks not supported yet: {aval_in.shape=},'
+        f' {block_spec.block_shape=}, {axes=}'
+    )
+  new_block_shape = tuple(
+      bd for i, bd in enumerate(block_spec.block_shape) if i not in axes
+  )
+
+  def new_index_map(*args):
+    idx = block_spec.index_map(*args)
+    return tuple(idx[i] for i in range(len(idx)) if i not in axes)
+
+  return block_spec.replace(
+      block_shape=tuple(new_block_shape), index_map=new_index_map
+  )
+
+
+@register_push_block_spec_rule(lax.broadcast_in_dim_p)
+def _broadcast_in_dim_push_rule(
+    ctx: PushRuleContext,
+    block_spec: pallas_core.BlockSpec,
+    *,
+    shape: tuple[int, ...],
+    broadcast_dimensions: tuple[int, ...],
+    sharding: jax.sharding.Sharding,
+):
+  del sharding
+  in_aval = ctx.avals_in[0]
+  assert isinstance(in_aval, core.ShapedArray)
+  in_shape = in_aval.shape
+
+  dim_map = {
+      out_dim: in_dim
+      for in_dim, out_dim in enumerate(broadcast_dimensions)
+  }
+
+  new_block_shape = []
+  for i, s in enumerate(shape):
+    if i in dim_map:
+      in_dim = dim_map[i]
+      if in_shape[in_dim] != s:
+        assert pallas_core.get_block_size(block_spec.block_shape[in_dim]) == 1
+        new_block_shape.append(s)
+      else:
+        new_block_shape.append(block_spec.block_shape[in_dim])
+    else:
+      new_block_shape.append(s)
+
+  def new_index_map(*args):
+    idx = block_spec.index_map(*args)
+    return tuple(
+        idx[dim_map[i]] if i in dim_map else 0 for i in range(len(shape))
+    )
+
+  return pallas_core.BlockSpec(tuple(new_block_shape), new_index_map)
+
+
+@register_push_block_spec_rule(lax.concatenate_p)
+def _concatenate_push_rule(
+    ctx: PushRuleContext,
+    *block_specs: pallas_core.BlockSpec,
+    dimension: int,
+):
+  avals_in = ctx.avals_in
+  block_shapes = [
+      pallas_core._canonicalize_block_shape(block_spec.block_shape)
+      for block_spec in block_specs
+  ]
+  # We only support concatenation if the entirety of the concat dimension is blocked.
+  assert all(hasattr(aval_in, 'shape') for aval_in in avals_in)
+  if not all(
+      block_shape[dimension] == pallas_core.Blocked(avals_in.shape[dimension])  # pytype: disable=attribute-error
+      for block_shape, avals_in in zip(block_shapes, avals_in)
+  ):
+    raise NotImplementedError(
+        f'concatenate not supported yet: {block_shapes=}, {avals_in=}'
+    )
+  def _new_index_map(*args):
+    all_indices = [block_spec.index_map(*args) for block_spec in block_specs]
+    # This is a very important check. We cannot actually construct a single BlockSpec
+    # for the output of concatenate if the indices are not identical across all the
+    # inputs. This is not something we can always enforce statically, but to be conservative
+    # we apply a very aggressive check. We can consider relaxing this later.
+    if not all(
+        (all_indices[0][i] is all_indices[j][i])
+        for i in range(len(all_indices[0]))
+        for j in range(len(all_indices))
+    ):
+      raise ValueError(
+          'Cannot statically prove that all input blocks to concatenate are the'
+          ' same.'
+      )
+    # If all block indices are the same, we are materializing the full concatenation along
+    # the concat dimension, so we use index 0.
+    base_indices = list(all_indices[0])
+    base_indices[dimension] = 0
+    return tuple(base_indices)
+
+  new_block_shape = list(block_specs[0].block_shape)
+  # Since the entirety of the concat dimension is materialized in the blocks,
+  # the new block size is the sum of the block sizes of the inputs along that
+  # dimension.
+  new_block_shape[dimension] = sum(
+      pallas_core.get_block_size(block_shape[dimension])
+      for block_shape in block_shapes
+  )
+  return pallas_core.BlockSpec(tuple(new_block_shape), _new_index_map)
