@@ -20,9 +20,10 @@ from contextlib import contextmanager
 import dataclasses
 import enum
 import functools
-from typing import Any, Union
+from typing import Any, Literal, Union
 
 import jax
+from jax import core as jax_core
 from jax import lax
 from jax import tree_util
 from jax._src import util as jax_util
@@ -31,15 +32,16 @@ from jax._src.pallas import primitives as primitives
 from jax._src.pallas.mosaic import core as tpu_core
 from jax._src.pallas.mosaic import helpers as tpu_helpers
 from jax._src.pallas.mosaic import primitives as tpu_primitives
+from jax._src.pallas.mosaic import tpu_info
+from jax._src.state import types as state_types
 from jax.experimental import pallas as pl
-from jax.extend.backend import get_default_device
 import jax.numpy as jnp
-import numpy as np
 
 
 SMEM = tpu_core.MemorySpace.SMEM
 VMEM = tpu_core.MemorySpace.VMEM
-ANY = tpu_core.MemorySpace.ANY
+HBM = tpu_core.MemorySpace.HBM
+ANY = pallas_core.MemorySpace.ANY
 REF = pallas_core.MemoryRef
 GridDimensionSemantics = tpu_core.GridDimensionSemantics
 PARALLEL = tpu_core.PARALLEL
@@ -76,28 +78,35 @@ def _broadcast_pytree_to(from_pytree, to_pytree):
   return tree_util.tree_unflatten(treedef, broadcast_leaves)
 
 
-@jax_util.cache(trace_context_in_key=False)
 def _get_tpu_generation() -> int:
-  kind = get_default_device().device_kind
-  if kind.endswith(' lite'):
-    kind = kind[:-len(' lite')]
-  if kind.startswith("TPU v"):
-    return int(kind[5])
-  else:
-    assert "TPU7x" in kind
-    return 7
+  return tpu_info.get_tpu_info().generation
+
+
+def _make_tiling(
+    shape: tuple[int, ...], ty: jax_core.AbstractValue
+) -> tuple[int | None, ...]:
+  """Compute a tiling for the given shape and type.
+
+  For a n-dimensional shape, returns (8, 128) for the last 2 dimensions
+  and 1 for the leading n - 2. For example, (256, 256) -> (8, 128) and
+  (2, 3, 128, 128) -> (1, 1, 8, 128).
+
+  Types are not required to have a dtype, so for such types we return None for
+  all dimensions because their tiling is unknown.
+  """
 
-def _make_tiling(shape: tuple[int, ...], dtype: np.dtype) -> tuple[int, ...]:
-  # For a n-dimensional shape, returns (8, 128) for the last 2 dimensions
-  # and 1 for the leading n - 2. For example, (256, 256) -> (8, 128) and
-  # (2, 3, 128, 128) -> (1, 1, 8, 128).
   if len(shape) < 2:
     raise ValueError(f"Shape must have at least 2 dimensions: {shape=}")
+
+  if not hasattr(ty, 'dtype'):
+    return (None,) * len(shape)
+
   leading_dims, final_dims = shape[:-2], shape[-2:]
   # We want to find the minimum power of 2 that fits the second-minor dimension
   # of shape, with maximum value 8.
   second_minor, _ = final_dims
-  packing = 4 // dtype.itemsize
+
+  packing = 4 // ty.dtype.itemsize
   max_tiling = _TILING[0]
   second_minor_tiling = (1 + int(_get_tpu_generation() < 4)) * packing
   while second_minor_tiling < min(second_minor, max_tiling):
@@ -122,13 +131,18 @@ def _make_block_ds(
   assert isinstance(out, pl.Slice)
   return out
 
-def _create_blocked_slice(block_index: jax.Array | int,
-                          block_size: int,
-                          dim_size: int,
-                          tiling: int):
+
+def _create_blocked_slice(
+    block_index: jax.Array | int,
+    block_size: int,
+    dim_size: int,
+    tiling: int | None,
+):
   block_start = block_size * block_index
   if (dim_rem := dim_size % block_size) == 0:
     return pl.ds(block_start, block_size)
+  if tiling is None:
+    raise ValueError("If tiling is None, block_size must divide dim_size.")
   if block_size % tiling != 0:
     raise ValueError(f"Block size must divide tiling: {block_size=}, {tiling=}")
   num_blocks = pl.cdiv(dim_size, block_size)
@@ -145,12 +159,15 @@ def _create_bounded_slice(slice_start: jax.Array | int,
                           slice_size: jax.Array | int,
                           block_size: int,
                           dim_size: int,
-                          tiling: int):
-  if block_size % tiling != 0:
+                          tiling: int | None):
+  if tiling is not None and block_size % tiling != 0:
     raise ValueError(f"Block size must divide tiling: {block_size=}, {tiling=}")
   # We assume by construction that slice_size <= block_size. We also assume
   # that the slice_start is already aligned to the tiling.
 
+  if tiling is None:
+    return pl.ds(slice_start, slice_size)
+
   # If we are out of bound, we need to round the slice size down to the nearest
   # multiple of the tiling.
   is_oob = slice_start + slice_size > dim_size
@@ -165,7 +182,7 @@ def _create_bounded_slice(slice_start: jax.Array | int,
 
 def _make_block_slice(
     block_index: jax.Array, block_size: pl.BlockDim | int | None, size: int,
-    tiling: int
+    tiling: int | None
 ) -> pl.Slice | slice | int | jax.Array:
   # Computes a slice given a block index and block size. In the default case,
   # we return slice(block_index * block_size, (block_index + 1) * block_size).
@@ -340,7 +357,7 @@ class BufferedRefBase:
   def compute_index(self):
     return self.spec.index_map
 
-  def get_dma_slice(self, src_shape, src_dtype, grid_indices):
+  def get_dma_slice(self, src_ty, grid_indices):
     # We need to handle blocks that might go OOB in the src array. An in bounds
     # block looks like this (for array shape (600, 600) and block shape
     # (256, 256)):
@@ -387,10 +404,14 @@ class BufferedRefBase:
     # Suppose A is now (601, 600), instead of picking a (88, 256)-sized block
     # for the last iteration on that dimension, we will pick the next highest
     # tile multiple, i.e. (96, 256).
+
+    if (src_shape := getattr(src_ty, "shape", None)) is None:
+      raise ValueError(f'Type {src_ty} does not have a type.')
+
     if len(src_shape) < 2:
       raise NotImplementedError("Must use >1D values.")
 
-    tiling = _make_tiling(src_shape, src_dtype)
+    tiling = _make_tiling(src_shape, src_ty)
     block_indices = self.compute_index(*grid_indices)
     return tuple(
         _make_block_slice(bi, bs, ss, t)
@@ -411,17 +432,24 @@ class BufferedRefBase:
     """Returns a new BufferedRefBase with the given block spec."""
     raise NotImplementedError()
 
+def _ref_to_value_aval(ref):
+  """Return the inner of a ref, or a ShapedArray for TransformedRefs."""
+  return (
+      jax_core.ShapedArray(shape=ref.shape, dtype=ref.dtype)
+      if isinstance(ref, state_types.TransformedRef)
+      else jax.typeof(ref).inner_aval
+  )
+
 
 # TODO(justinfu): Refactor and rename slot fields to reflect cumulative values
 # instead of slot index.
-@tree_util.register_pytree_node_class
+@tree_util.register_dataclass
 @dataclasses.dataclass(frozen=True)
 class BufferedRef(BufferedRefBase):
   """A helper class to automate VMEM double buffering in pallas pipelines.
 
   Attributes:
     spec: pallas blockspec.
-    dtype: dtype for buffers.
     buffer_type: enum indicating whether this is an input, output, or in/out
       accumulator buffered reference.
     window_ref: a multiple-buffer to hold the working and dirty buffers used
@@ -451,9 +479,8 @@ class BufferedRef(BufferedRefBase):
     swap: Tracks whether the BufferedRef slots need to be swapped before next
       copy.
   """
-  _spec: pl.BlockSpec       # static metadata
-  dtype: Any                # static metadata
-  _buffer_type: BufferType  # static metadata
+  _spec: pl.BlockSpec = dataclasses.field(metadata=dict(static=True))
+  _buffer_type: BufferType = dataclasses.field(metadata=dict(static=True))
   window_ref: ArrayRef | None
   accum_ref: ArrayRef | None
   copy_in_slot: ArrayRef | None
@@ -510,47 +537,28 @@ class BufferedRef(BufferedRefBase):
       raise ValueError("buffer count is undefined")
     return self.window_ref.shape[0]  # type: ignore[union-attr]
 
-  def tree_flatten(self):
-    return (
-        (
-            self.window_ref,
-            self.accum_ref,
-            self.copy_in_slot,
-            self.wait_in_slot,
-            self.copy_out_slot,
-            self.wait_out_slot,
-            self._copy_in_slot_reg,
-            self._wait_in_slot_reg,
-            self._copy_out_slot_reg,
-            self._wait_out_slot_reg,
-            self.next_fetch_smem,
-            self.next_fetch_sreg,
-            self.sem_recvs,
-            self.sem_sends,
-            self.swap,
-        ),
-        (self._spec, self.dtype, self._buffer_type),
-    )
-
-  @classmethod
-  def tree_unflatten(cls, meta, data):
-    return cls(*meta, *data)
-
   @staticmethod
   def buffer_types() -> type[BufferType]:
     return BufferType
 
   @classmethod
-  def create(cls, spec: pl.BlockSpec, dtype, buffer_type, buffer_count,
-             needs_swap_ref=True,
-             grid_rank=None,
-             use_lookahead=False,
-             source_memory_space: tpu_core.MemorySpace = ANY) -> BufferedRef:
+  def create(
+      cls,
+      spec: pl.BlockSpec,
+      dtype_or_type,
+      buffer_type,
+      buffer_count,
+      needs_swap_ref=True,
+      grid_rank=None,
+      use_lookahead=False,
+      source_memory_space: tpu_core.MemorySpace | Literal[ANY] = ANY,  # type: ignore[valid-type]
+  ) -> BufferedRef:
     """Create a BufferedRef.
 
     Args:
       spec: pallas blockspec.
-      dtype: dtype for buffers.
+      dtype_or_type: dtype or aval for buffers. If an aval, the shape is
+        ignored.
       buffer_type: enum indicating whether this is an input, output, or in/out
         accumulator buffered reference.
       needs_swap_ref: whether a swap slots tracker needs to be allocated.
@@ -561,21 +569,29 @@ class BufferedRef(BufferedRefBase):
     Returns:
       Initialized BufferedRef
     """
+
+    # (123, 456) is a dummy shape since we never use ty without
+    # calling .update(shape=...) first.
+    ty = (
+        dtype_or_type
+        if isinstance(dtype_or_type, jax_core.AbstractValue)
+        else jax_core.ShapedArray((123, 456), dtype_or_type)
+    )
+
     block_shape = _get_block_shape(spec)
     if buffer_type is BufferType.ACCUMULATOR:
-      accum_ref = VMEM(block_shape, dtype)
+      accum_ref = VMEM.from_type(ty.update(shape=block_shape))
     else:
       accum_ref = None
-    if source_memory_space == VMEM:
-      # We don't need to do any double-buffering in the case that our pipeline
-      # reference is already in VMEM, we just need allocate the accumulation
-      # buffer and we will refer to the original reference slices directly.
-      if spec.memory_space not in (VMEM, None):
-        raise ValueError(
-            f"Cannot hold a non-buffered ref in {spec.memory_space=}")
+    buffer_memory_space = (
+          VMEM if spec.memory_space is None else spec.memory_space)
+    if buffer_memory_space not in (SMEM, VMEM, HBM):
+      raise ValueError(
+          f"Unsupported buffer memory space: {buffer_memory_space}"
+      )
+    if source_memory_space is buffer_memory_space:
       return cls(
           _spec=spec,
-          dtype=dtype,
           _buffer_type=buffer_type,
           window_ref=None,  # to be bound to existing ref by the pipeline routine
           accum_ref=accum_ref,
@@ -594,21 +610,16 @@ class BufferedRef(BufferedRefBase):
           swap=None,
       )
     else:
-      buffer_memory_space = (
-          VMEM if spec.memory_space is None else spec.memory_space)
-      if buffer_memory_space not in (SMEM, VMEM):
-        raise ValueError(
-            f"Unsupported buffer memory space: {buffer_memory_space}"
-        )
       if use_lookahead and grid_rank is None:
         raise ValueError(
             "grid_rank must be specified when use_lookahead is True."
         )
+
+      buffer_ty = ty.update(shape=(buffer_count, *block_shape))
       return cls(
           _spec=spec,
-          dtype=dtype,
           _buffer_type=buffer_type,
-          window_ref=buffer_memory_space((buffer_count,) + block_shape, dtype),
+          window_ref=buffer_memory_space.from_type(buffer_ty),
           accum_ref=accum_ref,
           copy_in_slot=SMEM((1,), jnp.uint32) if buffer_type.is_input else None,
           wait_in_slot=SMEM((1,), jnp.uint32) if buffer_type.is_input else None,
@@ -635,22 +646,28 @@ class BufferedRef(BufferedRefBase):
       )
 
   @classmethod
-  def input(cls, spec, dtype, buffer_count=2, **kwargs):
-    return cls.create(spec, dtype, BufferType.INPUT, buffer_count, **kwargs)
+  def input(cls, spec, dtype_or_type, buffer_count=2, **kwargs):
+    return cls.create(
+        spec, dtype_or_type, BufferType.INPUT, buffer_count, **kwargs
+    )
 
   @classmethod
-  def output(cls, spec, dtype, buffer_count=2, **kwargs):
-    return cls.create(spec, dtype, BufferType.OUTPUT, buffer_count, **kwargs)
+  def output(cls, spec, dtype_or_type, buffer_count=2, **kwargs):
+    return cls.create(
+        spec, dtype_or_type, BufferType.OUTPUT, buffer_count, **kwargs
+    )
 
   @classmethod
-  def accumulator(cls, spec, dtype, buffer_count=2, **kwargs):
-    return cls.create(spec, dtype, BufferType.ACCUMULATOR, buffer_count,
-                      **kwargs)
+  def accumulator(cls, spec, dtype_or_type, buffer_count=2, **kwargs):
+    return cls.create(
+        spec, dtype_or_type, BufferType.ACCUMULATOR, buffer_count, **kwargs
+    )
 
   @classmethod
-  def input_output(cls, spec, dtype, buffer_count=2, **kwargs):
-    return cls.create(spec, dtype, BufferType.INPUT_OUTPUT, buffer_count,
-                      **kwargs)
+  def input_output(cls, spec, dtype_or_type, buffer_count=2, **kwargs):
+    return cls.create(
+        spec, dtype_or_type, BufferType.INPUT_OUTPUT, buffer_count, **kwargs
+    )
 
   @property
   def block_shape(self):
@@ -957,7 +974,7 @@ class BufferedRef(BufferedRefBase):
     if self.swap is not None:
       self.swap[0] = True
     slot = self.current_copy_in_slot
-    src_slice = self.get_dma_slice(src_ref.shape, src_ref.dtype, grid_indices)
+    src_slice = self.get_dma_slice(_ref_to_value_aval(src_ref), grid_indices)
     dst_slice = tuple(
         pl.ds(0, s.size)
         for s, bd in zip(src_slice, self.block_shape)
@@ -978,7 +995,7 @@ class BufferedRef(BufferedRefBase):
     if self.swap is not None:
       self.swap[0] = True
     slot = self.current_copy_out_slot
-    dst_slice = self.get_dma_slice(dst_ref.shape, dst_ref.dtype, grid_indices)
+    dst_slice = self.get_dma_slice(_ref_to_value_aval(dst_ref), grid_indices)
     src_slice = tuple(
         pl.ds(0, s.size)
         for s, bd in zip(dst_slice, self.block_shape)
@@ -996,7 +1013,7 @@ class BufferedRef(BufferedRefBase):
     if not self.is_buffered: return
     assert not (self.window_ref is None or isinstance(self.window_ref, REF))
     assert self.sem_recvs is not None
-    src_slice = self.get_dma_slice(src_ref.shape, src_ref.dtype, grid_indices)
+    src_slice = self.get_dma_slice(_ref_to_value_aval(src_ref), grid_indices)
     dst_slice = tuple(
         pl.ds(0, s.size)
         for s, bd in zip(src_slice, self.block_shape)
@@ -1018,7 +1035,7 @@ class BufferedRef(BufferedRefBase):
     assert not (self.window_ref is None or isinstance(self.window_ref, REF))
     assert self.sem_sends is not None
     wait_slot = self.current_wait_out_slot
-    dst_slice = self.get_dma_slice(dst_ref.shape, dst_ref.dtype, grid_indices)
+    dst_slice = self.get_dma_slice(_ref_to_value_aval(dst_ref), grid_indices)
     src_slice = tuple(
         pl.ds(0, s.size)
         for s, bd in zip(dst_slice, self.block_shape)
@@ -1068,15 +1085,25 @@ class BufferedRef(BufferedRefBase):
       ).astype(self.window_ref.dtype)
 
 
-def fetch_until_full_slots(buffered_ref, src_ref,
-                           grid,
-                           grid_offsets,
-                           predicate: jax.Array | bool = True,
-                           use_sreg_for_state: bool = False,
-                           update_slots: bool = True):
-  """Continually fetch values until all copy_in slots are full."""
-  if not buffered_ref.use_lookahead:
-    raise ValueError("fetch_until_full_slots requires lookahead.")
+def fetch_with_lookahead(buffered_ref, src_ref,
+                         grid,
+                         grid_offsets,
+                         predicate: jax.Array | bool = True,
+                         max_num_fetches: int | None = None,
+                         update_slots: bool = True):
+  """Fetch future blocks using unbounded lookahead.
+
+  Args:
+    buffered_ref: the BufferedRef to fetch for.
+    src_ref: the source Ref.
+    grid: the grid bounds.
+    grid_offsets: the grid offsets (used for megacore).
+    predicate: a boolean predicate for whether to perform the fetch.
+    max_num_fetches: the maximum number of fetches to perform. If None,
+      this will continually fetch until all copy_in slots are full.
+    update_slots: whether to update the register slot indices.
+  """
+  assert buffered_ref.use_lookahead
   add_offset = lambda x: tuple(
       i + j for i, j in zip(x, grid_offsets, strict=True))
   index_inbound = lambda x: _tuple_lt(x, grid)
@@ -1087,12 +1114,22 @@ def fetch_until_full_slots(buffered_ref, src_ref,
     else:
       return x.astype(jnp.uint32)
 
+  fetch_limit = buffered_ref.cumulative_wait_in + buffered_ref.buffer_count
+  if max_num_fetches is not None:
+    fetch_once_limit = buffered_ref.cumulative_copy_in + max_num_fetches
+    # We would like to write jnp.minimum(fetch_limit, fetch_once_limit)
+    # but this does not compile in Mosaic.
+    fetch_limit = lax.select(fetch_limit < fetch_once_limit,
+                             fetch_limit, fetch_once_limit)
+
+
   def _loop_cond(carry):
     _, next_indices, cumulative_copy_in = carry
-    has_slots = cumulative_copy_in < (
-        buffered_ref.cumulative_wait_in + buffered_ref.buffer_count)
+    # Don't fetch more blocks than we have buffers.
+    within_limit = cumulative_copy_in < fetch_limit
+    # Don't fetch past the end of the grid.
     in_bounds = index_inbound(next_indices)
-    return predicate & has_slots & in_bounds
+    return predicate & within_limit & in_bounds
 
   def _loop_body(carry):
     current_indices, next_indices, cumulative_copy_in = carry
@@ -1117,11 +1154,7 @@ def fetch_until_full_slots(buffered_ref, src_ref,
 
   buffered_ref = buffered_ref.with_next_fetch(final_indices)
   if update_slots:
-    if use_sreg_for_state:
-      buffered_ref = buffered_ref.with_slot_index(
-          copy_in_slot=final_copy_in_slot)
-    else:
-      buffered_ref.copy_in_slot[0] = final_copy_in_slot
+    buffered_ref = buffered_ref.with_slot_index(copy_in_slot=final_copy_in_slot)
   return buffered_ref, final_copy_in_slot
 
 
@@ -1216,7 +1249,6 @@ class Scheduler:
       last_cycle=None,
       init_accumulators=None,
       trace_scopes=True,
-      use_sreg_for_state: bool = False,
   ):
     """Initializes scheduler.
 
@@ -1231,8 +1263,6 @@ class Scheduler:
       init_accumulators: do we zero-initialize accumulator state for this
         invocation of the pipeline.
       trace_scopes: whether to use named_scope to trace blocks in the pipeline.
-      use_sreg_for_state: optional bool, indicates whether to use sregs for
-        current_slot state.
     """
     self.step = step
     self.grid = grid
@@ -1242,7 +1272,6 @@ class Scheduler:
     self.last_cycle = last_cycle
     self.init_accumulators = init_accumulators
     self.trace_scopes = trace_scopes
-    self.use_sreg_for_state = use_sreg_for_state
 
     # Total number of linear steps.
     self.num_steps = _grid_size(grid)
@@ -1301,7 +1330,7 @@ class Scheduler:
     # Currently this is based on the iteration, but if we want to support
     # lookahead this will depend on whether the lookahead reached the end.
     if not buffered_ref.is_buffered:
-      return False
+      return jnp.bool(False)
     return self.step >= (self.num_steps - buffered_ref.buffer_count + 1)
 
   def has_changed(self, buffered_ref):
@@ -1340,50 +1369,59 @@ class Scheduler:
   # Below is the sequence of conditional waits and copies used for inputs,
   # outputs, and in-out accumulators.
 
-  def initialize(self, buffered_ref, src_ref, schedule=None):
+  def initialize_step(self, buffered_ref, src_ref, schedule=None, step=0):
     if schedule is None:
       schedule = _default_schedule
+    # TODO(justinfu): Should cache this, but it doesn't actually do computation
+    # in both default & fixed schedules right now so it doesn't increase
+    # the Jaxpr size.
     do_copy = schedule["prologue_copy_in"](self, buffered_ref, src_ref)
 
-    with self._named_scope("ep_initialize"):
-      @pl.when(self.first_step_ever)
-      def _init_slots():
-        buffered_ref.init_slots()
-
-      if self.use_sreg_for_state:
+    with self._named_scope(f"ep_initialize_{step}"):
+      if step == 0:
+        @pl.when(self.first_step_ever)
+        def _init_slots():
+          buffered_ref.init_slots()
         buffered_ref = buffered_ref.load_slots()
 
       if not buffered_ref.is_input or not buffered_ref.is_buffered:
         return buffered_ref
 
-      if buffered_ref.use_lookahead:
-        @pl.when(self.first_step_ever & do_copy)
-        def _start():
-          buffered_ref.copy_in(src_ref,
-            self.add_offset(buffered_ref.next_fetch_indices))  # pylint: disable=cell-var-from-loop
-        buffered_ref = buffered_ref.advance_copy_in_slot(do_copy)
+      if (step + 1) >= buffered_ref.buffer_count:
+        return buffered_ref
 
-        buffered_ref, _ = fetch_until_full_slots(buffered_ref,
-                                              src_ref, self.grid,
-                                              self.grid_offsets,
-                                              self.first_step_ever & do_copy,
-                                              self.use_sreg_for_state)
-      else:
-        grid_indices = self.indices
-        @pl.when(do_copy)
-        def _start():
-          buffered_ref.copy_in(src_ref, grid_indices)
-        buffered_ref = buffered_ref.advance_copy_in_slot(do_copy)
-        for i in range(buffered_ref.buffer_count - 2):
-          next_grid_indices = self.fetch_indices[i+1]
-          block_indices = buffered_ref.compute_index(*grid_indices)
-          next_block_indices = buffered_ref.compute_index(*next_grid_indices)
-          will_change = _tuples_differ(block_indices, next_block_indices)
-          @pl.when(do_copy & will_change)  # pylint: disable=cell-var-from-loop
+      if buffered_ref.use_lookahead:
+        if step == 0:
+          # We always fetch the first block.
+          @pl.when(do_copy)
           def _start():
-            buffered_ref.copy_in(src_ref, next_grid_indices)  # pylint: disable=cell-var-from-loop
-          grid_indices = next_grid_indices
-          buffered_ref = buffered_ref.advance_copy_in_slot(do_copy & will_change)
+            buffered_ref.copy_in(src_ref,
+              self.add_offset(buffered_ref.next_fetch_indices))  # pylint: disable=cell-var-from-loop
+          buffered_ref = buffered_ref.advance_copy_in_slot(do_copy)
+        else:
+          buffered_ref, _ = fetch_with_lookahead(
+              buffered_ref,
+              src_ref,
+              self.grid,
+              self.grid_offsets,
+              predicate=self.first_step_ever & do_copy,
+              max_num_fetches=1,
+          )
+      else:
+        if step == 0:
+          predicate = do_copy
+          fetch_indices = self.fetch_indices[step]
+        else:
+          fetch_indices = self.fetch_indices[step]
+          prev_grid_indices = self.fetch_indices[step - 1]
+          block_indices = buffered_ref.compute_index(*fetch_indices)
+          prev_block_indices = buffered_ref.compute_index(*prev_grid_indices)
+          block_changed = _tuples_differ(block_indices, prev_block_indices)
+          predicate = do_copy & block_changed
+        @pl.when(predicate)  # pylint: disable=cell-var-from-loop
+        def _start():
+          buffered_ref.copy_in(src_ref, fetch_indices)  # pylint: disable=cell-var-from-loop
+        buffered_ref = buffered_ref.advance_copy_in_slot(predicate)
     return buffered_ref
 
   def wait_in(self, buffered_ref, src_ref, schedule=None) -> "BufferedRef":
@@ -1420,9 +1458,9 @@ class Scheduler:
       return buffered_ref
 
     if buffered_ref.use_lookahead:
-      buffered_ref, _ = fetch_until_full_slots(buffered_ref,
-        src_ref, self.grid, self.grid_offsets, predicate=True,
-        use_sreg_for_state=self.use_sreg_for_state)
+      buffered_ref, _ = fetch_with_lookahead(
+          buffered_ref, src_ref, self.grid, self.grid_offsets, predicate=True
+      )
     else:
       @pl.when(pred)
       @self._named_scope("ep_copy_in")
@@ -1454,18 +1492,18 @@ class Scheduler:
             src_ref, self.add_offset(buffered_ref.next_fetch_sreg))  # pylint: disable=cell-var-from-loop
       buffered_ref = buffered_ref.advance_copy_in_slot(pred)
 
-      buffered_ref, final_copy_in_slot = fetch_until_full_slots(
-          buffered_ref, src_ref, self.grid, self.grid_offsets,
-          predicate=pred, use_sreg_for_state=self.use_sreg_for_state,
-          update_slots=False)
+      buffered_ref, final_copy_in_slot = fetch_with_lookahead(
+          buffered_ref,
+          src_ref,
+          self.grid,
+          self.grid_offsets,
+          predicate=pred,
+          update_slots=False,
+      )
       @pl.when(pred)
       def _():
-        if self.use_sreg_for_state:
-          bref = buffered_ref.with_slot_index(copy_in_slot=final_copy_in_slot)
-          bref.save_slots()
-        else:
-          buffered_ref.copy_in_slot[0] = final_copy_in_slot
-          raise NotImplementedError()
+        bref = buffered_ref.with_slot_index(copy_in_slot=final_copy_in_slot)
+        bref.save_slots()
     else:
       pred = pred & self.last_step
       grid_indices = self.indices
@@ -1485,8 +1523,7 @@ class Scheduler:
           buffered_ref.copy_in(src_ref, next_grid_indices)  # pylint: disable=cell-var-from-loop
         buffered_ref = buffered_ref.advance_copy_in_slot(pred & should_prefetch)
         grid_indices = next_grid_indices
-      if self.use_sreg_for_state:
-        buffered_ref.save_slots()
+      buffered_ref.save_slots()
     return
 
   def wait_out(self, buffered_ref, dst_ref, schedule=None) -> "BufferedRef":
@@ -1547,8 +1584,7 @@ class Scheduler:
       if buffered_ref.is_output:
         buffered_ref.wait_out(dst_ref, self.indices)
 
-    if self.use_sreg_for_state:
-      buffered_ref.save_slots()
+    buffered_ref.save_slots()
 
   def advance_slots(self, buffered_ref, schedule=None):
     if schedule is None:
@@ -1653,12 +1689,11 @@ def get_pipeline_schedule(schedule) -> Any:
 
 def make_pipeline_allocations(
     *refs,
-    in_specs=None,
-    out_specs=None,
+    in_specs=(),
+    out_specs=(),
     should_accumulate_out=False,
     needs_swap_ref=True,
     grid=None,
-    use_lookahead=False,
 ):
   """Create BufferedRefs for the pipeline.
 
@@ -1673,15 +1708,11 @@ def make_pipeline_allocations(
       as accumulators.
     needs_swap_ref: whether a swap slots tracker needs to be allocated.
     grid: grid to use for the pipeline.
-    use_lookahead: whether to enable lookahead for the pipeline. Requires
-      the pipeline to use SREGs for state.
 
   Returns:
     A list of BufferedRefs, one corresponding to each ref specified in the
     in_specs and out_specs.
   """
-  if use_lookahead and grid is None:
-    raise ValueError("Grid must be specified when using lookahead.")
   # TODO(levskaya): generalize argument tree handling here and in emit_pipeline.
   num_in_specs = len(in_specs)
   if not isinstance(in_specs, (list, tuple)):
@@ -1696,9 +1727,15 @@ def make_pipeline_allocations(
   out_refs = refs[num_in_specs:]
   def make_input_bref(in_spec, in_ref):
     buffer_count = 2
+    use_lookahead = False
     if in_spec.pipeline_mode is not None:
       buffer_count = in_spec.pipeline_mode.buffer_count
-    return BufferedRef.input(in_spec, in_ref.dtype, buffer_count,
+      use_lookahead = in_spec.pipeline_mode.use_lookahead
+    if use_lookahead and grid is None:
+      raise ValueError("Grid must be specified when using lookahead.")
+
+    in_aval = _ref_to_value_aval(in_ref)
+    return BufferedRef.input(in_spec, in_aval, buffer_count,
                              needs_swap_ref=needs_swap_ref,
                              grid_rank=len(grid),
                              use_lookahead=use_lookahead,
@@ -1708,12 +1745,16 @@ def make_pipeline_allocations(
     buffer_count = 2
     if out_spec.pipeline_mode is not None:
       buffer_count = out_spec.pipeline_mode.buffer_count
+      if out_spec.pipeline_mode.use_lookahead:
+        raise ValueError("Output buffering does not support lookahead.")
+
+    out_aval = _ref_to_value_aval(out_ref)
 
     if accumulate:
-      return BufferedRef.accumulator(out_spec, out_ref.dtype, buffer_count,
+      return BufferedRef.accumulator(out_spec, out_aval, buffer_count,
                                      needs_swap_ref=needs_swap_ref,
                                      source_memory_space=out_ref.memory_space)
-    return BufferedRef.output(out_spec, out_ref.dtype, buffer_count,
+    return BufferedRef.output(out_spec, out_aval, buffer_count,
                               needs_swap_ref=needs_swap_ref,
                               source_memory_space=out_ref.memory_space)
   out_brefs = jax.tree.map(
@@ -1831,7 +1872,7 @@ def sync_copy(src: REF | BufferedRef, dst: REF | BufferedRef, indices):
     bref = dst
     hbm_ref = src
     copy_in = True
-  hbm_slice = bref.get_dma_slice(hbm_ref.shape, hbm_ref.dtype, indices)
+  hbm_slice = bref.get_dma_slice(_ref_to_value_aval(hbm_ref), indices)
   bref_slice = tuple(
       pl.ds(0, s.size)
       for s, bd in zip(hbm_slice, bref.block_shape)
@@ -1849,16 +1890,14 @@ def emit_pipeline(
     body,
     *,
     grid: tuple[int | jax.Array, ...],
-    in_specs=None,
-    out_specs=None,
+    in_specs=(),
+    out_specs=(),
     should_accumulate_out: bool = False,
     core_axis: int | None = None,
     core_axis_name: str | None = None,
     dimension_semantics: tuple[GridDimensionSemantics, ...] | None = None,
     trace_scopes: bool = True,
     no_pipelining: bool = False,
-    use_sreg_for_state: bool = True,
-    use_lookahead: bool = False,
 ):
   """Creates a function to emit a manual pallas pipeline.
 
@@ -1885,15 +1924,8 @@ def emit_pipeline(
       or ARBITRARY).
     trace_scopes: optional bool, indicates whether to annotate each region in
       the pipeline using named_scope.
-    no_pipelining: If True, turns off pipelining and all copies will be
-      made synchronous. This is useful for debugging multiple-buffering
-      related bugs.
-    use_sreg_for_state: optional bool, indicates whether to use sregs for
-      current_slot state.
-    use_lookahead: optional bool, indicates whether to use lookahead on
-      input buffers. Enabling lookahead allows the pipeline to begin fetching
-      the next changed block as soon as a slot is available, no matter how
-      many iterations ahead that block is. Requires use_sreg_for_state=True.
+    no_pipelining: If True, turns off pipelining and all copies will be made
+      synchronous. This is useful for debugging multiple-buffering related bugs.
   """
   if any(not isinstance(d, (int, jax.Array)) for d in grid):
     grid_types = tuple(type(d) for d in grid)
@@ -1987,7 +2019,6 @@ def emit_pipeline(
               should_accumulate_out=should_accumulate_out,
               needs_swap_ref=needs_swap_ref,
               grid=grid,
-              use_lookahead=use_lookahead,
           ),
       )
     if isinstance(allocations, list):
@@ -2013,7 +2044,6 @@ def emit_pipeline(
           last_cycle=last_cycle,
           init_accumulators=init_accumulators,
           trace_scopes=trace_scopes,
-          use_sreg_for_state=use_sreg_for_state,
       )
 
     def loop_body(step, carry):
@@ -2031,13 +2061,11 @@ def emit_pipeline(
         with scheduler._named_scope("ep_prefetch"):
           if prefetch is not None:
             do_prefetch = step == num_steps - 1
-            if use_sreg_for_state:
-              map_brefs(lambda x: x.save_slots(do_prefetch), brefs)
+            map_brefs(lambda x: x.save_slots(do_prefetch), brefs)
             lax.cond(do_prefetch,
                     lambda: prefetch(*brefs, scheduler),
                     lambda: None)
-            if use_sreg_for_state:
-              brefs = map_brefs(lambda x: x.load_slots(do_prefetch), brefs)
+            brefs = map_brefs(lambda x: x.load_slots(do_prefetch), brefs)
 
         # run the kernel!
         if body_prologue is not None:
@@ -2053,13 +2081,11 @@ def emit_pipeline(
         with scheduler._named_scope("ep_postyeet"):
           if postyeet is not None:
             do_postyeet = step == 0
-            if use_sreg_for_state:
-              map_brefs(lambda x: x.save_slots(do_postyeet), brefs)
+            map_brefs(lambda x: x.save_slots(do_postyeet), brefs)
             lax.cond(do_postyeet,
                     lambda: postyeet(*brefs, scheduler),
                     lambda: None)
-            if use_sreg_for_state:
-              brefs = map_brefs(lambda x: x.load_slots(do_postyeet), brefs)
+            brefs = map_brefs(lambda x: x.load_slots(do_postyeet), brefs)
 
         brefs = map_brefs(scheduler.advance_slots, brefs, schedule)
         # Unbind window_refs for VMEM-backed buffers. Without this
@@ -2106,8 +2132,15 @@ def emit_pipeline(
         # pipeline prologue
         initial_indices = (0,) * len(grid)
         scheduler = make_scheduler(0, initial_indices)
+        brefs = allocations
         with scheduler.grid_env():
-          brefs = map_brefs(scheduler.initialize, allocations, refs, schedule)
+          # We issue num_stages-1 prefetch copies per buffer.
+          # We iterate over steps in the outer loop because we want to
+          # queue all iteration 0 prefetches before iteration 1, and so on.
+          for step in range(scheduler.num_stages - 1):
+            brefs = map_brefs(functools.partial(
+                scheduler.initialize_step, step=step),
+                brefs, refs, schedule)
 
         # pipeline loop
         brefs, next_indices = lax.fori_loop(
@@ -2127,11 +2160,9 @@ def emit_pipeline_with_allocations(
     body,
     *,
     grid,
-    in_specs=None,
-    out_specs=None,
+    in_specs=(),
+    out_specs=(),
     should_accumulate_out=False,
-    use_sreg_for_state=True,
-    use_lookahead=False,
 ):
   """Creates pallas pipeline and top-level allocation preparation functions.
 
@@ -2142,34 +2173,24 @@ def emit_pipeline_with_allocations(
     out_specs: output pallas block specs
     should_accumulate_out: booleans to indicate which outputs should be treated
       as accumulators.
-    use_sreg_for_state: optional bool, indicates whether to use sregs for
-      current_slot state.
-    use_lookahead: optional bool, indicates whether to use lookahead on
-      input buffers. Enabling lookahead allows the pipeline to begin fetching
-      the next changed block as soon as a slot is available, no matter how
-      many iterations ahead that block is. Requires use_sreg_for_state=True.
 
   Returns:
-    (emit_pipeline, make_allocations) function pair, where:
-    emit_pipeline is the pallas pipeline function.
-    make_allocations is a function to create buffered refs for the inner
-      pipeline that can be created at the top-level of a pallas call to be
-      reused across multiple invocations of the inner pipeline.
-
+    (emit_pipeline, make_allocations) function pair, where
+      - emit_pipeline is the pallas pipeline function.
+      - make_allocations is a function to create buffered refs for the inner
+        pipeline that can be created at the top-level of a pallas call to be
+        reused across multiple invocations of the inner pipeline.
   """
   make_allocations = functools.partial(make_pipeline_allocations,
                     in_specs=in_specs,
                     out_specs=out_specs,
                     should_accumulate_out=should_accumulate_out,
-                    grid=grid,
-                    use_lookahead=use_lookahead)
+                    grid=grid)
   pipeline = emit_pipeline(
       body,
       grid=grid,
       in_specs=in_specs,
       out_specs=out_specs,
-      should_accumulate_out=should_accumulate_out,
-      use_sreg_for_state=use_sreg_for_state,
-      use_lookahead=use_lookahead)
+      should_accumulate_out=should_accumulate_out)
 
   return pipeline, make_allocations
