@@ -18,7 +18,7 @@ from __future__ import annotations
 
 import abc
 import collections
-from collections.abc import Callable, Iterable, Sequence
+from collections.abc import Callable, Hashable, Iterable, Mapping, Sequence
 import dataclasses
 import enum
 import functools
@@ -28,17 +28,18 @@ from typing import Any, ClassVar, Literal, Union
 
 import jax
 from jax._src import core as jax_core
+from jax._src import custom_batching
 from jax._src import dtypes
 from jax._src import effects
 from jax._src import frozen_dict
-from jax._src import state
+from jax._src import lax
 from jax._src import pretty_printer as pp
+from jax._src import state
 from jax._src import tree_util
+from jax._src import util
 from jax._src.lib.mlir.dialects import arith as arith_dialect
 from jax._src.pallas import core as pallas_core
-from jax._src.pallas import helpers as pallas_helpers
 from jax._src.pallas import primitives as pallas_primitives
-import jax._src.pallas.utils as pallas_utils
 from jax._src.state import discharge as state_discharge
 from jax._src.state import indexing
 from jax._src.state import types as state_types
@@ -110,7 +111,6 @@ class CompilerParams(pallas_core.CompilerParams):
   approx_math: bool = False
   dimension_semantics: Sequence[DimensionSemantics] | None = None
   max_concurrent_steps: int = 1
-  delay_release: int = 0
   unsafe_no_auto_barriers: bool = False
   profile_space: int = 0
   profile_dir: str = ""
@@ -142,7 +142,7 @@ class MemorySpace(enum.Enum):
 
   def __call__(
       self,
-      shape: tuple[int, ...],
+      shape: Sequence[int],
       dtype: jnp.dtype,
       *,
       transforms: Sequence[MemoryRefTransform] = (),
@@ -150,6 +150,8 @@ class MemorySpace(enum.Enum):
       collective: bool | None = None,
       layout: TMEMLayout | None = None,
   ) -> pallas_core.MemoryRef:
+    shape = tuple(shape)
+    # TODO(sharadmv): Add HiType constructor support.
     if self == MemorySpace.TMEM:
       if transforms:
         raise ValueError("transforms are not supported for TMEM")
@@ -157,7 +159,7 @@ class MemorySpace(enum.Enum):
         collective = False
       if layout is None:
         if packed is None:
-          if dtypes.bit_width(dtype) != 32:
+          if dtypes.itemsize_bits(dtype) != 32:
             raise ValueError(
                 "dtypes narrower than 32-bit require either the packed argument"
                 " or an explicit TMEM layout"
@@ -174,8 +176,9 @@ class MemorySpace(enum.Enum):
       if packed is not None or collective is not None or layout is not None:
         raise ValueError("packed, collective and layout arguments are only supported for TMEM.")
       mgpu_layout = None
-    return GPUMemoryRef(shape, dtype, memory_space=self, transforms=transforms,
-                        layout=mgpu_layout, collective=collective)
+    return GPUMemoryRef(jax_core.ShapedArray(shape, dtype), memory_space=self,
+                        transforms=transforms, layout=mgpu_layout,
+                        collective=collective)
 
 
 class SemaphoreType(enum.Enum):
@@ -188,7 +191,8 @@ class SemaphoreType(enum.Enum):
       dtype = pallas_core.BarrierSemaphore()
     else:
       dtype = pallas_core.Semaphore()
-    return pallas_core.MemoryRef(shape, dtype, MemorySpace.GMEM)
+    return pallas_core.MemoryRef(jax_core.ShapedArray(shape, dtype),
+                                 MemorySpace.GMEM)
 
   def get_array_aval(self) -> jax_core.ShapedArray:
     return self(()).get_array_aval()
@@ -213,40 +217,120 @@ WGxWG_SEMANTICS = (
     mgpu.LoweringSemantics.Warpgroup, PrimitiveSemantics.Warpgroup)
 
 
+# TODO(justinfu): Reconcile with pl.kernel.
 def kernel(
     body: Callable[..., None],
     out_shape: object,
     *,
     scratch_shapes: pallas_core.ScratchShapeTree = (),
     compiler_params: pallas_core.CompilerParams | None = None,
+    # Mesh kwargs
+    grid: tuple[int, ...] = (),
+    grid_names: tuple[str, ...] = (),
+    cluster: tuple[int, ...] = (),
+    cluster_names: tuple[str, ...] = (),
+    num_threads: int | None = None,
+    thread_name: str | None = None,
     **mesh_kwargs: object,
 ):
+  """Entry point for defining a Mosaic GPU kernel.
+
+  Args:
+    body: The kernel body, which should take as arguments the input, output,
+      and scratch Refs. The number of input Refs is determined by the number
+      of arguments passed into kernel returned by this function. The number of
+      output and scratch Refs are determined by `out_shape` and `scratch_shapes`
+      respectively.
+    out_shape: a PyTree of :class:`jax.ShapeDtypeStruct` describing the shape
+      and dtypes of the outputs.
+    scratch_shapes: an iterable (may be nested) of GPUMemoryRef describing
+      scratch Refs to allocate for this kernel.
+    compiler_params: Additional compiler options. See the `CompilerParams`
+      dataclass for more details.
+    grid: A tuple of integers specifying the size of the kernel grid.
+    grid_names: The axis names of the grid. Must be the same length as `grid`.
+    cluster: A tuple of integers specifying the size of the kernel cluster.
+    cluster_names: The axis names of the grid. Must be the same length as
+      `cluster`.
+    num_threads: The number of threads to launch per block. Note that these
+      do not correspond to CUDA threads, but rather to warpgroups on Hopper
+      and Blackwell GPUs.
+    thread_name: The axis name used to query the thread index.
+    **mesh_kwargs: Additional mesh kwargs. See `Mesh` for more details.
+
+  Returns:
+    A function that runs the kernel. It should take any number of input
+    operands and returns an output with the same PyTree structure as
+    `out_shape`.
+  """
   if unwrap_out := not isinstance(out_shape, (tuple, list)):
     out_shape = (out_shape,)
+
+  @custom_batching.custom_vmap
   def wrapper(*operands):
     def stateful(operand_and_out_refs):
       operand_refs, out_refs = operand_and_out_refs
-      mesh = Mesh(**mesh_kwargs)
-      thread_name = mesh.thread_name if mesh.thread_name is not None else ()
+      mesh = Mesh(
+          grid=grid,
+          grid_names=grid_names,
+          cluster=cluster,
+          cluster_names=cluster_names,
+          num_threads=num_threads,
+          thread_name=thread_name,
+          **mesh_kwargs)
+      _thread_name = mesh.thread_name if mesh.thread_name is not None else ()
       def cmap_body():
         pallas_primitives.run_scoped(
-            lambda *scratch_refs: body(*operand_refs, *out_refs, *scratch_refs),
-            *scratch_shapes,
-            collective_axes=thread_name,
+            functools.partial(body, *operand_refs, *out_refs),
+            *(scratch_shapes if isinstance(scratch_shapes, Sequence) else ()),
+            collective_axes=_thread_name,
+            **(scratch_shapes if isinstance(scratch_shapes, Mapping) else {}),
         )
       if mesh.kernel_name is not None:
         cmap_body.__name__ = mesh.kernel_name
       else:
         # The body function name is used to set the name of the kernel as a
         # fallback if the kernel name is not set explicitly.
-        cmap_body.__name__ = getattr(body, '__name__', 'anonymous')
-      pallas_core.core_map(
-          mesh, compiler_params=compiler_params
-      )(cmap_body)
-    _, outs = state_discharge.run_state(stateful)(
-        (operands, pallas_helpers.empty_like(out_shape, backend="mosaic_gpu"))
-    )
+        cmap_body.__name__ = getattr(body, "__name__", "anonymous")
+      pallas_core.core_map(mesh, compiler_params=compiler_params)(cmap_body)
+    _, outs = state_discharge.run_state(stateful)((
+        operands,
+        jax.tree.map(lambda s: jax.lax.empty(s.shape, s.dtype), out_shape),
+    ))
     return outs[0] if unwrap_out else outs
+
+  @wrapper.def_vmap
+  def _vmap_rule(axis_size, in_batched, *args):
+    axis_name = object()
+
+    def batched_body(*refs):
+      idx = lax.axis_index(axis_name)
+      lens = (len(args), len(out_shape))
+      operand_refs, out_refs, scratch_refs = util.split_list(refs, lens)
+      slice_ref = lambda r, b=True: (r.at[idx] if b else r)
+      operand_refs = tree_util.tree_map(slice_ref, operand_refs, in_batched)
+      out_refs = tree_util.tree_map(slice_ref, out_refs)
+      return body(*operand_refs, *out_refs, *scratch_refs)
+
+    out_shape_ = out_shape[0] if unwrap_out else out_shape
+    add_batch_dim = lambda x: x.update(shape=(axis_size, *x.shape))
+    mesh_kwargs_ = dict(mesh_kwargs)
+    out = kernel(
+        batched_body,
+        out_shape=tree_util.tree_map(add_batch_dim, out_shape_),
+        scratch_shapes=scratch_shapes,
+        compiler_params=compiler_params,
+        grid=(axis_size,) + grid,
+        grid_names=(axis_name,) + grid_names,
+        cluster=cluster,
+        cluster_names=cluster_names,
+        num_threads=num_threads,
+        thread_name=thread_name,
+        **mesh_kwargs_,
+    )(*args)
+    out_batched = tree_util.tree_map(lambda _: True, out_shape_)
+    return out, out_batched
+
   return wrapper
 
 
@@ -304,7 +388,10 @@ def _ref_group_size(refs: _GPUMemoryRefTree) -> int:
       raise NotImplementedError(f"Unsupported dtype: {ref.dtype}")
     ref_bits = math.prod(ref.shape) * nbits
     if ref_bits % 8:
-      raise ValueError("Only byte-aligned shapes are supported.")
+      raise ValueError(
+          "Only byte-aligned shapes are supported. Got shape:"
+          f" {ref.dtype}{ref.shape}"
+      )
     size += ref_bits // 8
   return size
 
@@ -314,7 +401,8 @@ def _ref_group_tmem_col_size(refs: _GPUMemoryRefTree) -> int:
   """
   ncols = 0
   for ref in jax.tree.leaves(refs):
-    ref_ncols = ref.layout.cols_in_shape(ref.shape, dtypes.bit_width(ref.dtype))
+    ref_ncols = ref.layout.cols_in_shape(ref.shape,
+                                         dtypes.itemsize_bits(ref.dtype))
     ncols += align_to(ref_ncols, TMEM_COL_ALIGNMENT)
   return ncols
 
@@ -327,7 +415,7 @@ def infer_tmem_layout(
     collective: bool) -> tcgen05.TMEMLayout:
   """Infers the number of columns used and layout for allocating TMEM Refs."""
   if packed:
-    packing = 32 // dtypes.bit_width(dtype)
+    packing = 32 // dtypes.itemsize_bits(dtype)
   else:
     packing = 1
   return tcgen05._infer_tmem_layout(shape, collective=collective, packing=packing)  # type: ignore
@@ -343,7 +431,8 @@ def flatten_ref_union(ref_union: AbstractRefUnion) -> tuple[_Ref, ...]:
     union_bytes = 0
     for ref_group in ref_union.refs:
       byte_offset = 0
-      for ref in jax.tree.leaves(ref_group):
+      def unflatten(ref):
+        nonlocal byte_offset
         byte_offset = align_to(byte_offset, SMEM_ALIGNMENT)
         assert isinstance(ref, state.AbstractRef) or isinstance(
             ref, pallas_core.TransformedRef
@@ -351,10 +440,8 @@ def flatten_ref_union(ref_union: AbstractRefUnion) -> tuple[_Ref, ...]:
         if not isinstance(ref, pallas_core.TransformedRef):
           ref = pallas_core.TransformedRef(ref, transforms=())
         transform = ExtractAliasedRef.from_transformed_ref(ref, byte_offset)
-        flat_refs.append(
-            pallas_core.TransformedRef(
-                ref_union, transforms=(transform, *ref.transforms)
-            )
+        result = pallas_core.TransformedRef(
+            ref_union, transforms=(transform, *ref.transforms)
         )
         if jnp.issubdtype(ref.dtype, jnp.integer):
           nbits = jnp.iinfo(ref.dtype).bits
@@ -364,27 +451,34 @@ def flatten_ref_union(ref_union: AbstractRefUnion) -> tuple[_Ref, ...]:
           raise NotImplementedError(f"Unsupported dtype: {ref.dtype}")
         ref_bits = math.prod(ref.shape) * nbits
         if ref_bits % 8:
-          raise ValueError("Only byte-aligned shapes are supported.")
+          raise ValueError(
+              "Only byte-aligned shapes are supported. Got shape:"
+              f" {ref.dtype}{ref.shape}"
+          )
         byte_offset += ref_bits // 8
+        return result
+      flat_refs.append(jax.tree.map(unflatten, ref_group))
       union_bytes = max(union_bytes, byte_offset)
     assert union_bytes == ref_union.shape[0]
   elif ref_union.memory_space == TMEM:
     union_cols = 0
     for ref_group in ref_union.refs:
       col_offset = 0
-      for ref in jax.tree.leaves(ref_group):
+      def unflatten(ref):
+        nonlocal col_offset
         col_offset = align_to(col_offset, TMEM_COL_ALIGNMENT)
         if not isinstance(ref, pallas_core.TransformedRef):
           ref = pallas_core.TransformedRef(ref, transforms=())
-        ncols = ref.layout.cols_in_shape(ref.shape, dtypes.bit_width(ref.dtype))
+        ncols = ref.layout.cols_in_shape(ref.shape,
+                                         dtypes.itemsize_bits(ref.dtype))
         transform = ExtractAliasedRef.from_transformed_ref(
             ref, col_offset, layout=ref.layout)
-        flat_refs.append(
-            pallas_core.TransformedRef(
-                ref_union, transforms=(transform, *ref.transforms)
-            )
+        result = pallas_core.TransformedRef(
+            ref_union, transforms=(transform, *ref.transforms)
         )
         col_offset += ncols
+        return result
+      flat_refs.append(jax.tree.map(unflatten, ref_group))
       union_cols = max(union_cols, col_offset)
     assert union_cols == ref_union.shape[1], (union_cols, ref_union.shape[1])
   else:
@@ -414,8 +508,8 @@ class AbstractRefUnion(state.AbstractRef):
     del tracer, index, value  # Unused.
     raise ValueError("Ref unions can't be assigned to.")
 
-  def update(self, inner_aval=None, memory_space=None):
-    ref = super().update(inner_aval, memory_space)
+  def update(self, inner_aval=None, memory_space=None, kind=None):
+    ref = super().update(inner_aval, memory_space, kind)
     return AbstractRefUnion(ref.inner_aval, self.refs, self.memory_space)
 
   @functools.cached_property
@@ -434,7 +528,6 @@ class AbstractRefUnion(state.AbstractRef):
     return first_ref.collective
 
 
-
 @dataclasses.dataclass(init=False, frozen=True)
 class RefUnion(GPUMemoryRef):
   """A sequence of trees of refs that are allowed to reuse the same memory.
@@ -455,8 +548,9 @@ class RefUnion(GPUMemoryRef):
       object.__setattr__(self, "refs", refs)
       num_bytes = max(map(_ref_group_size, self.refs))
       super().__init__(
-          shape=(num_bytes,),
-          dtype=jnp.int8,
+          inner_aval=jax_core.ShapedArray(
+              (num_bytes,), jnp.int8
+          ),
           memory_space=SMEM,
           transforms=(),
       )
@@ -469,8 +563,10 @@ class RefUnion(GPUMemoryRef):
             "Some aliased TMEM references are collective and some are not."
         )
       super().__init__(
-          shape=(128, max_cols,),
-          dtype=jnp.int32,
+          inner_aval=jax_core.ShapedArray(
+              shape=(128, max_cols,),
+              dtype=jnp.int32,
+          ),
           memory_space=TMEM,
           transforms=(),
           layout=tcgen05.tmem_default_layout(packing=1),
@@ -576,7 +672,8 @@ class UntileRef(state_types.Transform):
       self, dtype: jnp.dtype, shape: tuple[int, ...]
   ) -> tuple[tuple[int, ...], state_types.Transform]:
     del dtype
-    raise NotImplementedError("Reshapes don't commute with transposes.")
+    # TODO(slebedev): Support this.
+    raise NotImplementedError("Reshapes don't commute with tiling.")
 
   def untransform_index(
       self, dtype: jnp.dtype | ir.Type, idxs: tuple[Index, ...]
@@ -588,9 +685,16 @@ class UntileRef(state_types.Transform):
     for idx, tile in zip(tiled_idxs, self.tiling):
       if isinstance(idx, slice):
         if idx.step is not None and idx.step != 1:
-          raise NotImplementedError("Strided slices unsupported")
-        if (idx.start is not None and idx.start % tile) or (idx.stop is not None and idx.stop % tile):
-          raise ValueError("Non-empty slices must be tile aligned")
+          raise NotImplementedError(
+              f"Strided slices unsupported. Got stride: {idx.step}"
+          )
+        if (idx.start is not None and idx.start % tile) or (
+            idx.stop is not None and idx.stop % tile
+        ):
+          raise ValueError(
+              f"Expected slice start ({idx.start}) and slice stop ({idx.stop})"
+              f" to be divisible by the tile size ({tile})"
+          )
         idxs_after_tiling.append(slice(idx.start // tile, idx.stop // tile))
       elif isinstance(idx, mgpu.DynamicSlice):
         if idx.length % tile:
@@ -663,16 +767,7 @@ class TransposeTransform(MemoryRefTransform):
 
 @tree_util.register_dataclass
 @dataclasses.dataclass(frozen=True)
-class TransposeRef(state_types.Transform):
-  permutation: tuple[int, ...] = dataclasses.field(metadata=dict(static=True))
-
-  def transform_shape(self, shape):
-    if shape is None:
-      return None
-    return tuple(shape[i] for i in self.permutation)
-
-  def transform_dtype(self, dtype):
-    return dtype
+class TransposeRef(state_types.RefTransposer):
 
   def untransform_transpose(
       self, perm
@@ -705,9 +800,6 @@ class TransposeRef(state_types.Transform):
   def undo_to_gpu_transform(self) -> mgpu.MemRefTransform:
     return mgpu.TransposeTransform(_perm_inverse(self.permutation))
 
-  def pretty_print(self, context: jax_core.JaxprPpContext) -> pp.Doc:
-    return pp.text(f"{{transpose({list(self.permutation)})}}")
-
 
 @tree_util.register_pytree_node_class
 @dataclasses.dataclass
@@ -734,6 +826,30 @@ class PeerMemRef(state_types.Transform):
     return cls(arrays[0], metadata[0])
 
 
+@tree_util.register_pytree_node_class
+@dataclasses.dataclass
+class MulticastRef(state_types.Transform):
+  collective_axes: tuple[Hashable, ...]
+
+  def transform_shape(self, shape):
+    return shape
+
+  def transform_dtype(self, dtype):
+    return dtype
+
+  def untransform_index(
+      self, idxs: tuple[Index, ...]
+  ) -> tuple[tuple[Index, ...], state_types.Transform]:
+    return idxs, self
+
+  def tree_flatten(self):
+    return (), self.collective_axes
+
+  @classmethod
+  def tree_unflatten(cls, metadata, arrays):
+    return cls(metadata[0])
+
+
 def remote_ref(
     ref: _Ref,
     device_id: jax.typing.ArrayLike,
@@ -744,11 +860,36 @@ def remote_ref(
     if not isinstance(jax_core.get_aval(ref), state_types.AbstractRef):
       raise TypeError("ref must be a reference")
     ref = pallas_core.TransformedRef(ref, transforms=())
+  if any(isinstance(t, MulticastRef) for t in ref.transforms):
+    raise ValueError("Can't make a multicast reference into a peer reference.")
   return pallas_core.TransformedRef(
       ref.ref, (*ref.transforms, PeerMemRef(device_id, device_id_type)),
   )
 
 
+def multicast_ref(
+    ref: _Ref,
+    collective_axes: Hashable | tuple[Hashable, ...],
+) -> pallas_core.TransformedRef:
+  """Return a multicast reference for cross-device operations.
+
+  Args:
+    ref: The reference to transform.
+    collective_axes: The JAX mesh axes indicating the devices to operate on.
+  """
+  if not isinstance(collective_axes, tuple):
+    collective_axes = (collective_axes,)
+  if not isinstance(ref, pallas_core.TransformedRef):
+    if not isinstance(jax_core.get_aval(ref), state_types.AbstractRef):
+      raise TypeError("ref must be a reference")
+    ref = pallas_core.TransformedRef(ref, transforms=())
+  if any(isinstance(t, PeerMemRef) for t in ref.transforms):
+    raise ValueError("Can't make a peer reference into a multicast reference.")
+  return pallas_core.TransformedRef(
+      ref.ref, (*ref.transforms, MulticastRef(collective_axes)),
+  )
+
+
 def transform_ref(
     ref: pallas_core.TransformedRef,
     transform: state_types.Transform
@@ -765,7 +906,10 @@ def transpose_ref(
     ref: pallas_core.TransformedRef | Any,
     permutation: tuple[int, ...],
 ) -> pallas_core.TransformedRef:
-  return transform_ref(ref, TransposeRef(permutation))
+  assert hasattr(ref, "memory_space")
+  if ref.memory_space == MemorySpace.TMEM:
+    raise ValueError("Can't transpose a TMEM reference.")
+  return ref.transpose(permutation)
 
 def untile_ref(ref, tiling: tuple[int, ...]) -> pallas_core.TransformedRef:
   return transform_ref(ref, UntileRef(tiling))
@@ -841,7 +985,7 @@ class SwizzleTransform(MemoryRefTransform):
     raise NotImplementedError
 
   def __call__(self, aval: jax_core.ShapedArray) -> jax_core.ShapedArray:
-    swizzle_elems = (self.swizzle * 8) // pallas_utils.dtype_bitwidth(aval.dtype)
+    swizzle_elems = (self.swizzle * 8) // dtypes.itemsize_bits(aval.dtype)
     if swizzle_elems != aval.shape[-1]:
       raise ValueError(
           f"Swizzle {self.swizzle} requires the trailing dimension to be of"
@@ -906,7 +1050,23 @@ class UnswizzleRef(state_types.Transform):
 
 @dataclasses.dataclass
 class BlockSpec(pallas_core.BlockSpec):
+  r"""A GPU-specific ``BlockSpec``.
+
+  Attributes:
+    transforms: A sequence of transforms that will be applied to the
+      reference.
+    delay_release: used during pipelining to delay the release of
+      resources of a slot after it is used in the computation.
+    collective_axes: When set, all blocks along the specified axes must execute
+      the same sequence of pipeline operations (with the only exception being
+      the index_map in non-collective ``BlockSpec``\ s), and all of them must
+      return the same block from the index_map for this operand. This enables
+      the pipelining helpers to use collective async copies, which can improve
+      performance.
+  """
   transforms: Sequence[MemoryRefTransform] = ()
+  delay_release: int = 0
+  collective_axes: tuple[Hashable, ...] = ()
 
   def to_block_mapping(
       self,
@@ -916,16 +1076,21 @@ class BlockSpec(pallas_core.BlockSpec):
       index_map_avals: Sequence[jax_core.AbstractValue],
       index_map_tree: tree_util.PyTreeDef,
       grid: pallas_core.GridMappingGrid,
-      mapped_dims: tuple[int, ...],
+      vmapped_dims: tuple[int, ...],
       debug: bool = False,
   ) -> pallas_core.BlockMapping:
+    if self.collective_axes:
+      raise ValueError(
+          "collective_axes is not supported in pallas_call. Use plgpu.kernel"
+          " with plgpu.emit_pipeline_warp_specialized instead."
+      )
     bm = super().to_block_mapping(
         origin,
         array_aval,
         index_map_avals=index_map_avals,
         index_map_tree=index_map_tree,
         grid=grid,
-        mapped_dims=mapped_dims,
+        vmapped_dims=vmapped_dims,
         debug=debug,
     )
     block_inner_aval = bm.block_aval.inner_aval
@@ -967,6 +1132,8 @@ class ClusterBarrierType(dtypes.ExtendedDType):
   name: ClassVar[str] = "cluster_barrier"
 
   collective_axes: tuple[str | tuple[str, ...], ...]
+  num_arrivals: int
+  orders_tensor_core: bool
 
   def __str__(self):
     return self.name
@@ -989,6 +1156,9 @@ class Barrier:
   num_barriers: int = 1
   orders_tensor_core: bool = False
 
+  def get_array_aval(self) -> jax_core.ShapedArray:
+    raise ValueError("Barriers are not arrays")
+
   def get_ref_aval(self) -> state.AbstractRef:
     aval = jax_core.ShapedArray(
         [self.num_barriers],
@@ -1008,10 +1178,18 @@ class Barrier:
 class ClusterBarrier:
   collective_axes: tuple[str | tuple[str, ...], ...]
   num_barriers: int = 1
+  num_arrivals: int = 1
+  orders_tensor_core: bool = False
+
+  def get_array_aval(self) -> jax_core.ShapedArray:
+    raise ValueError("Cluster barriers are not arrays")
 
   def get_ref_aval(self) -> state.AbstractRef:
     aval = jax_core.ShapedArray(
-        [self.num_barriers], ClusterBarrierType(self.collective_axes)
+        [self.num_barriers],
+        ClusterBarrierType(
+            self.collective_axes, self.num_arrivals, self.orders_tensor_core
+        ),
     )
     return state.AbstractRef(aval, SMEM)
 
@@ -1050,8 +1228,8 @@ class WGMMAAbstractAccumulatorRef(state.AbstractRef):
   def __repr__(self) -> str:
     return f'Accumulator{{{self.inner_aval.str_short()}}}'
 
-  def update(self, inner_aval=None, memory_space=None):
-    ref = super().update(inner_aval, memory_space)
+  def update(self, inner_aval=None, memory_space=None, kind=None):
+    ref = super().update(inner_aval, memory_space, kind)
     return WGMMAAbstractAccumulatorRef(
         inner_aval=ref.inner_aval,
         memory_space=ref.memory_space,
@@ -1078,8 +1256,8 @@ class AbstractTMEMRef(state.AbstractRef):
   def __repr__(self) -> str:
     return f'TMEM({self.inner_aval.str_short()}, layout={self.layout}, collective={self.collective})'
 
-  def update(self, inner_aval=None, memory_space=None):
-    ref = super().update(inner_aval, memory_space)
+  def update(self, inner_aval=None, memory_space=None, kind=None):
+    ref = super().update(inner_aval, memory_space, kind)
     return AbstractTMEMRef(
         ref.inner_aval, ref.memory_space, self.layout, self.collective
     )
@@ -1114,10 +1292,12 @@ class Mesh:
           "num_threads and thread_name must be either both set or both None,"
           f" got {self}"
       )
-    if self.num_threads is not None and self.num_threads > 2048 // 128:
+    max_mosaic_threads = 2048 // 128
+    if self.num_threads is not None and self.num_threads > max_mosaic_threads:
       raise ValueError(
           "Requested too many CUDA threads per block. Each Mosaic thread"
-          " corresponds to 128 CUDA threads."
+          f" corresponds to 128 CUDA threads. At most {max_mosaic_threads}"
+          f" are supported, got {self}"
       )
     object.__setattr__(self, "grid", tuple(self.grid))
     object.__setattr__(self, "grid_names", tuple(self.grid_names))
@@ -1179,6 +1359,7 @@ def _gpu_mesh_discharge_rule(
     debug,
     cost_estimate,
     name,
+    metadata,
 ):
   if not isinstance(mesh, Mesh):
     raise TypeError(f"Mesh must be a `plgpu.Mesh`, got {type(mesh)}")
@@ -1189,6 +1370,11 @@ def _gpu_mesh_discharge_rule(
     )
   if not compiler_params:
     compiler_params = CompilerParams()
+  sa_avals = [a for a in in_avals if isinstance(a, jax_core.ShapedArray)]
+  if sa_avals:
+    raise NotImplementedError(
+        f"Cannot close over values in core_map: {sa_avals}"
+    )
   return pallas_core.default_mesh_discharge_rule(
       in_avals,
       out_avals,
@@ -1201,6 +1387,8 @@ def _gpu_mesh_discharge_rule(
       cost_estimate=cost_estimate,
       name=name,
       memory_space=GMEM,
+      metadata=metadata,
+      scratch_shapes=[],
   )
 
 
@@ -1252,7 +1440,7 @@ class SomeLayout:
 
 @dataclasses.dataclass(frozen=True)
 class ParameterizedLayout(SomeLayout):
-  layout_cls: Layout
+  layout_cls: Layout | TMEMLayout
   args: Sequence[Any]
   kwargs: Any
 
@@ -1279,15 +1467,27 @@ class ReducedLayout(SomeLayout):
 class Layout(SomeLayout, enum.Enum):
   #: [m, n] matrix, where m % 64 == 0 == n % 8.
   WGMMA = enum.auto()
+  WGMMA_8BIT = enum.auto()
+  WGMMA_UPCAST_2X = enum.auto()
+  WGMMA_UPCAST_4X = enum.auto()
   WGMMA_TRANSPOSED = enum.auto()
 
   WG_SPLAT = enum.auto()
   WG_STRIDED = enum.auto()
 
+  TILED = enum.auto()
+
   TCGEN05 = enum.auto()
   TCGEN05_TRANSPOSED = enum.auto()
   TCGEN05_M64_COLLECTIVE = enum.auto()
   TCGEN05_TMEM_NATIVE = enum.auto()
+  TCGEN05_M64_COLLECTIVE_NATIVE = enum.auto()
+
+  SMEM_GMEM_COPY = enum.auto()
+  TMA_GATHER_INDICES = enum.auto()
+
+  # TODO(b/435159109): Remove this once LLVM regression is addressed.
+  _WGMMA_ACC_32BIT = enum.auto()  # Temporarily exposed to work around LLVM bugs
 
   def __call__(self, *args, **kwargs) -> ParameterizedLayout:
     return ParameterizedLayout(self, args, kwargs)
@@ -1304,10 +1504,24 @@ class Layout(SomeLayout, enum.Enum):
       case Layout.WGMMA:
         check_no_args()
         return mgpu.WGMMA_LAYOUT
+      case Layout.WGMMA_8BIT:
+        check_no_args()
+        return mgpu.WGMMA_LAYOUT_8BIT
+      case Layout.WGMMA_UPCAST_2X:
+        check_no_args()
+        return mgpu.WGMMA_LAYOUT_UPCAST_2X
+      case Layout.WGMMA_UPCAST_4X:
+        check_no_args()
+        return mgpu.WGMMA_LAYOUT_UPCAST_4X
+      case Layout._WGMMA_ACC_32BIT:
+        check_no_args()
+        return mgpu.fragmented_array.WGMMA_LAYOUT_ACC_32BIT
       case Layout.WG_SPLAT:
         return mgpu.WGSplatFragLayout(*args, **kwargs)  # pytype: disable=missing-parameter
       case Layout.WG_STRIDED:
         return mgpu.WGStridedFragLayout(*args, **kwargs)  # pytype: disable=missing-parameter
+      case Layout.TILED:
+        return mgpu.TiledLayout(*args, **kwargs)
       case Layout.TCGEN05:
         check_no_args()
         return mgpu.TCGEN05_LAYOUT
@@ -1315,10 +1529,24 @@ class Layout(SomeLayout, enum.Enum):
         check_no_args()
         return mgpu.TCGEN05_TRANSPOSED_LAYOUT
       case Layout.TCGEN05_TMEM_NATIVE:
-        check_no_args()
-        return tcgen05.TMEM_NATIVE_LAYOUT
+        if args or kwargs:
+          return mgpu.tmem_native_layout(*args, **kwargs)
+        return mgpu.TMEM_NATIVE_LAYOUT
       case Layout.TCGEN05_M64_COLLECTIVE:
         return tcgen05.fa_m64_collective_layout(*args, **kwargs)  # pytype: disable=missing-parameter
+      case Layout.TCGEN05_M64_COLLECTIVE_NATIVE:
+        return tcgen05.tmem_m64_collective_layout(*args, **kwargs).as_tiled_layout()  # pytype: disable=missing-parameter
+      case Layout.SMEM_GMEM_COPY:
+        normalize_args = lambda shape, dtype, swizzle: (shape, dtype, swizzle)
+        shape, dtype, swizzle = normalize_args(*args, **kwargs)
+        bitwidth = dtypes.itemsize_bits(dtype)
+        tiling = (8, 8 * swizzle // bitwidth)
+        row_tiles, col_tiles = mgpu.tile_shape(shape, tiling)[-4:-2]
+        return mgpu.fragmented_array.tiled_copy_smem_gmem_layout(
+            row_tiles, col_tiles, swizzle, bitwidth
+        )
+      case Layout.TMA_GATHER_INDICES:
+        return mgpu.TMA_GATHER_INDICES_LAYOUT
 
 
 # TODO(apaszke): Adjust the users and remove these backfills.
@@ -1331,9 +1559,38 @@ Layout.TCGEN05_TMEM_NATIVE_ROW = Layout.TCGEN05_TMEM_NATIVE.reduce(1)
 
 class TMEMLayout(enum.Enum):
   """Layout for TMEM references."""
+  # TODO(apaszke): Remove the layout suffix.
   SCALES_LAYOUT = enum.auto()
+  SPARSE_METADATA_LAYOUT = enum.auto()
+  M64_COLLECTIVE_LAYOUT = enum.auto()
 
-  def to_mgpu(self) -> tcgen05.TMEMLayout:
+  def __call__(self, *args, **kwargs) -> ParameterizedLayout:
+    return ParameterizedLayout(self, args, kwargs)
+
+  def to_mgpu(self, *args, **kwargs) -> tcgen05.TMEMLayout:
     match self:
       case TMEMLayout.SCALES_LAYOUT:
-        return tcgen05.scales_layout()
+        return tcgen05.scales_layout(*args, **kwargs)
+      case TMEMLayout.SPARSE_METADATA_LAYOUT:
+        return tcgen05.sparse_meta_layout(*args, **kwargs)
+      case TMEMLayout.M64_COLLECTIVE_LAYOUT:
+        return tcgen05.tmem_m64_collective_layout(*args, **kwargs)  # pytype: disable=missing-parameter
+
+
+def TryClusterCancelResult(
+    num_buffers: int | None = None) -> pallas_core.MemoryRef:
+  """Helper function to create Refs for cluster launch control results.
+
+  Args:
+    num_buffers: Optional argument for specifying the number of buffers
+      to allocate. If None, will return a single 16-byte buffer. If specified,
+      will return a (num_buffers, 16)-shaped buffer.
+
+  Returns:
+    A MemoryRef with the correct shape for holding the opaque cluster launch
+    control result.
+  """
+  if num_buffers is None:
+    return SMEM((16,), jnp.int8)
+  else:
+    return SMEM((num_buffers, 16), jnp.int8)
