@@ -18,10 +18,11 @@ from __future__ import annotations
 
 from collections.abc import Callable, Sequence
 import dataclasses
+import enum
 import functools
 import itertools as it
 import math
-from typing import Any, Protocol, TypeVar, cast
+from typing import Any, Protocol, TypeVar, Union, cast
 
 import jax
 from jax import api_util
@@ -42,6 +43,10 @@ map = util.safe_map
 zip = util.safe_zip
 T = TypeVar('T')
 
+
+BlockSpecPytree = Sequence[Union[pl.BlockSpec, "BlockSpecPytree"]]
+AbstractRefPytree = Sequence[Union[state.AbstractRef, "AbstractRefPytree"]]
+
 def _get_block_size(
     bd: pl.Blocked | pl.Element | pl.Squeezed | pl.BoundedSlice | int | None,
 ) -> int:
@@ -54,13 +59,25 @@ def _get_block_size(
       raise NotImplementedError(f"Unsupported block size type: {type(bd)}")
 
 def _get_block_shape(spec: pallas_core.BlockSpec):
-  assert spec.block_shape is not None
-  return tuple(_get_block_size(bd) for bd in spec.block_shape)
+  if spec.block_shape is None:
+    raise ValueError("Block shape must be specified.")
+
+  block_shape = tuple(
+      _get_block_size(bd)
+      for bd in spec.block_shape
+      if not (bd is None or isinstance(bd, pl.Squeezed))
+  )
+  return block_shape
+
+
+map_brefs = functools.partial(
+    jax.tree.map, is_leaf=lambda x: isinstance(x, BufferedRef)
+)
 
 @jax.tree_util.register_dataclass
 @dataclasses.dataclass(frozen=True)
 class BufferedRef:
-  spec: pallas_core.BlockSpec = dataclasses.field(metadata={"static": True})
+  spec: gpu_core.BlockSpec = dataclasses.field(metadata={"static": True})
   is_index_invariant: bool = dataclasses.field(metadata={"static": True})
   gmem_ref: state.AbstractRef
   # ``None`` if the ref is pinned to GMEM; otherwise, has shape
@@ -74,21 +91,30 @@ class BufferedRef:
       return self.gmem_ref
     return self.smem_ref.at[slot]
 
-  def compute_gmem_slice(self, grid_indices) -> tuple[pl.Slice, ...]:
+  def compute_gmem_slice(self, grid_indices) -> tuple[pl.Slice | jax.Array, ...]:
     index_map = self.spec.index_map
     assert index_map is not None
+    assert self.spec.block_shape is not None
     # We don't allow Python scalars here, because they are interpreted
     # differently depending on the x32/x64 mode.
     assert all(i.dtype == jnp.dtype(jnp.int32) for i in grid_indices)
-    sizes = _get_block_shape(self.spec)
+
+    def _make_block_slice(block_index: jax.Array, bd: pl.BlockDim | int | None):
+      match bd:
+        case int():
+          return pl.Slice(block_index * bd, bd)
+        case pl.Blocked(block_size):
+          return pl.Slice(block_index * block_size, block_size)
+        case None | pl.Squeezed():
+          return block_index
+        case _:
+          raise ValueError(f"Unsupported block dimension type: {bd}")
+
     return tuple(
-        pl.Slice(idx * size, size)  # type: ignore[arg-type]
-        for idx, size in zip(
-            index_map(*grid_indices), sizes  # type: ignore[arg-type]
-        )
+        map(_make_block_slice, index_map(*grid_indices), self.spec.block_shape)
     )
 
-  def copy_in(self, slot, grid_indices, barrier_ref):
+  def copy_in(self, slot, grid_indices, barrier_ref, barrier_slot=None):
     if not _in_smem(self.spec):
       return
     assert self.smem_ref is not None
@@ -96,7 +122,8 @@ class BufferedRef:
     gpu_primitives.copy_gmem_to_smem(
         self.gmem_ref.at[gmem_slices],  # pytype: disable=unsupported-operands
         self.smem_ref.at[slot],  # pytype: disable=unsupported-operands
-        barrier_ref.at[slot],
+        barrier_ref.at[barrier_slot if barrier_slot is not None else slot],
+        collective_axes=getattr(self.spec, "collective_axes", ()),
     )
 
   def copy_out(self, slot, grid_indices, predicate=None):
@@ -175,6 +202,20 @@ jax.tree_util.register_dataclass(
 )
 
 
+def _downcast_spec(
+    spec: gpu_core.BlockSpec | pallas_core.BlockSpec,
+) -> gpu_core.BlockSpec:
+  if isinstance(spec, gpu_core.BlockSpec):
+    return spec
+
+  return gpu_core.BlockSpec(
+      block_shape=spec.block_shape,
+      index_map=spec.index_map,
+      memory_space=spec.memory_space,
+      pipeline_mode=spec.pipeline_mode,
+  )
+
+
 def emit_pipeline(
     body: Callable[..., T],
     *,
@@ -182,7 +223,6 @@ def emit_pipeline(
     in_specs: Sequence[pallas_core.BlockSpec] = (),
     out_specs: Sequence[pallas_core.BlockSpec] = (),
     max_concurrent_steps: int = 1,
-    delay_release: int = 0,
     init_carry: T | None = None,
 ):
   r"""Creates a function to emit a manual pipeline within a Pallas kernel.
@@ -203,9 +243,6 @@ def emit_pipeline(
     out_specs: A sequence of :class:`~jax.experimental.pallas.BlockSpec`\s
       for outputs.
     max_concurrent_steps: Maximum concurrently active pipeline stages.
-    delay_release: Number of steps to delay before reusing input/output
-      references. Must be ``< max_concurrent_steps``. Useful for hiding WGMMA
-      latency (typically set to 1).
     init_carry: Optional initial carry. If provided, ``body`` handles
       carry-over state between iterations, and the pipeline returns the
       final carry.
@@ -215,14 +252,30 @@ def emit_pipeline(
     pipeline and returns the final carry value (if ``init_carry`` was used),
     otherwise it returns None.
   """
-  if max_concurrent_steps <= delay_release:
+
+  in_specs = tuple(map(_downcast_spec, in_specs))
+  out_specs = tuple(map(_downcast_spec, out_specs))
+  for spec in in_specs:
+    if spec.collective_axes:
+      raise NotImplementedError(
+          "BlockSpecs with collective_axes are not supported in emit_pipeline"
+      )
+  for spec in out_specs:
+    if spec.collective_axes:
+      raise ValueError("Output BlockSpecs cannot have collective_axes")
+  # TODO(justinfu): Factor out common code between warp-specialized and
+  # normal pipelines.
+  delay_release_levels = sorted({s.delay_release for s in in_specs}) or [0]
+  if delay_release_levels and max_concurrent_steps <= delay_release_levels[0]:
     raise ValueError(
-        "max_concurrent_steps must be greater than delay_release, but"
-        f" {max_concurrent_steps=}, {delay_release=}"
+        "max_concurrent_steps must be greater than all delay_release values,"
+        f" but {max_concurrent_steps=} and {delay_release_levels=}."
     )
 
   num_steps = math.prod(grid)
   has_dynamic_grid = not isinstance(num_steps, int)
+  # Convert the grid to int32 explicitly to avoid dtype promotion errors.
+  grid = tuple(jnp.asarray(g, dtype=jnp.int32) for g in grid)
 
   # Shrink ``max_concurrent_steps`` if the total number of steps is lower to
   # reduce the size of the refs allocated in SMEM.
@@ -282,20 +335,29 @@ def emit_pipeline(
 
     # Initialize the pipeline.
     indices = (jnp.asarray(0, dtype=jnp.int32),) * len(grid)
-    fetch_indices = indices
-    for step in range(max_concurrent_steps):
+    if has_dynamic_grid:
+      prologue_steps = lax.min(max_concurrent_steps, num_steps)
+    else:
+      assert max_concurrent_steps <= num_steps
+      prologue_steps = max_concurrent_steps
+
+    def prologue(step, fetch_indices):
       for bref in in_brefs:
         bref.copy_in(step, fetch_indices, barrier_ref)
-      fetch_indices = _inc_grid_by_1(fetch_indices, grid)
-    del fetch_indices
+      return _inc_grid_by_1(fetch_indices, grid)
+    jax.lax.fori_loop(0, prologue_steps, prologue, indices, unroll=not has_dynamic_grid)
 
     # This is true if any of the outputs need to be transferred inside the loop.
-    copies_out_in_loop = not all(bref.is_index_invariant for bref in out_brefs)
-    needs_epilogue = any(bref.is_index_invariant for bref in out_brefs)
+    smem_out_brefs = [bref for bref in out_brefs if _in_smem(bref.spec)]
+    copies_out_in_loop = not all(bref.is_index_invariant for bref in smem_out_brefs)
+    needs_epilogue = any(bref.is_index_invariant for bref in smem_out_brefs)
 
+    # In the loop body, `max_concurrent_steps` may be larger than `num_steps` in
+    # the dynamic grid case. This is fine, since in that case, we will never
+    # need to fetch more data anyway.
     def loop_body(step, carry):
       slot = lax.rem(step, max_concurrent_steps)
-      indices, fetch_indices, last_store_slices, prev_body_carry = carry
+      indices, fetch_index_levels, last_store_slices, prev_body_carry = carry
 
       if barrier_ref is not None:
         # Wait for the current GMEM->SMEM copy to complete, if any.
@@ -303,7 +365,7 @@ def emit_pipeline(
       # Wait for the previous output SMEM->GMEM copy to complete.
       if copies_out_in_loop:
         gpu_primitives.wait_smem_to_gmem(
-            max_concurrent_steps - (1 + delay_release), wait_read_only=True
+            max_concurrent_steps - 1, wait_read_only=True
         )
 
       next_body_carry = body(
@@ -326,7 +388,8 @@ def emit_pipeline(
           continue
         assert last_store_slices[idx] is not None
         new_store_slices[idx] = tuple(
-            _Slice(s.start, s.size) for s in bref.compute_gmem_slice(indices)
+            _Slice(s.start, s.size) if isinstance(s, pl.Slice) else s
+            for s in bref.compute_gmem_slice(indices)
         )
         are_same_slices = map(
             lambda old, new: old == new,
@@ -347,43 +410,60 @@ def emit_pipeline(
       if copies_out_in_loop:
         gpu_primitives.commit_smem_to_gmem_group()
 
-      fetch_step = step + (max_concurrent_steps - delay_release)
-      fetch_slot = lax.rem(fetch_step, max_concurrent_steps)
+      for delay_release, fetch_indices in zip(
+          delay_release_levels, fetch_index_levels
+      ):
+        fetch_step = step + (max_concurrent_steps - delay_release)
+        fetch_slot = lax.rem(fetch_step, max_concurrent_steps)
 
-      def do_fetch():
-        for bref in in_brefs:
-          bref.copy_in(fetch_slot, fetch_indices, barrier_ref)
+        # pylint: disable=cell-var-from-loop
+        def do_fetch():
+          for bref in in_brefs:
+            if bref.spec.delay_release == delay_release:
+              bref.copy_in(fetch_slot, fetch_indices, barrier_ref)
+        # pylint: enable=cell-var-from-loop
 
-      jax.lax.cond(
-          lax.bitwise_and(step >= delay_release, fetch_step < num_steps),
-          do_fetch,
-          lambda: None,
-      )
+        jax.lax.cond(
+            lax.bitwise_and(step >= delay_release, fetch_step < num_steps),
+            do_fetch,
+            lambda: None,
+        )
 
+      next_fetch_indices_levels = [
+          _inc_grid_by_1(fetch_indices, grid)
+          for fetch_indices in fetch_index_levels
+      ]
       return (
           _inc_grid_by_1(indices, grid),
-          _inc_grid_by_1(fetch_indices, grid),
+          next_fetch_indices_levels,
           new_store_slices,
           next_body_carry if init_carry is not None else None,
       )
 
-    # Invariant: ``indices`` and ``fetch_indices`` are always
-    # ``max_concurrent_steps-delay_release`` apart.
-    fetch_indices = indices
-    for _ in range(max_concurrent_steps-delay_release):
-      fetch_indices = _inc_grid_by_1(fetch_indices, grid)
+    fetch_index_levels = []
+    for delay_release in delay_release_levels:
+      fetch_indices = indices
+      for _ in range(max_concurrent_steps - delay_release):
+        fetch_indices = _inc_grid_by_1(fetch_indices, grid)
+      fetch_index_levels.append(fetch_indices)
+
+    def _init_store_slice(bd):
+      if bd is None or isinstance(bd, pl.Squeezed):
+        return jnp.array(-1, dtype=jnp.int32)
+      return _Slice(-1, -1)
+
     # TODO(justinfu): Only store base pointer instead of all indices.
     last_store_slices = [
         None
         if bref.is_index_invariant
-        else (_Slice(-1, -1),) * len(bref.spec.block_shape)
+        else tuple(map(_init_store_slice, bref.spec.block_shape))
         for bref in out_brefs
     ]
     last_indices, _, _, final_carry = lax.fori_loop(
         0,
         num_steps,
         loop_body,
-        (indices, fetch_indices, last_store_slices, init_carry),
+        (indices, fetch_index_levels, last_store_slices, init_carry),
     )
 
     # Outputs invariant to the sequential axis are never written from inside the
@@ -399,7 +479,7 @@ def emit_pipeline(
 
       gpu_primitives.commit_smem_to_gmem_group()
 
-    if out_brefs:
+    if smem_out_brefs:
       # Finalize the pipeline.
       gpu_primitives.wait_smem_to_gmem(0)
     return final_carry if init_carry is not None else None
@@ -435,38 +515,55 @@ class ComputeContext(Protocol):
     ...
 
 
+class PipelinePipeline(enum.IntEnum):
+  START = 0
+  STEADY = 1
+  STOP = 2
+
+
+class WarpSpecializedPipeline(Protocol):
+  """Protocol for a warp specialized pipeline."""
+  def __call__(
+      self, *gmem_refs: Any, allocations: Any | None = None,
+  ) -> None:
+    ...
+
+  def get_allocations(self, *gmem_refs: Any) -> Any:
+    ...
+
+
 def emit_pipeline_warp_specialized(
     body: Callable[..., None],
     *,
     grid: pallas_core.TupleGrid,
     memory_registers: int,
-    in_specs: Sequence[pl.BlockSpec] = (),
-    out_specs: Sequence[pl.BlockSpec] = (),
+    in_specs: BlockSpecPytree = (),
+    out_specs: BlockSpecPytree = (),
     max_concurrent_steps: int = 2,
     wg_axis: str,
     num_compute_wgs: int,
+    pipeline_state: jax.Array | PipelinePipeline | None = None,
     manual_consumed_barriers: bool = False,
     compute_context: ComputeContext | None = None,
     memory_thread_idx: int | None = None,
-):
+) -> WarpSpecializedPipeline:
   """Creates a function to emit a warp-specialized pipeline.
 
   The ``body`` function should have the following signature (without carry).
   ``consumed_barriers`` is an optional argument that is only passed if the
-  ``manual_consumed_barriers`` argument is True.
+  ``manual_consumed_barriers`` argument is True::
 
-  ```
-  def body(indices, *input_refs, *output_refs, [consumed_barriers]) -> None:
-  ```
+    def body(indices, *input_refs, *output_refs, *consumed_barriers) -> None:
 
   or with a carries enabled (enabled via the ``compute_context`` argument),
-  where the body returns the next carry:
+  where the body returns the next carry::
 
-  ```
-  def body(
-      indices, *input_refs, *output_refs, [consumed_barriers], carry
-  ) -> Carry:
-  ```
+    def body(
+        indices, *input_refs, *output_refs, *consumed_barriers, carry
+    ) -> Carry:
+
+  When ``manual_consumed_barriers`` is True, the user must arrive on all the
+  consumed barriers from all compute warpgroups at each pipeline step.
 
   Args:
     body: The pipeline body.
@@ -493,9 +590,69 @@ def emit_pipeline_warp_specialized(
       final carry.
     memory_thread_idx: The index of the memory thread. If not specified,
       defaults to the last thread.
+    pipeline_state: If multiple pipelines that have almost the same parameters
+      (only in/out_specs and body can differ) are going to be evaluated
+      in sequence, this argument can be used to avoid pipeline bubbles between
+      their invocations. The first pipeline in the sequence should use the
+      ``START`` state, followed by an arbitrary number of ``STEADY`` states,
+      followed by a single ``STOP`` state. Note that until the pipeline with
+      ``STOP`` is done, the memory thread will not wait for the compute threads
+      to complete and fully consume their work. Any modification of their
+      operands other than invoking another pipeline is disallowed.
+
+      Important: To achieve bubble-free execution, it is important to also use
+      the manual allocation mode by calling ``get_allocations`` on the returned
+      function, passing the result to ``pl.run_scoped`` and the provided results
+      to the returned function as an ``allocations`` keyword argument.
+      Otherwise, the pipeline function will perform the scoped allocation itself
+      which can lead to synchronization that can still cause pipeline bubbles.
   """
+
   # TODO(justinfu): Factor out common code between warp-specialized and
   # normal pipelines.
+  if not isinstance(in_specs, (list, tuple)):
+    in_specs = (in_specs,)
+  if not isinstance(out_specs, (list, tuple)):
+    out_specs = (out_specs,)
+  if isinstance(in_specs, list):
+    in_specs = tuple(in_specs)
+  if isinstance(out_specs, list):
+    out_specs = tuple(out_specs)
+
+  flat_in_specs, in_specs_treedef = jax.tree.flatten(in_specs)
+  flat_in_specs = tuple(map(_downcast_spec, flat_in_specs))
+  for spec in flat_in_specs:
+    if len(spec.collective_axes) > 1:
+      raise ValueError(
+          "Only a single collective axis supported in input BlockSpecs, but"
+          f" got {spec.collective_axes}"
+      )
+  collective_axes = tuple(frozenset(
+      a for spec in flat_in_specs for a in spec.collective_axes
+  ))
+  flat_out_specs, out_specs_treedef = jax.tree.flatten(out_specs)
+  flat_out_specs = tuple(map(_downcast_spec, flat_out_specs))
+  for spec in flat_out_specs:
+    if spec.collective_axes:
+      raise ValueError("Output BlockSpecs cannot have collective_axes")
+  delay_release = None
+  for in_spec in in_specs:
+    if not isinstance(in_spec, gpu_core.BlockSpec):
+      delay_release = 0
+      continue
+    delay_release = in_spec.delay_release
+    if in_spec.delay_release != delay_release:
+      raise NotImplementedError(
+          "All inputs must have the same delay_release, but"
+          f" {in_spec.delay_release=} != {delay_release=}"
+      )
+
+  delay_release = delay_release or 0
+  if max_concurrent_steps <= delay_release:
+    raise ValueError(
+        "max_concurrent_steps must be greater than delay_release, but"
+        f" {max_concurrent_steps=}, {delay_release=}"
+    )
 
   if memory_thread_idx is None:
     memory_thread_idx = num_compute_wgs
@@ -509,10 +666,12 @@ def emit_pipeline_warp_specialized(
   # Trace the index maps to determine if they depend on the grid.
   # Grid-independent values will not be multiple-buffered.
   in_spec_has_seq_axis = [
-      not _is_index_invariant(spec, grid) for spec in in_specs]
+      not _is_index_invariant(spec, grid) for spec in flat_in_specs]
   out_spec_has_seq_axis = [
-      not _is_index_invariant(spec, grid) for spec in out_specs]
+      not _is_index_invariant(spec, grid) for spec in flat_out_specs]
   spec_has_seq_axis = [*in_spec_has_seq_axis, *out_spec_has_seq_axis]
+  if not all(in_spec_has_seq_axis):
+    raise NotImplementedError("Only inputs with a dependency on the grid are supported.")
 
   num_steps = math.prod(grid)
   has_dynamic_grid = not isinstance(num_steps, int)
@@ -529,39 +688,50 @@ def emit_pipeline_warp_specialized(
   if not has_dynamic_grid and max_concurrent_steps > num_steps:
     max_concurrent_steps = cast(int, num_steps)
 
-  def pipeline(*gmem_refs: state.AbstractRef):
-    in_gmem_refs, out_gmem_refs = util.split_list(gmem_refs, [len(in_specs)])
-    if len(out_gmem_refs) != len(out_specs):
+  def _get_scoped_allocs(*gmem_refs: AbstractRefPytree):
+    in_gmem_refs = gmem_refs[:len(in_specs)]
+    out_gmem_refs = gmem_refs[len(in_specs):]
+    flat_in_gmem_refs, in_gmem_refs_treedef = jax.tree.flatten(in_gmem_refs)
+    flat_out_gmem_refs, out_gmem_refs_treedef = jax.tree.flatten(out_gmem_refs)
+    if in_specs_treedef != in_gmem_refs_treedef:
       raise ValueError(
-          "Number of output refs does not match number of output specs."
+          "Input specs and input gmem refs must have the same pytree structure."
+          f" {in_specs_treedef} != {in_gmem_refs_treedef}"
       )
+    if out_specs_treedef != out_gmem_refs_treedef:
+      raise ValueError(
+          "Output specs and output gmem refs must have the same pytree structure."
+          f" {out_specs_treedef} != {out_gmem_refs_treedef}"
+      )
+    flat_gmem_refs = [*flat_in_gmem_refs, *flat_out_gmem_refs]
     smem_allocs = []
     for spec, has_seq_dim, gmem_ref in zip(
-        it.chain(in_specs, out_specs),
+        it.chain(flat_in_specs, flat_out_specs),
         spec_has_seq_axis,
-        gmem_refs):
+        flat_gmem_refs):
       slots = max_concurrent_steps if has_seq_dim else 1
       smem_allocs.append(
           gpu_core.SMEM(
-              (slots, *spec.block_shape),   # type: ignore
+              (slots, *_get_block_shape(spec)),   # type: ignore
               gmem_ref.dtype,
               transforms=getattr(spec, "transforms", ()),
           )
       )
-    in_smem_refs, out_smem_refs = util.split_list(
-        smem_allocs, [len(in_specs)])
-
-    in_smem_barriers = []
-    consumed_barriers = []
-    for has_seq_dim in in_spec_has_seq_axis:
-      num_barriers = max_concurrent_steps if has_seq_dim else 1
-      in_smem_barriers.append(
-          gpu_core.Barrier(
-          num_arrivals=1,
-          num_barriers=num_barriers))
+    flat_in_smem_refs, flat_out_smem_refs = util.split_list(
+        smem_allocs, [len(flat_in_specs)])
+    in_smem_barrier = gpu_core.Barrier(num_arrivals=len(flat_in_specs), num_barriers=max_concurrent_steps)
+    flat_consumed_barriers = []
+    consumed_barrier_type: Any
+    if collective_axes:
+      consumed_barrier_type = functools.partial(
+          gpu_core.ClusterBarrier, collective_axes=collective_axes  # type: ignore
+      )
+    else:
+      consumed_barrier_type = gpu_core.Barrier
+    for _ in flat_in_specs:
       if manual_consumed_barriers:
-        consumed_barriers.append(
-            gpu_core.Barrier(
+        flat_consumed_barriers.append(
+            consumed_barrier_type(
                 num_arrivals=num_compute_wgs,
                 num_barriers=max_concurrent_steps,
             )
@@ -569,44 +739,86 @@ def emit_pipeline_warp_specialized(
     if not manual_consumed_barriers:
       # We only allocated one consumed barrier for all inputs when using
       # automatic consumed barriers.
-      consumed_barriers = [
-          gpu_core.Barrier(
+      flat_consumed_barriers = [
+          consumed_barrier_type(
               num_arrivals=num_compute_wgs,
               num_barriers=max_concurrent_steps,
           )
       ]
-    return pl.run_scoped(
-        functools.partial(
-            scoped_pipeline,
-            in_gmem_refs=in_gmem_refs,
-            out_gmem_refs=out_gmem_refs,
-        ),
-        in_smem_refs=in_smem_refs,
-        out_smem_refs=out_smem_refs,
-        in_smem_barrier_refs=in_smem_barriers,
-        consumed_barrier_refs=consumed_barriers,
-        collective_axes=wg_axis,
+    return dict(
+        flat_in_smem_refs=flat_in_smem_refs,
+        flat_out_smem_refs=flat_out_smem_refs,
+        in_smem_barrier_ref=in_smem_barrier,
+        flat_consumed_barrier_refs=flat_consumed_barriers,
     )
 
+  def pipeline(*gmem_refs: AbstractRefPytree, allocations: Any | None = None):
+    """
+    Run the pipeline.
+
+    Args:
+      *gmem_refs: A list of pytrees of pallas refs
+      allocations: The allocation provided by ``pl.run_scoped`` when the result
+        of calling ``get_allocations(*gmem_refs)`` is passed to
+        ``pl.run_scoped``.
+    """
+    in_gmem_refs = gmem_refs[:len(in_specs)]
+    out_gmem_refs = gmem_refs[len(in_specs):]
+    flat_in_gmem_refs, in_gmem_refs_treedef = jax.tree.flatten(in_gmem_refs)
+    flat_out_gmem_refs, out_gmem_refs_treedef = jax.tree.flatten(out_gmem_refs)
+    if in_specs_treedef != in_gmem_refs_treedef:
+      raise ValueError(
+          "Input specs and input gmem refs must have the same pytree structure."
+          f" {in_specs_treedef} != {in_gmem_refs_treedef}"
+      )
+    if out_specs_treedef != out_gmem_refs_treedef:
+      raise ValueError(
+          "Output specs and output gmem refs must have the same pytree structure."
+          f" {out_specs_treedef} != {out_gmem_refs_treedef}"
+      )
+
+    if allocations is None:
+      if pipeline_state is not None:
+        raise ValueError(
+            "Pipeline state should not be set when using automatic allocation."
+        )
+      return pl.run_scoped(
+          functools.partial(
+              scoped_pipeline,
+              flat_in_gmem_refs=flat_in_gmem_refs,
+              flat_out_gmem_refs=flat_out_gmem_refs,
+          ),
+          **_get_scoped_allocs(*gmem_refs),
+          collective_axes=wg_axis,
+      )
+    else:
+      scoped_pipeline(
+          flat_in_gmem_refs=flat_in_gmem_refs,
+          flat_out_gmem_refs=flat_out_gmem_refs,
+          **allocations,
+      )
+
+  pipeline.get_allocations = _get_scoped_allocs
+
   def scoped_pipeline(
       *,
-      in_gmem_refs,
-      out_gmem_refs,
-      in_smem_refs,
-      out_smem_refs,
-      in_smem_barrier_refs,
-      consumed_barrier_refs,
+      flat_in_gmem_refs,
+      flat_out_gmem_refs,
+      flat_in_smem_refs,
+      flat_out_smem_refs,
+      in_smem_barrier_ref,
+      flat_consumed_barrier_refs,
   ):
-    in_brefs: Sequence[BufferedRef] = [
+    flat_in_brefs: Sequence[BufferedRef] = [
         BufferedRef(spec, not has_seq_axis, gmem_ref, smem_ref)
         for spec, has_seq_axis, gmem_ref, smem_ref in zip(
-            in_specs, in_spec_has_seq_axis, in_gmem_refs, in_smem_refs
+            flat_in_specs, in_spec_has_seq_axis, flat_in_gmem_refs, flat_in_smem_refs
         )
     ]
-    out_brefs: Sequence[BufferedRef] = [
+    flat_out_brefs: Sequence[BufferedRef] = [
         BufferedRef(spec, not has_seq_axis, gmem_ref, smem_ref)
         for spec, has_seq_axis, gmem_ref, smem_ref in zip(
-            out_specs, out_spec_has_seq_axis, out_gmem_refs, out_smem_refs
+            flat_out_specs, out_spec_has_seq_axis, flat_out_gmem_refs, flat_out_smem_refs
         )
     ]
 
@@ -616,52 +828,82 @@ def emit_pipeline_warp_specialized(
           action="increase")
 
       # This is true if any of the outputs need to be transferred inside the loop.
-      copies_out_in_loop = not all(bref.is_index_invariant for bref in out_brefs)
-      needs_epilogue = any(bref.is_index_invariant for bref in out_brefs)
+      smem_out_brefs = [bref for bref in flat_out_brefs if _in_smem(bref.spec)]
+      # The implementation below has races when we have multiple compute WGs.
+      # The problem is that we expect the compute WGs to deal with issuing the
+      # SMEM->GMEM copies, but (1) we never predicate them, so we repeat the
+      # same copy multiple times, and (2) we don't synchronize the compute WGs
+      # in any way. In the unlikely event that one of the compute WGs runs 2
+      # steps ahead, it might start overwriting the output buffer before the
+      # other WG has issued its copy.
+      #
+      # The best fix here would be to move the SMEM->GMEM copies into the memory
+      # WG and use proper barriers (with arrival_count=2) to ensure all WGs have
+      # produced their outputs before it is sent out to GMEM.
+      if smem_out_brefs and num_compute_wgs > 1:
+        raise NotImplementedError(
+            "SMEM outputs are not supported with multiple compute warpgroups"
+        )
+      copies_out_in_loop = not all(bref.is_index_invariant for bref in smem_out_brefs)
+      needs_epilogue = any(bref.is_index_invariant for bref in smem_out_brefs)
 
       def compute_loop_body(step, carry):
         indices, last_store_slices, prev_body_carry = carry
         slot = lax.rem(step, max_concurrent_steps)
+        consumed_slot = lax.rem(step - delay_release, max_concurrent_steps)
         # Wait for the current GMEM->SMEM copies to complete.
-        for in_barrier, has_seq_dim in zip(
-            in_smem_barrier_refs, in_spec_has_seq_axis):
-          # TODO(justinfu): Use a single barrier with
-          # num_arrivals=len(in_smem_barrier_refs)
-          gpu_primitives.barrier_wait(
-              in_barrier.at[_get_slot(slot, has_seq_dim)])
+        gpu_primitives.barrier_wait(in_smem_barrier_ref.at[_get_slot(slot, True)])
 
         # Wait for the previous output SMEM->GMEM copy to complete.
         if copies_out_in_loop:
-          gpu_primitives.wait_smem_to_gmem(max_concurrent_steps - 1)
+          gpu_primitives.wait_smem_to_gmem(
+              max_concurrent_steps - 1, wait_read_only=True
+          )
 
-        body_refs = []
-        for bref in it.chain(in_brefs, out_brefs):
-          buf_slot = _get_slot(slot, not bref.is_index_invariant)
-          body_refs.append(bref.get_ref_for_slot(buf_slot))
+        in_brefs = jax.tree.unflatten(in_specs_treedef, flat_in_brefs)
+        out_brefs = jax.tree.unflatten(out_specs_treedef, flat_out_brefs)
+        all_brefs = (*in_brefs, *out_brefs)
+        body_args = map_brefs(
+            lambda bref: bref.get_ref_for_slot(
+                _get_slot(slot, not bref.is_index_invariant)
+            ),
+            all_brefs,
+        )
 
-        body_args = body_refs
         if manual_consumed_barriers:
-          body_args += [consumed_barrier_ref.at[slot] for consumed_barrier_ref in consumed_barrier_refs]
+          barriers = jax.tree.unflatten(
+              in_specs_treedef,
+              [barrier.at[consumed_slot] for barrier in flat_consumed_barrier_refs],
+          )
+          body_args = (*body_args, *barriers)
         if has_carry:
-          body_args += [prev_body_carry]
+          body_args = (*body_args, prev_body_carry)
         next_body_carry = body(indices, *body_args)
 
         if not manual_consumed_barriers:
-          [consumed_barrier_ref] = consumed_barrier_refs
-          gpu_primitives.barrier_arrive(consumed_barrier_ref.at[slot])
+          [consumed_barrier_ref] = flat_consumed_barrier_refs
+          if delay_release > 0:
+            lax.cond(
+                step < delay_release,
+                lambda: None,
+                lambda: gpu_primitives.barrier_arrive(consumed_barrier_ref.at[consumed_slot]),
+            )
+          else:
+            gpu_primitives.barrier_arrive(consumed_barrier_ref.at[consumed_slot])
         # TODO(justinfu,apaszke): This should probably be done by the memory WG.
         # Copy the output from SMEM to GMEM.
         if copies_out_in_loop:
           gpu_primitives.commit_smem()
 
         new_store_slices = last_store_slices[:]
-        for idx, bref in enumerate(out_brefs):
+        for idx, bref in enumerate(flat_out_brefs):
           if bref.is_index_invariant:
             assert last_store_slices[idx] is None
             continue
           assert last_store_slices[idx] is not None
           new_store_slices[idx] = tuple(
-              _Slice(s.start, s.size) for s in bref.compute_gmem_slice(indices)
+              _Slice(s.start, s.size) if isinstance(s, pl.Slice) else s
+              for s in bref.compute_gmem_slice(indices)
           )
           are_same_slices = map(
               lambda old, new: old == new,
@@ -676,12 +918,18 @@ def emit_pipeline_warp_specialized(
         next_indices = _inc_grid_by_1(indices, grid)
         return (next_indices, new_store_slices, next_body_carry)
       init_indices = (jnp.asarray(0, dtype=jnp.int32),) * len(grid)
+
+      def _init_store_slice(bd):
+        if bd is None or isinstance(bd, pl.Squeezed):
+          return jnp.array(-1, dtype=jnp.int32)
+        return _Slice(-1, -1)
+
       # TODO(justinfu): Only store base pointer instead of all indices.
       last_store_slices = [
           None
           if bref.is_index_invariant
-          else (_Slice(-1, -1),) * len(bref.spec.block_shape)
-          for bref in out_brefs
+          else tuple(map(_init_store_slice, bref.spec.block_shape))
+          for bref in flat_out_brefs
       ]
 
       if has_carry:
@@ -714,18 +962,19 @@ def emit_pipeline_warp_specialized(
 
       if needs_epilogue:
         last_slot = lax.rem(num_steps - 1, max_concurrent_steps)
-        for bref in out_brefs:
+        for bref in flat_out_brefs:
           if bref.is_index_invariant:
             bref.copy_out(_get_slot(last_slot, has_seq_dim=False),
                           last_indices, predicate=None)
 
         gpu_primitives.commit_smem_to_gmem_group()
 
-      if out_brefs:
+      if smem_out_brefs:
         # Finalize the pipeline.
         gpu_primitives.wait_smem_to_gmem(0)
 
     # The memory thread executes this block which issues all pipelined DMAs.
+    # TODO(apaszke,justinfu): Use a single arrive_expect_tx for all transfers.
     def memory_block():
       gpu_primitives.set_max_registers(memory_registers, action="decrease")
       indices = (jnp.asarray(0, dtype=jnp.int32),) * len(grid)
@@ -734,12 +983,33 @@ def emit_pipeline_warp_specialized(
       else:
         assert max_concurrent_steps <= num_steps
         prologue_steps = max_concurrent_steps
+      pipeline_init_prologue_steps = prologue_steps
+      if pipeline_state is not None:
+        if has_dynamic_grid:
+          raise NotImplementedError(
+              "A pipeline of pipelines is not supported with dynamic grids"
+          )
+        if num_steps % max_concurrent_steps:
+          raise NotImplementedError(
+              "A pipeline of pipelines is only allowed when the number of steps"
+              f" (product of grid, here {num_steps}) is divisible by"
+              f" {max_concurrent_steps=}"
+          )
+        if delay_release:
+          raise NotImplementedError(
+              "A pipeline of pipelines is not supported with delay_release"
+          )
+        if isinstance(pipeline_state, PipelinePipeline):
+          prologue_steps = prologue_steps if pipeline_state == PipelinePipeline.START else 0
+        else:
+          prologue_steps = jnp.where(pipeline_state == PipelinePipeline.START, prologue_steps, 0)
 
       # Begin initial copies.
       def _init_step(step, indices):
-        for bref, barrier in zip(in_brefs, in_smem_barrier_refs):
+        for bref in flat_in_brefs:
           buf_slot = _get_slot(step, not bref.is_index_invariant)
-          bref.copy_in(buf_slot, indices, barrier)
+          barrier_slot = _get_slot(step, True)
+          bref.copy_in(buf_slot, indices, in_smem_barrier_ref, barrier_slot)
         return _inc_grid_by_1(indices, grid)
 
       indices = jax.lax.fori_loop(
@@ -754,28 +1024,35 @@ def emit_pipeline_warp_specialized(
         if not manual_consumed_barriers:
           # We only have one consumed barrier when using automatic consumed
           # barrier management.
-          [consumed_barrier_ref] = consumed_barrier_refs
+          [consumed_barrier_ref] = flat_consumed_barrier_refs
           gpu_primitives.barrier_wait(consumed_barrier_ref.at[slot])
-          consumed_barrier_it = [None] * len(in_brefs)
+          consumed_barrier_it = [None] * len(flat_in_brefs)
         else:
-          consumed_barrier_it = consumed_barrier_refs
+          consumed_barrier_it = flat_consumed_barrier_refs
 
-        for bref, barrier, consumed_barrier in zip(
-            in_brefs, in_smem_barrier_refs, consumed_barrier_it):
+        for bref, consumed_barrier in zip(flat_in_brefs, consumed_barrier_it):
           if manual_consumed_barriers:
             gpu_primitives.barrier_wait(consumed_barrier.at[slot])  # pytype: disable=attribute-error
-          bref.copy_in(
-              _get_slot(fetch_slot, not bref.is_index_invariant), indices, barrier)
+          buf_slot = _get_slot(fetch_slot, not bref.is_index_invariant)
+          barrier_slot = _get_slot(fetch_slot, True)
+          bref.copy_in(buf_slot, indices, in_smem_barrier_ref, barrier_slot)
         next_indices = _inc_grid_by_1(indices, grid)
         return (next_indices,)
-      lax.fori_loop(0, num_steps - max_concurrent_steps,
-                    memory_loop_body, (indices,))
+      lax.fori_loop(0, num_steps - prologue_steps, memory_loop_body, (indices,))
       # Await all the arrivals to not leave barriers in a bad state.
-      # We only need to account for the prologue steps.
-      @pl.loop(0, prologue_steps, unroll=not has_dynamic_grid)
-      def _epi_step(step):
-        for barrier in consumed_barrier_refs:
-          gpu_primitives.barrier_wait(barrier.at[step])
+      # We only need to account for the prologue steps, only the first
+      # delay_release of them skip arrivals, so we subtract them.
+      @pl.when(pipeline_state is None or pipeline_state == PipelinePipeline.STOP)
+      def _quiesce():
+        @pl.loop(
+            num_steps - pipeline_init_prologue_steps,
+            num_steps - delay_release,
+            unroll=not has_dynamic_grid,
+        )
+        def _epi_step(step):
+          consumed_slot = lax.rem(step, max_concurrent_steps)
+          for barrier in flat_consumed_barrier_refs:
+            gpu_primitives.barrier_wait(barrier.at[consumed_slot])
 
     wg_idx = lax.axis_index(wg_axis)
     lax.cond(
@@ -783,7 +1060,8 @@ def emit_pipeline_warp_specialized(
         compute_block,
         memory_block
     )
-  return pipeline
+  # Mypy doesn't notice the .get_allocations assignment above.
+  return pipeline  # type: ignore
 
 def _compute_registers(
     memory_registers: int,
