@@ -20,34 +20,37 @@ from collections.abc import Callable, Sequence
 from collections.abc import Hashable
 import enum
 import functools
+import math
 import string
 from typing import Any
 
-import jax
-from jax import lax
-from jax import tree_util
+import jax._src.lax as lax
+from jax._src import tree_util
 from jax._src import ad_util
 from jax._src import api_util
-from jax._src import callback
 from jax._src import core as jax_core
+from jax._src import config
+from jax._src import debugging
 from jax._src import dtypes
+from jax._src import typing as jax_typing
 from jax._src import effects
 from jax._src import linear_util as lu
 from jax._src import pretty_printer as pp
 from jax._src import state
 from jax._src import util
 from jax._src.interpreters import ad
-from jax._src.interpreters import batching
 from jax._src.interpreters import partial_eval as pe
+from jax._src.lib.mlir import ir
+from jax._src.lib.mlir.dialects import arith
 from jax._src.pallas import core as pallas_core
+from jax._src.pallas import utils as pallas_utils
 from jax._src.state import discharge as state_discharge
 from jax._src.state import indexing
 from jax._src.state import primitives as sp
 from jax._src.state import types as state_types
 from jax.interpreters import mlir
-import jax.numpy as jnp
+from jax._src import numpy as jnp
 
-partial = functools.partial
 Slice = indexing.Slice
 NDIndexer = indexing.NDIndexer
 
@@ -55,16 +58,15 @@ map, unsafe_map = util.safe_map, map
 zip, unsafe_zip = util.safe_zip, zip
 
 program_id_p = jax_core.Primitive("program_id")
-batching.ragged_prop_rules[program_id_p] = batching.ragged_mask_no_op_rule
 
-def program_id(axis: int) -> jax.Array:
+def program_id(axis: int) -> jax_typing.Array:
   """Returns the kernel execution position along the given axis of the grid.
 
-  For example, with a 2D `grid` in the kernel execution corresponding to the
-  grid coordinates `(1, 2)`,
-  `program_id(axis=0)` returns `1` and `program_id(axis=1)` returns `2`.
+  For example, with a 2D ``grid`` in the kernel execution corresponding to the
+  grid coordinates ``(1, 2)``,
+  ``program_id(axis=0)`` returns ``1`` and ``program_id(axis=1)`` returns ``2``.
 
-  The returned value is an array of shape `()` and dtype `int32`.
+  The returned value is an array of shape ``()`` and dtype ``int32``.
 
   Args:
     axis: the axis of the grid along which to count the program.
@@ -90,7 +92,7 @@ def _program_id_abstract_eval(**_):
 
 num_programs_p = jax_core.Primitive("num_programs")
 
-def num_programs(axis: int) -> int | jax.Array:
+def num_programs(axis: int) -> int | jax_typing.Array:
   """Returns the size of the grid along the given axis."""
   return num_programs_p.bind(axis=axis)
 
@@ -346,6 +348,8 @@ max_contiguous_p.def_impl(lambda x, **_: x)
 mlir.register_lowering(max_contiguous_p, lambda _, x, **__: [x])
 
 def max_contiguous(x, values):
+  """A compiler hint that asserts the ``values`` first values of ``x`` are contiguous.
+  """
   if not isinstance(values, (list, tuple)):
     values = (values,)
   return max_contiguous_p.bind(x, values=tuple(values))
@@ -359,7 +363,19 @@ multiple_of_p = jax_core.Primitive("multiple_of")
 multiple_of_p.def_impl(lambda x, **_: x)
 mlir.register_lowering(multiple_of_p, lambda _, x, **__: [x])
 
-def multiple_of(x: jax.Array, values: Sequence[int] | int) -> jax.Array:
+def multiple_of(x: jax_typing.Array, values: Sequence[int] | int) -> jax_typing.Array:
+  """A compiler hint that asserts a value is a static multiple of another.
+
+  Note that misusing this function, such as asserting ``x`` is a multiple of
+  ``N`` when it is not, can result in undefined behavior.
+
+  Args:
+    x: The input array.
+    values: A set of static divisors that ``x`` is a multiple of.
+
+  Returns:
+    A copy of ``x``.
+  """
   values = (values,) if isinstance(values, int) else tuple(values)
   return multiple_of_p.bind(x, values=values)
 
@@ -373,7 +389,7 @@ load_p = jax_core.Primitive('masked_load')
 @load_p.def_effectful_abstract_eval
 def _load_abstract_eval(*avals_flat, args_tree, **_):
   ref, transforms, _, _ = args_tree.unflatten(avals_flat)
-  assert transforms is not None and isinstance(transforms[-1], NDIndexer)
+  assert transforms is not None
   transformed_ref = pallas_core.TransformedRef(ref, transforms)
   return (
       jax_core.ShapedArray(transformed_ref.shape, transformed_ref.dtype),
@@ -469,14 +485,19 @@ def _pad_values_to_avoid_dynamic_slice_oob_shift(value,
                   padding_value=padding_value)
   return value
 
-_unpad_values_to_avoid_dynamic_slice_oob_shift = partial(
-  _pad_values_to_avoid_dynamic_slice_oob_shift, unpad=True)
+_unpad_values_to_avoid_dynamic_slice_oob_shift = functools.partial(
+    _pad_values_to_avoid_dynamic_slice_oob_shift, unpad=True
+)
 
 
 @state_discharge.register_discharge_rule(load_p)
 def _load_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
   del out_avals  # Unused.
   ref, transforms, mask, other = args_tree.unflatten(args_flat)
+  transforms = list(transforms)
+  if not transforms or not isinstance(transforms[-1], indexing.NDIndexer):
+    ref_shape = state.get_transforms_shape(transforms, in_avals[0].shape)
+    transforms.append(indexing.NDIndexer.make_trivial_indexer(ref_shape))
   *prev_transforms, idx = transforms
   assert isinstance(idx, NDIndexer)
   ref = state_discharge.transform_array(ref, prev_transforms)
@@ -493,7 +514,7 @@ def _load_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
     # of bounds, it will instead move the start_index backwards so the slice
     # will fit in memory.
     ref = _pad_values_to_avoid_dynamic_slice_oob_shift(ref, slice_sizes)
-    idx_dtype = dtypes.canonicalize_dtype(jnp.int64)
+    idx_dtype = dtypes.default_int_dtype()
     out_ones = lax.dynamic_slice(
         ref,
         [jnp.astype(s, idx_dtype) for s in slice_starts],
@@ -515,8 +536,8 @@ swap_p = jax_core.Primitive('masked_swap')
 
 @swap_p.def_effectful_abstract_eval
 def _swap_abstract_eval(*avals_flat, args_tree, **_):
-  ref, transforms, val, _ = args_tree.unflatten(avals_flat)
-  assert transforms is not None and isinstance(transforms[-1], NDIndexer)
+  ref, transforms, val, mask = args_tree.unflatten(avals_flat)
+  assert transforms is not None
   transformed_ref = pallas_core.TransformedRef(ref, transforms)
   expected_output_shape = transformed_ref.shape
   expected_output_dtype = transformed_ref.dtype
@@ -592,6 +613,10 @@ ad.primitive_jvps[swap_p] = _swap_jvp
 def _swap_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
   del out_avals  # Unused.
   ref, transforms, val, mask = args_tree.unflatten(args_flat)
+  transforms = list(transforms)
+  if not transforms or not isinstance(transforms[-1], indexing.NDIndexer):
+    ref_shape = state.get_transforms_shape(transforms, in_avals[0].shape)
+    transforms.append(indexing.NDIndexer.make_trivial_indexer(ref_shape))
   *prev_transforms, idx = transforms
   assert isinstance(idx, NDIndexer)
   ref = state_discharge.transform_array(ref, prev_transforms)
@@ -634,7 +659,7 @@ def _swap_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
 
 
 def load(x_ref_or_view, idx, *, mask=None, other=None, cache_modifier=None,
-         eviction_policy=None, volatile=False) -> jax.Array:
+         eviction_policy=None, volatile=False) -> jax_typing.Array:
   """Returns an array loaded from the given index.
 
   If neither ``mask`` nor ``other`` is specified, this function has the same
@@ -664,7 +689,7 @@ def load(x_ref_or_view, idx, *, mask=None, other=None, cache_modifier=None,
   )
 
 def swap(x_ref_or_view, idx, val, *, mask=None, eviction_policy=None,
-         _function_name="swap") -> jax.Array:
+         _function_name="swap") -> jax_typing.Array:
   """Swaps the value at the given index and returns the old value.
 
   See :func:`~jax.experimental.pallas.load` for the meaning of the arguments.
@@ -688,8 +713,36 @@ def store(x_ref_or_view, idx, val, *, mask=None, eviction_policy=None) -> None:
   _ = swap(x_ref_or_view, idx, val, mask=mask, eviction_policy=eviction_policy,
            _function_name="store")
 
+
+def _handle_small(dtype: jax_typing.DTypeLike):
+  """Ugly workaround to support types that don't allow automatic promotion."""
+  if dtype == jnp.int4:
+    return jnp.int8
+  if dtype == jnp.float8_e4m3b11fnuz:
+    return jnp.bfloat16
+  return dtype
+
+
 def dot(a, b, trans_a: bool = False, trans_b: bool = False,
         allow_tf32: bool | None = None, precision=None):
+  """Computes the dot product of two arrays.
+
+  The inputs can optionally be transposed before computing the
+  product. Depending on the hardware, this can be cheaper than
+  computing the transpose beforehand.
+
+  Args:
+    a: The left-hand size of the dot product, of shape ``(..., N)``.
+    b: The right-hand size of the dot product, of shape ``(...N, M)``.
+    trans_a: Whether to transpose ``a`` before the product.
+    trans_b: Whether to transpose ``b`` before the product.
+    allow_tf32: Whether to use tf32 precision.
+      Mutually exclusive with ``precision``.
+    precision: Specifies the precision of the dot product.
+
+  See Also:
+    :func:`jax.numpy.dot`
+  """
   if (a.ndim != 2) or (b.ndim != 2):
     raise ValueError("`a` and `b` must be 2D arrays.")
   lhs_contract_dim = 0 if trans_a else 1
@@ -699,15 +752,9 @@ def dot(a, b, trans_a: bool = False, trans_b: bool = False,
       raise ValueError("Only one of allow_tf32 and precision can be specified")
     precision = lax.Precision.HIGH if allow_tf32 else lax.Precision.HIGHEST
 
-  def _handle_f8(dtype: jax.typing.DTypeLike):
-    """Ugly workaround to support float8_e4m3b11fnuz in dot."""
-    if dtype == jnp.float8_e4m3b11fnuz:
-      return jnp.bfloat16
-    return dtype
-
-  dtype = jnp.promote_types(_handle_f8(a.dtype), _handle_f8(b.dtype))
+  dtype = jnp.promote_types(_handle_small(a.dtype), _handle_small(b.dtype))
   out_dtype = jnp.int32 if jnp.issubdtype(dtype, jnp.integer) else jnp.float32
-  return jax.lax.dot_general(
+  return lax.dot_general(
       a,
       b,
       dimension_numbers=(((lhs_contract_dim,), (rhs_contract_dim,)), ((), ())),
@@ -744,24 +791,7 @@ def _reciprocal_lowering_rule(
 mlir.register_lowering(reciprocal_p, _reciprocal_lowering_rule)
 
 
-class PrintEffect(effects.Effect):
-  __str__ = lambda self: "Print"
-
-
-debug_print_effect = PrintEffect()
-
-# TODO(slebedev): Consider making the effect ordered.
-effects.lowerable_effects.add_type(PrintEffect)
-effects.control_flow_allowed_effects.add_type(PrintEffect)
-effects.remat_allowed_effects.add_type(PrintEffect)
-effects.custom_derivatives_allowed_effects.add_type(PrintEffect)
-
-
-debug_print_p = jax_core.Primitive("debug_print")
-debug_print_p.multiple_results = True
-
-
-def debug_print(fmt: str, *args: jax.typing.ArrayLike):
+def debug_print(fmt: str, *args: jax_typing.ArrayLike):
   """Prints values from inside a Pallas kernel.
 
   Args:
@@ -772,7 +802,8 @@ def debug_print(fmt: str, *args: jax.typing.ArrayLike):
         (``{...}``), since it is always printed before any of the values.
       * On GPU, when using the experimental Mosaic GPU backend, ``fmt`` must
         contain a placeholder for each value to be printed. Format specs and
-        conversions are not supported. All values must be scalars.
+        conversions are not supported. If a single value is provided, the value
+        may be an array. Otherwise, all values must be scalars.
       * On TPU, if all inputs are scalars: If ``fmt`` contains placeholders,
         all values must be 32-bit integers. If there are no placeholders, the
         values are printed after the format string.
@@ -780,16 +811,12 @@ def debug_print(fmt: str, *args: jax.typing.ArrayLike):
         the format string. The format string must end with a single placeholder
         ``{}``.
     *args: The values to print.
-  """  # fmt: skip
-  has_placeholders = False
-  if fmt:
-    _, field_name, *_ = next(iter(string.Formatter().parse(fmt)))
-    has_placeholders = field_name is not None
-  return debug_print_p.bind(*args, fmt=fmt, has_placeholders=has_placeholders)
+  """
+  return debugging.debug_print(fmt, *args, skip_format_check=True)
 
 
 def check_debug_print_format(
-    fmt: str, *args: jax.typing.ArrayLike
+    fmt: str, *args: jax_typing.ArrayLike
 ):
   n_placeholders = 0
   for _, field, spec, conversion in string.Formatter().parse(fmt):
@@ -811,59 +838,6 @@ def check_debug_print_format(
     )
 
 
-@debug_print_p.def_impl
-def debug_print_impl(*args: Any, fmt: str, has_placeholders: bool):
-  if has_placeholders:
-    print(fmt.format(*args))
-  else:
-    print(fmt, *args)
-  return ()
-
-
-@debug_print_p.def_effectful_abstract_eval
-def debug_print_abstract_eval(*avals: Any, fmt: str, has_placeholders: bool):
-  del avals, fmt, has_placeholders  # Unused.
-  return [], {debug_print_effect}
-
-
-def debug_print_batching_rule(args, dims, **params):
-  """Unrolls the print primitive across the mapped axis."""
-  axis_size = next(x.shape[i] for x, i in zip(args, dims) if i is not None)
-
-  # TODO(sharadmv): implement in terms of rolled loop unstead of unrolled.
-  def get_arg_at_dim(i, dim, arg):
-    if dim is batching.not_mapped:
-      # Broadcast unmapped argument
-      return arg
-    return lax.index_in_dim(arg, i, axis=dim, keepdims=False)
-
-  outs = []
-  for i in range(axis_size):
-    args_idx = map(functools.partial(get_arg_at_dim, i), dims, args)
-    outs.append(debug_print_p.bind(*args_idx, **params))
-  outs = [jnp.stack(xs) for xs in zip(*outs)]
-  return outs, (0,) * len(outs)
-
-
-batching.primitive_batchers[debug_print_p] = functools.partial(
-    debug_print_batching_rule, debug_print_p
-)
-
-
-@functools.partial(mlir.register_lowering, debug_print_p)
-def debug_print_lowering_rule(ctx, *args, **params):
-  result, _, _ = callback.emit_python_callback(
-      ctx,
-      functools.partial(debug_print_p.impl, **params),
-      None,
-      list(args),
-      ctx.avals_in,
-      ctx.avals_out,
-      has_side_effect=True,
-  )
-  return result
-
-
 # All of those shenanigans are because we can't make TransformedRef a PyTree,
 # because they should appear as atomic JAX values to the users.
 # TODO(apaszke): This can be deleted once we make transforms in Mosaic GPU
@@ -880,6 +854,17 @@ def wrap_with_transforms(f, transforms, *args):
 run_scoped_p = jax_core.Primitive("run_scoped")
 run_scoped_p.multiple_results = True
 
+def _run_scoped_is_high(*avals, jaxpr, **params):
+  del avals, params
+  return jaxpr.is_high
+run_scoped_p.is_high = _run_scoped_is_high  # type: ignore[method-assign]
+
+def _run_scoped_to_lojax(*args, jaxpr, **params):
+  closed_hi_jaxpr = jax_core.ClosedJaxpr(jaxpr, args)
+  closed_lo_jaxpr = pe.lower_jaxpr(closed_hi_jaxpr)
+  consts = closed_lo_jaxpr.consts
+  return run_scoped_p.bind(*consts, jaxpr=closed_lo_jaxpr.jaxpr, **params)
+run_scoped_p.to_lojax = _run_scoped_to_lojax
 
 def run_scoped(
     f: Callable[..., Any],
@@ -893,9 +878,9 @@ def run_scoped(
   to allocate for each argument. Each backend has its own set of reference
   types in addition to :class:`jax.experimental.pallas.MemoryRef`.
 
-  When `collective_axes` is specified, the same allocation will be returned for
+  When ``collective_axes`` is specified, the same allocation will be returned for
   all programs that only differ in their program ids along the collective axes.
-  It is an error not to call the same `run_scoped` in all programs along that
+  It is an error not to call the same ``run_scoped`` in all programs along that
   axis.
   """
   if not isinstance(collective_axes, tuple):
@@ -922,7 +907,8 @@ def run_scoped(
   # parent scope). Jax can't reason about effects to references that
   # are not in the invars of an operation so we just put them all
   # there.
-  jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(flat_fun, avals)
+  with config.mutable_array_checks(False):
+    jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(flat_fun, avals)
   out = run_scoped_p.bind(*consts, jaxpr=jaxpr, collective_axes=collective_axes)
   return tree_util.tree_unflatten(out_tree_thunk(), out)
 
@@ -1021,6 +1007,48 @@ def _run_scoped_lowering_rule(ctx, *args, jaxpr, collective_axes):
   return mlir.lower_fun(_lower_fun, multiple_results=True)(ctx, *args)
 
 
+get_global_p = jax_core.Primitive("get_global")
+get_global_p.multiple_results = False
+get_global_p.ref_primitive = True
+jax_core._ref_allocating_primitives.add(get_global_p)
+
+def get_global(what: pallas_core.ScratchShape) -> jax_typing.Array:
+  """Returns a global reference that persists across all kernel invocations.
+
+  Each call to ``get_global`` returns a different and unique reference, but one that
+  is stable across invocations of the kernel body.
+
+  Args:
+    what: The reference type to allocate. Each backend has its own set of
+      reference types (e.g., :class:`jax.experimental.pallas.mosaic_gpu.SemaphoreType` for GPU).
+
+  Example::
+
+    sem_ref = pl.get_global(plgpu.SemaphoreType.REGULAR)
+    pl.semaphore_signal(sem_ref)
+    pl.semaphore_wait(sem_ref)
+  """
+  ref_aval = what.get_ref_aval()
+  return get_global_p.bind(what=ref_aval)
+
+
+@get_global_p.def_abstract_eval
+def _get_global_abstract_eval(*, what):
+  return what
+
+
+def _get_global_discharge_rule(in_avals, out_avals, *, what):
+  del in_avals, out_avals, what
+  raise NotImplementedError(
+      "get_global discharge is not supported in interpret mode."
+  )
+
+
+state_discharge.register_discharge_rule(get_global_p)(
+    _get_global_discharge_rule
+)
+
+
 def _get_ref_and_transforms(ref):
   if isinstance(ref, state.TransformedRef):
     return ref.ref, ref.transforms
@@ -1056,7 +1084,7 @@ def check_sem_avals(
   ):
     raise ValueError(
         f"Must {name} semaphores of the following types:"
-        f" {allowed_semaphore_types}."
+        f" {allowed_semaphore_types}. Got {sem_dtype}."
     )
 
 
@@ -1077,7 +1105,15 @@ semaphore_read_p = jax_core.Primitive("semaphore_read")
 semaphore_read_p.multiple_results = False
 
 
-def semaphore_read(sem_or_view):
+def semaphore_read(sem_or_view) -> jax_typing.Array:
+  """Reads the value of a semaphore.
+
+  Args:
+    sem_or_view: A Ref (or view) representing a semaphore.
+
+  Returns:
+    A scalar Array containing the value of the semaphore.
+  """
   ref, transforms = _get_ref_and_transforms(sem_or_view)
   args = [ref, transforms]
   flat_args, args_tree = tree_util.tree_flatten(args)
@@ -1105,18 +1141,39 @@ state_discharge.register_discharge_rule(semaphore_read_p)(
 )
 
 
+DeviceId = int | jax_typing.Array | None | tuple[int | jax_typing.Array, ...] | dict[Any, int | jax_typing.Array]
+
+
 semaphore_signal_p = jax_core.Primitive('semaphore_signal')
 semaphore_signal_p.multiple_results = True
 
 
 def semaphore_signal(
     sem_or_view,
-    inc: int | jax.Array = 1,
+    inc: int | jax_typing.Array = 1,
     *,
-    device_id: int | jax.Array | None | tuple[int | jax.Array, ...] = None,
+    device_id: DeviceId = None,
     device_id_type: DeviceIdType = DeviceIdType.MESH,
-    core_index: int | jax.Array | None = None,
+    core_index: int | jax_typing.Array | None = None,
 ):
+  """Increments the value of a semaphore.
+
+  This operation can also be performed remotely if ``device_id`` is specified,
+  in which ``sem_or_view`` refers to a Ref located on another device.
+  Note that it is assumed that ``sem_or_view`` is already allocated
+  (e.g. through the proper use of barriers), or else this operation could
+  result in undefined behavior.
+
+  Args:
+    sem_or_view: A Ref (or view) representing a semaphore.
+    inc: The value to increment by.
+    device_id (optional): Specifies which device to signal.
+      If not specified, ``sem_or_view`` is assumed to be local.
+    device_id_type (optional): The format in which
+      ``device_id`` should be specified.
+    core_index (optional): If on a multi-core device,
+      specifies which core to signal.
+  """
   ref, transforms = _get_ref_and_transforms(sem_or_view)
   inc = jnp.asarray(inc, dtype=jnp.int32)
   args = [ref, transforms, inc, device_id, core_index]
@@ -1128,43 +1185,40 @@ def semaphore_signal(
   )
 
 
-# TODO(sharadmv,ivyzheng): remove this once we use axis dicts primarily
-class CommsEffect(effects.Effect):
-  pass
-_comms_effect = CommsEffect()
-effects.lowerable_effects.add_type(CommsEffect)
-effects.control_flow_allowed_effects.add_type(CommsEffect)
-effects.remat_allowed_effects.add_type(CommsEffect)
-effects.custom_derivatives_allowed_effects.add_type(CommsEffect)
-
-
 @semaphore_signal_p.def_effectful_abstract_eval
 def _semaphore_signal_abstract_eval(
     *avals,
     args_tree,
     device_id_type: DeviceIdType,
 ):
-  del device_id_type
   (
       sem_aval,
       sem_transforms_avals,
       value_aval,
-      device_id_avals,
+      device_id_aval,
       core_index_aval,
   ) = tree_util.tree_unflatten(args_tree, avals)
   check_sem_avals(sem_aval, sem_transforms_avals, "signal")
   if value_aval.dtype != jnp.dtype("int32"):
-    raise ValueError("Must signal an int32 value.")
+    raise ValueError(f"Must signal an int32 value, but got {value_aval.dtype}")
   effs : set[effects.Effect] = set()
-  if device_id_avals is not None:
-    device_id_flat_avals = tree_util.tree_leaves(device_id_avals)
+  if device_id_aval is not None:
+    device_id_flat_avals = tree_util.tree_leaves(device_id_aval)
     for aval in device_id_flat_avals:
       if aval.dtype != jnp.dtype("int32"):
-        raise ValueError("`device_id`s must be an int32 value.")
-    effs.add(_comms_effect)
+        raise ValueError(
+            f"`device_id`s must be an int32 value, but got {aval.dtype}"
+        )
+    if device_id_type is DeviceIdType.MESH and isinstance(device_id_aval, dict):
+      for k in device_id_aval:
+        if not isinstance(k, tuple):
+          k = (k,)
+        for k_ in k:
+          effs.add(jax_core.NamedAxisEffect(k_))
+    else:
+      effs.add(pallas_core.comms_effect)
   return [], effs
 
-
 def _semaphore_signal_pp_eqn(eqn: jax_core.JaxprEqn,
                              context: jax_core.JaxprPpContext,
                              settings: jax_core.JaxprPpSettings):
@@ -1223,16 +1277,27 @@ state_discharge.register_discharge_rule(semaphore_signal_p)(
 semaphore_wait_p = jax_core.Primitive('semaphore_wait')
 semaphore_wait_p.multiple_results = True
 
-def semaphore_wait(sem_or_view, dec: int | jax.Array = 1):
+
+def semaphore_wait(
+    sem_or_view, value: int | jax_typing.Array = 1, *, decrement: bool = True
+):
+  """Blocks execution of the current thread until a semaphore reaches a value.
+
+  Args:
+    sem_or_view: A Ref (or view) representing a semaphore.
+    value: The target value that the semaphore should reach before unblocking.
+    decrement: Whether to decrement the value of the semaphore after
+      a successful wait.
+  """
   ref, transforms = _get_ref_and_transforms(sem_or_view)
-  dec = jnp.asarray(dec, dtype=jnp.int32)
-  args = [ref, transforms, dec]
+  value = jnp.asarray(value, dtype=jnp.int32)
+  args = [ref, transforms, value, decrement]
   flat_args, args_tree = tree_util.tree_flatten(args)
   semaphore_wait_p.bind(*flat_args, args_tree=args_tree)
 
 @semaphore_wait_p.def_abstract_eval
 def _semaphore_wait_abstract_eval(*avals, args_tree):
-  sem_aval, sem_transforms_avals, value_aval = tree_util.tree_unflatten(
+  sem_aval, sem_transforms_avals, value_aval, _ = tree_util.tree_unflatten(
       args_tree, avals
   )
   check_sem_avals(sem_aval, sem_transforms_avals, "wait")
@@ -1250,14 +1315,20 @@ def _semaphore_wait_pp_eqn(eqn: jax_core.JaxprEqn,
       sem,
       sem_transforms,
       value,
+      decrement,
   ) = tree_util.tree_unflatten(tree, invars)
-  return pp.concat([
+  parts = [
       pp.text("semaphore_wait"),
+  ]
+  if decrement:
+    parts.append(pp.text("[dec]"))
+  parts += [
       pp.text(" "),
       sp.pp_ref_transforms(context, sem, sem_transforms),
       pp.text(" "),
       pp.text(jax_core.pp_var(value, context)),
-  ])
+  ]
+  return pp.concat(parts)
 jax_core.pp_eqn_rules[semaphore_wait_p] = _semaphore_wait_pp_eqn
 
 def _semaphore_wait_discharge_rule(in_avals,
@@ -1265,13 +1336,135 @@ def _semaphore_wait_discharge_rule(in_avals,
                                      *flat_args,
                                      args_tree):
   del out_avals
-  [ref, transforms, dec] = args_tree.unflatten(flat_args)
+  [ref, transforms, value, decrement] = args_tree.unflatten(flat_args)
   sem_value = _transform_semaphore(ref, transforms, in_avals[0])
-  dec = dec.astype(pallas_core.SEMAPHORE_INTERPRET_DTYPE)
-  _, new_sem_value = state_discharge.transform_swap_array(
-      ref, transforms, sem_value - dec
-  )
+  value = value.astype(pallas_core.SEMAPHORE_INTERPRET_DTYPE)
+  if decrement:
+    _, new_sem_value = state_discharge.transform_swap_array(
+        ref, transforms, sem_value - value
+    )
+  else:
+    new_sem_value = sem_value
   return (new_sem_value,) + (None,) * (len(in_avals) - 1), ()
 state_discharge.register_discharge_rule(semaphore_wait_p)(
     _semaphore_wait_discharge_rule
 )
+
+
+def _device_id_dict_to_mesh(mesh_context: pallas_utils.MeshInfo | None, device_id_dict, get_axis_index):
+  i32 = ir.IntegerType.get_signless(32)
+  if mesh_context is None:
+    mesh_axis_sizes = {}
+  else:
+    mesh_axis_sizes = dict(
+        zip(mesh_context.axis_names, mesh_context.mesh_shape)
+    )
+  physical_axis_dict = {}
+  # Handle joint axes (i.e., one logical axis over >1 physical axes)
+  for axis_name, idx in device_id_dict.items():
+    if isinstance(axis_name, tuple) and any(
+        a in mesh_axis_sizes for a in axis_name
+    ):
+      if not all(a in mesh_axis_sizes for a in axis_name):
+        raise NotImplementedError(
+            f"{axis_name} mixes JAX mesh and Pallas mesh grid axes"
+        )
+      axes_dimensions = [mesh_axis_sizes[name] for name in axis_name]
+      for axis_index, axis_name in enumerate(axis_name):
+        axis_size = mesh_axis_sizes[axis_name]
+        inner_mesh_size = math.prod(axes_dimensions[axis_index + 1 :])
+        minor_divisor = arith.constant(i32, inner_mesh_size)
+
+        # Fast path for power of 2s
+        if inner_mesh_size & (inner_mesh_size - 1) == 0:
+          shift_len = (inner_mesh_size & -inner_mesh_size).bit_length() - 1
+          partial_device_idx = arith.shrui(idx, arith.constant(i32, shift_len))
+        else:
+          partial_device_idx = arith.divsi(idx, minor_divisor)
+
+        if axis_size & (axis_size - 1) == 0:
+          device_idx = arith.andi(
+              partial_device_idx,
+              arith.constant(i32, mesh_axis_sizes[axis_name] - 1),
+          )
+        else:
+          device_idx = arith.remsi(
+              partial_device_idx, arith.constant(i32, axis_size)
+          )
+        physical_axis_dict[axis_name] = device_idx
+    else:
+      physical_axis_dict[axis_name] = idx
+  device_id = []
+  for axis_name in mesh_axis_sizes:
+    if axis_name in physical_axis_dict:
+      device_id.append(physical_axis_dict[axis_name])
+    else:
+      device_id.append(get_axis_index(axis_name))
+  non_mesh_axes = {
+      k: v
+      for k, v in physical_axis_dict.items()
+      if k not in mesh_axis_sizes
+  }
+  return tuple(device_id), non_mesh_axes
+
+
+def device_id_to_logical(
+    mesh_context: pallas_utils.MeshInfo | None,
+    device_id: ir.Value | tuple[ir.Value, ...] | dict[Any, ir.Value],
+    device_id_type: DeviceIdType,
+    get_axis_index,
+) -> tuple[ir.Value, dict[Any, ir.Value]]:
+  """Normalizes a device id into a logical device id and axes that don't correspond to JAX mesh axes.
+
+  The indexing implied by the returned axis dict should be handled by the caller.
+  """
+  non_mesh_axes = {}
+  if isinstance(device_id, dict):
+    if device_id_type is not DeviceIdType.MESH:
+      raise ValueError(
+          "`device_id_type` must be MESH if `device_id` is a dict,"
+          f" got: {device_id_type = }."
+      )
+    device_id, non_mesh_axes = _device_id_dict_to_mesh(mesh_context, device_id, get_axis_index)
+  if device_id_type is DeviceIdType.MESH:
+    # Mesh means we are passed the mesh coordinates for the device
+    device_ids = tree_util.tree_leaves(device_id)
+    mesh_strides: tuple[int, ...]
+    if mesh_context is None:
+      mesh_strides = ()
+    else:
+      mesh_strides = mesh_context.mesh_strides
+    if len(device_ids) != len(mesh_strides):
+      raise ValueError(
+          "Number of device ids must match the number of mesh axes, but got"
+          f" {len(device_ids)} ids for a {len(mesh_strides)}D mesh."
+      )
+
+    i32 = ir.IntegerType.get_signless(32)
+    if len(device_ids) == 0:
+      return arith.constant(i32, 0), non_mesh_axes
+    return functools.reduce(
+        arith.addi,
+        (
+            arith.muli(a, arith.constant(i32, b))
+            for a, b in zip(device_ids, mesh_strides)
+        ),
+    ), non_mesh_axes
+  elif device_id_type is DeviceIdType.LOGICAL:
+    return device_id, non_mesh_axes
+  raise NotImplementedError(f"Unsupported device id type: {device_id_type}")
+
+
+delay_p = jax_core.Primitive("delay")
+delay_p.multiple_results = True
+
+
+@delay_p.def_abstract_eval
+def _delay_abstract_eval(nanos):
+  del nanos
+  return []
+
+
+def delay(nanos: int | jax_typing.Array) -> None:
+  """Sleeps for the given number of nanoseconds."""
+  delay_p.bind(nanos)
