@@ -27,6 +27,7 @@ from jax._src import core
 from jax._src import dispatch
 from jax._src import dtypes
 from jax._src import ffi
+from jax._src import literals
 from jax._src import numpy as jnp
 from jax._src import pretty_printer as pp
 from jax._src import source_info_util
@@ -38,10 +39,10 @@ from jax._src.interpreters import ad
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
 from jax._src.interpreters import pxla
-from jax._src.interpreters import xla
 from jax._src.lax import control_flow as lax_control_flow
 from jax._src.lax import lax
 from jax._src.lax import slicing as lax_slicing
+from jax._src.lib import jaxlib_extension_version
 from jax._src.lib import gpu_prng
 from jax._src.lib import xla_client as xc
 from jax._src.lib.mlir import ir
@@ -168,11 +169,12 @@ class PRNGKeyArray(Array):
     _check_prng_key_data(impl, key_data)
     self._impl = impl
     self._consumed = False  # TODO(jakevdp): default to True here?
-    if isinstance(key_data, np.ndarray):
+    if isinstance(key_data, (np.ndarray, literals.TypedNdArray)):
       aval = core.get_aval(key_data)
       device = pxla.get_default_device()
-      key_data = pxla.batched_device_put(aval, SingleDeviceSharding(device),
-                                         [key_data], [device], committed=False)
+      key_data = pxla.batched_device_put(
+          aval, SingleDeviceSharding(device), [np.asarray(key_data)], [device],
+          committed=False)
     self._base_array = key_data
 
   def _replace_with(self, value: PRNGKeyArray):
@@ -344,8 +346,8 @@ def seed_with_impl(impl: PRNGImpl, seed: int | typing.ArrayLike) -> PRNGKeyArray
 
 
 def keys_shaped_array(impl, shape, sharding, vma):
-  aval = core.ShapedArray(shape, KeyTy(impl), vma=vma)
-  return core.update_aval_with_sharding(aval, sharding)
+  aval = core.ShapedArray(shape, KeyTy(impl))
+  return core.update_aval_with_sharding(aval, sharding, vma=vma)
 
 def base_arr_shape_to_keys_shape(impl, base_arr_shape):
   base_ndim = len(impl.key_shape)
@@ -401,10 +403,16 @@ class KeyTyRules:
     phys_handler = phys_handler_maker(phys_aval, phys_sharding, phys_indices)
 
     # set up a handler that calls the physical one and wraps back up
-    def handler(bufs):
-      return PRNGKeyArray(aval.dtype._impl, phys_handler(bufs))
+    if jaxlib_extension_version >= 390:
+      def handler(arr):
+        return PRNGKeyArray(aval.dtype._impl, arr)
 
-    return handler
+      return phys_handler.wrap(handler)
+    else:
+      def handler(bufs):
+        return PRNGKeyArray(aval.dtype._impl, phys_handler(bufs))
+
+      return handler
 
   @staticmethod
   def global_sharded_result_handler(aval, out_sharding, committed):
@@ -413,9 +421,14 @@ class KeyTyRules:
 
     phys_sharding = physical_sharding(aval, out_sharding)
     phys_handler = phys_handler_maker(phys_aval, phys_sharding, committed)
-    def handler(bufs):
-      return PRNGKeyArray(aval.dtype._impl, phys_handler(bufs))
-    return handler
+    if jaxlib_extension_version >= 390:
+      def handler(bufs):
+        return PRNGKeyArray(aval.dtype._impl, bufs)
+      return phys_handler.wrap(handler)
+    else:
+      def handler(bufs):
+        return PRNGKeyArray(aval.dtype._impl, phys_handler(bufs))
+      return handler
 
   @staticmethod
   def make_sharded_array(aval, sharding, arrays, committed):
@@ -482,7 +495,7 @@ class KeyTy(dtypes.ExtendedDType):
 
 
 core.pytype_aval_mappings[PRNGKeyArray] = lambda x: x.aval
-xla.canonicalize_dtype_handlers[PRNGKeyArray] = lambda x: x
+dtypes.canonicalize_value_handlers[PRNGKeyArray] = lambda x: x
 
 
 def key_array_shard_arg_handler(xs: Sequence[PRNGKeyArray], shardings, layouts,
@@ -497,9 +510,9 @@ def key_array_shard_arg_handler(xs: Sequence[PRNGKeyArray], shardings, layouts,
 pxla.shard_arg_handlers[PRNGKeyArray] = key_array_shard_arg_handler
 
 
-def key_array_constant_handler(x):
+def key_array_constant_handler(x, aval):
   arr = x._base_array
-  return mlir.get_constant_handler(type(arr))(arr)
+  return mlir.get_constant_handler(type(arr))(arr, aval)
 mlir.register_constant_handler(PRNGKeyArray, key_array_constant_handler)
 
 
@@ -597,9 +610,13 @@ batching.defvectorized(random_split_p)
 def random_split_abstract_eval(keys_aval, *, shape):
   # TODO(yashkatariya): random_split should take sharding as an arg too so we
   # don't choose None here?
-  new_spec = (*keys_aval.sharding.spec, *[None] * len(shape))
+  if keys_aval.sharding.mesh.empty:
+    out_sharding = core.get_cur_mesh_sharding()
+  else:
+    new_spec = (*keys_aval.sharding.spec, *[None] * len(shape))
+    out_sharding = keys_aval.sharding.update(spec=new_spec)
   return keys_shaped_array(keys_aval.dtype._impl, (*keys_aval.shape, *shape),
-                           keys_aval.sharding.update(spec=new_spec), keys_aval.vma)
+                           out_sharding, keys_aval.vma)
 
 @random_split_p.def_impl
 def random_split_impl(keys, *, shape):
@@ -678,7 +695,14 @@ batching.defvectorized(random_bits_p)
 def random_bits_abstract_eval(keys_aval, *, bit_width, shape):
   out_shape = (*keys_aval.shape, *shape)
   out_dtype = dtypes.dtype(f'uint{bit_width}')
-  return core.ShapedArray(out_shape, out_dtype, vma=keys_aval.vma)
+  # TODO(yashkatariya): random_bits should take an out_sharding argument.
+  if keys_aval.sharding.mesh.empty:
+    out_sharding = core.get_cur_mesh_sharding()
+  else:
+    new_spec = (*keys_aval.sharding.spec, *[None] * len(shape))
+    out_sharding = keys_aval.sharding.update(spec=new_spec)
+  return core.ShapedArray(out_shape, out_dtype, sharding=out_sharding,
+                          vma=keys_aval.vma)
 
 @random_bits_p.def_impl
 def random_bits_impl(keys, *, bit_width, shape):
@@ -801,7 +825,7 @@ def threefry_seed(seed: typing.Array) -> typing.Array:
   """
   return _threefry_seed(seed)
 
-@partial(jit, inline=True)
+@jit(inline=True)
 def _threefry_seed(seed: typing.Array) -> typing.Array:
   if seed.shape:
     raise TypeError(f"PRNG key seed must be a scalar; got {seed!r}.")
@@ -1076,7 +1100,7 @@ def iota_2x32_shape_lowering(ctx, *, shape):
 mlir.register_lowering(iota_2x32_shape_p, iota_2x32_shape_lowering)
 
 
-@partial(jit, inline=True)
+@jit(inline=True)
 def threefry_2x32(keypair, count):
   """Apply the Threefry 2x32 hash.
 
@@ -1127,20 +1151,20 @@ def threefry_split(key: typing.Array, shape: Shape) -> typing.Array:
   shape = tuple(unsafe_map(core.concrete_dim_or_error, shape))
   return _threefry_split(key, shape)
 
-@partial(jit, static_argnums=(1,))
+@jit(static_argnums=(1,))
 def _threefry_split(key, shape) -> typing.Array:
   if config.threefry_partitionable.value:
     return _threefry_split_foldlike(key, shape)
   else:
     return _threefry_split_original(key, shape)
 
-@partial(jit, static_argnums=(1,), inline=True)
+@jit(static_argnums=(1,), inline=True)
 def _threefry_split_original(key, shape) -> typing.Array:
   num = math.prod(shape)
   counts = lax.iota(np.uint32, num * 2)
   return lax.reshape(threefry_2x32(key, counts), (*shape, 2))
 
-@partial(jit, static_argnums=(1,), inline=True)
+@jit(static_argnums=(1,), inline=True)
 def _threefry_split_foldlike(key, shape) -> typing.Array:
   k1, k2 = key
   counts1, counts2 = iota_2x32_shape(shape)
@@ -1187,7 +1211,7 @@ def _threefry_random_bits_partitionable(key: typing.Array, bit_width, shape):
   else:
     return lax.convert_element_type(bits1 ^ bits2, dtype)
 
-@partial(jit, static_argnums=(1, 2), inline=True)
+@jit(static_argnums=(1, 2), inline=True)
 def _threefry_random_bits_original(key: typing.Array, bit_width, shape):
   size = math.prod(shape)
   # Compute ceil(bit_width * size / 32) in a way that is friendly to shape
