@@ -35,16 +35,21 @@ from jax._src.lax import lax
 from jax._src.state import indexing
 from jax._src.state.types import (
     AbstractRef,
+    AbstractLinVal,
     AccumEffect,
     ReadEffect,
     Transform,
     TransformedRef,
     WriteEffect,
 )
-from jax._src.typing import Array
+from jax._src.typing import Array, ArrayLike
 from jax._src.util import safe_map, safe_zip
 
 
+# Stand-in for hi-jax inputs to Ref.
+HijaxType = Any
+
+
 ## General utilities
 
 ## JAX utilities
@@ -65,7 +70,17 @@ traceback_util.register_exclusion(__file__)
 get_p = core.Primitive("get")
 get_p.is_effectful = lambda params: True  # type: ignore
 get_p.def_impl(partial(dispatch.apply_primitive, get_p))
-batching.ragged_prop_rules[get_p] = batching.ragged_mask_transfer_identity
+
+get_p.is_high = lambda ref_aval, *_, tree: ref_aval.is_high  # type: ignore
+def _get_to_lojax(ref, *idx, tree):
+  val_ty = core.typeof(ref._refs)
+  transforms = tree_util.tree_unflatten(tree, idx)
+  if transforms:
+    ref = TransformedRef(ref, transforms[:-1])
+    idx = transforms[-1]
+    return val_ty.ref_get_to_lojax(ref, idx)
+  return val_ty.raise_val(*map(ref_get, val_ty.lower_val(ref._refs)))
+get_p.to_lojax = _get_to_lojax  # type: ignore
 
 Indexer = Union[int, slice, Array, types.EllipsisType]
 
@@ -74,7 +89,6 @@ def get_ref_and_transforms(
     ref_or_view: Any,
     idx: Indexer | tuple[Indexer, ...] | None,
     function_name: str,
-    force_trailing_indexer: bool = True,  # TODO(apaszke): Clean this up.
 ) -> tuple[Any, tuple[Transform, ...]]:
   if isinstance(ref_or_view, TransformedRef):
     ref, transforms = ref_or_view.ref, ref_or_view.transforms
@@ -83,7 +97,8 @@ def get_ref_and_transforms(
   ref_aval = core.get_aval(ref)
   if not isinstance(ref_aval, AbstractRef):
     raise ValueError(f"Can only call `{function_name}` on a `Ref`: {ref}.")
-  if not isinstance(ref_aval.inner_aval, core.ShapedArray):
+  if (not isinstance(ref_aval.inner_aval, core.ShapedArray)
+      and not ref_aval.inner_aval.is_high):
     return ref, ()
 
   if idx is None or idx is Ellipsis:
@@ -91,19 +106,50 @@ def get_ref_and_transforms(
   elif not isinstance(idx, tuple):
     idx = (idx,)
 
-  if not idx and not force_trailing_indexer:
+  if not idx:
     return ref, transforms
   if not idx and transforms and isinstance(transforms[-1], indexing.NDIndexer):
     return ref, transforms
   nd_indexer = indexing.NDIndexer.from_indices_shape(idx, ref_or_view.shape)
   return ref, (*transforms, nd_indexer)
 
-
+@partial(traceback_util.api_boundary, repro_api_name="jax.ref.get")
 def ref_get(
-    ref_or_view: Any, idx: Indexer | tuple[Indexer, ...] | None = None
-) -> Array:
-  """Reads a value from a `Ref`, a.k.a. value <- ref[idx]."""
-  ref, transforms = get_ref_and_transforms(ref_or_view, idx, "ref_get")
+    ref: core.Ref | TransformedRef,
+    idx: Indexer | tuple[Indexer, ...] | None = None
+) -> Array | HijaxType:
+  """Read a value from an Ref.
+
+  This is equivalent to ``ref[idx]`` for a NumPy-style indexer ``idx``.
+  For more on mutable array refs, refer to the `Ref guide`_.
+
+  Args:
+    ref: a :class:`jax.ref.Ref` object.
+    idx: a NumPy-style indexer
+
+  Returns:
+    A :class:`jax.Array` object (note, not a :class:`jax.ref.Ref`) containing
+    the indexed elements of the mutable reference.
+
+  Examples:
+    >>> import jax
+    >>> ref = jax.new_ref(jax.numpy.arange(5))
+    >>> jax.ref.get(ref, slice(1, 3))
+    Array([1, 2], dtype=int32)
+
+    Equivalent operation via indexing syntax:
+
+    >>> ref[1:3]
+    Array([1, 2], dtype=int32)
+
+    Use ``...`` to extract the full buffer:
+
+    >>> ref[...]
+    Array([0, 1, 2, 3, 4], dtype=int32)
+
+  .. _Ref guide: https://docs.jax.dev/en/latest/array_refs.html
+  """
+  ref, transforms = get_ref_and_transforms(ref, idx, "ref_get")
   flat_transforms, tree = tree_util.tree_flatten(transforms)
   return get_p.bind(ref, *flat_transforms, tree=tree)
 
@@ -128,27 +174,76 @@ swap_p = core.Primitive("swap")
 swap_p.is_effectful = lambda params: True  # type: ignore
 swap_p.def_impl(partial(dispatch.apply_primitive, swap_p))
 
-
-def swap_ragged_prop_rule(eqn_params, invar_raggedness, outvars):
-  assert len(invar_raggedness) == 2
-  invar_raggedness_lhs = invar_raggedness[0]
-  invar_raggedness_rhs = invar_raggedness[1]
-
-  return [invar_raggedness_rhs, invar_raggedness_lhs], [None]
+swap_p.is_high = lambda ref_aval, *_, tree: ref_aval.is_high  # type: ignore
+def _swap_to_lojax(ref, val, *idx, tree):
+  ref_val_ty = core.typeof(ref._refs)
+  val_ty = core.typeof(val)
+  transforms = tree_util.tree_unflatten(tree, idx)
+  if transforms:
+    ref = TransformedRef(ref, transforms[:-1])
+    idx = transforms[-1]
+    return ref_val_ty.ref_swap_to_lojax(ref, val, idx)
+  lo_refs = ref_val_ty.lower_val(ref._refs)
+  lo_vals = val_ty.lower_val(val)
+  outs = [ref_swap(lo_ref, idx, lo_val) for lo_ref, lo_val
+          in zip(lo_refs, lo_vals)]
+  return val_ty.raise_val(*outs)
+swap_p.to_lojax = _swap_to_lojax  # type: ignore
 
 
-batching.ragged_prop_rules[swap_p] = swap_ragged_prop_rule
-
+@partial(traceback_util.api_boundary, repro_api_name="jax.ref.swap")
 def ref_swap(
-    ref_or_view: AbstractRef | TransformedRef,
+    ref: core.Ref | TransformedRef,
     idx: Indexer | tuple[Indexer, ...] | None,
-    value: Array,
+    value: ArrayLike | HijaxType,
     _function_name: str = "ref_swap",
-) -> Array:
-  """Sets a `Ref`'s value and returns the original value."""
-  if hasattr(ref_or_view, 'dtype'):
-    value = _maybe_implicit_cast(ref_or_view.dtype, value)
-  ref, transforms = get_ref_and_transforms(ref_or_view, idx, _function_name)
+) -> Array | HijaxType:
+  """Update an array value inplace while returning the previous value.
+
+  This is equivalent to ``ref[idx], prev = value, ref[idx]`` while returning
+  ``prev``, for a NumPy-style indexer ``idx``.
+  For more on mutable array refs, refer to the `Ref guide`_.
+
+  Args:
+    ref: a :class:`jax.ref.Ref` object. On return, the buffer will be
+      mutated by this operation.
+    idx: a NumPy-style indexer
+    value: a :class:`jax.Array` object (note, not a :class:`jax.ref.Ref`)
+      containing the values to set in the array.
+
+  Returns:
+    A :class:`jax.Array` containing the previous value at `idx`.
+
+  Examples:
+    >>> import jax
+    >>> ref = jax.new_ref(jax.numpy.arange(5))
+    >>> jax.ref.swap(ref, 3, 10)
+    Array(3, dtype=int32)
+    >>> ref
+    Ref([ 0,  1,  2, 10,  4], dtype=int32)
+
+    Equivalent operation via indexing syntax:
+
+    >>> ref = jax.new_ref(jax.numpy.arange(5))
+    >>> ref[3], prev = 10, ref[3]
+    >>> prev
+    Array(3, dtype=int32)
+    >>> ref
+    Ref([ 0,  1,  2, 10,  4], dtype=int32)
+
+    Use ``...`` to swap the value of a scalar ref:
+
+    >>> ref = jax.new_ref(jax.numpy.int32(5))
+    >>> jax.ref.swap(ref, ..., 10)
+    Array(5, dtype=int32)
+    >>> ref
+    Ref(10, dtype=int32)
+
+  .. _Ref guide: https://docs.jax.dev/en/latest/array_refs.html
+  """
+  if hasattr(ref, 'dtype'):
+    value = _maybe_implicit_cast(ref.dtype, value)
+  ref, transforms = get_ref_and_transforms(ref, idx, _function_name)
   flat_transforms, tree = tree_util.tree_flatten(transforms)
   return swap_p.bind(ref, value, *flat_transforms, tree=tree)
 
@@ -157,6 +252,8 @@ def ref_swap(
 #     value == np.array(value, dtype).item()): return cast
 def _maybe_implicit_cast(dtype, value):
   aval = core.typeof(value)
+  if not isinstance(aval, core.ShapedArray):
+    return value
   if (aval.weak_type and
       (dtypes.issubdtype(dtype, np.floating) and
        dtypes.issubdtype(aval.dtype, np.floating)) or
@@ -166,13 +263,51 @@ def _maybe_implicit_cast(dtype, value):
   return value
 
 
+@partial(traceback_util.api_boundary, repro_api_name="jax.ref.set")
 def ref_set(
-    ref_or_view: AbstractRef | TransformedRef,
+    ref: core.Ref | TransformedRef,
     idx: Indexer | tuple[Indexer, ...] | None,
-    value: Array,
+    value: ArrayLike | HijaxType,
 ) -> None:
-  """Sets a `Ref`'s value, a.k.a. ref[idx] <- value."""
-  ref_swap(ref_or_view, idx, value, _function_name="ref_set")
+  """Set a value in an Ref in-place.
+
+  This is equivalent to ``ref[idx] = value`` for a NumPy-style indexer
+  ``idx``. For more on mutable array refs, refer to the `Ref guide`_.
+
+  Args:
+    ref: a :class:`jax.ref.Ref` object. On return, the buffer will be
+      mutated by this operation.
+    idx: a NumPy-style indexer
+    value: a :class:`jax.Array` object (note, not a :class:`jax.ref.Ref`)
+      containing the values to set in the array.
+
+  Returns:
+    None
+
+  Examples:
+    >>> import jax
+    >>> ref = jax.new_ref(jax.numpy.zeros(5))
+    >>> jax.ref.set(ref, 1, 10.0)
+    >>> ref
+    Ref([ 0., 10.,  0.,  0.,  0.], dtype=float32)
+
+    Equivalent operation via indexing syntax:
+
+    >>> ref = jax.new_ref(jax.numpy.zeros(5))
+    >>> ref[1] = 10.0
+    >>> ref
+    Ref([ 0., 10.,  0.,  0.,  0.], dtype=float32)
+
+    Use ``...`` to set the value of a scalar ref:
+
+    >>> ref = jax.new_ref(jax.numpy.int32(0))
+    >>> ref[...] = 4
+    >>> ref
+    Ref(4, dtype=int32)
+
+  .. _Ref guide: https://docs.jax.dev/en/latest/array_refs.html
+  """
+  ref_swap(ref, idx, value, _function_name="ref_set")
 
 
 # `addupdate_p` mutates a `Ref`, adding a value to its existing value.
@@ -193,14 +328,54 @@ addupdate_p.def_impl(partial(dispatch.apply_primitive, addupdate_p))
 
 
 def ref_addupdate(
-    ref_or_view: AbstractRef,
+    ref: core.Ref | TransformedRef,
     idx: Indexer | tuple[Indexer, ...] | None,
-    x: Array,
+    x: ArrayLike | HijaxType,
 ) -> None:
-  """Mutates a ref with an additive update i.e. `ref[idx] += x`."""
-  ref, transforms = get_ref_and_transforms(ref_or_view, idx, "ref_addupdate")
+  """Add to an element in an Ref in-place.
+
+  This is analogous to ``ref[idx] += value`` for a NumPy array ``ref`` and
+  NumPy-style indexer ``idx``. However, for an Ref ``ref``, executing
+  ``ref[idx] += value`` actually performs a ``ref_get``, add, and ``ref_set``,
+  so using this function can be more efficient under autodiff. For more on
+  mutable array refs, refer to the `Ref guide`_.
+
+  Args:
+    ref: a :class:`jax.ref.Ref` object. On return, the buffer will be
+      mutated by this operation.
+    idx: a NumPy-style indexer
+    x: a :class:`jax.Array` object (note, not a :class:`jax.ref.Ref`)
+      containing the values to add at the specified indices.
+
+  Returns:
+    None
+
+  Examples:
+    >>> import jax
+    >>> ref = jax.new_ref(jax.numpy.arange(5))
+    >>> jax.ref.addupdate(ref, 2, 10)
+    >>> ref
+    Ref([ 0,  1, 12,  3,  4], dtype=int32)
+
+    Equivalent operation via indexing syntax:
+
+    >>> ref = jax.new_ref(jax.numpy.arange(5))
+    >>> ref[2] += 10
+    >>> ref
+    Ref([ 0,  1, 12,  3,  4], dtype=int32)
+
+    Use ``...`` to add to a scalar ref:
+
+    >>> ref = jax.new_ref(jax.numpy.int32(2))
+    >>> ref[...] += 10
+    >>> ref
+    Ref(12, dtype=int32)
+
+  .. _Ref guide: https://docs.jax.dev/en/latest/array_refs.html
+  """
+  ref, transforms = get_ref_and_transforms(ref, idx, "ref_addupdate")
   flat_transforms, tree = tree_util.tree_flatten(transforms)
-  return addupdate_p.bind(ref, x, *flat_transforms, tree=tree)
+  addupdate_p.bind(ref, x, *flat_transforms, tree=tree)
 
 
 ## get/set/addupdate abstract evaluation rules
@@ -234,6 +409,8 @@ def _sharding_after_transforming(sharding, transforms):
 def _get_abstract_eval(ref_aval: AbstractRef, *args,
                        tree):
   transforms = tree_util.tree_unflatten(tree, args)
+  if transforms and ref_aval.inner_aval.is_high:
+    return ref_aval.inner_aval.ref_get_abstract_eval(ref_aval, *args, tree=tree)
   if not isinstance(ref_aval, AbstractRef):
     raise ValueError(f"`get` must be called on `Ref` types: {ref_aval}.")
   if isinstance(ref_aval.inner_aval, core.ShapedArray):
@@ -253,6 +430,9 @@ def _swap_abstract_eval(ref_aval: AbstractRef,
                         val_aval: core.AbstractValue,
                         *args: Any, tree):
   transforms = tree_util.tree_unflatten(tree, args)
+  if transforms and ref_aval.inner_aval.is_high:
+    return ref_aval.inner_aval.ref_swap_abstract_eval(
+        ref_aval, val_aval, *args, tree=tree)
   out_aval: core.AbstractValue
   if not isinstance(ref_aval, AbstractRef):
     raise ValueError(f"`swap` must be called on `Ref` types: {ref_aval}.")
@@ -293,6 +473,7 @@ def _addupdate_abstract_eval(ref_aval: AbstractRef,
   if isinstance(ref_aval.inner_aval, core.ShapedArray):
     out_shape = _shape_after_transforming(ref_aval.shape, transforms)
     out_dtype = _dtype_after_transforming(ref_aval.dtype, transforms)
+    out_sharding = _sharding_after_transforming(ref_aval.sharding, transforms)
     assert isinstance(val_aval, core.ShapedArray)
     if out_shape != val_aval.shape:
       raise ValueError(
@@ -306,6 +487,12 @@ def _addupdate_abstract_eval(ref_aval: AbstractRef,
       raise ValueError("Invalid dtype for `addupdate`. "
                        f"Ref dtype: {ref_aval.dtype}. "
                        f"Value shape: {val_aval.dtype}. ")
+    if ((out_sharding.mesh._any_axis_explicit or
+         val_aval.sharding.mesh._any_axis_explicit) and
+        out_sharding != val_aval.sharding):
+      raise ValueError("Invalid sharding for `addupdate`. "
+                       f"Ref sharding: {ref_aval.sharding}. "
+                       f"Value sharding: {val_aval.sharding}. ")
   else:
     # Check that the transforms are valid
     if transforms:
@@ -404,6 +591,8 @@ def _swap_jvp(primals: list[Any], tangents: list[Any], **params: Any):
   out_primal = swap_p.bind(ref_primal, x_primal, *idx, **params)
   if isinstance(ref_tangent, ad_util.Zero) and isinstance(x_tangent, ad_util.Zero):
     out_tangent = ad_util.Zero(core.typeof(out_primal).to_tangent_aval())
+  elif ref_tangent.aval.kind == "anselm_ref":
+    out_tangent = ad_util.Zero(core.typeof(out_primal).to_tangent_aval())
   else:
     if isinstance(ref_tangent, ad_util.Zero):
       raise Exception("performing a set/swap operation with a differentiated "
@@ -419,8 +608,9 @@ def addupdate_jvp_rule(primals: list[Any], tangents: list[Any], **params: Any):
   ref_primal, x_primal, *idx = primals
   ref_tangent, x_tangent, *_ = tangents
   x_tangent = ad_util.instantiate(x_tangent)
-  addupdate_p.bind(ref_primal, x_primal, *idx, **params)
-  addupdate_p.bind(ref_tangent, x_tangent, *idx, **params)
+  if ref_tangent.aval.kind != "anselm_ref":
+    addupdate_p.bind(ref_primal, x_primal, *idx, **params)
+    addupdate_p.bind(ref_tangent, x_tangent, *idx, **params)
   return [], []
 ad.primitive_jvps[addupdate_p] = addupdate_jvp_rule
 
@@ -446,23 +636,91 @@ def addupdate_transpose(cts_in, ref, x, *idx, **params):
   return [None, g] + [None] * len(idx)
 ad.primitive_transposes[addupdate_p] = addupdate_transpose
 
+
+def _get_transpose_fancy(g, ref_, *idx, **params):
+  if idx and type(g) is not ad_util.Zero:
+    addupdate_p.bind(ref_.inst().ref, g, *idx, **params)
+  else:
+    ref_.accum(g)
+ad.fancy_transposes[get_p] = _get_transpose_fancy
+
+def _swap_transpose_fancy(g, ref_, x, *idx, **params):
+  if ref_.ref is None and type(g) is ad_util.Zero:
+    return
+  elif ref_.ref is None:
+    swap_p.bind(ref_.inst().ref, ad_util.instantiate(g), *idx, **params)
+  else:
+    x_bar = swap_p.bind(ref_.inst().ref, ad_util.instantiate(g), *idx, **params)
+    x.accum(x_bar)
+ad.fancy_transposes[swap_p] = _swap_transpose_fancy
+
+def addupdate_transpose_fancy(cts_in, ref_, x, *idx, **params):
+  if ref_.ref is not None and isinstance(x, ad.GradAccum):
+    x_bar = get_p.bind(ref_.ref, *idx, **params)
+    x.accum(x_bar)
+ad.fancy_transposes[addupdate_p] = addupdate_transpose_fancy
+
 ## get/swap/addupdate partial_eval_custom rules
 
-def _state_partial_eval_custom(prim, saveable, unks_in, inst_in, eqn):
-  if any(unks_in):
-    res = [v for v, inst in zip(eqn.invars, inst_in) if not inst]
-    return None, eqn, [True] * len(eqn.outvars), [True] * len(eqn.outvars), res
-  elif saveable(prim, *[var.aval for var in eqn.invars], **eqn.params):
-    return eqn, None, [False] * len(eqn.outvars), [False] * len(eqn.outvars), []
-  res = [v for v, inst in zip(eqn.invars, inst_in) if not inst]
-  return eqn, eqn, [False] * len(eqn.outvars), [True] * len(eqn.outvars), res
+def _array_ref_partial_eval_custom(saveable, unks_in, inst_in, eqn):
+  del saveable  # ignored, always full remat array_ref on known input
+  unk, = unks_in
+  inst, = inst_in
+  invar, = eqn.invars
+  res = [invar] if not inst else []
+  if unk:
+    return None, eqn, [True], [True], res  # tangent operation
+  else:
+    return eqn, eqn, [False], [True], res  # full remat
+pe.partial_eval_jaxpr_custom_rules[core.ref_p] = _array_ref_partial_eval_custom
 
-pe.partial_eval_jaxpr_custom_rules[get_p] = partial(_state_partial_eval_custom,
-                                                    get_p)
-pe.partial_eval_jaxpr_custom_rules[swap_p] = partial(_state_partial_eval_custom,
-                                                     swap_p)
-pe.partial_eval_jaxpr_custom_rules[addupdate_p] = partial(
-    _state_partial_eval_custom, addupdate_p)
+def _array_ref_batched(axis_data, vals_in, dims_in, memory_space, kind):
+  val, = vals_in
+  dim, = dims_in
+  if dim is None:
+    # We defensively batch the ref, b/c it could later be hit with a batched val
+    val2 = batching.broadcast(val, axis_data.size, 0,
+                              axis_data.explicit_mesh_axis)
+    return core.ref_p.bind(val2, memory_space=memory_space, kind=kind), 0
+  else:
+    return core.ref_p.bind(val, memory_space=memory_space, kind=kind), dim
+batching.fancy_primitive_batchers[core.ref_p] = _array_ref_batched
+
+def _freeze_batched(axis_data, vals_in, dims_in):
+  ref, = vals_in
+  dim, = dims_in
+  return core.freeze_p.bind(ref), dim
+batching.fancy_primitive_batchers[core.freeze_p] = _freeze_batched
+
+def _state_partial_eval_custom(saveable, unks_in, inst_in, eqn):
+  del saveable  # ignored, always full remat state ops on known inputs
+                # (except for anselm_ref)
+  ref_unk, *_ = unks_in
+  ref_inst, *inst_in = inst_in
+  _, *val_vars = eqn.invars
+  assert ref_inst
+  res = [v for v, inst in zip(val_vars, inst_in) if not inst]
+  if ref_unk:
+    return None, eqn, [True], [True], res  # tangent operation
+  elif eqn.invars[0].aval.kind == "anselm_ref":
+    return eqn, None, [False], [False], res
+  else:
+    return eqn, eqn, [False], [True], res  # full remat
+pe.partial_eval_jaxpr_custom_rules[get_p] = _state_partial_eval_custom
+pe.partial_eval_jaxpr_custom_rules[swap_p] = _state_partial_eval_custom
+
+def _addupdate_partial_eval_custom(saveable, unks_in, inst_in, eqn):
+  del saveable  # ignored, always full remat state ops on known inputs
+  ref_unk, *_ = unks_in
+  ref_inst, *inst_in = inst_in
+  _, *val_vars = eqn.invars
+  assert ref_inst
+  res = [v for v, inst in zip(val_vars, inst_in) if not inst]
+  if ref_unk:
+    return None, eqn, [], [], res  # tangent operation
+  else:
+    return eqn, eqn, [], [], res  # full remat
+pe.partial_eval_jaxpr_custom_rules[addupdate_p] = _addupdate_partial_eval_custom
 
 ##  get/swap/addupdate batching rules
 
@@ -561,8 +819,13 @@ def _batch_indexer(
                                       bcast_dims)
       new_indices.append(idx)
   if ref_dim is not batching.not_mapped:
-    iota = lax.broadcasted_iota(np.dtype('int32'), new_integer_indexer_shape, 0)
-    new_indices.insert(ref_dim, iota)
+    if indexer.int_indexer_shape:
+      batch_idx = lax.broadcasted_iota(
+          np.dtype('int32'), new_integer_indexer_shape, 0)
+    else:
+      batch_idx = indexing.Slice(0, axis_size)  # type: ignore
+      new_integer_indexer_shape = ()
+    new_indices.insert(ref_dim, batch_idx)
   return indexing.NDIndexer(
       tuple(new_indices), ref_shape, new_integer_indexer_shape, validate=True
   )
@@ -573,12 +836,15 @@ def _get_vmap(batched_args, batched_dims, *, tree):
   ref, *flat_idxs = batched_args
   ref_dim, *flat_idx_dims = batched_dims
   indexers = tree_util.tree_unflatten(tree, flat_idxs)
+  if not indexers:
+    return get_p.bind(ref, *flat_idxs, tree=tree), ref_dim
   indexers_dims = tree_util.tree_unflatten(tree, flat_idx_dims)
 
   idx_is_batched = any(i_dim is not batching.not_mapped
                        for i_dim in flat_idx_dims)
   if len(indexers) > 1:
     raise NotImplementedError("Batching with multiple indexers not supported.")
+
   # TODO(sharadmv): handle vmap of multiple indexers
   new_indexers = tuple(_batch_indexer(indexer, dims, axis_size,
                                   ref.shape, ref_dim, idx_is_batched)
@@ -589,38 +855,56 @@ def _get_vmap(batched_args, batched_dims, *, tree):
   int_indexers_contiguous = bool(
       np.all(np.diff(np.where(is_int_indexing)[0]) == 1)
   )
+  # Note: _batch_indexer will add a slice for the batch dim if the int_indexer
+  # shape is empty, else it will use advanced/int indexing.
+  will_add_int_batcher = bool(indexers[0].int_indexer_shape)
+
   is_new_int_indexing, _, _ = indexing.unpack_ndindexer(new_indexers[0])
   new_int_indexers_contiguous = bool(
       np.all(np.diff(np.where(is_new_int_indexing)[0]) == 1)
   )
 
   out = get_p.bind(ref, *flat_indexers, tree=tree)
-  if not int_indexers_contiguous:  # will always be moved to the front
+  should_transpose = (int_indexers_contiguous and
+                      not new_int_indexers_contiguous)
+  if will_add_int_batcher and should_transpose:
+    original_pos = is_int_indexing.index(True)
+    array_indexer_shape = new_indexers[0].int_indexer_shape
+    array_indexer_len = len(array_indexer_shape)
+
+    transpose_order = list(range(len(out.shape)))
+    transpose_order = (
+        transpose_order[0],
+        *transpose_order[array_indexer_len:array_indexer_len+original_pos],
+        *transpose_order[1:array_indexer_len],
+        *transpose_order[array_indexer_len+original_pos:],
+    )
+    out = lax.transpose(out, transpose_order)
     out_bdim = 0
-  else:  # originally not going to be moved to the front
-    if new_int_indexers_contiguous:  # now not going to be moved to the front
-      out_bdim = is_new_int_indexing.index(True)
-    else:  # now going to be moved to the front
-      original_pos = is_int_indexing.index(True)
-      array_indexer_shape = new_indexers[0].int_indexer_shape
-      array_indexer_len = len(array_indexer_shape)
-
-      transpose_order = list(range(len(out.shape)))
-      transpose_order = (
-          transpose_order[0],
-          *transpose_order[array_indexer_len:array_indexer_len+original_pos],
-          *transpose_order[1:array_indexer_len],
-          *transpose_order[array_indexer_len+original_pos:],
-      )
-
-      out = lax.transpose(out, transpose_order)
+  else:
+    if ref_dim is not batching.not_mapped:
+      if will_add_int_batcher:
+        if not int_indexers_contiguous:
+          # In this case the indexer is always moved to the front.
+          out_bdim = 0
+        else:
+          # In this case the indexer is not moved to the front.
+          out_bdim = is_new_int_indexing.index(True)
+      else:
+        # We only trigger this case when the int_indexer shape is empty,
+        # so we don't need to account for int_indexer_shape.
+        int_indexers_before_ref_dim = int(np.sum(is_new_int_indexing[:ref_dim]))
+        out_bdim = ref_dim - int_indexers_before_ref_dim
+    else:
       out_bdim = 0
+      if any(is_int_indexing):
+        # The batch dim is the indexer's batch dim.
+        original_pos = is_int_indexing.index(True)
+        out_bdim = original_pos
   return out, out_bdim
 batching.primitive_batchers[get_p] = _get_vmap
 
-def _swap_vmap(batched_args, batched_dims, *, tree):
-  axis_size, = {x.shape[d] for x, d in zip(batched_args, batched_dims)
-                if d is not batching.not_mapped}
+def _swap_vmap(axis_data, batched_args, batched_dims, *, tree):
   ref, val, *flat_idxs = batched_args
   ref_dim, val_dim, *flat_idx_dims = batched_dims
   indexers = tree_util.tree_unflatten(tree, flat_idxs)
@@ -636,11 +920,15 @@ def _swap_vmap(batched_args, batched_dims, *, tree):
                     f"an unbatched array reference of type {core.typeof(ref)}. "
                     "Move the array reference to be an argument to the vmapped "
                     "function?")
-
+  if not indexers:
+    if ref_is_batched and not val_is_batched:
+      val = batching.broadcast(val, axis_data.size, ref_dim,
+                               axis_data.explicit_mesh_axis)
+    return swap_p.bind(ref, val, *flat_idxs, tree=tree), ref_dim
   if len(indexers) > 1:
     raise NotImplementedError("Batching with multiple indexers not supported.")
   # TODO(sharadmv): handle vmap of multiple indexers
-  new_indexers = tuple(_batch_indexer(indexer, dims, axis_size,
+  new_indexers = tuple(_batch_indexer(indexer, dims, axis_data.size,
                                   ref.shape, ref_dim, idx_is_batched)
                      for indexer, dims in zip(indexers, indexers_dims))
   flat_indexers, tree = tree_util.tree_flatten(new_indexers)
@@ -657,11 +945,15 @@ def _swap_vmap(batched_args, batched_dims, *, tree):
   if not new_int_indexers_contiguous:  # will be moved to the front
     batched_dim_in_result = 0
   else:
-    batched_dim_in_result = is_new_int_indexing.index(True) + 0
+    try:
+      batched_dim_in_result = is_new_int_indexing.index(True) + 0
+    except ValueError:
+      batched_dim_in_result = ref_dim
 
   if not val_is_batched:
     if ref_is_batched or idx_is_batched:
-      val = batching.broadcast(val, axis_size, batched_dim_in_result)
+      val = batching.broadcast(val, axis_data.size, batched_dim_in_result,
+                               axis_data.explicit_mesh_axis)
   else:
     val = batching.moveaxis(val, val_dim, batched_dim_in_result)
 
@@ -693,11 +985,9 @@ def _swap_vmap(batched_args, batched_dims, *, tree):
     out = out.transpose(transpose_order_inversed)
 
   return out, batched_dim_in_result
-batching.primitive_batchers[swap_p] = _swap_vmap
+batching.fancy_primitive_batchers[swap_p] = _swap_vmap
 
-def _addupdate_vmap(batched_args, batched_dims, *, tree):
-  axis_size, = {x.shape[d] for x, d in zip(batched_args, batched_dims)
-                if d is not batching.not_mapped}
+def _addupdate_vmap(axis_data, batched_args, batched_dims, *, tree):
   ref, val, *flat_idxs = batched_args
   ref_dim, val_dim, *flat_idx_dims = batched_dims
   indexers = tree_util.tree_unflatten(tree, flat_idxs)
@@ -707,10 +997,21 @@ def _addupdate_vmap(batched_args, batched_dims, *, tree):
   val_is_batched = val_dim is not batching.not_mapped
   idx_is_batched = any(i_dim is not batching.not_mapped
                        for i_dim in flat_idx_dims)
+
+  if not ref_is_batched:
+    raise Exception("performing an addupdate operation with vmapped value on "
+                    f"an unbatched array reference of type {core.typeof(ref)}. "
+                    "Move the array reference to be an argument to the vmapped "
+                    "function?")
+  if not indexers:
+    if val_dim != ref_dim:
+      val = batching.matchaxis2(axis_data, val_dim, ref_dim, val)
+    return addupdate_p.bind(ref, val, *flat_idxs, tree=tree), []
   if len(indexers) > 1:
     raise NotImplementedError("Batching with multiple indexers not supported.")
+
   # TODO(sharadmv): handle vmap of multiple indexers
-  new_indexers = tuple(_batch_indexer(indexer, dims, axis_size,
+  new_indexers = tuple(_batch_indexer(indexer, dims, axis_data.size,
                                   ref.shape, ref_dim, idx_is_batched)
                      for indexer, dims in zip(indexers, indexers_dims))
   flat_indexers, tree = tree_util.tree_flatten(new_indexers)
@@ -727,11 +1028,15 @@ def _addupdate_vmap(batched_args, batched_dims, *, tree):
   if not new_int_indexers_contiguous:  # will be moved to the front
     batched_dim_in_result = 0
   else:
-    batched_dim_in_result = is_new_int_indexing.index(True)
+    try:
+      batched_dim_in_result = is_new_int_indexing.index(True)
+    except ValueError:
+      batched_dim_in_result = ref_dim
 
   if not val_is_batched:
     if ref_is_batched or idx_is_batched:
-      val = batching.broadcast(val, axis_size, batched_dim_in_result)
+      val = batching.broadcast(val, axis_data.size, batched_dim_in_result,
+                               axis_data.explicit_mesh_axis)
   else:
     val = batching.moveaxis(val, val_dim, batched_dim_in_result)
 
@@ -752,7 +1057,7 @@ def _addupdate_vmap(batched_args, batched_dims, *, tree):
     val = val.transpose(transpose_order)
 
   return addupdate_p.bind(ref, val, *flat_indexers, tree=tree), []
-batching.primitive_batchers[addupdate_p] = _addupdate_vmap
+batching.fancy_primitive_batchers[addupdate_p] = _addupdate_vmap
 
 # Currently, JAX doesn't have a primitive that does an equal-rank broadcast.
 # We could use `jnp.broadcast_to` but that lowers to squeezing,
@@ -763,6 +1068,18 @@ batching.primitive_batchers[addupdate_p] = _addupdate_vmap
 broadcast_to_p = core.Primitive('broadcast_to')
 
 def broadcast_to(a: Array, shape: tuple[int, ...]) -> Array:
+  """Broadcasts an array to a new shape.
+
+  Args:
+    a: The array to broadcast.
+    shape: The desired shape to broadcast to.
+
+  Returns:
+    An array of shape ``shape``.
+
+  See Also:
+    :func:`jax.numpy.broadcast_to`
+  """
   import jax.numpy as jnp  # pytype: disable=import-error
   a = jnp.asarray(a)
   if a.shape == shape:
@@ -784,25 +1101,90 @@ mlir.register_lowering(
 
 # === AD rules for mutable arrays ===
 
-def _mut_jvp(primals, tangents, *, memory_space):
-  (init_val,), (init_val_dot,) = primals, tangents
-  primal_out = core.mutable_array_p.bind(init_val, memory_space=memory_space)
-  if type(init_val_dot) is ad_util.Zero:
-    tangent_out = core.mutable_array_p.bind(
-        ad_util.zeros_like_aval(init_val_dot.aval), memory_space=memory_space)
+def _ref_jvp(primals, tangents, *, memory_space, kind):
+  (init_val,), (init_dot,) = primals, tangents
+  primal_out = core.ref_p.bind(init_val, memory_space=memory_space, kind=kind)
+  if type(init_dot) is ad_util.Zero:
+    zero = ad_util.zeros_like_aval(init_dot.aval)
+    tangent_out = core.ref_p.bind(zero, memory_space=memory_space, kind=kind)
   else:
-    tangent_out = core.mutable_array_p.bind(init_val_dot,
-                                            memory_space=memory_space)
+    tangent_out = core.ref_p.bind(init_dot, memory_space=memory_space, kind=kind)
   return primal_out, tangent_out
 
-def _mut_lin(nzs, x, *, memory_space):
+def _ref_lin(nzs, x, *, memory_space, kind):
   nz, = nzs
-  x_ref = core.mutable_array_p.bind(x, memory_space=memory_space)
+  x_ref = core.ref_p.bind(x, memory_space=memory_space, kind=kind)
   def mut_lin(_, x_dot):
-    return core.mutable_array_p.bind(ad_util.instantiate(x_dot),
-                                     memory_space=memory_space)
+    zero = ad_util.instantiate(x_dot)
+    return core.ref_p.bind(zero, memory_space=memory_space, kind=kind)
   return x_ref, True, None, mut_lin
 
-ad.primitive_jvps[core.mutable_array_p] = _mut_jvp
+ad.primitive_jvps[core.ref_p] = _ref_jvp
+ad.primitive_linearizations[core.ref_p] = _ref_lin
+# TODO(mattjj): lin rule for freeze and accum_grad_in_ref?
 ad.defjvp(core.freeze_p, lambda g, _: core.freeze(g))
-ad.primitive_linearizations[core.mutable_array_p] = _mut_lin
+ad.defjvp(core.accum_grad_in_ref_p, lambda g, _: core.accum_grad_in_ref_p.bind(g))
+
+# === pinned, chained LinearVals ===
+
+def create_linear(ty, memory_space=None):
+  return create_linear_p.bind(ty=ty, memory_space=memory_space)
+create_linear_p = core.Primitive('create_linear')
+
+@create_linear_p.def_abstract_eval
+def _create_linear_abstract_eval(*, ty, memory_space):
+  if not isinstance(ty, core.ShapedArray): raise NotImplementedError(ty)
+  return AbstractLinVal(ty, memory_space)
+
+def _lower_create_linear(ctx):
+  out_aval, = ctx.avals_out
+  return mlir.custom_call(
+      "CreateBuffer",
+      operands=[],
+      result_types=[mlir.aval_to_ir_type(out_aval)],
+  ).results
+mlir.register_lowering(create_linear_p, _lower_create_linear)
+
+
+def pin(x):
+  return pin_p.bind(x)
+pin_p = core.Primitive('pin')
+
+@pin_p.def_abstract_eval
+def _pin_abstract_eval(aval):
+  if not isinstance(aval, core.ShapedArray): raise NotImplementedError(aval)
+  return AbstractLinVal(aval)
+
+def _lower_pin(ctx, x_op):
+  out_aval, = ctx.avals_out
+  return mlir.custom_call(
+      "Pin",
+      operands=mlir.flatten_ir_values([x_op]),
+      result_types=[mlir.aval_to_ir_type(out_aval)],
+  ).results
+mlir.register_lowering(pin_p, _lower_pin)
+
+
+def unpin(x):
+  return unpin_p.bind(x)
+unpin_p = core.Primitive('unpin')
+
+@unpin_p.def_abstract_eval
+def _unpin_abstract_eval(aval):
+  if not isinstance(aval, AbstractLinVal): raise TypeError(aval)
+  return aval.inner_aval
+
+def _lower_unpin(ctx, x_op):
+  out_aval, = ctx.avals_out
+  return mlir.custom_call(
+      "Unpin",
+      operands=mlir.flatten_ir_values([x_op]),
+      result_types=[mlir.aval_to_ir_type(out_aval)],
+  ).results
+mlir.register_lowering(unpin_p, _lower_unpin)
+
+
+def _linval_to_mlir_type(a):
+  return mlir.ir.MemRefType.get(a.shape, mlir.dtype_to_ir_type(a.dtype),
+                                memory_space=a.memory_space)
+mlir.ir_type_handlers[AbstractLinVal] = _linval_to_mlir_type
