@@ -36,8 +36,6 @@ import unittest
 import zlib
 
 from absl.testing import parameterized
-import jax
-from jax import lax
 from jax._src import api
 from jax._src import compilation_cache
 from jax._src import config
@@ -46,14 +44,16 @@ from jax._src import deprecations
 from jax._src import dispatch
 from jax._src import dtypes as _dtypes
 from jax._src import lib as _jaxlib
+from jax._src import mesh as mesh_lib
 from jax._src import monitoring
+from jax._src import sharding_impls
 from jax._src import test_warning_util
-from jax._src.typing import ArrayLike, DTypeLike
 from jax._src import xla_bridge
 from jax._src import util
-from jax._src import mesh as mesh_lib
 from jax._src.cloud_tpu_init import running_in_cloud_tpu_vm
 from jax._src.interpreters import mlir
+from jax._src.lax import lax
+from jax._src.lib import cuda_versions
 from jax._src.lib.mlir.dialects import hlo
 from jax._src.numpy.util import promote_dtypes, promote_dtypes_inexact
 from jax._src.public_test_util import (  # noqa: F401
@@ -63,11 +63,15 @@ from jax._src.test_loader import thread_unsafe_test as thread_unsafe_test
 from jax._src.test_loader import thread_unsafe_test_class as thread_unsafe_test_class
 from jax._src.test_loader import JaxTestLoader as JaxTestLoader
 from jax._src.test_loader import TEST_NUM_THREADS as TEST_NUM_THREADS
+from jax._src.tree_util import tree_all, tree_flatten, tree_map, tree_unflatten
+from jax._src.typing import ArrayLike, DTypeLike
 from jax._src.util import unzip2
-from jax.tree_util import tree_all, tree_flatten, tree_map, tree_unflatten
 import numpy as np
 import numpy.random as npr
 
+# When running tests, install the ABSL failure signal handler. This dumps a
+# C++ back trace on fatal signals, which is helpful for debugging.
+util.install_failure_signal_handler()
 
 
 # This submodule includes private test utilities that are not exported to
@@ -131,8 +135,8 @@ def to_default_dtype(arr: ArrayLike) -> np.ndarray:
   ``jax_default_dtype_bits`` flags.
   """
   arr = np.asarray(arr)
-  dtype = _dtypes._default_types.get(arr.dtype.kind)
-  return arr.astype(_dtypes.canonicalize_dtype(dtype)) if dtype else arr
+  dtype_fn = _dtypes.default_types.get(arr.dtype.kind)
+  return arr.astype(dtype_fn()) if dtype_fn else arr
 
 def with_jax_dtype_defaults(func: Callable[..., Any], use_defaults: bool = True):
   """Return a version of a function with outputs that match JAX's default dtypes.
@@ -293,11 +297,12 @@ def count_events(event):
 
 count_device_put = count_events("batched_device_put")
 count_device_put_fast_path_hit = count_events("batched_copy_array")
-count_pjit_cpp_cache_miss = count_events("pjit_lower")
+count_pjit_cpp_cache_miss = count_events("jit_cpp_cache_miss")
 count_jit_tracing_cache_miss = count_events("create_pjit_jaxpr")
 count_aot_jit_cpp_cache_miss = count_events("stages_compiled_call")
 count_jit_and_pmap_lowerings = count_events("lower_jaxpr_to_module")
 count_jit_compilation_cache_miss = count_events("pxla_cached_compilation")
+count_compilation_after_persistent_cache_miss = count_events("compile_after_persistent_compilation_miss")
 count_jax_array_shard_arg_calls = count_events("_array_shard_arg")
 
 
@@ -356,12 +361,12 @@ def assert_num_jit_and_pmap_compilations(times):
 
 @contextmanager
 def count_internal_device_puts():
-  before = jax._src.lib._jax.get_internal_device_put_info()
+  before = _jaxlib._jax.get_internal_device_put_info()
   counts = {}
   try:
     yield lambda: counts
   finally:
-    after = jax._src.lib._jax.get_internal_device_put_info()
+    after = _jaxlib._jax.get_internal_device_put_info()
     for k, v in after.items():
       diff = v - before.get(k, 0)
       if diff != 0:
@@ -407,11 +412,26 @@ def is_device_cuda():
 def is_cloud_tpu():
   return running_in_cloud_tpu_vm
 
+def is_optimized_build():
+  return _jaxlib._jax.is_optimized_build()
+
+def is_asan():
+  return _jaxlib._jax.is_asan()
+
+def is_msan():
+  return _jaxlib._jax.is_msan()
+
+def is_tsan():
+  return _jaxlib._jax.is_tsan()
+
+def is_sanitized():
+  return _jaxlib._jax.is_sanitized()
+
 # Returns True if it is not cloud TPU. If it is cloud TPU, returns True if it is
 # built at least `date``.
 # TODO(b/327203806): after libtpu adds a XLA version and the oldest support
 # libtpu contains the XLA version, remove using built time to skip tests.
-def if_cloud_tpu_at_least(year: int, month: int, day: int):
+def is_cloud_tpu_at_least(year: int, month: int, day: int):
   date = datetime.date(year, month, day)
   if not is_cloud_tpu():
     return True
@@ -443,7 +463,7 @@ def stablehlo_version_at_least(required_version: str):
 def get_tpu_version() -> int:
   if device_under_test() != "tpu":
     raise ValueError("Device is not TPU")
-  kind = jax.devices()[0].device_kind
+  kind = xla_bridge.devices()[0].device_kind
   match = re.match(r"TPU[^\d]*(\d+)", kind)
   if match is None:
     raise ValueError(f"Device kind {kind} is not supported")
@@ -459,19 +479,87 @@ def is_device_tpu(version: int | None = None, variant: str = "") -> bool:
     return False
   if version is None:
     return True
-  device_kind = jax.devices()[0].device_kind
+  device_kind = xla_bridge.devices()[0].device_kind
   expected_version = f"v{version}{variant}"
   # Special case v5e until the name is updated in device_kind
   if expected_version == "v5e":
     return "v5 lite" in device_kind
   elif expected_version == "v6e":
     return "v6 lite" in device_kind
+  elif expected_version == "v5p":
+    return device_kind.endswith("v5")
+  elif expected_version == "v7x":
+    return "TPU7x" in device_kind
   return expected_version in device_kind
 
+def pattern_search(patterns: str | Sequence[str], string: str):
+  if not isinstance(patterns, tuple):
+    patterns = (patterns,)  # type: ignore
+
+  for pattern in patterns:
+    if re.search(pattern, string):
+      return pattern
+  return None
+
+def device_kind_match(device_patterns: str | Sequence[str]):
+  device_kind = xla_bridge.devices()[0].device_kind
+  matching_pattern = pattern_search(device_patterns, device_kind)
+  return matching_pattern
+
+def skip_if_errors(
+    *,
+    error_patterns: str | Sequence[str],
+    device_patterns: str | Sequence[str],
+    reason: str | Callable[[str, str], str],
+):
+  """Skip if both error message and device kind match a corresponding pattern."""
+  def skip(test_method):
+    @functools.wraps(test_method)
+    def test_method_wrapper(self, *args, **kwargs):
+      device_kind = xla_bridge.devices()[0].device_kind
+      try:
+        return test_method(self, *args, **kwargs)
+      except Exception as e:
+        matching_error_pattern = pattern_search(error_patterns, str(e))
+        matching_device_pattern = pattern_search(device_patterns, device_kind)
+        if matching_error_pattern and matching_device_pattern:
+          if not isinstance(reason, str):
+            reason_str = reason(matching_error_pattern, matching_device_pattern)
+          else:
+            reason_str = reason
+          self.skipTest(reason_str)
+        raise
+    return test_method_wrapper
+  return skip
+
+skip_if_mosaic_gpu_exceeds_shared_memory = functools.partial(
+  skip_if_errors,
+  error_patterns="kernel exceeds available shared memory",
+  reason=lambda err, dev: f"Mosaic GPU kernel exceeds shared memory on {dev}",
+)
+
+skip_if_triton_exceeds_shared_memory = functools.partial(
+  skip_if_errors,
+  error_patterns="Shared memory size limit exceeded",
+  reason=lambda err, dev: f"Triton kernel exceeds shared memory on {dev}",
+)
+
+def get_cuda_nonportable_max_cluster_size():
+  # Per-device nonportable maximum cluster sizes for Jetson Thor and DGX
+  # Spark (GB10) determined by querying cuOccupancyMaxPotentialClusterSize
+  if device_kind_match("Thor$"):
+    return 8
+  elif device_kind_match("GB10$"):
+    return 12
+  # 16 is the nonportable maximum cluster size on:
+  # - Hopper: https://docs.nvidia.com/cuda/hopper-tuning-guide/index.html#:~:text=cluster%20size%20of-,16,-by%20opting%20in
+  # - Blackwell: https://docs.nvidia.com/cuda/blackwell-tuning-guide/index.html#:~:text=cluster%20size%20of-,16,-by%20opting%20in
+  return 16
+
 def is_cuda_compute_capability_at_least(capability: str) -> bool:
   if not is_device_cuda():
     return False
-  d, *_ = jax.local_devices(backend="gpu")
+  d, *_ = xla_bridge.local_devices(backend="gpu")
   target = tuple(int(x) for x in capability.split("."))
   current = tuple(int(x) for x in d.compute_capability.split("."))
   return current >= target
@@ -479,11 +567,19 @@ def is_cuda_compute_capability_at_least(capability: str) -> bool:
 def is_cuda_compute_capability_equal(capability: str) -> bool:
   if not is_device_cuda():
     return False
-  d, *_ = jax.local_devices(backend="gpu")
+  d, *_ = xla_bridge.local_devices(backend="gpu")
   target = tuple(int(x) for x in capability.split("."))
   current = tuple(int(x) for x in d.compute_capability.split("."))
   return current == target
 
+def is_cuda_version_at_least(major: int, minor: int):
+  assert 0 <= major
+  assert 0 <= minor < 100
+  return (
+      cuda_versions is not None
+      and cuda_versions.cuda_runtime_get_version() >= major * 1000 + minor * 10
+  )
+
 
 class CudaArchSpecificTest:
   """A mixin with methods allowing to skip arch specific tests."""
@@ -500,14 +596,13 @@ class CudaArchSpecificTest:
 def _get_device_tags():
   """returns a set of tags defined for the device under test"""
   if is_device_rocm():
-    device_tags = {device_under_test(), "rocm"}
+    return {device_under_test(), "rocm"}
   elif is_device_cuda():
-    device_tags = {device_under_test(), "cuda"}
+    return {device_under_test(), "cuda"}
   elif device_under_test() == "METAL":
-    device_tags = {device_under_test(), "gpu"}
+    return {device_under_test(), "gpu"}
   else:
-    device_tags = {device_under_test()}
-  return device_tags
+    return {device_under_test()}
 
 def test_device_matches(device_types: Iterable[str]) -> bool:
   assert not isinstance(
@@ -1124,9 +1219,9 @@ class NotPresent:
 @contextmanager
 def assert_global_configs_unchanged():
   starting_cache = compilation_cache._cache
-  starting_config = jax.config.values.copy()
+  starting_config = config.config.values.copy()
   yield
-  ending_config = jax.config.values
+  ending_config = config.config.values
   ending_cache = compilation_cache._cache
 
   if starting_config != ending_config:
@@ -1153,36 +1248,28 @@ class JaxTestCase(parameterized.TestCase):
     'jax_legacy_prng_key': 'error',
   }
 
-  _context_stack: ExitStack | None = None
-
-
   def setUp(self):
     super().setUp()
-    self.enter_context(assert_global_configs_unchanged())
+    self.enterContext(assert_global_configs_unchanged())
 
     # We use the adler32 hash for two reasons.
     # a) it is deterministic run to run, unlike hash() which is randomized.
     # b) it returns values in int32 range, which RandomState requires.
     self._rng = npr.RandomState(zlib.adler32(self._testMethodName.encode()))
 
-    # TODO(phawkins): use TestCase.enterContext once Python 3.11 is the minimum
-    # version.
-    self._context_stack = ExitStack()
-    self.addCleanup(self._context_stack.close)
-    stack = self._context_stack
-    stack.enter_context(global_config_context(**self._default_global_config))
+    self.enterContext(global_config_context(**self._default_global_config))
     for config_name, value in self._default_thread_local_config.items():
-      stack.enter_context(jax._src.config.config_states[config_name](value))
+      self.enterContext(config.config_states[config_name](value))
 
     if TEST_WITH_PERSISTENT_COMPILATION_CACHE.value:
       assert TEST_NUM_THREADS.value <= 1, "Persistent compilation cache is not thread-safe."
-      stack.enter_context(config.enable_compilation_cache(True))
-      stack.enter_context(config.raise_persistent_cache_errors(True))
-      stack.enter_context(config.persistent_cache_min_compile_time_secs(0))
-      stack.enter_context(config.persistent_cache_min_entry_size_bytes(0))
-      tmp_dir = stack.enter_context(tempfile.TemporaryDirectory())
-      stack.enter_context(config.compilation_cache_dir(tmp_dir))
-      stack.callback(compilation_cache.reset_cache)
+      self.enterContext(config.enable_compilation_cache(True))
+      self.enterContext(config.raise_persistent_cache_errors(True))
+      self.enterContext(config.persistent_cache_min_compile_time_secs(0))
+      self.enterContext(config.persistent_cache_min_entry_size_bytes(0))
+      tmp_dir = self.enterContext(tempfile.TemporaryDirectory())
+      self.enterContext(config.compilation_cache_dir(tmp_dir))
+      self.addCleanup(compilation_cache.reset_cache)
 
   def tearDown(self) -> None:
     assert core.reset_trace_state()
@@ -1252,14 +1339,16 @@ class JaxTestCase(parameterized.TestCase):
                             rtol=rtol, canonicalize_dtypes=canonicalize_dtypes,
                             err_msg=err_msg)
     elif is_sequence(actual) and not hasattr(actual, '__array__'):
-      self.assertTrue(is_sequence(desired) and not hasattr(desired, '__array__'))
+      self.assertTrue(is_sequence(desired) and not hasattr(desired, '__array__'),
+                      msg=f"Expected sequence, got {desired}")
       self.assertEqual(len(actual), len(desired))
       for actual_elt, desired_elt in zip(actual, desired):
         self.assertAllClose(actual_elt, desired_elt, check_dtypes=check_dtypes, atol=atol,
                             rtol=rtol, canonicalize_dtypes=canonicalize_dtypes,
                             err_msg=err_msg)
     elif hasattr(actual, '__array__') or np.isscalar(actual):
-      self.assertTrue(hasattr(desired, '__array__') or np.isscalar(desired))
+      self.assertTrue(hasattr(desired, '__array__') or np.isscalar(desired),
+                      msg=f"Expected array-like, got {desired}")
       if check_dtypes:
         self.assertDtypesMatch(actual, desired, canonicalize_dtypes=canonicalize_dtypes)
       actual = np.asarray(actual)
@@ -1307,7 +1396,7 @@ class JaxTestCase(parameterized.TestCase):
 
   @contextmanager
   def assertWarnsRegex(self, warning, regex):
-    if regex is not None:
+    if regex is not None and not isinstance(regex, re.Pattern):
         regex = re.compile(regex)
 
     with test_warning_util.record_warnings() as ws:
@@ -1318,8 +1407,8 @@ class JaxTestCase(parameterized.TestCase):
       if regex is not None and not regex.search(str(w.message)):
         continue
       return
-    self.fail(f"Expected warning not found {warning}:'{regex}', got "
-              f"{ws}")
+    self.fail(f"Expected warning not found {warning}:'{regex}', "
+              f"got warnings: {[str(w.message) for w in ws]}")
 
 
   def _CompileAndCheck(self, fun, args_maker, *, check_dtypes=True, tol=None,
@@ -1379,7 +1468,32 @@ class JaxTestCase(parameterized.TestCase):
                         atol=atol or tol, rtol=rtol or tol,
                         canonicalize_dtypes=canonicalize_dtypes)
 
-_PJIT_IMPLEMENTATION = jax.jit
+  def assertCacheMisses(self,
+                        func: Callable[[], Any], *,
+                        cpp: int | None = None,
+                        aot_call: int | None = None,
+                        tracing: int | None = None,
+                        lowering: int | None = None,
+                        compilation_after_persistent_cache_miss: int | None = None):
+    with (count_pjit_cpp_cache_miss() as cpp_count,
+          count_aot_jit_cpp_cache_miss() as aot_call_count,
+          count_jit_tracing_cache_miss() as tracing_count,
+          count_jit_and_pmap_lowerings() as lowering_count,
+          count_compilation_after_persistent_cache_miss() as compilation_count):
+      func()
+    if cpp is not None:
+      self.assertEqual(cpp, cpp_count())
+    if aot_call is not None:
+      self.assertEqual(aot_call, aot_call_count())
+    if tracing is not None:
+      self.assertEqual(tracing, tracing_count())
+    if lowering is not None:
+      self.assertEqual(lowering, lowering_count())
+    if compilation_after_persistent_cache_miss is not None:
+      self.assertEqual(compilation_after_persistent_cache_miss,
+                       compilation_count())
+
+_PJIT_IMPLEMENTATION = api.jit
 _PJIT_IMPLEMENTATION._name = "jit"
 _NOOP_JIT_IMPLEMENTATION = lambda x, *args, **kwargs: x
 _NOOP_JIT_IMPLEMENTATION._name = "noop"
@@ -1409,11 +1523,11 @@ def with_mesh(named_shape: MeshSpec) -> Generator[None, None, None]:
   # This is similar to the `with_mesh` function above, but isn't a decorator.
   axis_names, shape = unzip2(named_shape)
   size = math.prod(shape)
-  local_devices = list(jax.local_devices())
+  local_devices = list(xla_bridge.local_devices())
   if len(local_devices) < size:
     raise unittest.SkipTest(f"Test requires {size} local devices")
   mesh_devices = np.array(local_devices[:size]).reshape(shape)  # type: ignore
-  with jax.sharding.Mesh(mesh_devices, axis_names):
+  with mesh_lib.Mesh(mesh_devices, axis_names):
     yield
 
 def with_mesh_from_kwargs(f):
@@ -1433,7 +1547,7 @@ def with_explicit_mesh(sizes, names, axis_types=None, iota_order=False):
   def decorator(fn):
     def mesh_fn(*args, **kwargs):
       mesh = create_mesh(sizes, names, iota_order, axis_types=axis_types)
-      with jax.sharding.use_mesh(mesh):
+      with sharding_impls.set_mesh(mesh):
         return fn(*args, **kwargs, mesh=mesh)
     return mesh_fn
   return decorator
@@ -1441,14 +1555,16 @@ def with_explicit_mesh(sizes, names, axis_types=None, iota_order=False):
 
 def create_mesh(mesh_shape, axis_names, iota_order=False, axis_types=None):
   size = math.prod(mesh_shape)
-  if len(jax.devices()) < size:
-    raise unittest.SkipTest(f"Test requires {size} global devices.")
+  if len(xla_bridge.devices()) < size:
+    raise unittest.SkipTest(f"Test requires {size} global devices and found {len(xla_bridge.devices())}.")
   if iota_order:
-    devices = sorted(jax.devices(), key=lambda d: d.id)
+    devices = sorted(xla_bridge.devices(), key=lambda d: d.id)
     mesh_devices = np.array(devices[:size]).reshape(mesh_shape)
-    return jax.sharding.Mesh(mesh_devices, axis_names, axis_types=axis_types)
+    return mesh_lib.Mesh(mesh_devices, axis_names, axis_types=axis_types)
   else:
-    return jax.make_mesh(mesh_shape, axis_names, axis_types)
+    if axis_types is None:
+      axis_types = (mesh_lib.AxisType.Auto,) * len(mesh_shape)
+    return sharding_impls.make_mesh(mesh_shape, axis_names, axis_types)
 
 class _cached_property:
   null = object()
@@ -1548,8 +1664,8 @@ def strict_promotion_if_dtypes_match(dtypes):
   and enable standard dtype promotion otherwise.
   """
   if all(dtype == dtypes[0] for dtype in dtypes):
-    return jax.numpy_dtype_promotion('strict')
-  return jax.numpy_dtype_promotion('standard')
+    return config.numpy_dtype_promotion('strict')
+  return config.numpy_dtype_promotion('standard')
 
 _version_regex = re.compile(r"([0-9]+(?:\.[0-9]+)*)(?:(rc|dev).*)?")
 def parse_version(v: str) -> tuple[int, ...]:
@@ -1627,7 +1743,7 @@ def register_event_duration_listener(callback):
     monitoring.register_event_duration_secs_listener(callback)
     yield
   finally:
-    monitoring._unregister_event_duration_listener_by_callback(callback)
+    monitoring.unregister_event_duration_listener(callback)
 
 
 @contextmanager
@@ -1664,22 +1780,12 @@ def set_env(**kwargs):
     os.environ.update({k: v for k, v in original.items() if v is not None})
 
 def fwd_bwd_jaxprs(f, *example_args):
-  fwd_jaxpr, (y_shape, res_shape) = jax.make_jaxpr(
-      lambda *args: jax.vjp(f, *args), return_shape=True)(*example_args)
-  bwd_jaxpr = jax.make_jaxpr(lambda res, outs: res(outs))(res_shape, y_shape)
+  fwd_jaxpr, (y_shape, res_shape) = api.make_jaxpr(
+      lambda *args: api.vjp(f, *args), return_shape=True)(*example_args)
+  bwd_jaxpr = api.make_jaxpr(lambda res, outs: res(outs))(res_shape, y_shape)
   return fwd_jaxpr, bwd_jaxpr
 
 
-def numpy_vecdot(x, y, axis):
-  """Implementation of numpy.vecdot for testing on numpy < 2.0.0"""
-  if numpy_version() >= (2, 0, 0):
-    raise ValueError("should be calling vecdot directly on numpy 2.0.0")
-  x = np.moveaxis(x, axis, -1)
-  y = np.moveaxis(y, axis, -1)
-  x, y = np.broadcast_arrays(x, y)
-  return np.matmul(np.conj(x[..., None, :]), y[..., None])[..., 0, 0]
-
-
 def complex_plane_sample(dtype, size_re=10, size_im=None):
   """Return a 2-D array of complex numbers that covers the complex plane
      with a grid of samples.
@@ -1781,8 +1887,8 @@ class vectorize_with_mpmath(np.vectorize):
     self.extra_prec_multiplier = kwargs.pop('extra_prec_multiplier', 0)
     self.extra_prec = kwargs.pop('extra_prec', 0)
     self.mpmath = mpmath
-    self.contexts = dict()
-    self.contexts_inv = dict()
+    self.contexts = {}
+    self.contexts_inv = {}
     for fp_format, prec in self.float_prec.items():
       ctx = self.mpmath.mp.clone()
       ctx.prec = prec
@@ -2235,6 +2341,17 @@ class numpy_with_mpmath:
       assert 0  # unreachable
 
 # Hypothesis testing support
+def hypothesis_is_thread_safe() -> bool:
+  """Returns True if the installed hypothesis version is thread-safe.
+
+  Hypothesis versions >= 6.136.9 are thread-safe.
+  """
+  try:
+    import hypothesis as hp  # pytype: disable=import-error
+    return tuple(int(x) for x in hp.__version__.split('.')) >= (6, 136, 9)
+  except (ModuleNotFoundError, ImportError):
+    return True
+
 def setup_hypothesis(max_examples=30) -> None:
   """Sets up the hypothesis profiles.
 
@@ -2294,3 +2411,13 @@ def setup_hypothesis(max_examples=30) -> None:
   profile = HYPOTHESIS_PROFILE.value
   logging.info("Using hypothesis profile: %s", profile)
   hp.settings.load_profile(profile)
+
+
+def runtime_environment() -> str | None:
+  """Returns None, "bazel" or "pytest"."""
+  if sys.executable is None:
+    return None
+  elif 'bazel-out' in sys.executable:
+    return "bazel"
+  else:
+    return "pytest"
