@@ -22,20 +22,20 @@ import collections.abc
 from collections.abc import Callable, Sequence
 import dataclasses
 import enum
-import functools
 import io
 import json
 from typing import Any, TypedDict
 
-import jax
+from jax._src import api
 from jax._src import config
 from jax._src import core
+from jax._src import dispatch
 from jax._src import sharding_impls
 from jax._src.cloud_tpu_init import is_cloud_tpu_older_than
 from jax._src.frozen_dict import FrozenDict
+from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
 from jax._src.lib import tpu
-from jax.interpreters import xla
 from jaxlib.mlir import ir
 from jaxlib.mlir.passmanager import PassManager
 
@@ -58,28 +58,41 @@ _MOSAIC_ALLOW_HLO = config.bool_state(
 # mode: for 1 month when exporting, or when using old cloud TPU.
 #
 # This can be achieved by adding:
-#    if ctx.is_forward_compat() or is_cloud_tpu_older_than(<today>):
+#    if ctx.is_forward_compat() or backend is None or is_cloud_tpu_older_than(<today>):
 #       return <previous_serialization_version>
 #    return None
 #
 # We should also add a TODO to remove the conditional one month later.
 def get_ir_version(ctx: mlir.LoweringRuleContext) -> int | None:
-  # TODO: b/423649694 - remove the forward compatibility check after 2025-07-18.
-  if ctx.is_forward_compat() or is_cloud_tpu_older_than(2025, 6, 19):
-    return 4
-  # TODO: b/425259894 - remove the forward compatibility check after 2025-07-24
-  if ctx.is_forward_compat() or is_cloud_tpu_older_than(2025, 7, 25):
-    return 5
-  # TODO: b/427422863 - remove the forward compatibility check after 2025-07-26.
-  if ctx.is_forward_compat() or is_cloud_tpu_older_than(2025, 7, 27):
-    return 6
+  backend = ctx.module_context.get_backend(optional=True)
+  # TODO(apaszke): remove the forward compatibility check after 2025-12-5.
+  if (
+      ctx.is_forward_compat()
+      or backend is None
+      or is_cloud_tpu_older_than(2025, 11, 5, backend)
+  ):
+    return 8
   return None
 
 
 tpu_custom_call_p = core.Primitive("tpu_custom_call")
-tpu_custom_call_p.def_impl(
-    functools.partial(xla.apply_primitive, tpu_custom_call_p))
 tpu_custom_call_p.multiple_results = True
+dispatch.simple_impl(tpu_custom_call_p)
+
+
+def tpu_custom_call_batcher(axis_data, args, dims, **kwargs):
+  if axis_data.size != 1:
+    raise NotImplementedError(
+        "tpu_custom_call does not support non-trivial batching."
+    )
+  unbatched_args = tuple(
+      a if (d is batching.not_mapped or d is None) else a[d]
+      for a, d in zip(args, dims, strict=True)
+  )
+  out_unbatched = tpu_custom_call_p.bind(*unbatched_args, **kwargs)
+  out = tuple(o[None] for o in out_unbatched)
+  return out, (0,) * len(out)
+batching.fancy_primitive_batchers[tpu_custom_call_p] = tpu_custom_call_batcher
 
 
 class MemorySpace(enum.Enum):
@@ -87,6 +100,8 @@ class MemorySpace(enum.Enum):
   VMEM = enum.auto()
   SEMAPHORE_MEM = enum.auto()
   SMEM = enum.auto()
+  HOST = enum.auto()
+  SC_SCALAR_SEMAPHORE_MEM = enum.auto()
 
   @property
   def color(self) -> int:
@@ -96,8 +111,12 @@ class MemorySpace(enum.Enum):
       return 1
     elif self == MemorySpace.SEMAPHORE_MEM:
       return 2
+    elif self == MemorySpace.SC_SCALAR_SEMAPHORE_MEM:
+      return 8
     elif self == MemorySpace.SMEM:
       return 4
+    elif self == MemorySpace.HOST:
+      return 5
     else:
       raise ValueError("invalid memory space: " + str(self))
 
@@ -106,6 +125,29 @@ class CostEstimate(TypedDict):
   flops: int
   transcendentals: int
   bytes_accessed: int
+  remote_bytes_transferred: int = 0
+
+  def to_json(self) -> bytes:
+    return (
+        f'{{"flops": {self["flops"]}, "transcendentals":'
+        f' {self["transcendentals"]}, "bytes_accessed":'
+        f' {self["bytes_accessed"]}, "remote_bytes_transferred":'
+        f' {self["remote_bytes_transferred"]}}}'
+    ).encode("ascii")
+
+
+class TpuSideEffectType(enum.Enum):
+  # No side effects, can be deduplicated / removed if unused.
+  PURE = "pure"
+  # Cannot be deduplicated, but can be removed if unused.
+  DATAFLOW_SIDE_EFFECTING = "dataflow_side_effecting"
+  # Cannot be deduplicated or removed.
+  SIDE_EFFECTING = "side_effecting"
+
+
+class Tiling(enum.Enum):
+  COMPACT = "TILING_COMPACT"
+  SPARSE_CORE = "TILING_SPARSE_CORE"
 
 
 @dataclasses.dataclass(frozen=True)
@@ -127,6 +169,9 @@ class CustomCallBackendConfig:
   disable_bounds_checks: bool
   active_core_count: int | None
   input_memory_spaces: tuple[MemorySpace | None, ...] | None
+  skip_device_barrier: bool
+  shape_invariant_numerics: bool
+  tiling: Tiling | None = None  # Only used for SparseCore.
 
   def __post_init__(self):
     if self.allow_input_fusion is not None:
@@ -156,9 +201,7 @@ class CustomCallBackendConfig:
       config.write(str(self.collective_id).encode("ascii"))
     if self.cost_estimate is not None:
       config.write(b', "cost_estimate": ')
-      config.write(
-          json.dumps(dict(self.cost_estimate), sort_keys=True).encode("ascii")
-      )
+      config.write(_compact_json_object(**self.cost_estimate))
     if self.needs_hlo_passes:
       config.write(b', "needs_hlo_passes": ')
       config.write(str(self.needs_hlo_passes).lower().encode("ascii"))
@@ -168,11 +211,13 @@ class CustomCallBackendConfig:
     if self.needs_layout_passes:
       config.write(b', "needs_layout_passes": ')
       config.write(str(self.needs_layout_passes).lower().encode("ascii"))
+    if not self.shape_invariant_numerics:
+      config.write(b', "shape_invariant_numerics": ')
+      config.write(str(self.shape_invariant_numerics).lower().encode("ascii"))
     if self.allow_input_fusion is not None:
       config.write(b', "allow_input_fusion": [')
       for i, value in enumerate(self.allow_input_fusion):
         config.write(b"true" if value else b"false")
-        # config.write(str(value).lower().encode("ascii"))
         if i + 1 != len(self.allow_input_fusion):
           config.write(b",")
       config.write(b"]")
@@ -218,7 +263,13 @@ class CustomCallBackendConfig:
     if self.disable_bounds_checks:
       config.write(b', "disable_bounds_checks": ')
       config.write(str(self.disable_bounds_checks).lower().encode("ascii"))
+    if self.skip_device_barrier:
+      config.write(b', "skip_device_barrier": ')
+      config.write(str(self.skip_device_barrier).lower().encode("ascii"))
     config.write(b"}")  # End of custom_call_config.
+    if self.tiling is not None:
+      config.write(b', "sparse_core_config": ')
+      config.write(_compact_json_object(tiling=self.tiling.value))
     if self.device_type is not None:
       config.write(b', "device_type": ')
       config.write(
@@ -236,7 +287,7 @@ class CustomCallBackendConfig:
       for i, (flag, value) in enumerate(self.flags.items()):
         config.write(b'{"flag_type": "')
         config.write(flag.encode("ascii"))
-        config.write(b'", value: {')
+        config.write(b'", "value": {')
         if isinstance(value, bool):
           config.write(b'"boolean_value": ')
           config.write(b"true" if value else b"false")
@@ -258,6 +309,12 @@ class CustomCallBackendConfig:
     return config.getvalue()
 
 
+def _compact_json_object(**kwargs: Any) -> bytes:
+  return json.dumps(
+      kwargs, sort_keys=True, indent=0, separators=(",", ":")
+  ).encode("ascii")
+
+
 @tpu_custom_call_p.def_abstract_eval
 def _tpu_custom_call_abstract_eval(*_, out_avals, **__):
   return out_avals
@@ -271,7 +328,7 @@ def _tpu_custom_call_lowering(
     ctx: mlir.LoweringRuleContext,
     *in_nodes,  # pylint: disable=missing-function-docstring
     config: CustomCallBackendConfig,
-    has_side_effects: bool,
+    has_side_effects: TpuSideEffectType,
     kernel_name: str | None,
     out_avals: Any,
     input_output_aliases: tuple[tuple[int, int], ...],
@@ -280,7 +337,8 @@ def _tpu_custom_call_lowering(
   result_types = [mlir.aval_to_ir_type(aval) for aval in out_avals]
   axis_context = ctx.module_context.axis_context
   if isinstance(axis_context, sharding_impls.SPMDAxisContext):
-    if axis_context.manual_axes != frozenset(axis_context.mesh.axis_names):
+    if (axis_context.manual_axes and
+        axis_context.manual_axes != frozenset(axis_context.mesh.axis_names)):
       raise NotImplementedError(
           "Mosaic kernels cannot be automatically partitioned. Please wrap the"
           " call in a shard_map."
@@ -307,24 +365,29 @@ def _tpu_custom_call_lowering(
   # information.
   if kernel_name is not None:
     extra_attributes = dict(kernel_name=ir.StringAttr.get(kernel_name))
-  has_side_effects = has_side_effects if has_side_effects is not None else False
   call = mlir.custom_call(
       "tpu_custom_call",
       result_types=result_types,
       operands=in_nodes,
       backend_config=config.to_json(),
       api_version=1,
-      has_side_effect=has_side_effects,
+      has_side_effect=has_side_effects != TpuSideEffectType.PURE,
       operand_output_aliases=dict(input_output_aliases),
       operand_layouts=_avals_to_layouts(ctx.avals_in),
       result_layouts=_avals_to_layouts(ctx.avals_out),
       result_shapes=result_shapes,
       extra_attributes=extra_attributes,
   )
+  metadata_dict = {}
   if metadata is not None:
-    call.attributes["mhlo.frontend_attributes"] = ir.DictAttr.get(
-        dict(kernel_metadata=ir.StringAttr.get(json.dumps(metadata)))
+    metadata_dict["kernel_metadata"] = ir.StringAttr.get(
+        _compact_json_object(**metadata)
     )
+  assert isinstance(has_side_effects, TpuSideEffectType)
+  if has_side_effects == TpuSideEffectType.DATAFLOW_SIDE_EFFECTING:
+    metadata_dict["xla_allow_dce_side_effecting_op"] = ir.StringAttr.get("true")
+  if metadata_dict:
+    call.attributes["mhlo.frontend_attributes"] = ir.DictAttr.get(metadata_dict)
   return call.results
 
 
@@ -335,14 +398,11 @@ mlir.register_lowering(tpu_custom_call_p, _tpu_custom_call_lowering,
 def _lower_mosaic_module_to_asm(
     module: ir.Module,
     *,
-    device_type: str | None,
     ir_version: int | None = None,
-) -> tuple[ir.Module, tuple[bool, bool, bool, bool]]:
+) -> tuple[ir.Module, tuple[bool, bool]]:
   has_communication, has_custom_barrier = tpu.private_has_communication(
       module.operation
   )
-  needs_hlo_passes = _MOSAIC_ALLOW_HLO.value
-  needs_layout_passes = not device_type
   # We'll mutate the module, so clone it
   with module.context as ctx, module.operation.location as _:
     module_op = module.operation.clone()
@@ -364,8 +424,6 @@ def _lower_mosaic_module_to_asm(
     return asm, (
         has_communication,
         has_custom_barrier,
-        needs_hlo_passes,
-        needs_layout_passes,
     )
 
 
@@ -496,16 +554,21 @@ def _lower_to_custom_call_config(
     ir_version: int | None = None,
     disable_bounds_checks: bool = False,
     input_memory_spaces: tuple[MemorySpace | None, ...] | None = None,
+    skip_device_barrier: bool = False,
+    allow_collective_id_without_custom_barrier: bool = False,
+    shape_invariant_numerics: bool = False,
+    needs_layout_passes: bool | None = None,
+    tiling: Tiling | None = None,
 ) -> CustomCallBackendConfig:
   device_type = _get_device_type(module)
+  needs_hlo_passes = _MOSAIC_ALLOW_HLO.value
+  if needs_layout_passes is None:
+    needs_layout_passes = not device_type
   lowered_module_asm, (
       has_communication,
       has_custom_barrier,
-      needs_hlo_passes,
-      needs_layout_passes,
   ) = _lower_mosaic_module_to_asm(
       module,
-      device_type=device_type,
       ir_version=ir_version,
   )
   active_core_count = _get_active_core_count(module)
@@ -527,6 +590,10 @@ def _lower_to_custom_call_config(
       disable_bounds_checks=disable_bounds_checks,
       active_core_count=active_core_count,
       input_memory_spaces=input_memory_spaces,
+      skip_device_barrier=skip_device_barrier,
+      allow_collective_id_without_custom_barrier=allow_collective_id_without_custom_barrier,
+      shape_invariant_numerics=shape_invariant_numerics,
+      tiling=tiling,
   )
 
 
@@ -549,13 +616,17 @@ def _lowered_to_custom_call_config(
     disable_bounds_checks: bool = False,
     active_core_count: int | None = None,
     input_memory_spaces: tuple[MemorySpace | None, ...] | None = None,
+    skip_device_barrier: bool = False,
+    allow_collective_id_without_custom_barrier: bool = False,
+    shape_invariant_numerics: bool = False,
+    tiling: Tiling | None = None,
 ):
   if has_custom_barrier:
     if collective_id is None:
       raise ValueError(
           "collective_id has to be specified when using a custom barrier"
       )
-  elif collective_id is not None:
+  elif collective_id is not None and not allow_collective_id_without_custom_barrier:
     raise ValueError(
         "collective_id has to be unspecified or None when not using a custom"
         " barrier"
@@ -565,7 +636,7 @@ def _lowered_to_custom_call_config(
         "vmem_limit_bytes must be an int: provided with a"
         f" {type(vmem_limit_bytes)}."
     )
-  config = CustomCallBackendConfig(
+  return CustomCallBackendConfig(
       lowered_module_asm,
       has_communication,
       collective_id,
@@ -582,8 +653,10 @@ def _lowered_to_custom_call_config(
       disable_bounds_checks,
       active_core_count=active_core_count,
       input_memory_spaces=input_memory_spaces,
+      skip_device_barrier=skip_device_barrier,
+      shape_invariant_numerics=shape_invariant_numerics,
+      tiling=tiling,
   )
-  return config
 
 
 def lower_module_to_custom_call(
@@ -599,13 +672,24 @@ def lower_module_to_custom_call(
     input_output_aliases: tuple[tuple[int, int], ...],
     internal_scratch_in_bytes: int | None,
     collective_id: int | None,
-    has_side_effects: bool,
+    has_side_effects: bool | TpuSideEffectType,
     serialization_format: int | None,
     output_memory_spaces: tuple[MemorySpace | None, ...] | None,
     disable_bounds_checks: bool = False,
     input_memory_spaces: tuple[MemorySpace | None, ...] | None,
     metadata: Any | None = None,
+    skip_device_barrier: bool = False,
+    allow_collective_id_without_custom_barrier: bool = False,
+    shape_invariant_numerics: bool = False,
+    needs_layout_passes: bool | None = None,
+    tiling: Tiling | None = None,
 ) -> Sequence[ir.Value]:
+  if isinstance(has_side_effects, bool):
+    has_side_effects = (
+        TpuSideEffectType.PURE
+        if not has_side_effects
+        else TpuSideEffectType.SIDE_EFFECTING
+    )
   config = _lower_to_custom_call_config(
       module,
       vmem_limit_bytes=vmem_limit_bytes,
@@ -619,6 +703,11 @@ def lower_module_to_custom_call(
       ir_version=get_ir_version(ctx),
       disable_bounds_checks=disable_bounds_checks,
       input_memory_spaces=input_memory_spaces,
+      skip_device_barrier=skip_device_barrier,
+      allow_collective_id_without_custom_barrier=allow_collective_id_without_custom_barrier,
+      shape_invariant_numerics=shape_invariant_numerics,
+      needs_layout_passes=needs_layout_passes,
+      tiling=tiling,
   )
   return _tpu_custom_call_lowering(
       ctx,
@@ -644,12 +733,15 @@ def as_tpu_kernel(
     input_output_aliases: tuple[tuple[int, int], ...] = (),
     internal_scratch_in_bytes: int | None = None,
     collective_id: int | None = None,
-    has_side_effects: bool = False,
+    has_side_effects: TpuSideEffectType = TpuSideEffectType.PURE,
     serialization_format: int | None = 1,
     output_memory_spaces: tuple[MemorySpace | None, ...] | None = None,
     disable_bounds_checks: bool = False,
     input_memory_spaces: tuple[MemorySpace | None, ...] | None = None,
+    shape_invariant_numerics: bool = False,
+    needs_layout_passes: bool | None = None,
     metadata: Any | None = None,
+    _ir_version: int | None = None,
 ) -> Callable[..., Any]:
   """Turns an MLIR Mosaic kernel into a JAX-compatible function."""
   config = _lower_to_custom_call_config(
@@ -664,6 +756,9 @@ def as_tpu_kernel(
       output_memory_spaces=output_memory_spaces,
       disable_bounds_checks=disable_bounds_checks,
       input_memory_spaces=input_memory_spaces,
+      shape_invariant_numerics=shape_invariant_numerics,
+      needs_layout_passes=needs_layout_passes,
+      ir_version=_ir_version,
   )
   return _as_jax_callable(
       config,
@@ -684,7 +779,7 @@ def lowered_as_tpu_kernel(
     needs_hlo_passes: bool = False,
     needs_layout_passes: bool = False,
     has_communication: bool = False,
-    has_side_effects: bool = False,
+    has_side_effects: bool | TpuSideEffectType = False,
     has_custom_barrier: bool = False,
     kernel_name: str | None = None,
     vmem_limit_bytes: int | None = None,
@@ -695,11 +790,18 @@ def lowered_as_tpu_kernel(
     internal_scratch_in_bytes: int | None = None,
     disable_bounds_checks: bool = False,
     metadata: Any | None = None,
+    allow_collective_id_without_custom_barrier: bool = False,
 ) -> Callable[..., Any]:
   device_type = _get_device_type(lowered_module)
   lowered_module_asm = lowered_module.operation.get_asm(
       binary=True, enable_debug_info=True
   )
+  if isinstance(has_side_effects, bool):
+    has_side_effects = (
+        TpuSideEffectType.PURE
+        if not has_side_effects
+        else TpuSideEffectType.DATAFLOW_SIDE_EFFECTING
+    )
   config = _lowered_to_custom_call_config(
       lowered_module_asm,
       vmem_limit_bytes=vmem_limit_bytes,
@@ -715,6 +817,7 @@ def lowered_as_tpu_kernel(
       needs_hlo_passes=needs_hlo_passes,
       needs_layout_passes=needs_layout_passes,
       disable_bounds_checks=disable_bounds_checks,
+      allow_collective_id_without_custom_barrier=allow_collective_id_without_custom_barrier,
   )
   return _as_jax_callable(
       config,
@@ -728,7 +831,7 @@ def lowered_as_tpu_kernel(
 
 def _as_jax_callable(
     config: CustomCallBackendConfig,
-    has_side_effects: bool,
+    has_side_effects: TpuSideEffectType,
     out_type: Any,
     *,
     kernel_name: str | None,
@@ -754,4 +857,4 @@ def _as_jax_callable(
     )
     return result[0] if unpack else result
 
-  return jax.jit(apply_kernel)
+  return api.jit(apply_kernel)
