@@ -41,7 +41,6 @@ from jax._src import hardware_utils
 from jax._src import traceback_util
 from jax._src import util
 from jax._src.cloud_tpu_init import get_tpu_library_path
-from jax._src.lib import jaxlib_extension_version
 from jax._src.lib import xla_client
 from jax._src.lib import _jax
 from jax._src.lib import _profiler
@@ -59,8 +58,6 @@ except ImportError as e:
 
 traceback_util.register_exclusion(__file__)
 
-XlaBackend = xla_client.Client
-
 # The platforms in this set will force forward compatibility for lowering.
 FORCE_FORWARD_COMPAT_LOWERING_PLATFORMS: set[str] = set()
 
@@ -69,24 +66,26 @@ MIN_COMPUTE_CAPABILITY = 52
 # TODO(phawkins): Remove jax_xla_backend.
 _XLA_BACKEND = config.string_flag(
     'jax_xla_backend', '',
-    'Deprecated, please use --jax_platforms instead.')
+    help='Deprecated, please use --jax_platforms instead.')
 BACKEND_TARGET = config.string_flag(
     'jax_backend_target',
     os.getenv('JAX_BACKEND_TARGET', '').lower(),
-    'Either "local" or "rpc:address" to connect to a remote service target.')
+    help='Either "local" or "rpc:address" to connect to a remote service target.')
 # TODO(skye): warn when this is used once we test out --jax_platforms a bit
 _PLATFORM_NAME = config.string_flag(
     'jax_platform_name',
     os.getenv('JAX_PLATFORM_NAME', '').lower(),
-    'Deprecated, please use --jax_platforms instead.')
+    help='Deprecated, please use --jax_platforms instead.')
 CUDA_VISIBLE_DEVICES = config.string_flag(
     'jax_cuda_visible_devices', 'all',
-    'Restricts the set of CUDA devices that JAX will use. Either "all", or a '
-    'comma-separate list of integer device IDs.')
+    help=(
+      'Restricts the set of CUDA devices that JAX will use. Either "all", or a '
+      'comma-separate list of integer device IDs.'))
 _ROCM_VISIBLE_DEVICES = config.string_flag(
     'jax_rocm_visible_devices', 'all',
-    'Restricts the set of ROCM devices that JAX will use. Either "all", or a '
-    'comma-separate list of integer device IDs.')
+    help=(
+      'Restricts the set of ROCM devices that JAX will use. Either "all", or a '
+      'comma-separate list of integer device IDs.'))
 
 MOCK_NUM_GPU_PROCESSES = config.int_flag(
     name="mock_num_gpu_processes",
@@ -102,12 +101,6 @@ MOCK_GPU_TOPOLOGY = config.string_flag(
          '<number-of-devices-per-host>". Empty string turns off mocking.',
 )
 
-_CPU_ENABLE_GLOO_COLLECTIVES = config.bool_flag(
-    name="jax_cpu_enable_gloo_collectives",
-    default=False,
-    help="Deprecated, please use jax_cpu_collectives_implementation instead.",
-)
-
 _CPU_ENABLE_ASYNC_DISPATCH = config.bool_flag(
     name="jax_cpu_enable_async_dispatch",
     default=True,
@@ -134,14 +127,16 @@ CROSS_HOST_TRANSPORT_ADDRESSES = config.string_flag(
 CROSS_HOST_TRANSFER_TIMEOUT_SECONDS = config.int_flag(
     "jax_cross_host_transfer_timeout_seconds",
     None,
-    "Timeout for cross host transfer metadata exchange through KV store. "
-    "Default is one minute.",
+    help=(
+      "Timeout for cross host transfer metadata exchange through KV store. "
+      "Default is one minute."
+    ),
 )
 
 CROSS_HOST_TRANSFER_TRANSFER_SIZE = config.int_flag(
     "jax_cross_host_transfer_transfer_size",
     None,
-    "Chunk size for chunked transfer requests."
+    help="Chunk size for chunked transfer requests."
 )
 
 # Warn the user if they call fork(), because it's not going to go well for them.
@@ -157,6 +152,29 @@ _at_fork_handler_installed = False
 
 _NameValueMapping = Mapping[str, Union[str, int, list[int], float, bool]]
 
+def _make_transfer_server_factory(
+) -> xla_client._xla.TransferServerInterfaceFactory | None:
+  """Creates a transfer server interface factory."""
+  if (not CROSS_HOST_TRANSFER_SOCKET_ADDRESS.value or not
+      hasattr(_jax, "make_transfer_server_interface_factory")):
+    return None
+  transport_addresses = []
+  if CROSS_HOST_TRANSPORT_ADDRESSES.value:
+    transport_addresses = CROSS_HOST_TRANSPORT_ADDRESSES.value.split(",")
+  transfer_server_kwargs = {
+      "distributed_client": distributed.global_state.client,
+      "socket_address": CROSS_HOST_TRANSFER_SOCKET_ADDRESS.value,
+      "transport_addresses": transport_addresses,
+  }
+  if CROSS_HOST_TRANSFER_TIMEOUT_SECONDS.value is not None:
+    transfer_server_kwargs["cross_host_transfer_timeout_seconds"] = (
+        CROSS_HOST_TRANSFER_TIMEOUT_SECONDS.value)
+  if CROSS_HOST_TRANSFER_TRANSFER_SIZE.value is not None:
+    transfer_server_kwargs["transfer_size"] = (
+        CROSS_HOST_TRANSFER_TRANSFER_SIZE.value)
+  return _jax.make_transfer_server_interface_factory(**transfer_server_kwargs)  # type: ignore
+
+
 def make_tpu_client(
     library_path: str | None = None, options: _NameValueMapping | None = None
 ):
@@ -171,25 +189,11 @@ def make_tpu_client(
     _jax.initialize_pjrt_plugin('tpu')
   if options is None:
     options = {}
-  if jaxlib_extension_version < 363:
-    return _jax.get_c_api_client("tpu", options)
-  transfer_server_factory = None
-  if (CROSS_HOST_TRANSFER_SOCKET_ADDRESS.value and
-      hasattr(_jax, "make_transfer_server_interface_factory")):
-    transport_addresses = []
-    if CROSS_HOST_TRANSPORT_ADDRESSES.value:
-      transport_addresses = CROSS_HOST_TRANSPORT_ADDRESSES.value.split(",")
-    transfer_server_factory = _jax.make_transfer_server_interface_factory(
-        distributed_client=distributed.global_state.client,
-        socket_address=CROSS_HOST_TRANSFER_SOCKET_ADDRESS.value,
-        transport_addresses=transport_addresses,
-        transfer_size=CROSS_HOST_TRANSFER_TRANSFER_SIZE.value,
-    )
   return _jax.get_c_api_client(
       "tpu",
       options,
       distributed.global_state.client,
-      transfer_server_factory,
+      _make_transfer_server_factory(),
   )
 
 
@@ -308,14 +312,6 @@ def make_cpu_client(
   # https://github.com/jax-ml/jax/pull/26172 goes in.
   if collectives is None and distributed.global_state.client is not None:
     collectives_impl = config.cpu_collectives_implementation.value
-    if _CPU_ENABLE_GLOO_COLLECTIVES.value:
-      collectives_impl = 'gloo'
-      warnings.warn('Setting `jax_cpu_enable_gloo_collectives` is '
-                      'deprecated. Please use `jax.config.update('
-                      '"jax_cpu_collectives_implementation", "gloo")` instead.',
-                      DeprecationWarning,
-                      )
-
     if collectives_impl == 'gloo':
       collectives = xla_client._xla.make_gloo_tcp_collectives(
         distributed_client=distributed.global_state.client,
@@ -329,29 +325,6 @@ def make_cpu_client(
       assert collectives_impl is None
 
   num_devices = num_cpu_devices.value if num_cpu_devices.value >= 0 else None
-  if jaxlib_extension_version < 363:
-    return xla_client.make_cpu_client(
-        asynchronous=_CPU_ENABLE_ASYNC_DISPATCH.value,
-        distributed_client=distributed.global_state.client,
-        node_id=distributed.global_state.process_id,
-        num_nodes=distributed.global_state.num_processes,
-        collectives=collectives,
-        num_devices=num_devices,
-        get_local_topology_timeout_minutes=cpu_get_local_topology_timeout_minutes.value,
-        get_global_topology_timeout_minutes=cpu_get_global_topology_timeout_minutes.value,
-    )
-  transfer_server_factory = None
-  if (CROSS_HOST_TRANSFER_SOCKET_ADDRESS.value and
-      hasattr(_jax, "make_transfer_server_interface_factory")):
-    transport_addresses = []
-    if CROSS_HOST_TRANSPORT_ADDRESSES.value:
-      transport_addresses = CROSS_HOST_TRANSPORT_ADDRESSES.value.split(",")
-    transfer_server_factory = _jax.make_transfer_server_interface_factory(
-        socket_address=CROSS_HOST_TRANSFER_SOCKET_ADDRESS.value,
-        transport_addresses=transport_addresses,
-        distributed_client=distributed.global_state.client,
-        transfer_size=CROSS_HOST_TRANSFER_TRANSFER_SIZE.value,
-    )
   return xla_client.make_cpu_client(
       asynchronous=_CPU_ENABLE_ASYNC_DISPATCH.value,
       distributed_client=distributed.global_state.client,
@@ -361,7 +334,7 @@ def make_cpu_client(
       num_devices=num_devices,
       get_local_topology_timeout_minutes=cpu_get_local_topology_timeout_minutes.value,
       get_global_topology_timeout_minutes=cpu_get_global_topology_timeout_minutes.value,
-      transfer_server_factory=transfer_server_factory,
+      transfer_server_factory=_make_transfer_server_factory(),
   )
 
 
@@ -542,8 +515,43 @@ def _options_from_jax_configs(plugin_name):
 OptionsDict = Mapping[str, str | int | list[int] | float | bool]
 
 
-# TODO(b/261345120): decide on a public name and expose a public method which is
-# an alias of this method.
+def make_pjrt_c_api_client(
+    plugin_name: str,
+    options: OptionsDict | Callable[[], OptionsDict] | None = None,
+) -> xla_client.Client:
+  """Creates a PjRt client for the given plugin.
+
+  Args:
+    plugin_name: the name of the plugin.
+    options: Optional. It is used when creating a PJRT plugin client. Can be a
+      callable, in which case it will be invoked upon plugin initialization
+      time, and will be expected to return an option dictionary.
+  """
+  if not xla_client.pjrt_plugin_initialized(plugin_name):
+    xla_client.initialize_pjrt_plugin(plugin_name)
+  updated_options: dict[str, Any] = {}
+  if options is not None:
+    updated_options.update(options() if callable(options) else options)
+  updated_options.update(_options_from_jax_configs(plugin_name))
+  if distributed.global_state.client is None:
+    return xla_client.make_c_api_client(plugin_name, updated_options, None)
+
+  distribute_options = {
+      'node_id': distributed.global_state.process_id,
+      'num_nodes': distributed.global_state.num_processes,
+  }
+  if (partition_index := distributed.global_state.partition_index) is not None:
+    distribute_options['partition_index'] = partition_index
+  if options is not None:
+    distribute_options.update(updated_options)
+  return xla_client.make_c_api_client(
+      plugin_name,
+      distribute_options,
+      distributed.global_state.client,
+      _make_transfer_server_factory(),
+  )
+
+
 def register_plugin(
     plugin_name: str,
     *,
@@ -551,6 +559,8 @@ def register_plugin(
     library_path: str | None = None,
     options: OptionsDict | Callable[[], OptionsDict] | None = None,
     c_api: Any | None = None,
+    factory: BackendFactory | None = None,
+    make_topology: TopologyFactory | None = None,
 ) -> Any:
   """Registers a backend factory for the PJRT plugin.
 
@@ -564,48 +574,9 @@ def register_plugin(
       callable, in which case it will be invoked upon plugin initialization
       time, and will be expected to return an option dictionary.
     c_api: Optional. The plugin can provide a PJRT C API to be registered.
+    factory: Optional. A factory function that creates a PJRT client. If not
+      provided, a default factory will be used.
   """
-  def factory():
-    if not xla_client.pjrt_plugin_initialized(plugin_name):
-      xla_client.initialize_pjrt_plugin(plugin_name)
-    updated_options = {}
-    if options is not None:
-      updated_options.update(options() if callable(options) else options)
-    updated_options.update(_options_from_jax_configs(plugin_name))
-    if distributed.global_state.client is None:
-      return xla_client.make_c_api_client(plugin_name, updated_options, None)
-
-    distribute_options = {
-        'node_id': distributed.global_state.process_id,
-        'num_nodes': distributed.global_state.num_processes,
-    }
-    if (slice_index := distributed.global_state.slice_index) is not None:
-      distribute_options['slice_index'] = slice_index
-    if options is not None:
-      distribute_options.update(updated_options)
-    if jaxlib_extension_version < 363:
-      return xla_client.make_c_api_client(
-          plugin_name, distribute_options, distributed.global_state.client)
-    cross_host_transfer_server_factory = None
-    if (CROSS_HOST_TRANSFER_SOCKET_ADDRESS.value and
-        hasattr(_jax, "make_transfer_server_interface_factory")):
-      transport_addresses = []
-      if CROSS_HOST_TRANSPORT_ADDRESSES.value:
-        transport_addresses = CROSS_HOST_TRANSPORT_ADDRESSES.value.split(",")
-      cross_host_transfer_server_factory = (
-          _jax.make_transfer_server_interface_factory(
-              distributed_client=distributed.global_state.client,
-              socket_address=CROSS_HOST_TRANSFER_SOCKET_ADDRESS.value,
-              transport_addresses=transport_addresses,
-              transfer_size=CROSS_HOST_TRANSFER_TRANSFER_SIZE.value,
-          )
-      )
-    return xla_client.make_c_api_client(
-        plugin_name,
-        distribute_options,
-        distributed.global_state.client,
-        cross_host_transfer_server_factory,
-    )
 
   if library_path and c_api:
     logger.error(
@@ -622,6 +593,15 @@ def register_plugin(
     )
     return
 
+  if factory is not None and options is not None:
+    raise ValueError(
+        "Cannot provide both 'factory' and 'options' when registering PJRT"
+        " plugin. When providing a custom factory, the factory's must handle"
+        " its own options."
+    )
+  if factory is None:
+    factory = partial(make_pjrt_c_api_client, plugin_name, options=options)
+
   logger.debug(
       'registering PJRT plugin %s from %s', plugin_name, library_path
   )
@@ -632,7 +612,7 @@ def register_plugin(
     assert c_api is not None
     xla_client.load_pjrt_plugin_with_c_api(plugin_name, c_api)
 
-  make_topology = partial(xla_client.make_c_api_device_topology, c_api)
+  make_topology = make_topology or partial(xla_client.make_c_api_device_topology, c_api)
   experimental = plugin_name not in _nonexperimental_plugins
   register_backend_factory(plugin_name, factory, priority=priority,
                            fail_quietly=False, experimental=experimental,
@@ -1121,7 +1101,8 @@ def process_count(
     backend: str | xla_client.Client | None = None
 ) -> int:
   """Returns the number of JAX processes associated with the backend."""
-  return max(d.process_index for d in devices(backend)) + 1
+  gen = (d.process_index for d in devices(backend))
+  return max(gen, default=0) + 1
 
 
 # TODO: remove this sometime after jax 0.2.13 is released
@@ -1187,10 +1168,12 @@ def make_pjrt_tpu_topology(topology_name='', **kwargs):
       topology_name, **kwargs
   )
 
-def _validate_backend_not_initialized(new_val):
+def _validate_backend_not_initialized(name, new_val):
   if backends_are_initialized():
+    if getattr(config.config, name) == new_val:
+      return
     raise RuntimeError(
-        "jax_num_cpu_devices config should be updated before backends are"
+        f"{name} config should be updated before backends are"
         " initialized i.e. before any JAX operation is executed. You should"
         " initialize this config immediately after `import jax`.")
 
@@ -1201,7 +1184,7 @@ num_cpu_devices = config.int_state(
         "Number of CPU devices to use. If not provided, the value of "
         "the XLA flag --xla_force_host_platform_device_count is used."
         " Must be set before JAX is initialized."),
-    validator=_validate_backend_not_initialized,
+    validator=partial(_validate_backend_not_initialized, "jax_num_cpu_devices"),
 )
 
 cpu_get_local_topology_timeout_minutes = config.int_state(
@@ -1211,7 +1194,8 @@ cpu_get_local_topology_timeout_minutes = config.int_state(
         "Timeout in minutes for getting the local topology of each CPU device"
         " when building the global topology."
     ),
-    validator=_validate_backend_not_initialized,
+    validator=partial(_validate_backend_not_initialized,
+                      "jax_cpu_get_local_topology_timeout_minutes"),
 )
 
 cpu_get_global_topology_timeout_minutes = config.int_state(
@@ -1222,5 +1206,6 @@ cpu_get_global_topology_timeout_minutes = config.int_state(
         " should be strictly greater than"
         " `--jax_cpu_get_local_topology_timeout_minutes`."
     ),
-    validator=_validate_backend_not_initialized,
+    validator=partial(_validate_backend_not_initialized,
+                      "jax_cpu_get_global_topology_timeout_minutes"),
 )
