@@ -13,28 +13,34 @@
 # limitations under the License.
 # ==============================================================================
 
-from collections.abc import Sequence
+from collections.abc import Callable, Iterable, Sequence
 import contextlib
 import ctypes
 import dataclasses
 import enum
+import functools
 import hashlib
+import io
+import itertools
 import math
 import os
 import pathlib
 import time
 from typing import Any, Generic, TypeVar
-from collections.abc import Callable
 import weakref
 
-import itertools
 import jax
+from jax._src import core as jax_core
+from jax._src import dtypes
 from jax._src import lib
+from jax._src import mesh as mesh_lib
 from jax._src import sharding_impls
+from jax._src import util as jax_util
 from jax._src.interpreters import mlir
 from jax._src.lib import mosaic_gpu_dialect as dialect
 from jaxlib.mlir import ir
 from jaxlib.mlir import passmanager
+from jaxlib.mlir.dialects import _gpu_ops_gen
 from jaxlib.mlir.dialects import arith
 from jaxlib.mlir.dialects import builtin
 from jaxlib.mlir.dialects import func
@@ -44,15 +50,12 @@ from jaxlib.mlir.dialects import memref
 from jaxlib.mlir.dialects import nvvm
 import numpy as np
 
-
-# mypy: ignore-errors
-
 from . import dialect_lowering
 from . import launch_context
 from . import layout_inference
+from . import layouts
 from . import profiler
 from . import tcgen05
-from . import transform_inference
 from . import utils
 
 # MLIR can't find libdevice unless we point it to the CUDA path
@@ -61,7 +64,7 @@ os.environ["CUDA_ROOT"] = cuda_root
 PYTHON_RUNFILES = os.environ.get("PYTHON_RUNFILES")
 
 # This tracks the latest Mosaic GPU IR version with a monthly delay.
-FWD_COMPAT_IR_VERSION = 1
+FWD_COMPAT_IR_VERSION = 2
 
 c = utils.c  # This is too common to fully qualify.
 
@@ -92,8 +95,8 @@ except ImportError:
     )
     if os.path.exists(libdevice_path):
       os.environ["MOSAIC_GPU_NVSHMEM_BC_PATH"] = libdevice_path
-    for root, _, files in os.walk(os.path.join(os.getcwd(), "_solib_local")):
-      if "libnvshmem_host.so.3" in files:
+    for root, _, files in os.walk(os.getcwd()):
+      if "/_solib" in root and "libnvshmem_host.so.3" in files:
         os.environ["MOSAIC_GPU_NVSHMEM_SO_PATH"] = os.path.join(
             root, "libnvshmem_host.so.3"
         )
@@ -130,15 +133,15 @@ def supports_cross_device_collectives():
   )
 
 
-mosaic_gpu_p = jax._src.core.Primitive("mosaic_gpu_p")
+mosaic_gpu_p = jax_core.Primitive("mosaic_gpu_p")
 mosaic_gpu_p.multiple_results = True
 
 
 @mosaic_gpu_p.def_abstract_eval
 def _mosaic_gpu_abstract_eval(*_, module, out_types, inout_types):
-  del module  # Unused.
+  del module # Unused.
   return [
-      jax._src.core.ShapedArray(t.shape, t.dtype)
+      jax_core.ShapedArray(t.shape, t.dtype)
       for t in itertools.chain(out_types, inout_types)
   ]
 
@@ -152,7 +155,8 @@ def _has_communication(module, **_):
 
 
 # TODO(apaszke): Implement a proper system for managing kernel lifetimes
-KNOWN_KERNELS = {}
+# Maps kernel ID to the compiled kernel ASM.
+KNOWN_KERNELS: dict[bytes, bytes] = {}
 
 
 def _mosaic_gpu_lowering_rule(
@@ -172,15 +176,18 @@ def _mosaic_gpu_lowering_rule(
     # to physical translation, which is currently not implemented.
     if isinstance(axis_context, sharding_impls.SPMDAxisContext):
       mesh = axis_context.mesh
-      if not np.array_equal(mesh.device_ids.ravel(), np.arange(mesh.size)):
+      # Skip the check for AbstractMesh
+      if (isinstance(mesh, mesh_lib.Mesh) and
+          not np.array_equal(mesh.device_ids.ravel(), np.arange(mesh.size))):
         raise NotImplementedError(
             "Mosaic GPU only supports meshes with device ordering that follows"
-            " row-major device ids."
+            f" row-major device ids. Got: {mesh.device_ids.ravel()} device ids."
         )
     elif isinstance(axis_context, sharding_impls.ShardingContext):
       if axis_context.num_devices != 1:
         raise NotImplementedError(
             "Mosaic GPU only supports single-device meshes in ShardingContext."
+            f" Got: {axis_context.num_devices} devices."
         )
     else:
       raise NotImplementedError(f"Unsupported sharding context: {axis_context}")
@@ -203,45 +210,33 @@ def _mosaic_gpu_lowering_rule(
       serialize=True,
       ir_version=FWD_COMPAT_IR_VERSION if ctx.is_forward_compat() else None,
   )
-  module_asm = module.operation.get_asm(binary=True, enable_debug_info=True)
+  bytecode_buffer = io.BytesIO()
+  module.operation.write_bytecode(bytecode_buffer, desired_version=0)
+  module_asm = bytecode_buffer.getvalue()
   kernel_id = hashlib.sha256(module_asm).digest()
   # Note that this is technically only a half measure. Someone might load a
   # compiled module with a hash collision from disk. But that's so unlikely with
   # SHA256 that it shouldn't be a problem.
   if (kernel_text := KNOWN_KERNELS.get(kernel_id, None)) is not None:
     if kernel_text != module_asm:
-      raise RuntimeError("Hash collision!")
+      raise RuntimeError("Kernel hash collision!")
   else:
     KNOWN_KERNELS[kernel_id] = module_asm
 
-  if ctx.is_forward_compat():
-    if use_custom_barrier:
-      raise ValueError("Barrier semaphore is not supported in forward compatibility mode. "
-                       "Please, use 'export_ignore_forward_compatibility=True'.")
-    op = mlir.custom_call(
-        "mosaic_gpu",
-        result_types=[mlir.aval_to_ir_type(aval) for aval in ctx.avals_out],
-        operands=args,
-        operand_layouts=[list(reversed(range(a.ndim))) for a in ctx.avals_in],
-        result_layouts=[list(reversed(range(a.ndim))) for a in ctx.avals_out],
-        backend_config=kernel_id + module_asm,
-        operand_output_aliases=dict(input_output_aliases),
-    )
-  else:
-    op = mlir.custom_call(
-        "mosaic_gpu_v2",
-        result_types=[mlir.aval_to_ir_type(aval) for aval in ctx.avals_out],
-        operands=args,
-        operand_layouts=[list(reversed(range(a.ndim))) for a in ctx.avals_in],
-        result_layouts=[list(reversed(range(a.ndim))) for a in ctx.avals_out],
-        backend_config=dict(
-            kernel_hash=ir.StringAttr.get(kernel_id),
-            module=ir.StringAttr.get(module_asm),
-            use_custom_barrier=ir.BoolAttr.get(use_custom_barrier),
-        ),
-        operand_output_aliases=dict(input_output_aliases),
-        api_version=4,
-    )
+  op = mlir.custom_call(
+      "mosaic_gpu_v2",
+      result_types=[mlir.aval_to_ir_type(aval) for aval in ctx.avals_out],
+      operands=args,
+      operand_layouts=[list(reversed(range(a.ndim))) for a in ctx.avals_in],
+      result_layouts=[list(reversed(range(a.ndim))) for a in ctx.avals_out],
+      backend_config=dict(
+          kernel_hash=ir.StringAttr.get(kernel_id),
+          module=ir.StringAttr.get(module_asm),
+          use_custom_barrier=ir.BoolAttr.get(use_custom_barrier),
+      ),
+      operand_output_aliases=dict(input_output_aliases),
+      api_version=4,
+  )
   return op.results
 
 
@@ -279,6 +274,7 @@ class Barrier:
 @dataclasses.dataclass(frozen=True)
 class ClusterBarrier:
   collective_dims: Sequence[gpu.Dimension]
+  arrival_count: int = 1
   num_barriers: int = 1
 
 @dataclasses.dataclass(frozen=True)
@@ -300,7 +296,7 @@ class TMEM:
 
 
 def _count_buffer_bytes(shape_dtype: jax.ShapeDtypeStruct) -> int:
-  return math.prod(shape_dtype.shape) * np.dtype(shape_dtype.dtype).itemsize
+  return math.prod(shape_dtype.shape) * dtypes.itemsize_bits(dtypes.dtype(shape_dtype.dtype)) // 8
 
 
 class LoweringSemantics(enum.Enum):
@@ -330,6 +326,36 @@ class _TMEMAlloc:
     )
 
 
+@dataclasses.dataclass()
+class _TMEMDialectAlloc:
+  addr_ref: ir.Value
+  shape: tuple[int, int]
+  dtype: ir.Type
+  packing: int
+  collective: bool
+  tmem_ref: ir.Value | None = dataclasses.field(init=False, default=None)
+
+  def alloc(self) -> int:
+    """Allocates TMEM and returns the number of columns allocated."""
+    result_type = ir.MemRefType.get(
+        self.shape,
+        self.dtype,
+        memory_space=utils.tmem(),
+    )
+    self.tmem_ref = dialect.tmem_alloc(
+        result_type,
+        self.addr_ref,
+        collective=self.collective,
+        packing=self.packing,
+    )
+    ncols = self.shape[1] // self.packing
+    return tcgen05.tmem_alloc_exact_ncols(ncols, exact=False)
+
+  def dealloc(self):
+    assert self.tmem_ref is not None
+    dialect.tmem_dealloc(self.tmem_ref)
+
+
 def _slice_smem(
     result: ir.Type,
     smem_base: ir.Value,
@@ -347,7 +373,9 @@ def _construct_smem_reftree(
     cluster_shape: tuple[int, int, int],
     dynamic_smem: ir.Value,
     smem_buffers: ShapeTree,
-    tmem_allocs: list[_TMEMAlloc],  # Mutated by this function!
+    tmem_allocs: list[
+        _TMEMAlloc | _TMEMDialectAlloc
+    ],  # Mutated by this function!
     lowering_semantics: LoweringSemantics,
     dynamic_smem_offset: int = 0,
 ) -> Callable[[], RefTree]:
@@ -377,6 +405,7 @@ def _construct_smem_reftree(
         )
       dynamic_smem_offset += num_barriers * utils.MBARRIER_BYTES
       return barrier_memref
+    ref: Any
     match ref_ty:
       case Union(members):
         member_thunks = [
@@ -397,20 +426,22 @@ def _construct_smem_reftree(
           return Union([t() for t in member_thunks])
 
       case TMABarrier(num_barriers):
-        init_fn = utils.DialectBarrierRef.initialize if (
-            lowering_semantics == LoweringSemantics.Warpgroup
-        ) else utils.BarrierRef.initialize
+        init_fn: Callable[..., Any] = (
+            utils.DialectBarrierRef.initialize
+            if lowering_semantics == LoweringSemantics.Warpgroup
+            else utils.BarrierRef.initialize
+        )
         ref = init_fn(barrier_memref(num_barriers), arrival_count=1)
       case Barrier(arrival_count, num_barriers):
-        init_fn = utils.DialectBarrierRef.initialize if (
-            lowering_semantics == LoweringSemantics.Warpgroup
-        ) else utils.BarrierRef.initialize
+        init_fn = (
+            utils.DialectBarrierRef.initialize
+            if lowering_semantics == LoweringSemantics.Warpgroup
+            else utils.BarrierRef.initialize
+        )
         ref = init_fn(barrier_memref(num_barriers), arrival_count=arrival_count)
-      case ClusterBarrier(collective_dims, num_barriers):
+      case ClusterBarrier(collective_dims, arrival_count, num_barriers):
         ref = utils.CollectiveBarrierRef.initialize(
-            barrier_memref(num_barriers),
-            collective_dims,
-            cluster_shape,
+            barrier_memref(num_barriers), arrival_count, collective_dims, cluster_shape
         )
       case TMEM(shape, dtype, layout=layout, collective=collective, packing=packing):
         addr_ref = _slice_smem(
@@ -419,19 +450,33 @@ def _construct_smem_reftree(
             c(dynamic_smem_offset, index),
             lowering_semantics,
         )
-        if layout is None:
-          layout = tcgen05._infer_tmem_layout(
-              shape, collective, 1 if packing is None else packing
-          )
-        num_cols = layout.cols_in_shape(
-            shape, utils.bitwidth(utils.dtype_to_ir_type(dtype))
-        )
-        tmem_allocs.append(_TMEMAlloc(addr_ref, num_cols, collective))
-        def ref(addr_ref=addr_ref, shape=shape, dtype=dtype, layout=layout):
-          addr = memref.load(addr_ref, [])
-          return tcgen05.TMEMRef(
-              addr, shape, utils.dtype_to_ir_type(dtype), layout
+        packing = 1 if packing is None else packing
+        ir_dtype = utils.dtype_to_ir_type(dtype)
+        if lowering_semantics == LoweringSemantics.Warpgroup:
+          if layout is not None:
+            packing = layout.vector_length
+
+          alloc = _TMEMDialectAlloc(
+              addr_ref, shape, ir_dtype, packing, collective
           )
+          tmem_allocs.append(alloc)
+          def ref(alloc=alloc, layout=layout):
+            assert alloc.tmem_ref is not None
+            if layout is not None:
+              layout_attr = layouts.to_layout_attr(layout)
+              return dialect.tmem_layout_cast(alloc.tmem_ref, layout_attr)
+            else:
+              return alloc.tmem_ref
+
+        else:
+          if layout is None:
+            layout = tcgen05._infer_tmem_layout(shape, collective, packing)
+          num_cols = layout.cols_in_shape(shape, utils.bitwidth(ir_dtype))
+          tmem_allocs.append(_TMEMAlloc(addr_ref, num_cols, collective))
+          def ref(addr_ref=addr_ref, shape=shape, ir_dtype=ir_dtype, layout=layout):
+            addr = memref.load(addr_ref, [])
+            return tcgen05.TMEMRef(addr, shape, ir_dtype, layout)
+
         dynamic_smem_offset += 4  # i32 takes up 4 bytes
       case _:
         mlir_dtype = utils.dtype_to_ir_type(ref_ty.dtype)
@@ -465,11 +510,15 @@ def _smem_tree_size(smem_buffers: ShapeTree) -> int:
         size += max(_smem_tree_size(s) for s in members)
       case (
           TMABarrier(num_barriers)
-          | ClusterBarrier(_, num_barriers=num_barriers)
+          | ClusterBarrier(_, _, num_barriers=num_barriers)
           | Barrier(_, num_barriers=num_barriers)
       ):
         if size % utils.MBARRIER_BYTES:
-          raise NotImplementedError("Misaligned barrier allocation")
+          raise NotImplementedError(
+              "Misaligned barrier allocation. Expected smem size"
+              f" ({size} bytes) to be divisible by the size of the barrier:"
+              f" {utils.MBARRIER_BYTES} bytes."
+          )
         size += num_barriers * utils.MBARRIER_BYTES
       case TMEM(_):
         # TODO(justinfu): This can trigger misaligned barrier allocations
@@ -494,7 +543,10 @@ def _launch(
     maybe_prof_buffer: ir.Value | None = None,
 ):
   if (profiler_spec is None) != (maybe_prof_buffer is None):
-    raise ValueError
+    raise ValueError(
+        "Both profiler_spec and maybe_prof_buffer must be specified or"
+        " left unspecified."
+    )
   index = ir.IndexType.get()
   i32 = ir.IntegerType.get_signless(32)
   i8 = ir.IntegerType.get_signless(8)
@@ -513,12 +565,21 @@ def _launch(
     profiler_start = (smem_bytes + align - 1) & ~(align - 1)
     smem_bytes = profiler_start + profiler_spec.smem_bytes(block=block)
 
-  # TODO(cperivol): Query the shared memory size programmatically.
-  if smem_bytes > 228 * 1024:
-    raise ValueError(f"Mosaic GPU kernel exceeds available shared memory {smem_bytes=} > 228000")
+  device = jax.local_devices()[0]
+  # For ahead-of-time compilation purposes, that is when a CUDA device
+  # isn't available to query directly, we default to 227 KB, the
+  # maximum amount of shared memory per thread block available in
+  # compute capabilities 9.0 and 10.x:
+  # https://docs.nvidia.com/cuda/cuda-c-programming-guide/#features-and-technical-specifications-technical-specifications-per-compute-capability
+  # Note in either case we assume all devices have the same amount of
+  # shared memory.
+  max_smem_bytes = getattr(device, "shared_memory_per_block_optin", 227 * 1024)
+  if smem_bytes > max_smem_bytes:
+    raise ValueError("Mosaic GPU kernel exceeds available shared memory: "
+                     f"{smem_bytes=} > {max_smem_bytes=}")
   if math.prod(cluster) != 1:
     if len(cluster) != 3:
-      raise ValueError("Clusters must be 3D")
+      raise ValueError(f"Clusters must be 3D. Got: {cluster}")
     cluster_kwargs = {
         "clusterSize" + d: c(s, index) for s, d in zip(cluster, "XYZ")
     }
@@ -530,9 +591,14 @@ def _launch(
         )
   else:
     cluster_kwargs = {}
-  launch_op = gpu.LaunchOp(
-      token.type, [token], *grid_vals, *block_vals,
-      dynamicSharedMemorySize=c(smem_bytes, i32), **cluster_kwargs)
+  launch_op = _gpu_ops_gen.LaunchOp(
+      token.type,
+      [token],
+      *grid_vals,
+      *block_vals,
+      dynamicSharedMemorySize=c(smem_bytes, i32),
+      **cluster_kwargs,
+  )
   launch_op.body.blocks.append(*([index] * (12 + 2 * len(cluster_kwargs))))  # Append an empty block
   with ir.InsertionPoint(launch_op.body.blocks[0]):
     dynamic_smem = gpu.dynamic_shared_memory(
@@ -550,8 +616,16 @@ def _launch(
           c(profiler_start, index),
           lowering_semantics,
       )
+      if lowering_semantics == LoweringSemantics.Warpgroup:
+        prof_smem = dialect.with_transforms(prof_smem, ir.ArrayAttr.get([]))
+        wrap_in_custom_primitive = True
+      else:
+        wrap_in_custom_primitive = False
       prof = profiler.OnDeviceProfiler(
-          profiler_spec, prof_smem, maybe_prof_buffer
+          profiler_spec,
+          prof_smem,
+          maybe_prof_buffer,
+          wrap_in_custom_primitive,
       )
     else:
       prof = None
@@ -560,7 +634,7 @@ def _launch(
         module, launch_context.Scratch(launch_op), cluster, prof
     )
     with ctx.named_region("Init"):
-      tmem_allocs: list[_TMEMAlloc] = []
+      tmem_allocs: list[_TMEMAlloc | _TMEMDialectAlloc] = []
       smem_ref_tree_thunk = _construct_smem_reftree(
           cluster, dynamic_smem, smem_buffers, tmem_allocs, lowering_semantics
       )
@@ -571,9 +645,14 @@ def _launch(
         nvvm.cluster_arrive_relaxed(aligned=ir.UnitAttr.get())
         nvvm.cluster_wait(aligned=ir.UnitAttr.get())
       if tmem_allocs:
-        eq = arith.CmpIPredicate.eq
-        is_init_warp = arith.cmpi(eq, utils.warp_idx(sync=False), c(0, i32))
-        with utils.when(is_init_warp):
+        init_warp_ctx: contextlib.AbstractContextManager
+        if lowering_semantics == LoweringSemantics.Warpgroup:
+          init_warp_ctx = contextlib.nullcontext()
+        else:
+          eq = arith.CmpIPredicate.eq
+          is_init_warp = arith.cmpi(eq, utils.warp_idx(sync=False), c(0, i32))
+          init_warp_ctx = utils.when(is_init_warp)
+        with init_warp_ctx:
           cols_used = 0
           for alloc in tmem_allocs:
             cols_used += alloc.alloc()
@@ -583,10 +662,23 @@ def _launch(
                 f"Requested {cols_used} columns which exceeds limit of "
                 f"{tcgen05.TMEM_MAX_COLS}."
             )
-          if any(alloc.collective for alloc in tmem_allocs):
-            tcgen05.tmem_relinquish_alloc_permit(collective=True)
-          if any(not alloc.collective for alloc in tmem_allocs):
-            tcgen05.tmem_relinquish_alloc_permit(collective=False)
+          collective_types = {alloc.collective for alloc in tmem_allocs}
+          if len(collective_types) > 1:
+            raise ValueError(
+                "Can't mix collective and non-collective TMEM allocations"
+                " within the same kernel."
+            )
+          collective = True in collective_types
+          if collective and math.prod(cluster) % 2:
+            raise ValueError(
+                "Collective TMEM allocations are only supported for clusters"
+                " with an even number of blocks in them. Got cluster:"
+                f" {cluster}"
+            )
+          if lowering_semantics == LoweringSemantics.Warpgroup:
+            dialect.tmem_relinquish_alloc_permit(collective=collective)
+          else:
+            tcgen05.tmem_relinquish_alloc_permit(collective=collective)
       gpu.barrier()  # Make sure the init is visible to all threads.
       smem_ref_tree = smem_ref_tree_thunk()
 
@@ -597,7 +689,11 @@ def _launch(
       if any(alloc.collective for alloc in tmem_allocs):
         nvvm.cluster_arrive_relaxed(aligned=ir.UnitAttr.get())
         nvvm.cluster_wait(aligned=ir.UnitAttr.get())
-      with utils.when(is_init_warp):
+      if lowering_semantics == LoweringSemantics.Warpgroup:
+        init_warp_ctx = contextlib.nullcontext()
+      else:
+        init_warp_ctx = utils.when(is_init_warp)
+      with init_warp_ctx:
         for alloc in tmem_allocs:
           alloc.dealloc()
     if prof is not None:
@@ -616,7 +712,7 @@ def _lower_as_gpu_kernel(
     smem_scratch_shape: ShapeTree | Union[ShapeTree],
     lowering_semantics: LoweringSemantics,
     module_name: str,
-    kernel_name: str | None = None,
+    kernel_name: str,
     prof_spec: profiler.ProfilerSpec | None = None,
 ):
   ptr_ty = ir.Type.parse("!llvm.ptr")
@@ -644,8 +740,6 @@ def _lower_as_gpu_kernel(
   dialect.register_dialect(module.context)
   attrs = module.operation.attributes
   attrs["sym_name"] = ir.StringAttr.get(module_name)
-  if kernel_name is None:
-    kernel_name = module_name
 
   # These are needed as nonlocal below.
   launch_ctx = None
@@ -719,7 +813,7 @@ def _declare_runtime_functions():
   )
 
 
-def as_gpu_kernel(
+def _kernel_to_module(
     body,
     grid: tuple[int, int, int],
     block: tuple[int, int, int],
@@ -730,7 +824,6 @@ def as_gpu_kernel(
     cluster: tuple[int, int, int] = (1, 1, 1),
     module_name: str = "unknown",
     kernel_name: str | None = None,
-    ir_version: int | None = None,
     thread_semantics: LoweringSemantics = LoweringSemantics.Lane,
     inout_shape = (),
 ):
@@ -742,6 +835,8 @@ def as_gpu_kernel(
     inout_shape = tuple(inout_shape)
   elif not isinstance(inout_shape, tuple):
     inout_shape = (inout_shape,)
+  if kernel_name is None:
+    kernel_name = jax_util.fun_name(body, "anonymous")
 
   inout_shape = jax.tree.map(lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype),
                              inout_shape)
@@ -764,13 +859,42 @@ def as_gpu_kernel(
     # Run Python lowering passes. The remaining passes will be run in C++ in
     # jax/jaxlib/mosaic/gpu/custom_call.cc
     layout_inference.infer_layout(module)  # pytype: disable=attribute-error
-    transform_inference.infer_transforms(module)  # pytype: disable=attribute-error
     dialect_lowering.lower_mgpu_dialect(module, launch_ctx)  # pytype: disable=attribute-error
 
   launch_ctx.scratch.finalize_size()
   module.operation.verify()
 
-  if launch_ctx.is_device_collective and not supports_cross_device_collectives():
+  return (
+      module,
+      in_shape,
+      inout_shape,
+      out_shape,
+      unwrap_output_tuple,
+      launch_ctx.is_device_collective,
+  )
+
+
+def as_gpu_kernel(
+    body,
+    grid: tuple[int, int, int],
+    block: tuple[int, int, int],
+    in_shape,
+    out_shape,
+    smem_scratch_shape: ShapeTree | Union[ShapeTree],
+    prof_spec: profiler.ProfilerSpec | None = None,
+    cluster: tuple[int, int, int] = (1, 1, 1),
+    module_name: str = "unknown",
+    kernel_name: str | None = None,
+    ir_version: int | None = None,
+    thread_semantics: LoweringSemantics = LoweringSemantics.Lane,
+    inout_shape = (),
+):
+  module, in_shape, inout_shape, out_shape, unwrap_output_tuple, is_device_collective = _kernel_to_module(
+      body, grid, block, in_shape, out_shape, smem_scratch_shape, prof_spec,
+      cluster, module_name, kernel_name, thread_semantics, inout_shape
+  )
+
+  if is_device_collective and not supports_cross_device_collectives():
     raise RuntimeError("Kernel is a cross-device collective but no support is available.")
 
   expected_arg_tys, expected_arg_treedef = jax.tree.flatten((*in_shape, *inout_shape))
@@ -797,7 +921,12 @@ def as_gpu_kernel(
         )
 
   def bind(*args) -> Any:
-    return mosaic_gpu_p.bind(*args, module=module, out_types=out_shape, inout_types=inout_shape)
+    return mosaic_gpu_p.bind(
+        *args,
+        module=module,
+        out_types=out_shape,
+        inout_types=inout_shape,
+    )
 
   if prof_spec is not None:
     @jax.jit
@@ -805,10 +934,7 @@ def as_gpu_kernel(
       _check_args(*args)
       *results, prof_buffer = bind(*args)
       def dump_profile(prof_buffer):
-        out_file = os.path.join(
-            os.getenv("TEST_UNDECLARED_OUTPUTS_DIR", "/tmp"),
-            f"{time.time_ns()}-trace.json",
-        )
+        out_file = os.path.join(prof_spec.dump_path, f"{time.time_ns()}-trace.json")
         try:
           with open(out_file, "x") as f:
             prof_spec.dump(prof_buffer, f, grid=grid, block=block)
@@ -837,61 +963,58 @@ def as_torch_gpu_kernel(
     cluster: tuple[int, int, int] = (1, 1, 1),
     module_name: str = "unknown",
     kernel_name: str | None = None,
-    lowering_semantics: LoweringSemantics = LoweringSemantics.Lane,
+    thread_semantics: LoweringSemantics = LoweringSemantics.Lane,
+    inout_shape=(),
 ):
-  try:
-    import torch  # pytype: disable=import-error
-  except ImportError:
-    raise RuntimeError("as_torch_gpu_kernel requires PyTorch")
-  torch.cuda.init()  # Make sure CUDA context is set up.
-
-  if isinstance(in_shape, list):
-    in_shape = tuple(in_shape)
-  elif not isinstance(in_shape, tuple):
-    in_shape = (in_shape,)
-
-  # TODO(slebedev): Make this a parameter.
-  inout_shape = ()
-
-  flat_out_types, out_treedef = jax.tree.flatten(out_shape)
-  expected_arg_treedef = jax.tree.structure(in_shape)
-
-  module, out_shape, unwrap_output_tuple, launch_ctx = (
-      _lower_as_gpu_kernel(
-          body, grid, cluster, block, in_shape, out_shape, inout_shape,
-          smem_scratch_shape, lowering_semantics, module_name, kernel_name,
-          prof_spec
-      )
+  (
+      module,
+      in_shape,
+      inout_shape,
+      out_shape,
+      unwrap_output_tuple,
+      is_device_collective,
+  ) = _kernel_to_module(
+      body,
+      grid,
+      block,
+      in_shape,
+      out_shape,
+      smem_scratch_shape,
+      prof_spec,
+      cluster,
+      module_name,
+      kernel_name,
+      thread_semantics,
+      inout_shape,
+  )
+  module = _run_serde_pass(module, serialize=True, ir_version=None)
+  return _as_torch_gpu_kernel(
+      module.operation.get_asm(binary=True, enable_debug_info=True),
+      in_shape,
+      out_shape,
+      inout_shape,
+      unwrap_output_tuple=unwrap_output_tuple,
   )
 
-  if lowering_semantics == LoweringSemantics.Warpgroup and dialect is not None:
-    # We need to run a pass that removes dead-code for which layout inference
-    # does not work.
-    pm = mlir.passmanager.PassManager.parse("builtin.module(canonicalize)", module.context)
-    pm.run(module.operation)
 
-    # Run Python lowering passes. The remaining passes will be run in C++ in
-    # jax/jaxlib/mosaic/gpu/custom_call.cc
-    layout_inference.infer_layout(module)  # pytype: disable=attribute-error
-    transform_inference.infer_transforms(module)  # pytype: disable=attribute-error
-    dialect_lowering.lower_mgpu_dialect(module, launch_ctx)  # pytype: disable=attribute-error
+def _compile_as_torch_gpu_kernel(module_asm: bytes):
+  try:
+    import torch  # type: ignore[import-not-found]  # pytype: disable=import-error
+  except ImportError:
+    raise RuntimeError("Can't compile for PyTorch: import torch failed") from None
 
-  launch_ctx.scratch.finalize_size()
-  module.operation.verify()
-
-  if launch_ctx.is_device_collective:
-    raise RuntimeError("Kernel is a cross-device collective but no support is available.")
+  torch.cuda.init()  # Make sure CUDA context is set up.
 
   # Get our hands on the compilation and unload functions
   try:
     try:
-      import jax_plugins.xla_cuda13 as cuda_plugin  # pytype: disable=import-error
+      import jax_plugins.xla_cuda13 as cuda_plugin  # type: ignore[import-not-found]  # pytype: disable=import-error
     except ImportError:
-      import jax_plugins.xla_cuda12 as cuda_plugin  # pytype: disable=import-error
+      import jax_plugins.xla_cuda12 as cuda_plugin  # type: ignore[import-not-found]  # pytype: disable=import-error
   except ImportError:
-    raise RuntimeError("as_torch_gpu_kernel only works with recent jaxlib builds "
-                       "that use backend plugins")
-  dll = ctypes.CDLL(cuda_plugin._get_library_path())
+    dll = ctypes.CDLL(None)
+  else:
+    dll = ctypes.CDLL(cuda_plugin._get_library_path())
   compile_func = dll.MosaicGpuCompile
   compile_func.argtypes = [ctypes.c_void_p]
   compile_func.restype = ctypes.POINTER(ctypes.c_void_p)
@@ -899,13 +1022,47 @@ def as_torch_gpu_kernel(
   unload_func.argtypes = [compile_func.restype]
   unload_func.restype = None
 
-  module_asm = module.operation.get_asm(binary=True, enable_debug_info=True)
-  compiled = compile_func(ctypes.c_char_p(module_asm))
-  if compiled is None:
+  compiled = compile_func(ctypes.c_char_p(module_asm), ctypes.c_int(len(module_asm)))
+  if not compiled:
     raise RuntimeError("Failed to compile the module")
   ctx, launch_ptr = compiled[0], compiled[1]
   ctx_ptr_ptr = ctypes.pointer(ctypes.c_void_p(ctx))
-  launch = ctypes.CFUNCTYPE(None, ctypes.c_void_p)(launch_ptr)
+  launch_c = ctypes.CFUNCTYPE(None, ctypes.c_void_p)(launch_ptr)
+
+  def launch(arg_ptrs, device):
+    # Allocate another buffer for args of the host-side program. This is sadly
+    # the default MLIR calling convention.
+    launch_args_ptr = (ctypes.POINTER(ctypes.c_void_p) * 3)()
+    launch_args_ptr[0] = ctx_ptr_ptr
+    launch_args_ptr[1] = ctypes.pointer(
+        torch.cuda.default_stream(device)._as_parameter_
+    )
+    launch_args_ptr[2] = ctypes.cast(
+        ctypes.pointer(ctypes.pointer(arg_ptrs)),
+        ctypes.POINTER(ctypes.c_void_p),
+    )
+    launch_c(launch_args_ptr)
+
+  return launch, functools.partial(unload_func, compiled)
+
+
+def _as_torch_gpu_kernel(
+    module_asm: bytes,
+    in_shape: Iterable[object],
+    out_shape: Iterable[object],
+    inout_shape: Iterable[object] = (),
+    *,
+    unwrap_output_tuple: bool = False,
+    _prepare_args = None,
+    _prepare_results = None,
+):
+  flat_arg_types, expected_arg_treedef = jax.tree.flatten((*in_shape, *inout_shape))
+  flat_out_types, _ = jax.tree.flatten(out_shape)
+  out_treedef = jax.tree.structure((*out_shape, *inout_shape))
+
+  launch, unload = _compile_as_torch_gpu_kernel(module_asm)
+  # _compile_as_torch_gpu_kernel checks that this succeeds
+  import torch  # type: ignore[import-not-found]  # pytype: disable=import-error
 
   def as_torch_dtype(dtype):
     # torch contains NumPy-compatible dtypes in its top namespace
@@ -918,6 +1075,17 @@ def as_torch_gpu_kernel(
           f"Invalid argument structure: expected {expected_arg_treedef}, got"
           f" {arg_treedef}, ({args=})"
       )
+    for arg, expected_ty in zip(flat_args, flat_arg_types):
+      if arg.shape != expected_ty.shape:
+        raise ValueError(
+            f"Argument shape mismatch: expected {expected_ty.shape}, got"
+            f" {arg.shape}"
+        )
+      if arg.dtype != as_torch_dtype(expected_ty.dtype):
+        raise ValueError(
+            "Argument dtype mismatch: expected"
+            f" {as_torch_dtype(expected_ty.dtype)}, got {arg.dtype}"
+        )
 
     # Construct a device pointer list like in the XLA calling convention
     buffers = (ctypes.c_void_p * (arg_treedef.num_leaves + out_treedef.num_leaves))()
@@ -931,19 +1099,13 @@ def as_torch_gpu_kernel(
       out = torch.empty(t.shape, dtype=as_torch_dtype(t.dtype), device=device)
       flat_outs.append(out)
       buffers[i] = out.data_ptr()
-    # Allocate another buffer for args of the host-side program. This is sadly
-    # the default MLIR calling convention.
-    args_ptr = (ctypes.POINTER(ctypes.c_void_p) * 3)()
-    args_ptr[0] = ctx_ptr_ptr
-    args_ptr[1] = ctypes.pointer(torch.cuda.default_stream(device)._as_parameter_)
-    args_ptr[2] = ctypes.cast(ctypes.pointer(ctypes.pointer(buffers)),
-                              ctypes.POINTER(ctypes.c_void_p))
-    launch(args_ptr)
-    return jax.tree.unflatten(out_treedef, flat_outs)
+    if num_inout_args := jax.tree.structure(inout_shape).num_leaves:
+      flat_outs += flat_args[-num_inout_args:]
+    launch(buffers, device)
+    out = jax.tree.unflatten(out_treedef, flat_outs)
+    return out[0] if unwrap_output_tuple else out
 
   # Unload the compiled code when the Python function is destroyed.
-  def unload(_):
-    unload_func(compiled)
-  apply.destructor = weakref.ref(apply, unload)
+  apply.destructor = weakref.ref(apply, lambda _weak_ref: unload)
 
   return apply
