@@ -21,7 +21,6 @@ import functools
 import jax
 from jax import lax
 import jax.numpy as jnp
-from jax._src.lax.control_flow.for_loop import for_loop
 
 from jax.experimental import pallas as pl
 from jax.experimental.pallas import triton as plgpu
@@ -32,40 +31,55 @@ def layer_norm_forward_kernel(
     *, eps: float, block_size: int):
   n_col = x_ref.shape[0]
 
-  def mean_body(i, acc_ref):
+  def mean_body(i, acc):
     col_idx = i * block_size + jnp.arange(block_size)
     mask = col_idx < n_col
-    a = pl.load(x_ref, (col_idx,), mask=mask, other=0.,
-                eviction_policy="evict_last").astype(jnp.float32)
-    acc_ref[:] += a
-  mean = for_loop(pl.cdiv(n_col, block_size), mean_body,
-                  jnp.zeros(block_size)).sum() / n_col
+    a = plgpu.load(
+        x_ref.at[col_idx], mask=mask, other=0.0, eviction_policy="evict_last"
+    ).astype(jnp.float32)
+    return acc + a
 
-  def var_body(i, acc_ref):
+  mean = lax.fori_loop(
+      0,
+      pl.cdiv(n_col, block_size),
+      mean_body,
+      init_val=jnp.zeros(block_size),
+  ).sum()
+  mean /= n_col
+
+  def var_body(i, acc):
     col_idx = i * block_size + jnp.arange(block_size)
     mask = col_idx < n_col
-    a = pl.load(x_ref, (col_idx,), mask=mask, other=0.,
-                eviction_policy="evict_last").astype(jnp.float32)
+    a = plgpu.load(
+        x_ref.at[col_idx], mask=mask, other=0.0, eviction_policy="evict_last"
+    ).astype(jnp.float32)
     a = jnp.where(mask, a - mean, 0.)
-    acc_ref[:] += a * a
-  var = for_loop(pl.cdiv(n_col, block_size), var_body,
-                 jnp.zeros(block_size)).sum() / n_col
+    return acc + a * a
+
+  var = lax.fori_loop(
+      0,
+      pl.cdiv(n_col, block_size),
+      var_body,
+      init_val=jnp.zeros(block_size),
+  ).sum()
+  var /= n_col
   rstd = 1 / jnp.sqrt(var + eps)
   if mean_ref is not None:
     mean_ref[...] = mean.astype(mean_ref.dtype)
   if rstd_ref is not None:
     rstd_ref[...] = rstd.astype(rstd_ref.dtype)
 
-  def body(i, _):
+  @pl.loop(0, pl.cdiv(n_col, block_size))
+  def body(i):
     col_idx = i * block_size + jnp.arange(block_size)
     mask = col_idx < n_col
-    weight = pl.load(weight_ref, (col_idx,), mask=mask)
-    bias = pl.load(bias_ref, (col_idx,), mask=mask)
-    x = pl.load(x_ref, (col_idx,), mask=mask, other=0.,
-                eviction_policy="evict_first").astype(jnp.float32)
+    weight = plgpu.load(weight_ref.at[col_idx], mask=mask)
+    bias = plgpu.load(bias_ref.at[col_idx], mask=mask)
+    x = plgpu.load(
+        x_ref.at[col_idx], mask=mask, other=0.0, eviction_policy="evict_first"
+    ).astype(jnp.float32)
     out = (x - mean) * rstd * weight + bias
-    pl.store(o_ref, (col_idx,), out.astype(o_ref.dtype), mask=mask)
-  for_loop(pl.cdiv(n_col, block_size), body, ())
+    plgpu.store(o_ref.at[col_idx], out.astype(o_ref.dtype), mask=mask)
 
 
 def layer_norm_forward(
@@ -116,40 +130,54 @@ def layer_norm_backward_kernel_dx(
     *, eps: float, block_size: int):
   n_col = x_ref.shape[0]
 
-  def mean_body(i, acc_ref):
+  def mean_body(i, acc):
     col_idx = i * block_size + jnp.arange(block_size)
     mask = col_idx < n_col
-    a = pl.load(x_ref, (col_idx,), mask=mask, other=0.,
-                eviction_policy="evict_last").astype(jnp.float32)
-    dout = pl.load(do_ref, (col_idx,), mask=mask, other=0.,
-                eviction_policy="evict_last").astype(jnp.float32)
-    weight = pl.load(weight_ref, (col_idx,), mask=mask, other=0.,
-                eviction_policy="evict_last").astype(jnp.float32)
+    a = plgpu.load(
+        x_ref.at[col_idx], mask=mask, other=0.0, eviction_policy="evict_last"
+    ).astype(jnp.float32)
+    dout = plgpu.load(
+        do_ref.at[col_idx], mask=mask, other=0.0, eviction_policy="evict_last"
+    ).astype(jnp.float32)
+    weight = plgpu.load(
+        weight_ref.at[col_idx],
+        mask=mask,
+        other=0.0,
+        eviction_policy="evict_last",
+    ).astype(jnp.float32)
     a_hat = (a - mean_ref[...]) * rstd_ref[...]
     wdout = weight * dout
-    mean1_acc_ref, mean2_acc_ref = acc_ref
-    mean1_acc_ref[:] += a_hat * wdout
-    mean2_acc_ref[:] += wdout
-  mean = for_loop(pl.cdiv(n_col, block_size), mean_body,
-                  (jnp.zeros(block_size), jnp.zeros(block_size)))
-  mean1, mean2 = mean
+    mean1_acc, mean2_acc = acc
+    return mean1_acc + a_hat * wdout, mean2_acc + wdout
+  mean1, mean2 = lax.fori_loop(
+      0,
+      pl.cdiv(n_col, block_size),
+      mean_body,
+      init_val=(jnp.zeros(block_size), jnp.zeros(block_size)),
+  )
   mean1 = mean1.sum() / n_col
   mean2 = mean2.sum() / n_col
 
-  def dx_body(i, acc_ref):
+  @pl.loop(0, pl.cdiv(n_col, block_size))
+  def dx_body(i):
     col_idx = i * block_size + jnp.arange(block_size)
     mask = col_idx < n_col
-    a = pl.load(x_ref, (col_idx,), mask=mask, other=0.,
-                eviction_policy="evict_last").astype(jnp.float32)
-    dout = pl.load(do_ref, (col_idx,), mask=mask, other=0.,
-                eviction_policy="evict_last").astype(jnp.float32)
-    weight = pl.load(weight_ref, (col_idx,), mask=mask, other=0.,
-                eviction_policy="evict_last").astype(jnp.float32)
+    a = plgpu.load(
+        x_ref.at[col_idx], mask=mask, other=0.0, eviction_policy="evict_last"
+    ).astype(jnp.float32)
+    dout = plgpu.load(
+        do_ref.at[col_idx], mask=mask, other=0.0, eviction_policy="evict_last"
+    ).astype(jnp.float32)
+    weight = plgpu.load(
+        weight_ref.at[col_idx],
+        mask=mask,
+        other=0.0,
+        eviction_policy="evict_last",
+    ).astype(jnp.float32)
     a_hat = (a - mean_ref[...]) * rstd_ref[...]
     wdout = weight * dout
     da = (wdout - (a_hat * mean1 + mean2)) * rstd_ref[...]
-    pl.store(dx_ref, (col_idx,), da.astype(dx_ref.dtype), mask=mask)
-  for_loop(pl.cdiv(n_col, block_size), dx_body, ())
+    plgpu.store(dx_ref.at[col_idx], da.astype(dx_ref.dtype), mask=mask)
 
 
 def layer_norm_backward_kernel_dw_db(
@@ -164,25 +192,36 @@ def layer_norm_backward_kernel_dw_db(
   col_idx = j * block_n + jnp.arange(block_n)
   col_mask = col_idx < n_col
 
-  def body(i, acc_ref):
+  def body(i, acc):
     row_idx = i * block_m + jnp.arange(block_m)
     row_mask = row_idx < m
     mask = row_mask[:, None] & col_mask[None, :]
-    a = pl.load(
-        x_ref, (row_idx[:, None], col_idx[None]), mask=mask, other=0.0
+    a = plgpu.load(
+        x_ref.at[row_idx[:, None], col_idx[None]], mask=mask, other=0.0
     ).astype(jnp.float32)
-    dout = pl.load(
-        do_ref, (row_idx[:, None], col_idx[None]), mask=mask, other=0.0
+    dout = plgpu.load(
+        do_ref.at[row_idx[:, None], col_idx[None]], mask=mask, other=0.0
     ).astype(jnp.float32)
-    mean = pl.load(mean_ref, (row_idx,), mask=row_mask, other=0.).astype(jnp.float32)
-    rstd = pl.load(rstd_ref, (row_idx,), mask=row_mask, other=0.).astype(jnp.float32)
+    mean = plgpu.load(mean_ref.at[row_idx], mask=row_mask, other=0.0).astype(
+        jnp.float32
+    )
+    rstd = plgpu.load(rstd_ref.at[row_idx], mask=row_mask, other=0.0).astype(
+        jnp.float32
+    )
     a_hat = (a - mean[:, None]) * rstd[:, None]
-    dw_acc_ref, db_acc_ref = acc_ref
-    dw_acc_ref[:] += (dout * a_hat).sum(axis=0)
-    db_acc_ref[:] += dout.sum(axis=0)
-  dw_acc, db_acc = for_loop(pl.cdiv(m, block_m), body, (jnp.zeros(block_n), jnp.zeros(block_n)))
-  pl.store(dw_ref, (col_idx,), dw_acc.astype(dw_ref.dtype), mask=col_mask)
-  pl.store(db_ref, (col_idx,), db_acc.astype(db_ref.dtype), mask=col_mask)
+    dw_acc_ref, db_acc_ref = acc
+    return dw_acc_ref + (dout * a_hat).sum(axis=0), db_acc_ref + dout.sum(
+        axis=0
+    )
+
+  dw_acc, db_acc = lax.fori_loop(
+      0,
+      pl.cdiv(m, block_m),
+      body,
+      init_val=(jnp.zeros(block_n), jnp.zeros(block_n)),
+  )
+  plgpu.store(dw_ref.at[col_idx], dw_acc.astype(dw_ref.dtype), mask=col_mask)
+  plgpu.store(db_ref.at[col_idx], db_acc.astype(db_ref.dtype), mask=col_mask)
 
 
 def layer_norm_backward(
