@@ -80,6 +80,7 @@ SplashResidualsType = tuple[
     jax.Array,  # k
     jax.Array,  # v
     Optional[SegmentIds],  # segment_ids
+    Optional[jax.Array],   # sinks
     jax.Array,  # out
     jax.Array,  # logsumexp
     Optional[mask_info_lib.MaskInfo],  # dq_mask_info
@@ -122,6 +123,7 @@ def _attention_reference(
     k: jax.Array,
     v: jax.Array,
     segment_ids: SegmentIds | None,
+    sinks: jax.Array | None,
     save_residuals: Literal[False],
     mask_value: float,
     custom_type: str,
@@ -137,6 +139,7 @@ def _attention_reference(
     k: jax.Array,
     v: jax.Array,
     segment_ids: SegmentIds | None,
+    sinks: jax.Array | None,
     save_residuals: Literal[True],
     mask_value: float,
     custom_type: str,
@@ -151,6 +154,7 @@ def _attention_reference(
     k: jax.Array,  # [kv_seq_len, head_dim]
     v: jax.Array,  # [kv_seq_len, head_dim]
     segment_ids: SegmentIds | None,
+    sinks: jax.Array | None,
     mask_value: float,
     save_residuals: bool,
     custom_type: str,
@@ -162,6 +166,7 @@ def _attention_reference(
       k,
       v,
       segment_ids,
+      sinks,
       mask_value,
       save_residuals,
       custom_type,
@@ -175,6 +180,7 @@ def _attention_reference_default(
     k: jax.Array,  # [kv_seq_len, head_dim]
     v: jax.Array,  # [kv_seq_len, head_dim]
     segment_ids: SegmentIds | None,
+    sinks: jax.Array | None,  # [] one scalar per qhead
     mask_value: float,
     save_residuals: bool,
     custom_type: str,
@@ -194,8 +200,10 @@ def _attention_reference_default(
 
   logits = jnp.where(mask, logits, mask_value)
   m = logits.max(axis=-1)
+  sinks = None if sinks is None else sinks.astype(logits.dtype)
+  m = m if sinks is None else jnp.maximum(m, sinks)
   s = jnp.exp(logits - m[..., None])
-  l = s.sum(axis=-1)
+  l = s.sum(axis=-1) + (0 if sinks is None else jnp.exp(sinks - m))
   s = s / l[..., None]
 
   o = jnp.einsum("st,td->sd", s, v.astype(jnp.float32))
@@ -212,6 +220,7 @@ def attention_reference(
     k: jax.Array,  # [kv_seq_len, head_dim]
     v: jax.Array,  # [kv_seq_len, head_dim]
     segment_ids: SegmentIds | None,
+    sinks: jax.Array | None = None,
     *,
     mask_value: float = DEFAULT_MASK_VALUE,
     save_residuals: bool = False,
@@ -224,6 +233,7 @@ def attention_reference(
       k,
       v,
       segment_ids,
+      sinks,
       mask_value=mask_value,
       save_residuals=save_residuals,
       custom_type=custom_type,
@@ -237,6 +247,7 @@ def _attention_reference_custom_fwd(
     k: jax.Array,  # [kv_seq_len, head_dim]
     v: jax.Array,  # [kv_seq_len, head_dim]
     segment_ids: SegmentIds | None,
+    sinks: jax.Array | None,
     mask_value: float,
     save_residuals: bool,
     custom_type: str,
@@ -251,12 +262,13 @@ def _attention_reference_custom_fwd(
       k,
       v,
       segment_ids,
+      sinks,
       mask_value=mask_value,
       save_residuals=True,
       custom_type=custom_type,
       attn_logits_soft_cap=attn_logits_soft_cap,
   )
-  return o, (mask, q, k, v, segment_ids, o, logsumexp)
+  return o, (mask, q, k, v, segment_ids, sinks, o, logsumexp)
 
 
 def _attention_reference_custom_bwd(
@@ -266,9 +278,9 @@ def _attention_reference_custom_bwd(
     attn_logits_soft_cap: float | None,
     res,
     do: jax.Array,
-) -> tuple[None, jax.Array, jax.Array, jax.Array, None]:
+) -> tuple[None, jax.Array, jax.Array, jax.Array, None, jax.Array | None]:
   del save_residuals
-  mask, q, k, v, segment_ids, o, logsumexp = res
+  mask, q, k, v, segment_ids, sinks, o, logsumexp = res
 
   uncapped_logits = jnp.einsum(
       "qc,kc->qk", q, k, preferred_element_type=jnp.float32)
@@ -306,11 +318,17 @@ def _attention_reference_custom_bwd(
     ds = g + g * d
   dk = jnp.einsum("sd,st->td", q.astype(jnp.float32), ds).astype(k.dtype)
   dq = jnp.einsum("st,td->sd", ds, k.astype(jnp.float32)).astype(q.dtype)
-  return None, dq, dk, dv, None
+  dsinks = None
+  if sinks is not None:  # the gradient is ``sum(-exp(s) / exp(lse) * o * do)``
+    sinks_exp = -jnp.exp(sinks[..., None, None].astype(jnp.float32)
+                         - logsumexp[..., None].astype(jnp.float32))
+    dsinks = jnp.sum(sinks_exp.astype(o.dtype) * do * o)
+  return None, dq, dk, dv, None, dsinks
 
 
 _attention_reference_custom = jax.custom_vjp(
-    _attention_reference, nondiff_argnums=(5, 6, 7, 8)
+  _attention_reference, nondiff_argnames=(
+    "mask_value", "save_residuals", "custom_type", "attn_logits_soft_cap")
 )
 _attention_reference_custom.defvjp(_attention_reference_custom_fwd,
                                    _attention_reference_custom_bwd)
@@ -322,6 +340,7 @@ def attention_reference_custom(
     k: jax.Array,  # [kv_seq_len, head_dim]
     v: jax.Array,  # [kv_seq_len, head_dim]
     segment_ids: SegmentIds | None,
+    sinks: jax.Array | None = None,
     *,
     mask_value: float = DEFAULT_MASK_VALUE,
     save_residuals: bool = False,
@@ -334,6 +353,7 @@ def attention_reference_custom(
       k,
       v,
       segment_ids,
+      sinks,
       mask_value,
       save_residuals,
       custom_type=custom_type,
@@ -361,6 +381,7 @@ def make_attention_reference(
       k: jax.Array,
       v: jax.Array,
       segment_ids: SegmentIds | None = None,
+      sinks: jax.Array | None = None,
       *,
       mask_value: float = DEFAULT_MASK_VALUE,
       save_residuals: bool = False,
@@ -385,7 +406,7 @@ def make_attention_reference(
     )
 
     if is_mqa:
-      func = jax.vmap(func, in_axes=(0, 0, None, None, None))
+      func = jax.vmap(func, in_axes=(0, 0, None, None, None, 0))
       is_grouped = False
     else:
       # In grouped attention (1 < num_kv_heads && num_kv_heads < num_q_heads).
@@ -412,14 +433,16 @@ def make_attention_reference(
         q_heads_per_kv_head = q_heads // kv_heads
         q = q.reshape((kv_heads, q_heads_per_kv_head, q_seq_len, head_dim))
         mask = mask.reshape((kv_heads, q_heads_per_kv_head, *mask.shape[1:]))
+        if sinks is not None:
+          sinks = sinks.reshape((kv_heads, q_heads_per_kv_head))
 
         # Inner-most vmap: iterate over the q heads.
-        func = jax.vmap(func, in_axes=(0, 0, None, None, None))
+        func = jax.vmap(func, in_axes=(0, 0, None, None, None, 0))
 
       # Outer-most vmap: iterate over the kv heads.
-      func = jax.vmap(func, in_axes=(0, 0, 0, 0, None))
+      func = jax.vmap(func, in_axes=(0, 0, 0, 0, None, 0))
 
-    out = func(mask, q, k, v, segment_ids)
+    out = func(mask, q, k, v, segment_ids, sinks)
 
     if is_grouped:
 
@@ -687,6 +710,7 @@ def flash_attention_kernel(
     v_ref,
     q_segment_ids_ref,
     kv_segment_ids_ref,
+    sinks_ref,
     mask_ref,
     q_sequence_ref,
     # Outputs
@@ -710,20 +734,21 @@ def flash_attention_kernel(
 ):
   float32 = jnp.float32
   HEAD_DIM_MINOR = QKVLayout.HEAD_DIM_MINOR
-
-  head_dim_v_repeats, rem = divmod(head_dim_v, NUM_LANES)
-  if rem != 0:
-    raise NotImplementedError(
-        f"{head_dim_v=} should be a multiple of {NUM_LANES}"
-    )
+  head_dim_v_repeats = pl.cdiv(head_dim_v, NUM_LANES)
 
   h, i, j = pl.program_id(0), pl.program_id(1), pl.program_id(2)
 
   @pl.when(j == 0)
   def init():
     o_scratch_ref[...] = jnp.zeros_like(o_scratch_ref)
-    m_scratch_ref[...] = jnp.full_like(m_scratch_ref, mask_value)
-    l_scratch_ref[...] = jnp.zeros_like(l_scratch_ref)
+    if sinks_ref is not None:
+      sinks = sinks_ref[0, h].astype(m_scratch_ref.dtype)
+      # initialize `max = sinks`, so `exp(sinks - max = 0) = 1`
+      m_scratch_ref[...] = sinks * jnp.ones_like(m_scratch_ref)
+      l_scratch_ref[...] = jnp.ones_like(l_scratch_ref)
+    else:
+      m_scratch_ref[...] = jnp.full_like(m_scratch_ref, mask_value)
+      l_scratch_ref[...] = jnp.zeros_like(l_scratch_ref)
 
   global_kv_index, _, should_run, should_not_mask = _next_nonzero(
       h,
@@ -800,7 +825,8 @@ def flash_attention_kernel(
     v = v.astype(float32)
     o_curr = lax.dot_general(s_curr, v, sv_dims)
 
-    alpha_o = pltpu.repeat(alpha, head_dim_v_repeats, axis=1)
+    alpha_o = pltpu.repeat(
+        alpha, head_dim_v_repeats, axis=1)[..., :o_scratch_ref.shape[-1]]
     o_scratch_ref[:] = alpha_o * o_scratch_ref[:] + o_curr
 
   @pl.when(should_run)
@@ -814,7 +840,8 @@ def flash_attention_kernel(
   @pl.when(j == grid_width - 1)
   def end():
     l = l_scratch_ref[...]
-    l_inv = pltpu.repeat(1.0 / l, head_dim_v_repeats, axis=1)
+    l_inv = pltpu.repeat(
+        1.0 / l, head_dim_v_repeats, axis=1)[..., :o_scratch_ref.shape[-1]]
     o_ref[...] = (o_scratch_ref[...] * l_inv).astype(o_ref.dtype)
     if logsumexp_ref is not None:
       assert logsumexp_ref.shape == (bq, NUM_LANES)
@@ -852,6 +879,7 @@ def _splash_attention_forward(
     k: jax.Array,
     v: jax.Array,
     segment_ids: SegmentIds | None,
+    sinks: jax.Array | None,
     mask_value: float,
     is_mqa: bool,
     block_sizes: BlockSizes,
@@ -876,6 +904,7 @@ def _splash_attention_forward(
     k: jax.Array,
     v: jax.Array,
     segment_ids: SegmentIds | None,
+    sinks: jax.Array | None,
     mask_value: float,
     is_mqa: bool,
     block_sizes: BlockSizes,
@@ -1001,11 +1030,13 @@ def _splash_attention_forward(
   # Convert the logical shape from head-minor to sequence-minor.
   in_specs = [
       pl.BlockSpec(
-          from_head_minor((None, bq, head_dim_qk), q_layout), q_index_map
+          from_head_minor((None, bq, head_dim_qk), q_layout),
+          q_index_map
       ),
       pl.BlockSpec(
           from_head_minor(
-              (bkv, head_dim_qk) if is_mqa else (None, bkv, head_dim_qk), k_layout
+              (bkv, head_dim_qk)
+              if is_mqa else (None, bkv, head_dim_qk), k_layout
           ),
           k_index_map,
       ),
@@ -1031,6 +1062,18 @@ def _splash_attention_forward(
     in_specs += [None, None]
     q_segment_ids = kv_segment_ids = None
 
+  if sinks is not None:
+    assert sinks.shape == (num_q_heads,)
+    # align sinks to sublanes to allow vmap and shard_map over the kernel
+    in_specs += [
+        pl.BlockSpec((NUM_SUBLANES, num_q_heads), lambda h, i, j, *_: (0, 0),
+                     memory_space=pltpu.SMEM)
+    ]
+    sinks = jnp.broadcast_to(sinks.astype(jnp.float32)[None, :],
+                             (NUM_SUBLANES, num_q_heads))
+  else:
+    in_specs += [None]
+
   if fwd_mask_info.partial_mask_blocks is not None:
     in_specs.append(pl.BlockSpec((None, bq, bkv), mask_index_map))
   else:
@@ -1133,6 +1176,7 @@ def _splash_attention_forward(
         v if v_layout == QKVLayout.HEAD_DIM_MINOR else v.swapaxes(-1, -2),
         q_segment_ids,
         kv_segment_ids,
+        sinks,
         fwd_mask_info.partial_mask_blocks,
         q_sequence,
     )
@@ -1160,7 +1204,11 @@ def _splash_attention_forward(
   return out
 
 
-@partial(jax.custom_vjp, nondiff_argnums=(7, 8, 9, 10, 11, 12, 13, 14))
+@partial(jax.custom_vjp, nondiff_argnames=(
+  "save_residuals", "mask_value", "is_mqa", "block_sizes",
+  "residual_checkpoint_name", "mask_function", "attn_logits_soft_cap",
+  "interpret")
+)
 def _splash_attention_custom(
     fwd_mask_info: mask_info_lib.MaskInfo,
     dq_mask_info: mask_info_lib.MaskInfo | None,
@@ -1169,6 +1217,7 @@ def _splash_attention_custom(
     k: jax.Array,
     v: jax.Array,
     segment_ids: SegmentIds | None,
+    sinks: jax.Array | None,
     save_residuals: bool,
     mask_value: float,
     is_mqa: bool,
@@ -1195,6 +1244,7 @@ def _splash_attention_custom(
       k,
       v,
       segment_ids,
+      sinks=sinks,
       mask_value=mask_value,
       is_mqa=is_mqa,
       block_sizes=block_sizes,
@@ -1214,6 +1264,7 @@ def _splash_attention_fwd(
     k: jax.Array,
     v: jax.Array,
     segment_ids: SegmentIds | None,
+    sinks: jax.Array | None,
     save_residuals: bool,
     mask_value: float,
     is_mqa: bool,
@@ -1235,6 +1286,7 @@ def _splash_attention_fwd(
       k,
       v,
       segment_ids,
+      sinks,
       mask_value=mask_value,
       is_mqa=is_mqa,
       block_sizes=block_sizes,
@@ -1249,6 +1301,7 @@ def _splash_attention_fwd(
       k,
       v,
       segment_ids,
+      sinks,
       out,
       logsumexp,
       dq_mask_info,
@@ -1267,6 +1320,7 @@ def _flash_attention_dq_kernel(
     v_ref,
     q_segment_ids_ref,
     kv_segment_ids_ref,
+    sinks_ref,
     logsumexp_ref,
     do_ref,
     di_ref,
@@ -1286,6 +1340,7 @@ def _flash_attention_dq_kernel(
     v_layout: QKVLayout,
     mask_function: MaskFunctionType | None,
 ):
+  del sinks_ref  # potentially fuse dsinks computation into the kernel later
   float32 = jnp.float32
   HEAD_DIM_MINOR = QKVLayout.HEAD_DIM_MINOR
 
@@ -1357,6 +1412,7 @@ def _splash_attention_bwd_dq(
     k,
     v,
     segment_ids,
+    sinks,
     logsumexp,
     do,
     di,
@@ -1492,6 +1548,18 @@ def _splash_attention_bwd_dq(
     q_segment_spec = kv_segment_spec = None
     q_segment_ids = kv_segment_ids = None
 
+  if sinks is not None:
+    assert sinks.shape == (num_q_heads,)
+    # align sinks to sublanes to allow vmap and shard_map over the kernel
+    sinks_spec = pl.BlockSpec(
+        (NUM_SUBLANES, num_q_heads), lambda h, i, j, *_: (0, 0),
+        memory_space=pltpu.SMEM
+    )
+    sinks = jnp.broadcast_to(sinks.astype(jnp.float32)[None, :],
+                             (NUM_SUBLANES, num_q_heads))
+  else:
+    sinks_spec = None
+
   do_spec = o_spec
 
   def logsumexp_index_map(h, i, *_):
@@ -1511,6 +1579,7 @@ def _splash_attention_bwd_dq(
       v_spec,
       q_segment_spec,
       kv_segment_spec,
+      sinks_spec,
       logsumexp_spec,
       do_spec,
       di_spec,
@@ -1591,6 +1660,7 @@ def _splash_attention_bwd_dq(
         v if v_layout == QKVLayout.HEAD_DIM_MINOR else v.swapaxes(-1, -2),
         q_segment_ids,
         kv_segment_ids,
+        sinks,
         logsumexp,
         do,
         di,
@@ -1611,6 +1681,7 @@ def _flash_attention_dkv_kernel(
     v_ref,
     q_segment_ids_ref,
     kv_segment_ids_ref,
+    sinks_ref,
     logsumexp_ref,
     do_ref,
     di_ref,
@@ -1638,6 +1709,7 @@ def _flash_attention_dkv_kernel(
     bkv: int,
     mask_function: MaskFunctionType | None,
 ):
+  del sinks_ref  # potentially fuse dsinks computation into the kernel later
   HEAD_DIM_MINOR = QKVLayout.HEAD_DIM_MINOR
   kv_index, q_head_index, q_index = (
       pl.program_id(0),
@@ -1791,6 +1863,7 @@ def _splash_attention_bwd_dkv(
     k,
     v,
     segment_ids,
+    sinks,
     logsumexp,
     do,
     di,
@@ -1945,8 +2018,7 @@ def _splash_attention_bwd_dkv(
   )
 
   dv_spec = pl.BlockSpec(
-      (bkv, head_dim_v) if is_mqa else (None, bkv, head_dim_v),
-      dkv_index_map,
+      (bkv, head_dim_v) if is_mqa else (None, bkv, head_dim_v), dkv_index_map,
   )
 
   def mask_index_map(
@@ -2005,6 +2077,18 @@ def _splash_attention_bwd_dkv(
     q_segment_spec = kv_segment_spec = None
     q_segment_ids = kv_segment_ids = None
 
+  if sinks is not None:
+    assert sinks.shape == (num_q_heads,)
+    # align sinks to sublanes to allow vmap and shard_map over the kernel
+    sinks_spec = pl.BlockSpec(
+        (NUM_SUBLANES, num_q_heads), lambda h, i, j, *_: (0, 0),
+        memory_space=pltpu.SMEM
+    )
+    sinks = jnp.broadcast_to(sinks.astype(jnp.float32)[None, :],
+                             (NUM_SUBLANES, num_q_heads))
+  else:
+    sinks_spec = None
+
   do_spec = o_spec
 
   def logsumexp_index_map(
@@ -2044,6 +2128,7 @@ def _splash_attention_bwd_dkv(
       v_spec,
       q_segment_spec,
       kv_segment_spec,
+      sinks_spec,
       logsumexp_spec,
       do_spec,
       di_spec,
@@ -2140,6 +2225,7 @@ def _splash_attention_bwd_dkv(
         v if v_layout == QKVLayout.HEAD_DIM_MINOR else v.swapaxes(-1, -2),
         q_segment_ids,
         kv_segment_ids,
+        sinks,
         logsumexp,
         do,
         di,
@@ -2174,6 +2260,7 @@ def _splash_attention_bwd(
     jax.Array,  # k
     jax.Array,  # v
     SegmentIds | None,  # segmend_ids
+    jax.Array | None,  # sinks
 ]:
   del save_residuals, residual_checkpoint_name
   if not block_sizes.has_backward_blocks:
@@ -2190,6 +2277,7 @@ def _splash_attention_bwd(
       k,
       v,
       segment_ids,
+      sinks,
       o,
       logsumexp,
       dq_mask_info,
@@ -2203,6 +2291,7 @@ def _splash_attention_bwd(
       k,
       v,
       segment_ids,
+      sinks,
       logsumexp,
       do,
       di,
@@ -2227,6 +2316,7 @@ def _splash_attention_bwd(
         k,
         v,
         segment_ids,
+        sinks,
         logsumexp,
         do,
         di,
@@ -2244,6 +2334,11 @@ def _splash_attention_bwd(
     )
   # Match the signature of the fwd function.
   assert dq is not None
+  dsinks = None
+  if sinks is not None:
+    sinks_exp = -jnp.exp(sinks[..., None, None].astype(jnp.float32)
+                         - logsumexp[..., None].astype(jnp.float32))
+    dsinks = jnp.sum(sinks_exp.astype(o.dtype) * o * do, axis=(-1, -2))
   return (
       None,  # fwd_mask_info
       None,  # dq_mask_info
@@ -2252,6 +2347,7 @@ def _splash_attention_bwd(
       dk,  # k
       dv,  # v
       None,  # segment_ids
+      dsinks,  # sinks
   )
 
 
@@ -2279,6 +2375,7 @@ def _splash_attention(
     k: jax.Array,
     v: jax.Array,
     segment_ids: SegmentIds | None = None,
+    sinks: jax.Array | None = None,
     *,
     is_mqa: bool,
     block_sizes: BlockSizes | None,
@@ -2298,7 +2395,7 @@ def _splash_attention(
   """
   def _collapse_partial_mask_blocks(mask_info: mask_info_lib.MaskInfo | None):
     if mask_info is None or mask_info.partial_mask_blocks is None:
-        return mask_info
+      return mask_info
 
     return mask_info._replace(
         partial_mask_blocks=mask_info.partial_mask_blocks.reshape(
@@ -2317,6 +2414,7 @@ def _splash_attention(
       k,
       v,
       segment_ids,
+      sinks,
       mask_value=mask_value,
       is_mqa=is_mqa,
       block_sizes=block_sizes,
