@@ -18,6 +18,8 @@
 from jax._src.lax.lax import (
   DotDimensionNumbers as DotDimensionNumbers,
   RaggedDotDimensionNumbers as RaggedDotDimensionNumbers,
+  AccuracyMode as AccuracyMode,
+  Tolerance as Tolerance,
   Precision as Precision,
   PrecisionLike as PrecisionLike,
   DotAlgorithm as DotAlgorithm,
@@ -232,7 +234,8 @@ from jax._src.lax.lax import (
   transpose as transpose,
   transpose_p as transpose_p,
   xor_p as xor_p,
-  zeros_like_array as zeros_like_array,
+  empty as empty,
+  zeros_like_array as _deprecated_zeros_like_array,
 )
 from jax._src.lax.special import (
   bessel_i0e as bessel_i0e,
@@ -354,7 +357,7 @@ from jax._src.lax.fft import (
 )
 from jax._src.lax.parallel import (
   all_gather as all_gather,
-  all_gather_invariant as all_gather_invariant,
+  pcast as pcast,
   all_gather_p as all_gather_p,
   all_to_all as all_to_all,
   all_to_all_p as all_to_all_p,
@@ -380,7 +383,7 @@ from jax._src.lax.parallel import (
   ragged_all_to_all_p as ragged_all_to_all_p,
 )
 from jax._src.core import (
-    pvary as pvary,
+    pvary as _deprecated_pvary,
 )
 from jax._src.lax.other import (
   conv_general_dilated_local as conv_general_dilated_local,
@@ -397,3 +400,30 @@ from jax.lax import linalg as linalg
 from jax._src.pjit import with_sharding_constraint as with_sharding_constraint
 from jax._src.pjit import sharding_constraint_p as sharding_constraint_p
 from jax._src.dispatch import device_put_p as device_put_p
+
+_deprecations = {
+    # Added on July 24th 2025.
+    "zeros_like_array": (
+        (
+            "jax.lax.zeros_like_array is deprecated. Use jax.numpy.zeros_like"
+            " instead."
+        ),
+        _deprecated_zeros_like_array,
+    ),
+    # Added on Dec 1, 2025
+    "pvary": (
+        "jax.lax.pvary is deprecated. Use `jax.lax.pcast(..., to='varying')",
+        _deprecated_pvary,
+    ),
+}
+
+import typing as _typing
+if _typing.TYPE_CHECKING:
+  zeros_like_array = _deprecated_zeros_like_array
+  pvary = _deprecated_pvary
+else:
+  from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
+  __getattr__ = _deprecation_getattr(__name__, _deprecations)
+  del _deprecation_getattr
+del _deprecated_zeros_like_array, _deprecated_pvary
+del _typing
