@@ -12,95 +12,62 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from jax._src.lib import xla_client as _xc
+from jax._src import deprecations as _deps
+
+_deps.warn(
+    'jax-lib-module',
+    (
+        'jax.lib.xla_client module will be removed in JAX v0.9.0;'
+        ' all its APIs were deprecated and removed by JAX v0.8.0.'
+    ),
+    stacklevel=4
+)
 
 _deprecations = {
-    # Finalized for JAX v0.7.0
-    "heap_profile": (
-        (
-            "jax.lib.xla_client.heap_profile was deprecated in JAX v0.6.0 and"
-            " removed in JAX v0.7.0"
-        ),
-        None,
-    ),
-    "get_topology_for_devices": (
-        (
-            "jax.lib.xla_client.get_topology_for_devices was deprecated in JAX"
-            " v0.6.0 and removed in JAX v0.7.0"
-        ),
-        None,
-    ),
-    "mlir_api_version": (
-        (
-            "jax.lib.xla_client.mlir_api_version was deprecated in JAX v0.6.0"
-            " and removed in JAX v0.7.0"
-        ),
-        None,
-    ),
-    "DeviceAssignment": (
-        (
-            "jax.lib.xla_client.DeviceAssignment was deprecated in JAX v0.6.0"
-            " and removed in JAX v0.7.0"
-        ),
-        None,
-    ),
-    # Added April 4 2025.
+    # Finalized in JAX v0.8.0; remove these messages in v0.9.0.
     "Client": (
         (
             "jax.lib.xla_client.Client was deprecated in JAX v0.6.0 and will be"
-            " removed in JAX v0.7.0"
+            " removed in JAX v0.8.0"
         ),
-        _xc.Client,
+        None,
     ),
     "CompileOptions": (
         (
             "jax.lib.xla_client.CompileOptions was deprecated in JAX v0.6.0 and"
-            " will be removed in JAX v0.7.0"
+            " will be removed in JAX v0.8.0"
         ),
-        _xc.CompileOptions,
+        None,
     ),
     "Frame": (
         (
             "jax.lib.xla_client.Frame was deprecated in JAX v0.6.0 and will be"
-            " removed in JAX v0.7.0"
+            " removed in JAX v0.8.0"
         ),
-        _xc.Frame,
+        None,
     ),
     "HloSharding": (
         (
             "jax.lib.xla_client.HloSharding was deprecated in JAX v0.6.0 and"
-            " will be removed in JAX v0.7.0"
+            " will be removed in JAX v0.8.0"
         ),
-        _xc.HloSharding,
+        None,
     ),
     "OpSharding": (
         (
             "jax.lib.xla_client.OpSharding was deprecated in JAX v0.6.0 and"
-            " will be removed in JAX v0.7.0"
+            " will be removed in JAX v0.8.0"
         ),
-        _xc.OpSharding,
+        None,
     ),
     "Traceback": (
         (
             "jax.lib.xla_client.Traceback was deprecated in JAX v0.6.0 and will"
-            " be removed in JAX v0.7.0"
+            " be removed in JAX v0.8.0"
         ),
-        _xc.Traceback,
+        None,
     ),
 }
 
-import typing as _typing
-
-if _typing.TYPE_CHECKING:
-  Client = _xc.Client
-  CompileOptions = _xc.CompileOptions
-  Frame = _xc.Frame
-  HloSharding = _xc.HloSharding
-  OpSharding = _xc.OpSharding
-  Traceback = _xc.Traceback
-else:
-  from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
-  __getattr__ = _deprecation_getattr(__name__, _deprecations)
-  del _deprecation_getattr
-del _typing
-del _xc
+__getattr__ = _deps.deprecation_getattr(__name__, _deprecations)
+del _deps
